% (c) 2017-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

 :- module(alloy2b,[load_model/1,translate_model/2,run_tests/0]).

:- use_module(library(lists)).

:- use_module(probsrc(bmachine)).
:- use_module(probsrc(translate)).
:- use_module(library(plunit)).

:- dynamic singleton_set/1, total_function/1, ordered_signature/1, command_counter/1.
:- volatile singleton_set/1, total_function/1, ordered_signature/1, command_counter/1.

:- meta_predicate map_translate(3,-,-,-).

command_counter(0).

% An automated translation from Alloy to classical B.
% The Alloy abstract syntax tree is translated to an untyped B AST as supported by ProB.
% Afterwards, the untyped AST can be typechecked and loaded by ProB.
% The used positions are from the Alloy parser and thus refer to the Alloy model.

% Usage:
% load_model/1: input is an Alloy model represented as a Prolog term as provided by the Kotlin translation; the model is translated to an untyped B AST and loaded by ProB
% get_loaded_machine_as_atom/1: returns the currently loaded B machine as an atom (requires a call to load_model/1 beforehand)
% translate_model/2: input is an Alloy model represented as a Prolog term as provided by the Kotlin translation, output is an untyped B AST

load_model(AlloyTerm) :- 
    translate_model(AlloyTerm,UntypedBAst) , ! , 
    UntypedBAst = machine(generated(none,AbstractMachine)) , 
    b_load_machine_from_term('alloytranslation',complete_machine('alloytranslation',[AbstractMachine],['alloytranslation'])).

get_loaded_machine_as_atom(MachineAtom) :- 
    full_b_machine(FullBMachine) , 
    % TODO: use atelierb_pp_mode 
    translate_machine(FullBMachine,MachineCodes,_) , 
    atom_codes(MachineAtom,MachineCodes).

translate_model(alloy_model(facts(Facts),assertions(Assertions),commands(Commands),functions(Functions),signatures(Signatures)),UntypedBAst) :- 
    % singleton sets are asserted at runtime using singleton_set/1
    % accumulate all translations, afterwards we build the untyped machine ast
    empty_machine_acc(MAcc) , 
    map_translate(translate_signature,MAcc,Signatures,MAcc1) , 
    map_translate(translate_assertion,MAcc1,Assertions,MAcc2) , 
    map_translate(translate_command,MAcc2,Commands,MAcc3) , 
    map_translate(translate_function,MAcc3,Functions,MAcc4) , 
    map_translate(translate_fact,MAcc4,Facts,MAcc5) , ! , 
    build_machine_ast(MAcc5,UntypedBAst) , 
    retract_state.

% Map the translation over a list and accumulate the results. 
% Type is one of signature, assertion, command, function, field or fact.
map_translate(_,MAcc,[],MAcc).
map_translate(TypePred,MAcc,[Part|T],Res) :- 
    call(TypePred,MAcc,Part,NewMAcc) , 
    map_translate(TypePred,NewMAcc,T,Res).

% signature
translate_signature(MAcc,signature(Name,Fields,Facts,Options,pos(Col,Row)),NewMAcc) :- 
    % assert signatures for singleton checks
    assert_singleton_set(Options,Name) , 
    extend_machine_acc(signatures,MAcc,[Name],MAcc1) , 
    translate_signature_aux(Name,Options,pos(Col,Row),MAcc1,MAcc2) ,  
    translate_signature_fields(MAcc2,Name,Fields,MAcc3) , 
    translate_e_p(Name,TName) , 
    map_translate(translate_signature_fact(TName),MAcc3,Facts,NewMAcc).

translate_signature_aux(Name,Options,Pos,MAcc,NewMAcc) :- 
    % ordered signatures are defined as distinct sets of integer when translating a command
    member(ordered,Options) , ! , 
    asserta(ordered_signature(Name)) , 
    define_ordered_signature_functions(Pos,MAcc,Name,MAcc1) , 
    extend_machine_acc(signatures,MAcc1,[Name],MAcc2) , 
    extend_machine_acc(properties,MAcc2,member(none,identifier(none,Name),pow_subset(none,integer_set(none))),MAcc3) , 
    define_sig_as_set_or_constant_aux(constants,MAcc3,Name,Options,Pos,NewMAcc).
translate_signature_aux(Name,Options,Pos,MAcc,NewMAcc) :- 
    define_sig_as_set_or_constant(MAcc,Name,Options,Pos,NewMAcc).

define_ordered_signature_functions(pos(Col,Row),MAcc,Name,NewMAcc) :- 
    TPos = pos(0,0,Row,Col,0,0) , TIDX = identifier(TPos,x) ,  TIDS = identifier(TPos,s) ,
    translate_e_p(Name,TName) ,  MemberX = member(TPos,TIDX,TName) , 
    % next_Sig(s)  == {x|x=s+1 & x:Sig}
    atom_concat(next_,Name,NextName) , 
    NextBody = conjunct(TPos,equal(TPos,TIDX,add(TPos,TIDS,integer(TPos,1))),MemberX) , 
    extend_machine_acc(definitions,MAcc,expression_definition(TPos,NextName,[TIDS],comprehension_set(TPos,[TIDX],NextBody)),MAcc1) , 
    % nexts_Sig(s) == {x|x>s & x:Sig}
    atom_concat(nexts_,Name,NextsName) , 
    NextsBody = conjunct(TPos,greater(TPos,TIDX,TIDS),MemberX) , 
    extend_machine_acc(definitions,MAcc1,expression_definition(TPos,NextsName,[TIDS],comprehension_set(TPos,[TIDX],NextsBody)),MAcc2) , 
    atom_concat(prev_,Name,PrevName) , 
    % prev_Sig(s)  == {x|x=s-1 & x:Sig}
    PrevBody = conjunct(TPos,equal(TPos,TIDX,minus(TPos,TIDS,integer(TPos,1))),MemberX) , 
    extend_machine_acc(definitions,MAcc2,expression_definition(TPos,PrevName,[TIDS],comprehension_set(TPos,[TIDX],PrevBody)),MAcc3) , 
    % prevs_Sig(s) == {x|x<s & x:Sig}
    atom_concat(prevs_,Name,PrevsName) , 
    PrevsBody = conjunct(TPos,less(TPos,TIDX,TIDS),MemberX) , 
    extend_machine_acc(definitions,MAcc3,expression_definition(TPos,PrevsName,[TIDS],comprehension_set(TPos,[TIDX],PrevsBody)),NewMAcc).

% list of fields from a specific signature
translate_signature_fields(MAcc,SignatureName,Fields,NewMAcc) :- 
    translate_e_p(SignatureName,TSignatureName) , 
    translate_signature_fields_aux(MAcc,TSignatureName,Fields,NewMAcc).

translate_signature_fields_aux(MAcc,_,[],MAcc).
translate_signature_fields_aux(MAcc,TSignatureName,[Field|T],NewMAcc) :- 
    translate_signature_field(TSignatureName,MAcc,Field,MAcc1) , 
    translate_signature_fields_aux(MAcc1,TSignatureName,T,NewMAcc).

% TODO: support disj keyword for field and parameter declarations

% field
translate_signature_field(TSignatureName,MAcc,Field,NewMAcc) :- 
    translate_signature_field_aux(TSignatureName,MAcc,Field,TField,MAcc1) , 
    extend_machine_acc(properties,MAcc1,TField,NewMAcc).

translate_signature_field_aux(TSignatureName,MAcc,field(Name,Expr,type(_Type,_Arity),pos(Col,Row)),TField,NewMAcc) :- 
    translate_e_p(Name,TName) , 
    extend_machine_acc(constants,MAcc,TName,NewMAcc) , 
    field_decl_special_cases(pos(0,0,Row,Col,0,0),TSignatureName,Expr,TName,TField) , !.

translate_function_field(field(Name,Expr,type(_Type,_Arity),pos(Col,Row)),TField,Name) :- 
    assert_singleton_set(Name) , 
    translate_e_p(Name,TName) , 
    fun_or_pred_decl_special_cases(pos(0,0,Row,Col,0,0),Expr,TName,TField) , !.

% fact
translate_fact(MAcc,fact(Expr,_Pos),NewMAcc) :- 
    translate_e_p(Expr,TExpr) , 
    extend_machine_acc(properties,MAcc,TExpr,NewMAcc).

% Signature facts: Identifiers may refer to the signature and are joined with 'this'.
% We then need a universal quantification using 'this'.
translate_signature_fact(TSignatureName,MAcc,Expr,NewMAcc) :- 
    translate_e_p(Expr,TExpr) , 
    alloy_expr_contains_join_with_this(Expr) , 
    ThisID = identifier(none,'this') , 
    TImplication = implication(none,subset(none,set_extension(none,[ThisID]),TSignatureName),TExpr) , 
    extend_machine_acc(properties,MAcc,forall(none,[ThisID],TImplication),NewMAcc).
translate_signature_fact(_TSignatureName,MAcc,Expr,NewMAcc) :- 
    translate_e_p(Expr,TExpr) , 
    extend_machine_acc(properties,MAcc,TExpr,NewMAcc).

alloy_expr_contains_join_with_this(AndOr) :- 
    AndOr =.. [Functor,ListOfAsts,_Pos] , 
    member(Functor,[and,or]) , 
    findall(Ast,(member(Ast,ListOfAsts) , alloy_expr_contains_join_with_this(Ast)),Temp) , 
    length(Temp,LTemp) , ! , LTemp > 0.
alloy_expr_contains_join_with_this(join(Arg1,Arg2,_Type,_Pos)) :- 
    Arg1 = identifier('this',_,_) ; Arg2 = identifier('this',_,_).
alloy_expr_contains_join_with_this(Expr) :- 
    Expr =.. [_Functor,Arg1,Arg2,_Type,_Pos] , 
    (alloy_expr_contains_join_with_this(Arg1) ;
     alloy_expr_contains_join_with_this(Arg2)).
alloy_expr_contains_join_with_this(Expr) :- 
    Expr =.. [_Functor,Arg,_Type,_Pos] , 
    alloy_expr_contains_join_with_this(Arg).

% function or predicate
translate_function(MAcc,FunctionOrPredicate,NewMAcc) :- 
    FunctionOrPredicate =.. [Functor,Name,Params,Decls,Body,pos(Col,Row)] , 
    alloy_to_b_operator(Functor,BFunctor) , 
    maplist(translate_e_p,Params,TParams) , 
    maplist(translate_function_field,Decls,TDecls,SingletonSetNames) , 
    translate_e_p(Body,TBody) , 
    (Functor = function 
    ->  append(TDecls,[member(none,identifier(none,'temp'),TBody)],BehaviorList) , 
        join_predicates_aux(conjunct,BehaviorList,Behavior) , 
        % define functions as comprehension sets
        UAst =.. [BFunctor,pos(0,0,Row,Col,0,0),Name,TParams,comprehension_set(pos(0,0,Row,Col,0,0),[identifier(none,'temp')],Behavior)]
    ;   append(TDecls,[TBody],BehaviorList) , 
        join_predicates_aux(conjunct,BehaviorList,Behavior) , 
        UAst =.. [BFunctor,pos(0,0,Row,Col,0,0),Name,TParams,Behavior]) , 
    extend_machine_acc(definitions,MAcc,UAst,NewMAcc) , 
    maplist(retract_singleton_set,SingletonSetNames).

% assertion
translate_assertion(MAcc,Assertion,NewMAcc) :- 
    translate_fact(MAcc,Assertion,NewMAcc).

% check and run command
translate_command(MAcc,Command,NewMAcc) :- 
    Command =.. [Functor,Body,GlobalScope,ExactScopes,BitWidth,pos(Col,Row)] , 
    (Functor = check ; Functor = run) , 
    translate_e_p(Body,TBody) , 
    % we need all signature names to define the global scope
    get_signature_names_from_machine_acc(MAcc,SignatureNames) , 
    translate_scopes(SignatureNames,GlobalScope,ExactScopes,BitWidth,TScopesPred) , 
    TPos = pos(0,0,Row,Col,0,0) , 
    Precondition = conjunct(TPos,TScopesPred,TBody) , 
    get_command_counter_atom(CommandCounter) , 
    atom_concat(Functor,CommandCounter,OperationName) , 
    Operation = operation(TPos,identifier(TPos,OperationName),[],[],precondition(TPos,Precondition,skip(none))) , 
    extend_machine_acc(operations,MAcc,Operation,NewMAcc).

% global scope, exact scopes and bitwidth
% We do not need to set the bitwidth since we have real integers in B. 
% Note: If only one command is defined in the Alloy model we could set min and max int of ProB in the definitions.
translate_scopes(SignatureNames,global_scope(GlobalScope),exact_scopes(ExactScopes),_BitWidth,conjunct(none,TExactScopes,TGlobalScopes)) :- 
    % translate exact scopes first
    translate_exact_scopes(0,ExactScopes,ExactSignatureNames,TExactScopes,OrderedSignatureCounter) ,  
    % if present, define the global scope for the remaining signatures
    translate_global_scope(OrderedSignatureCounter,GlobalScope,SignatureNames,ExactSignatureNames,TGlobalScopes,_).

% no exact scopes defined
translate_exact_scopes(OrderedSignatureCounter,[],_,truth(none),OrderedSignatureCounter) :- !.
translate_exact_scopes(OrderedSignatureCounter,ExactScopes,ExactSignatureNames,TExactScopes,NewOrderedSignatureCounter) :- 
    translate_exact_scopes_aux(OrderedSignatureCounter,ExactScopes,[],ExactSignatureNames,[],TempTExactScopes,NewOrderedSignatureCounter) , 
    join_predicates(conjunct,TempTExactScopes,TExactScopes).

translate_exact_scopes_aux(OrderedSignatureCounter,[],NamesAcc,NamesAcc,TScopeAcc,TScopeAcc,OrderedSignatureCounter).
translate_exact_scopes_aux(OrderedSignatureCounter,[(SignatureName,Scope)|T],NameAcc,ExactSignatureNames,TScopeAcc,TempTExactScopes,NewOrderedSignatureCounter) :- 
    is_ordered_signature(SignatureName) , ! , 
    define_ordered_signature_as_integer_set(OrderedSignatureCounter,SignatureName,Scope,TScope,TNewOrderedSignatureCounter) , 
    translate_exact_scopes_aux(TNewOrderedSignatureCounter,T,[SignatureName|NameAcc],ExactSignatureNames,[TScope|TScopeAcc],TempTExactScopes,NewOrderedSignatureCounter).
translate_exact_scopes_aux(OrderedSignatureCounter,[(SignatureName,Scope)|T],NameAcc,ExactSignatureNames,TScopeAcc,TempTExactScopes,NewOrderedSignatureCounter) :- 
    translate_e_p(SignatureName,TSignatureName) , 
    TScope = equal(none,card(none,TSignatureName),integer(none,Scope)) , 
    translate_exact_scopes_aux(OrderedSignatureCounter,T,[SignatureName|NameAcc],ExactSignatureNames,[TScope|TScopeAcc],TempTExactScopes,NewOrderedSignatureCounter).

define_ordered_signatures_as_integer_set(OrderedSignatureCounter,[],_,[],OrderedSignatureCounter).
define_ordered_signatures_as_integer_set(OrderedSignatureCounter,[TSignatureName|T],Scope,[TScope|TScopes],NewOrderedSignatureCounter) :- 
    define_ordered_signature_as_integer_set(OrderedSignatureCounter,TSignatureName,Scope,TScope,TNewOrderedSignatureCounter) , 
    define_ordered_signatures_as_integer_set(TNewOrderedSignatureCounter,T,Scope,TScopes,NewOrderedSignatureCounter).

define_ordered_signature_as_integer_set(OrderedSignatureCounter,TSignatureName,Scope,TScope,NewOrderedSignatureCounter) :- 
    NewOrderedSignatureCounter is OrderedSignatureCounter + Scope , 
    NewOrderedSignatureCounter1 is NewOrderedSignatureCounter - 1 , 
    TScope = equal(none,identifier(none,TSignatureName),interval(none,integer(none,OrderedSignatureCounter),integer(none,NewOrderedSignatureCounter1))).

translate_global_scope(OrderedSignatureCounter,-1,_,_,truth(none),OrderedSignatureCounter) :- !. % no global scope defined
translate_global_scope(OrderedSignatureCounter,GlobalScope,SignatureNames,ExactSignatureNames,TGlobalScopes,NewOrderedSignatureCounter) :- 
    findall(SignatureName,(member(SignatureName,SignatureNames) , \+member(SignatureName,ExactSignatureNames) , is_ordered_signature(SignatureName)),OrderedSignatureNames) , 
    findall(SignatureName,(member(SignatureName,SignatureNames) , \+member(SignatureName,ExactSignatureNames) , \+is_ordered_signature(SignatureName)),RestSignatureNames) , 
    translate_global_scope_aux(RestSignatureNames,GlobalScope,[],TempTGlobalScopes) , 
    join_predicates(conjunct,TempTGlobalScopes,TTGlobalScopes) , 
    define_ordered_signatures_as_integer_set(OrderedSignatureCounter,OrderedSignatureNames,GlobalScope,TOrderedScopes,NewOrderedSignatureCounter) , 
    join_predicates(conjunct,[TTGlobalScopes|TOrderedScopes],TGlobalScopes).

translate_global_scope_aux([],_,Acc,Acc).
translate_global_scope_aux([SignatureName|T],GlobalScope,Acc,TGlobalScopes) :- 
    translate_e_p(SignatureName,TSignatureName) , 
    TScope = less_equal(none,card(none,TSignatureName),integer(none,GlobalScope)) , 
    translate_global_scope_aux(T,GlobalScope,[TScope|Acc],TGlobalScopes).

% signature in (subset)
define_sig_as_set_or_constant(MAcc,Name,Options,pos(Col,Row),NewMAcc) :- 
    memberchk(subset(Parents),Options) , ! , 
    define_sig_as_set_or_constant_aux(constants,MAcc,Name,Options,pos(Col,Row),MAcc1) , 
    % TODO: consider several parents -> we need the universe type
    Parents = [Parent|_] , 
    translate_e_p(Name,TName) , 
    translate_e_p(Parent,TParent) , 
    TNode =.. [subset,pos(0,0,Row,Col,0,0),TName,TParent] , 
    extend_machine_acc(properties,MAcc1,TNode,NewMAcc).
% signature extends
define_sig_as_set_or_constant(MAcc,Name,Options,pos(Col,Row),NewMAcc) :- 
    member(subsig(Parent),Options) , ! , 
    define_sig_as_set_or_constant_aux(constants,MAcc,Name,Options,pos(Col,Row),MAcc1) ,
    translate_e_p(Name,TName) , 
    translate_e_p(Parent,TParent) , 
    TNode =.. [subset,pos(0,0,Row,Col,0,0),TName,TParent] , 
    extend_machine_acc(properties,MAcc1,TNode,NewMAcc).
% default signature
define_sig_as_set_or_constant(MAcc,Name,Options,Pos,NewMAcc) :- 
    define_sig_as_set_or_constant_aux(sets,MAcc,Name,Options,Pos,NewMAcc).

define_sig_as_set_or_constant_aux(sets,MAcc,Name,_Options,pos(Col,Row),NewMAcc) :- 
    % use deferred_set/2 instead of identifier/2 like for constants
    extend_machine_acc(sets,MAcc,deferred_set(pos(0,0,Row,Col,0,0),Name),NewMAcc).
define_sig_as_set_or_constant_aux(constants,MAcc,Name,_Options,pos(Col,Row),NewMAcc) :- 
    extend_machine_acc(constants,MAcc,identifier(pos(0,0,Row,Col,0,0),Name),NewMAcc).

% translate expression or predicate
translate_e_p(A,TA) :- 
    % check quantifiers first
    translate_quantifier_e(A,TA) , !.
translate_e_p(A,TA) :- 
    translate_unary_e_p(A,TA) , !.
translate_e_p(A,TA) :- 
    translate_binary_e_p(A,TA) , !.
translate_e_p(A,TA) :- 
    translate_if_then_else(A,TA).
translate_e_p(A,_) :- 
    format("Translation failed for ~w.~n",[A]).

translate_if_then_else(if_then_else(ConditionPred,TruthExpr,FalsityExpr,_Type,pos(Col,Row)),TIfThenElse) :- 
    translate_e_p(ConditionPred,TConditionPred) , 
    translate_e_p(TruthExpr,TTruthExpr) , 
    translate_e_p(FalsityExpr,TFalsityExpr) , 
    TIfThenElse = conjunct(pos(0,0,Row,Col,0,0),
        implication(pos(0,0,Row,Col,0,0),TConditionPred,TTruthExpr),
        implication(pos(0,0,Row,Col,0,0),negation(pos(0,0,Row,Col,0,0),TConditionPred),TFalsityExpr)).

% quantifiers
translate_quantifier_e(Quantifier,TQuantifier) :- 
    Quantifier =.. [Functor,Params,Fields,Body,_Type,pos(Col,Row)] , 
    member(Functor,[all,no,some,one,lone]) , 
    maplist(translate_e_p,Params,TParams) , 
    maplist(assert_singleton_set,Params) , 
    maplist(translate_function_field,Fields,TFieldsList,SingletonSetNames) , 
    join_predicates_aux(conjunct,TFieldsList,TFields) , 
    translate_e_p(Body,TBody) , 
    maplist(retract_singleton_set,Params) , 
    translate_quantifier_e_aux(pos(0,0,Row,Col,0,0),Functor,TParams,TFields,TBody,TQuantifier) , 
    maplist(retract_singleton_set,SingletonSetNames).

translate_quantifier_e_aux(Pos,all,TParams,TFields,TBody,forall(Pos,TParams,implication(none,TFields,TBody))).
translate_quantifier_e_aux(Pos,no,TParams,TFields,TBody,negation(Pos,exists(none,TParams,conjunct(none,TFields,TBody)))).
translate_quantifier_e_aux(Pos,some,TParams,TFields,TBody,exists(Pos,TParams,conjunct(none,TFields,TBody))).
translate_quantifier_e_aux(Pos,one,TParams,TFields,TBody,equal(Pos,card(none,comprehension_set(none,TParams,conjunct(none,TFields,TBody))),integer(none,1))).
translate_quantifier_e_aux(Pos,lone,TParams,TFields,TBody,less_equal(Pos,card(none,comprehension_set(none,TParams,conjunct(none,TFields,TBody))),integer(none,1))).

% unary expressions and predicates
translate_unary_e_p(Int,integer(none,Int)) :- integer(Int) , !.
translate_unary_e_p(identifier(ID,_Type,pos(Col,Row)),set_extension(pos(0,0,Row,Col,0,0),[identifier(none,ID)])) :- 
    is_singleton_set(ID) , !.
translate_unary_e_p(ID,set_extension(none,[identifier(none,ID)])) :- 
    is_singleton_set(ID) , !.
translate_unary_e_p(identifier('Int_',_Type,pos(Col,Row)),integer_set(pos(0,0,Row,Col,0,0))) :- !.
translate_unary_e_p('Int_',integer_set(none)) :- !.
translate_unary_e_p(identifier(ID,_Type,pos(Col,Row)),identifier(pos(0,0,Row,Col,0,0),ID)) :- !.
translate_unary_e_p(ID,identifier(none,ID)) :- atom(ID) , !.
translate_unary_e_p(integer(A,pos(Col,Row)),integer(pos(0,0,Row,Col,0,0),A)) :- !.
translate_unary_e_p(boolean(true,pos(Col,Row)),boolean_true(pos(0,0,Row,Col,0,0))) :- !.
translate_unary_e_p(boolean(false,pos(Col,Row)),boolean_false(pos(0,0,Row,Col,0,0))) :- !.
translate_unary_e_p(UnaryP,TUnaryP) :- 
    % and/or defines a list of ast nodes
    UnaryP =.. [Op,ArgList|_] , 
    memberchk(Op,[and,or]) , 
    is_list(ArgList) , ! , 
    maplist(translate_e_p,ArgList,TArgList) , 
    join_predicates(Op,TArgList,TUnaryP).
translate_unary_e_p(UnaryP,TUnaryP) :- 
    UnaryP =.. [Op,Arg,_Type,pos(Col,Row)] , 
    member(Op,[no,one,some,lone]) , ! ,
    translate_e_p(Arg,TArg) ,  
    translate_quantified_e(pos(0,0,Row,Col,0,0),Op,TArg,TUnaryP).
translate_unary_e_p(UnaryP,TUnaryP) :- 
    UnaryP =.. [Op,Arg,_Type,pos(Col,Row)] , 
    translate_e_p(Arg,TArg) , 
    alloy_to_b_operator(Op,BOp) , 
    TUnaryP =.. [BOp,pos(0,0,Row,Col,0,0),TArg].

integer_function_to_b('integer_\'minus_',minus).
integer_function_to_b('integer_\'plus_',add).
integer_function_to_b('integer_\'div_',div).
integer_function_to_b('integer_\'mul_',multiplication).
integer_function_to_b('integer_\'rem_',modulo).

ordering_function_to_b(type([SignatureName|_],_),OrderingFunctionName,FunctionName) :- 
    atom_concat(Prefix,TempFunctionName,OrderingFunctionName) , 
    (Prefix = 'ordering_\'' ; Prefix = SignatureName) , 
    ((TempFunctionName = first_ , FunctionName = min) ; (TempFunctionName = last_ , FunctionName = max)) , !.
ordering_function_to_b(type([SignatureName|_],_),OrderingFunctionName,FunctionName) :- 
    atom_concat('ordering_\'',TempFunctionName,OrderingFunctionName) , 
    atom_concat(TempFunctionName,SignatureName,FunctionName).

% binary expressions and predicates
translate_binary_e_p(join(Arg1,Arg2,_Type,pos(Col,Row)),TBinaryJoin) :- ! , 
    translate_join(pos(0,0,Row,Col,0,0),Arg1,Arg2,TBinaryJoin).
translate_binary_e_p(let(VarName,Expr,Sub,_Type,pos(Col,Row)),let_expression(BPos,[TVarName],equal(BPos,TVarName,TExpr),TSub)) :- ! , 
    translate_e_p(VarName,TVarName) , 
    translate_e_p(Expr,TExpr) , 
    translate_e_p(Sub,TSub) , 
    BPos = pos(0,0,Row,Col,0,0).
translate_binary_e_p(Call,TCall) :- 
    Call =.. [Functor,Name,Params,Type,pos(Col,Row)] , 
    (Functor = pred_call ; Functor = fun_call ) , ! , 
    translate_function_call(Name,Params,Type,pos(Col,Row),TCall).
translate_binary_e_p(Binary,TBinary) :- 
    Binary =.. [Op,Arg1,Arg2,_Type,pos(Col,Row)] , 
    translate_e_p(Arg1,TArg1) , 
    translate_e_p(Arg2,TArg2) , 
    alloy_to_b_operator(Op,BOp) , 
    TBinary =.. [BOp,pos(0,0,Row,Col,0,0),TArg1,TArg2].

translate_function_call(Name,_Params,type([Type|_],Arity),pos(Col,Row),TCall) :- 
    % function calls like stakes/last
    % first and last from util/ordering are translated to min and max since we use sets of integer
    ordering_function_to_b(type([Type|_],Arity),Name,BFunctor) , 
    (BFunctor = max ; BFunctor = min) , ! , 
    % the types of ordering calls have been preprocessed in Kotlin Alloy2Prolog, i.e., 
    % n-ary types have been flattened to the ordered signature name
    Type = SignatureName , 
    translate_e_p(SignatureName,TSignatureName) , 
    TCall =.. [BFunctor,pos(0,0,Row,Col,0,0),TSignatureName].
translate_function_call(Name,Params,_Type,pos(Col,Row),TCall) :- 
    % utility functions from integer
    integer_function_to_b(Name,BOp) , ! , 
    maplist(translate_e_p,Params,TParams) , 
    TCall =.. [BOp,pos(0,0,Row,Col,0,0)|TParams].
translate_function_call(Name,Params,Type,pos(Col,Row),TCall) :- 
    % utility functions from ordering
    ordering_function_to_b(Type,Name,BOp) , ! , 
    maplist(translate_e_p,Params,TParams) , 
    TCall =.. [definition,pos(0,0,Row,Col,0,0),BOp,TParams].
translate_function_call(Name,Params,_Type,pos(Col,Row),TCall) :- 
    % predicate and function call
    maplist(translate_e_p,Params,TParams) , 
    maplist(strip_singleton_set,TParams,TTParams) , 
    TCall =.. [definition,pos(0,0,Row,Col,0,0),Name,TTParams].

strip_singleton_set(set_extension(_,[identifier(Pos,Name)]),identifier(Pos,Name)) :- !.
strip_singleton_set(A,A).

% Translation of the dot join operator has several special cases depending on the arity of the arguments.
translate_join(Pos,Arg1,Arg2,TBinaryJoin) :- 
    translate_e_p(Arg1,TArg1) , 
    translate_e_p(Arg2,TArg2) , 
    translate_join_aux(Pos,Arg1,Arg2,TArg1,TArg2,TBinaryJoin).

% univ._
translate_join_aux(Pos,identifier('univ_',type(['univ_'],1),_),_Arg2,_TArg1,TArg2,ran(Pos,TArg2)).
% _.univ
translate_join_aux(Pos,_Arg1,identifier('univ_',type(['univ_'],1),_),TArg1,_TArg2,dom(Pos,TArg1)).
% [first,last]._
translate_join_aux(_Pos,Arg1,_Arg2,TArg1,_TArg2,TArg1) :- 
    Arg1 = fun_call(OrderingFunction,_Params,Type,_Pos1) , 
    ordering_function_to_b(Type,OrderingFunction,_BFunction) , !.
% _.[next,nexts,prev,prevs]
translate_join_aux(_Pos,Arg1,Arg2,_TArg1,_TArg2,TJoin) :- 
    % set rhs as the parameter of lhs function call
    Arg2 = fun_call(OrderingFunction,_Params,Type,Pos1) , 
    ordering_function_to_b(Type,OrderingFunction,_BFunction) , ! , 
    translate_e_p(fun_call(OrderingFunction,[Arg1],Type,Pos1),TJoin).
% unary._
translate_join_aux(Pos,Arg1,_Arg2,TArg1,TArg2,function(Pos,TArg2,[TTArg1])) :- 
    % function call if rhs is a total function and lhs a unary relation
    clear_pos_from_untyped_b_ast(TArg2,TArg2Clean) , 
    is_unary_relation(Arg1) , is_total_function(TArg2Clean) , ! , 
    remove_singleton_set(TArg1,TTArg1).
translate_join_aux(Pos,Arg1,_Arg2,TArg1,TArg2,image(Pos,TArg2,TArg1)) :- 
    is_unary_relation(Arg1).
% binary.unary
translate_join_aux(Pos,Arg1,Arg2,TArg1,TArg2,function(Pos,reverse(Pos,TArg1),[TTArg2])) :- 
    % function call if lhs is a total function and rhs a unary relation
    clear_pos_from_untyped_b_ast(TArg1,TArg1Clean) , 
    is_binary_relation(Arg1) , is_unary_relation(Arg2) , is_total_function(TArg1Clean) , ! , 
    remove_singleton_set(TArg2,TTArg2).
translate_join_aux(Pos,Arg1,Arg2,TArg1,TArg2,image(Pos,reverse(Pos,TArg1),TArg2)) :- 
    is_binary_relation(Arg1) , is_unary_relation(Arg2).
% binary.binary
translate_join_aux(Pos,Arg1,Arg2,TArg1,TArg2,composition(Pos,TArg1,TArg2)) :-
    is_binary_relation(Arg1) , is_binary_relation(Arg2).
translate_join_aux(Pos,Arg1,Arg2,_TArg1,_TArg2,empty_set(Pos)) :- 
    format("~nJoin not supported this way:~nLeft: ~w~nRight: ~w~n",[Arg1,Arg2]).

% Unary quantified expressions: no, one, some, lone
translate_quantified_e(Pos,no,TArg,equal(Pos,TArg,empty_set(none))).
translate_quantified_e(Pos,one,TArg,equal(Pos,card(none,TArg),integer(none,1))).
translate_quantified_e(Pos,some,TArg,greater(Pos,card(none,TArg),integer(none,0))).
translate_quantified_e(Pos,lone,TArg,less_equal(Pos,card(none,TArg),integer(none,1))).

% Field declarations have several special cases depending on the keywords set, one, some or lone.
field_decl_special_cases(Pos,TSignatureName,DeclTerm,TFieldID,TField) :- 
    DeclTerm =.. [_,SetID|_] , 
    translate_e_p(SetID,TSetID) , 
    field_decl_special_cases_aux(Pos,DeclTerm,TSignatureName,TSetID,TFieldID,TField).

field_decl_special_cases_aux(Pos,setof(_,_,_),TSignatureName,TSetID,TFieldID,member(Pos,TFieldID,relations(none,TSignatureName,TSetID))).
field_decl_special_cases_aux(Pos,oneof(_,_,_),TSignatureName,TSetID,TFieldID,member(Pos,TFieldID,total_function(none,TSignatureName,TSetID))) :- 
    asserta(total_function(TFieldID)).
field_decl_special_cases_aux(Pos,loneof(_,_,_),TSignatureName,TSetID,TFieldID,member(Pos,TFieldID,partial_function(none,TSignatureName,TSetID))).
field_decl_special_cases_aux(Pos,Function,_TSignatureName,_TSetID,TFieldID,member(Pos,TFieldID,TFunction)) :- 
    translate_e_p(Function,TFunction).
field_decl_special_cases_aux(Pos,Term,_,_,_,_) :- 
    format("~nField declaration not implemented: ~w~nPosition ~w~n",[Term,Pos]).

% In function or predicate definitions one can use either set or one.
fun_or_pred_decl_special_cases(Pos,setof(Expr,_,_),TFieldID,subset(Pos,TFieldID,TExpr)) :- 
    translate_e_p(Expr,TExpr).
fun_or_pred_decl_special_cases(Pos,oneof(Expr,_,_),TFieldID,subset(Pos,TFieldID,TExpr)) :- 
    translate_e_p(Expr,TExpr).
fun_or_pred_decl_special_cases(Pos,identifier(Name,Type,PosID),TFieldID,subset(Pos,TFieldID,TSignatureName)) :- 
    translate_e_p(identifier(Name,Type,PosID),TSignatureName).
fun_or_pred_decl_special_cases(_,Term,_,_) :- 
    format("~nField declaration for function or predicate not implemented: ~w~n",[Term]).

% Most of the operators can be translated straightforwardly from Alloy to B.
alloy_to_b_operator(Op,BOp) :- 
    alloy_to_b_operator_aux(Op,BOp) , !.

alloy_to_b_operator_aux(in,subset).
alloy_to_b_operator_aux(plus,union).
alloy_to_b_operator_aux(not,negation).
alloy_to_b_operator_aux(or,disjunct).
alloy_to_b_operator_aux(and,conjunct).
alloy_to_b_operator_aux(inverse,reverse).
alloy_to_b_operator_aux(minus,set_subtraction).
alloy_to_b_operator_aux(closure1,reflexive_closure).
alloy_to_b_operator_aux(cartesian,cartesian_product).
alloy_to_b_operator_aux(function,expression_definition).
alloy_to_b_operator_aux(predicate,predicate_definition).
alloy_to_b_operator_aux(Op,Op).

%%% 
% Accumulate the translated machine parts and signature names during the translation and build the machine AST afterwards.
% We may need the signature names later on if translating a global scope.
build_machine_ast(b_machine(ListOfMachineParts,_SignatureNames),machine(generated(none,AbstractMachine))) :- 
    % filter empty machine parts
    findall(MachinePart,(member(MachinePart,ListOfMachineParts) , MachinePart =.. [_,_,L] , L \= []),TempListOfUsedMachineParts) , 
    % properties need to be conjoined
    select(properties(none,L),TempListOfUsedMachineParts,RestListOfUsedMachineParts) , 
    join_predicates(conjunct,L,FlatL) , 
    AbstractMachine =.. [abstract_machine,none,machine(none),machine_header(none,alloytranslation,[]),[properties(none,FlatL)|RestListOfUsedMachineParts]] , !.

empty_machine_acc(b_machine([sets(none,[]),constants(none,[]),definitions(none,[]),properties(none,[]),assertions(none,[]),operations(none,[])],[])).

extend_machine_acc(signatures,b_machine(MachineParts,SignatureNames),New,b_machine(MachineParts,NewSignatureNames)) :- 
    append(New,SignatureNames,NewSignatureNames) , !.
extend_machine_acc(Functor,b_machine(MachineParts,SignatureNames),New,b_machine([NewMachinePart|RestMachineParts],SignatureNames)) :- 
    MachinePart =.. [Functor,none,List] , 
    select(MachinePart,MachineParts,RestMachineParts) , 
    NewMachinePart =.. [Functor,none,[New|List]].

get_signature_names_from_machine_acc(b_machine(_MachineParts,SignatureNames),SignatureNames).

% Get the value of command_counter/1 as codes and assert the increased counter.
get_command_counter_atom(CommandCounterAtom) :- 
    command_counter(CommandCounter) , 
    number_codes(CommandCounter,CommandCounterCodes) , 
    atom_codes(CommandCounterAtom,CommandCounterCodes) , 
    retractall(command_counter(_)) , 
    NewCommandCounter is CommandCounter + 1 , 
    asserta(command_counter(NewCommandCounter)).
%%%

% Assert signatures for singleton checks.
assert_singleton_set(Options,Name) :- 
    member(one,Options) , ! , 
    assert_singleton_set(Name).
assert_singleton_set(_,_).

assert_singleton_set(Name) :- 
    (on_exception(_,singleton_set(Name),fail) ; 
     asserta(singleton_set(Name))).
assert_singleton_set(_).

% The singleton set annotation is local for a quantifier or parameters of a function definition.
retract_singleton_set(Name) :- 
    retractall(singleton_set(Name)).

retract_state :- 
    retractall(command_counter(_)) , 
    asserta(command_counter(0)) , 
    retractall(singleton_set(_)) , 
    retractall(ordered_signature(_)) , 
    retractall(total_function(_)).

is_singleton_set(IDName) :- 
    on_exception(_,singleton_set(IDName),fail).

is_total_function(IDName) :- 
    on_exception(_,total_function(IDName),fail).

is_ordered_signature(IDName) :- 
    on_exception(_,ordered_signature(IDName),fail).

is_unary_relation(AlloyTerm) :- 
    get_type_and_arity_from_alloy_term(AlloyTerm,_Type,Arity) , 
    Arity == 1.

is_binary_relation(AlloyTerm) :- 
    get_type_and_arity_from_alloy_term(AlloyTerm,_Type,Arity) , 
    Arity == 2.

% Only used for comparison checks between untyped ASTs. All output ASTs have to be ground.
clear_pos_from_untyped_b_ast(UntypedBAst,Clean) :- 
    UntypedBAst =.. [Functor,_Pos|Args] , 
    Clean =.. [Functor,_|Args].

remove_singleton_set(set_extension(_,[identifier(Pos,Name)]),identifier(Pos,Name)).
remove_singleton_set(AST,AST).

get_type_and_arity_from_alloy_term(AlloyTerm,Type,Arity) :- 
    AlloyTerm =.. [_|Args] , 
    member(type(Type,Arity),Args) , !.

atom_or_identifier_term(ID,ID) :- atom(ID).
atom_or_identifier_term(identifier(IDName,_,_),IDName).

% Join a list of untyped B ASTs using either conjunct/3 or disjunct/3.
join_predicates(Op,TArgList,TBinaryP) :- 
    alloy_to_b_operator(Op,BOp) , 
    (BOp = conjunct ; BOp = disjunct) , 
    join_predicates_aux(BOp,TArgList,TBinaryP).

join_predicates_aux(_,[],truth(none)).
join_predicates_aux(BOp,[H|T],Conjoined) :- 
    join_predicates_aux(BOp,T,H,Conjoined).

join_predicates_aux(_,[],Acc,Acc).
join_predicates_aux(BOp,[H|T],Acc,Conjoined) :- 
    NewAcc =.. [BOp,none,H,Acc] , 
    join_predicates_aux(BOp,T,NewAcc,Conjoined).

% Type safe atom concat.
atom_concat_safe(A,B,C) :- 
    atom(A) , atom(B) , 
    atom_concat(A,B,C).

:- begin_tests(full_machine_translation).

test(cards,[]) :- 
    AlloyTerm = alloy_model(facts([]),assertions([]),commands([]),functions([predicate('ThreeOfAKind_',[identifier('hand_',type(['Card_'],1),pos(20,3))],[field('hand_',setof(identifier('Card_',type(['Card_'],1),pos(5,1)),type(['Card_'],1),pos(26,3)),type(['Card_'],1),pos(20,3))],and([equal(card(join(identifier('hand_',type(['Card_'],1),pos(20,3)),identifier('suit_',type(['Card->this_''Suit_'],2),pos(11,1)),type(['Suit_'],1),pos(8,4)),type(['Int_'],1),pos(3,4)),integer(1,pos(14,4)),type([untyped],0),pos(13,4)), equal(card(identifier('hand_',type(['Card_'],1),pos(20,3)),type(['Int_'],1),pos(20,4)),integer(3,pos(26,4)),type([untyped],0),pos(25,4))],pos(16,4)),pos(1,3))]),signatures([signature('Card_',[field('suit_',oneof(identifier('Suit_',type(['Suit_'],1),pos(5,2)),type(['Suit_'],1),pos(16,1)),type(['Suit_'],1),pos(11,1))],[],[],pos(5,1)),signature('Suit_',[],[],[],pos(5,2))])) , 
    translate_model(AlloyTerm,Translated) , ! , 
    Translated == machine(generated(none,abstract_machine(none,machine(none),machine_header(none,alloytranslation,[]),[properties(none,member(pos(0,0,1,11,0,0),identifier(none,suit_),total_function(none,identifier(none,'Card_'),identifier(pos(0,0,2,5,0,0),'Suit_')))),definitions(none,[predicate_definition(pos(0,0,3,1,0,0),'ThreeOfAKind_',[identifier(pos(0,0,3,20,0,0),hand_)],conjunct(none,conjunct(none,equal(pos(0,0,4,25,0,0),card(pos(0,0,4,20,0,0),set_extension(pos(0,0,3,20,0,0),[identifier(none,hand_)])),integer(pos(0,0,4,26,0,0),3)),equal(pos(0,0,4,13,0,0),card(pos(0,0,4,3,0,0),function(pos(0,0,4,8,0,0),identifier(pos(0,0,1,11,0,0),suit_),[identifier(none,hand_)])),integer(pos(0,0,4,14,0,0),1))),subset(pos(0,0,3,20,0,0),set_extension(none,[identifier(none,hand_)]),identifier(pos(0,0,1,5,0,0),'Card_'))))]),sets(none,[deferred_set(pos(0,0,2,5,0,0),'Suit_'),deferred_set(pos(0,0,1,5,0,0),'Card_')]),constants(none,[identifier(none,suit_)])]))) , 
    load_model(AlloyTerm).

test(filesystem,[]) :- 
    AlloyTerm = alloy_model(facts([fact(all(['d_', 'o_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(15,11)),type(['Dir_'],1),pos(12,11)), field('o_',oneof(join(identifier('d_',type(['Dir_'],1),pos(12,11)),identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['FSObject_'],1),pos(24,11)),type(['FSObject_'],1),pos(23,11)),type(['FSObject_'],1),pos(20,11))],equal(join(identifier('o_',type(['FSObject_'],1),pos(20,11)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(16,2)),type(['Dir_'],1),pos(37,11)),identifier('d_',type(['Dir_'],1),pos(12,11)),type([untyped],0),pos(45,11)),type([untyped],0),pos(8,11)),(1,11)),fact(equal(plus(identifier('File_',type(['File_'],1),pos(5,8)),identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['File_', 'Dir_'],1),pos(13,14)),identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),type([untyped],0),pos(19,14)),(1,14)),fact(in(identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),join(identifier('Root_',type(['Root_'],1),pos(9,17)),closure(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['univ->univ_'],2),pos(25,20)),type(['univ_'],1),pos(24,20)),type([untyped],0),pos(17,20)),(1,20))]),assertions([fact(no(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(24,23)),type(['Dir_'],1),pos(21,23))],in(identifier('d_',type(['Dir_'],1),pos(21,23)),join(identifier('d_',type(['Dir_'],1),pos(21,23)),closure1(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['Dir->this_''FSObject_'],2),pos(37,23)),type(['FSObject_'],1),pos(36,23)),type([untyped],0),pos(32,23)),type([untyped],0),pos(18,23)),(1,23)),fact(one(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(25,29)),type(['Dir_'],1),pos(22,29))],no(join(identifier('d_',type(['Dir_'],1),pos(22,29)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(16,2)),type(['Dir_'],1),pos(35,29)),type([untyped],0),pos(31,29)),type([untyped],0),pos(18,29)),(1,29)),fact(all(['o_'],[field('o_',oneof(identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),type(['FSObject_'],1),pos(29,35)),type(['FSObject_'],1),pos(26,35))],lone(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(48,35)),type(['Dir_'],1),pos(45,35))],in(identifier('o_',type(['FSObject_'],1),pos(26,35)),join(identifier('d_',type(['Dir_'],1),pos(45,35)),identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['FSObject_'],1),pos(60,35)),type([untyped],0),pos(56,35)),type([untyped],0),pos(40,35)),type([untyped],0),pos(22,35)),(1,35))]),commands([check(and([all(['d_', 'o_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(15,11)),type(['Dir_'],1),pos(12,11)), field('o_',oneof(join(identifier('d_',type(['Dir_'],1),pos(12,11)),identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['FSObject_'],1),pos(24,11)),type(['FSObject_'],1),pos(23,11)),type(['FSObject_'],1),pos(20,11))],equal(join(identifier('o_',type(['FSObject_'],1),pos(20,11)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(16,2)),type(['Dir_'],1),pos(37,11)),identifier('d_',type(['Dir_'],1),pos(12,11)),type([untyped],0),pos(45,11)),type([untyped],0),pos(8,11)), equal(plus(identifier('File_',type(['File_'],1),pos(5,8)),identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['File_', 'Dir_'],1),pos(13,14)),identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),type([untyped],0),pos(19,14)), in(identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),join(identifier('Root_',type(['Root_'],1),pos(9,17)),closure(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['univ->univ_'],2),pos(25,20)),type(['univ_'],1),pos(24,20)),type([untyped],0),pos(17,20)), not(no(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(24,23)),type(['Dir_'],1),pos(21,23))],in(identifier('d_',type(['Dir_'],1),pos(21,23)),join(identifier('d_',type(['Dir_'],1),pos(21,23)),closure1(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['Dir->this_''FSObject_'],2),pos(37,23)),type(['FSObject_'],1),pos(36,23)),type([untyped],0),pos(32,23)),type([untyped],0),pos(18,23)),type([untyped],0),pos(1,23))],pos(1,11)),global_scope(5),exact_scopes([]),bitwidth(-1),pos(1,26)),check(and([all(['d_', 'o_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(15,11)),type(['Dir_'],1),pos(12,11)), field('o_',oneof(join(identifier('d_',type(['Dir_'],1),pos(12,11)),identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['FSObject_'],1),pos(24,11)),type(['FSObject_'],1),pos(23,11)),type(['FSObject_'],1),pos(20,11))],equal(join(identifier('o_',type(['FSObject_'],1),pos(20,11)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(16,2)),type(['Dir_'],1),pos(37,11)),identifier('d_',type(['Dir_'],1),pos(12,11)),type([untyped],0),pos(45,11)),type([untyped],0),pos(8,11)), equal(plus(identifier('File_',type(['File_'],1),pos(5,8)),identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['File_', 'Dir_'],1),pos(13,14)),identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),type([untyped],0),pos(19,14)), in(identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),join(identifier('Root_',type(['Root_'],1),pos(9,17)),closure(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['univ->univ_'],2),pos(25,20)),type(['univ_'],1),pos(24,20)),type([untyped],0),pos(17,20)), not(one(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(25,29)),type(['Dir_'],1),pos(22,29))],no(join(identifier('d_',type(['Dir_'],1),pos(22,29)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(16,2)),type(['Dir_'],1),pos(35,29)),type([untyped],0),pos(31,29)),type([untyped],0),pos(18,29)),type([untyped],0),pos(1,29))],pos(1,11)),global_scope(5),exact_scopes([]),bitwidth(-1),pos(1,32)),check(and([all(['d_', 'o_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(15,11)),type(['Dir_'],1),pos(12,11)), field('o_',oneof(join(identifier('d_',type(['Dir_'],1),pos(12,11)),identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['FSObject_'],1),pos(24,11)),type(['FSObject_'],1),pos(23,11)),type(['FSObject_'],1),pos(20,11))],equal(join(identifier('o_',type(['FSObject_'],1),pos(20,11)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(16,2)),type(['Dir_'],1),pos(37,11)),identifier('d_',type(['Dir_'],1),pos(12,11)),type([untyped],0),pos(45,11)),type([untyped],0),pos(8,11)), equal(plus(identifier('File_',type(['File_'],1),pos(5,8)),identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['File_', 'Dir_'],1),pos(13,14)),identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),type([untyped],0),pos(19,14)), in(identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),join(identifier('Root_',type(['Root_'],1),pos(9,17)),closure(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['univ->univ_'],2),pos(25,20)),type(['univ_'],1),pos(24,20)),type([untyped],0),pos(17,20)), not(all(['o_'],[field('o_',oneof(identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),type(['FSObject_'],1),pos(29,35)),type(['FSObject_'],1),pos(26,35))],lone(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(48,35)),type(['Dir_'],1),pos(45,35))],in(identifier('o_',type(['FSObject_'],1),pos(26,35)),join(identifier('d_',type(['Dir_'],1),pos(45,35)),identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(28,5)),type(['FSObject_'],1),pos(60,35)),type([untyped],0),pos(56,35)),type([untyped],0),pos(40,35)),type([untyped],0),pos(22,35)),type([untyped],0),pos(1,35))],pos(1,11)),global_scope(5),exact_scopes([]),bitwidth(-1),pos(1,38))]),functions([]),signatures([signature('FSObject_',[field('parent_',loneof(identifier('Dir_',type(['Dir_'],1),pos(5,5)),type(['Dir_'],1),pos(24,2)),type(['Dir_'],1),pos(16,2))],[],[],pos(5,2)),signature('Dir_',[field('contents_',setof(identifier('FSObject_',type(['FSObject_'],1),pos(5,2)),type(['FSObject_'],1),pos(38,5)),type(['FSObject_'],1),pos(28,5))],[],[subsig('FSObject_')],pos(5,5)),signature('File_',[],[],[subsig('FSObject_')],pos(5,8)),signature('Root_',[],[no(join(identifier('Root_',type(['Root_'],1),pos(9,17)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(16,2)),type(['Dir_'],1),pos(35,17)),type([untyped],0),pos(32,17))],[one, subsig('Dir_')],pos(9,17))])) , 
    translate_model(AlloyTerm,Translated) , ! , 
    Translated == machine(generated(none,abstract_machine(none,machine(none),machine_header(none,alloytranslation,[]),[properties(none,conjunct(none,member(pos(0,0,2,16,0,0),identifier(none,parent_),partial_function(none,identifier(none,'FSObject_'),identifier(pos(0,0,5,5,0,0),'Dir_'))),conjunct(none,subset(pos(0,0,5,5,0,0),identifier(none,'Dir_'),identifier(none,'FSObject_')),conjunct(none,member(pos(0,0,5,28,0,0),identifier(none,contents_),relations(none,identifier(none,'Dir_'),identifier(pos(0,0,2,5,0,0),'FSObject_'))),conjunct(none,subset(pos(0,0,8,5,0,0),identifier(none,'File_'),identifier(none,'FSObject_')),conjunct(none,subset(pos(0,0,17,9,0,0),set_extension(none,[identifier(none,'Root_')]),identifier(none,'Dir_')),conjunct(none,equal(pos(0,0,17,32,0,0),image(pos(0,0,17,35,0,0),identifier(pos(0,0,2,16,0,0),parent_),set_extension(pos(0,0,17,9,0,0),[identifier(none,'Root_')])),empty_set(none)),conjunct(none,negation(pos(0,0,23,18,0,0),exists(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,23,21,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,5,5,0,0),'Dir_')),subset(pos(0,0,23,32,0,0),set_extension(pos(0,0,23,21,0,0),[identifier(none,d_)]),image(pos(0,0,23,36,0,0),reflexive_closure(pos(0,0,23,37,0,0),identifier(pos(0,0,5,28,0,0),contents_)),set_extension(pos(0,0,23,21,0,0),[identifier(none,d_)])))))),conjunct(none,equal(pos(0,0,29,18,0,0),card(none,comprehension_set(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,29,22,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,5,5,0,0),'Dir_')),equal(pos(0,0,29,31,0,0),image(pos(0,0,29,35,0,0),identifier(pos(0,0,2,16,0,0),parent_),set_extension(pos(0,0,29,22,0,0),[identifier(none,d_)])),empty_set(none))))),integer(none,1)),conjunct(none,forall(pos(0,0,35,22,0,0),[identifier(none,o_)],implication(none,subset(pos(0,0,35,26,0,0),set_extension(none,[identifier(none,o_)]),identifier(pos(0,0,2,5,0,0),'FSObject_')),less_equal(pos(0,0,35,40,0,0),card(none,comprehension_set(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,35,45,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,5,5,0,0),'Dir_')),subset(pos(0,0,35,56,0,0),set_extension(pos(0,0,35,26,0,0),[identifier(none,o_)]),image(pos(0,0,35,60,0,0),identifier(pos(0,0,5,28,0,0),contents_),set_extension(pos(0,0,35,45,0,0),[identifier(none,d_)])))))),integer(none,1)))),conjunct(none,forall(pos(0,0,11,8,0,0),[identifier(none,d_),identifier(none,o_)],implication(none,conjunct(none,subset(pos(0,0,11,20,0,0),set_extension(none,[identifier(none,o_)]),image(pos(0,0,11,24,0,0),identifier(pos(0,0,5,28,0,0),contents_),set_extension(pos(0,0,11,12,0,0),[identifier(none,d_)]))),subset(pos(0,0,11,12,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,5,5,0,0),'Dir_'))),equal(pos(0,0,11,45,0,0),image(pos(0,0,11,37,0,0),identifier(pos(0,0,2,16,0,0),parent_),set_extension(pos(0,0,11,20,0,0),[identifier(none,o_)])),set_extension(pos(0,0,11,12,0,0),[identifier(none,d_)])))),conjunct(none,equal(pos(0,0,14,19,0,0),union(pos(0,0,14,13,0,0),identifier(pos(0,0,8,5,0,0),'File_'),identifier(pos(0,0,5,5,0,0),'Dir_')),identifier(pos(0,0,2,5,0,0),'FSObject_')),subset(pos(0,0,20,17,0,0),identifier(pos(0,0,2,5,0,0),'FSObject_'),image(pos(0,0,20,24,0,0),closure(pos(0,0,20,25,0,0),identifier(pos(0,0,5,28,0,0),contents_)),set_extension(pos(0,0,17,9,0,0),[identifier(none,'Root_')]))))))))))))))),operations(none,[operation(pos(0,0,38,1,0,0),identifier(pos(0,0,38,1,0,0),check2),[],[],precondition(pos(0,0,38,1,0,0),conjunct(pos(0,0,38,1,0,0),conjunct(none,truth(none),truth(none)),conjunct(none,negation(pos(0,0,35,1,0,0),forall(pos(0,0,35,22,0,0),[identifier(none,o_)],implication(none,subset(pos(0,0,35,26,0,0),set_extension(none,[identifier(none,o_)]),identifier(pos(0,0,2,5,0,0),'FSObject_')),less_equal(pos(0,0,35,40,0,0),card(none,comprehension_set(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,35,45,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,5,5,0,0),'Dir_')),subset(pos(0,0,35,56,0,0),set_extension(pos(0,0,35,26,0,0),[identifier(none,o_)]),image(pos(0,0,35,60,0,0),identifier(pos(0,0,5,28,0,0),contents_),set_extension(pos(0,0,35,45,0,0),[identifier(none,d_)])))))),integer(none,1))))),conjunct(none,subset(pos(0,0,20,17,0,0),identifier(pos(0,0,2,5,0,0),'FSObject_'),image(pos(0,0,20,24,0,0),closure(pos(0,0,20,25,0,0),identifier(pos(0,0,5,28,0,0),contents_)),set_extension(pos(0,0,17,9,0,0),[identifier(none,'Root_')]))),conjunct(none,equal(pos(0,0,14,19,0,0),union(pos(0,0,14,13,0,0),identifier(pos(0,0,8,5,0,0),'File_'),identifier(pos(0,0,5,5,0,0),'Dir_')),identifier(pos(0,0,2,5,0,0),'FSObject_')),forall(pos(0,0,11,8,0,0),[identifier(none,d_),identifier(none,o_)],implication(none,conjunct(none,subset(pos(0,0,11,20,0,0),set_extension(none,[identifier(none,o_)]),image(pos(0,0,11,24,0,0),identifier(pos(0,0,5,28,0,0),contents_),set_extension(pos(0,0,11,12,0,0),[identifier(none,d_)]))),subset(pos(0,0,11,12,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,5,5,0,0),'Dir_'))),equal(pos(0,0,11,45,0,0),image(pos(0,0,11,37,0,0),identifier(pos(0,0,2,16,0,0),parent_),set_extension(pos(0,0,11,20,0,0),[identifier(none,o_)])),set_extension(pos(0,0,11,12,0,0),[identifier(none,d_)])))))))),skip(none))),operation(pos(0,0,32,1,0,0),identifier(pos(0,0,32,1,0,0),check1),[],[],precondition(pos(0,0,32,1,0,0),conjunct(pos(0,0,32,1,0,0),conjunct(none,truth(none),truth(none)),conjunct(none,negation(pos(0,0,29,1,0,0),equal(pos(0,0,29,18,0,0),card(none,comprehension_set(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,29,22,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,5,5,0,0),'Dir_')),equal(pos(0,0,29,31,0,0),image(pos(0,0,29,35,0,0),identifier(pos(0,0,2,16,0,0),parent_),set_extension(pos(0,0,29,22,0,0),[identifier(none,d_)])),empty_set(none))))),integer(none,1))),conjunct(none,subset(pos(0,0,20,17,0,0),identifier(pos(0,0,2,5,0,0),'FSObject_'),image(pos(0,0,20,24,0,0),closure(pos(0,0,20,25,0,0),identifier(pos(0,0,5,28,0,0),contents_)),set_extension(pos(0,0,17,9,0,0),[identifier(none,'Root_')]))),conjunct(none,equal(pos(0,0,14,19,0,0),union(pos(0,0,14,13,0,0),identifier(pos(0,0,8,5,0,0),'File_'),identifier(pos(0,0,5,5,0,0),'Dir_')),identifier(pos(0,0,2,5,0,0),'FSObject_')),forall(pos(0,0,11,8,0,0),[identifier(none,d_),identifier(none,o_)],implication(none,conjunct(none,subset(pos(0,0,11,20,0,0),set_extension(none,[identifier(none,o_)]),image(pos(0,0,11,24,0,0),identifier(pos(0,0,5,28,0,0),contents_),set_extension(pos(0,0,11,12,0,0),[identifier(none,d_)]))),subset(pos(0,0,11,12,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,5,5,0,0),'Dir_'))),equal(pos(0,0,11,45,0,0),image(pos(0,0,11,37,0,0),identifier(pos(0,0,2,16,0,0),parent_),set_extension(pos(0,0,11,20,0,0),[identifier(none,o_)])),set_extension(pos(0,0,11,12,0,0),[identifier(none,d_)])))))))),skip(none))),operation(pos(0,0,26,1,0,0),identifier(pos(0,0,26,1,0,0),check0),[],[],precondition(pos(0,0,26,1,0,0),conjunct(pos(0,0,26,1,0,0),conjunct(none,truth(none),truth(none)),conjunct(none,negation(pos(0,0,23,1,0,0),negation(pos(0,0,23,18,0,0),exists(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,23,21,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,5,5,0,0),'Dir_')),subset(pos(0,0,23,32,0,0),set_extension(pos(0,0,23,21,0,0),[identifier(none,d_)]),image(pos(0,0,23,36,0,0),reflexive_closure(pos(0,0,23,37,0,0),identifier(pos(0,0,5,28,0,0),contents_)),set_extension(pos(0,0,23,21,0,0),[identifier(none,d_)]))))))),conjunct(none,subset(pos(0,0,20,17,0,0),identifier(pos(0,0,2,5,0,0),'FSObject_'),image(pos(0,0,20,24,0,0),closure(pos(0,0,20,25,0,0),identifier(pos(0,0,5,28,0,0),contents_)),set_extension(pos(0,0,17,9,0,0),[identifier(none,'Root_')]))),conjunct(none,equal(pos(0,0,14,19,0,0),union(pos(0,0,14,13,0,0),identifier(pos(0,0,8,5,0,0),'File_'),identifier(pos(0,0,5,5,0,0),'Dir_')),identifier(pos(0,0,2,5,0,0),'FSObject_')),forall(pos(0,0,11,8,0,0),[identifier(none,d_),identifier(none,o_)],implication(none,conjunct(none,subset(pos(0,0,11,20,0,0),set_extension(none,[identifier(none,o_)]),image(pos(0,0,11,24,0,0),identifier(pos(0,0,5,28,0,0),contents_),set_extension(pos(0,0,11,12,0,0),[identifier(none,d_)]))),subset(pos(0,0,11,12,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,5,5,0,0),'Dir_'))),equal(pos(0,0,11,45,0,0),image(pos(0,0,11,37,0,0),identifier(pos(0,0,2,16,0,0),parent_),set_extension(pos(0,0,11,20,0,0),[identifier(none,o_)])),set_extension(pos(0,0,11,12,0,0),[identifier(none,d_)])))))))),skip(none)))]),constants(none,[identifier(pos(0,0,17,9,0,0),'Root_'),identifier(pos(0,0,8,5,0,0),'File_'),identifier(none,contents_),identifier(pos(0,0,5,5,0,0),'Dir_'),identifier(none,parent_)]),sets(none,[deferred_set(pos(0,0,2,5,0,0),'FSObject_')])]))) , 
    load_model(AlloyTerm).

test(filesystem2,[]) :- 
    AlloyTerm = alloy_model(facts([fact(and([no(join(identifier('Root_',type(['Root_'],1),pos(9,13)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(5,4)),type(['Dir_'],1),pos(12,16)),type([untyped],0),pos(5,16)), equal(identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(5,4)),inverse(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['FSObject->this_''Dir_'],2),pos(14,17)),type([untyped],0),pos(12,17)), equal(plus(identifier('File_',type(['File_'],1),pos(5,11)),identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['File_', 'Dir_'],1),pos(10,18)),identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),type([untyped],0),pos(16,18)), in(identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),join(identifier('Root_',type(['Root_'],1),pos(9,13)),closure(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['univ->univ_'],2),pos(22,19)),type(['univ_'],1),pos(21,19)),type([untyped],0),pos(14,19))],pos(1,1)),(1,15))]),assertions([fact(no(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['Dir_'],1),pos(12,23)),type(['Dir_'],1),pos(8,23))],in(identifier('d_',type(['Dir_'],1),pos(8,23)),join(identifier('d_',type(['Dir_'],1),pos(8,23)),closure1(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['Dir->this_''FSObject_'],2),pos(25,23)),type(['FSObject_'],1),pos(24,23)),type([untyped],0),pos(20,23)),type([untyped],0),pos(5,23)),(1,22)),fact(one(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['Dir_'],1),pos(13,27)),type(['Dir_'],1),pos(9,27))],no(join(identifier('d_',type(['Dir_'],1),pos(9,27)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(5,4)),type(['Dir_'],1),pos(23,27)),type([untyped],0),pos(19,27)),type([untyped],0),pos(5,27)),(1,26)),fact(all(['f_'],[field('f_',oneof(identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),type(['FSObject_'],1),pos(13,31)),type(['FSObject_'],1),pos(9,31))],lone(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['Dir_'],1),pos(33,31)),type(['Dir_'],1),pos(29,31))],in(identifier('f_',type(['FSObject_'],1),pos(9,31)),join(identifier('d_',type(['Dir_'],1),pos(29,31)),identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['FSObject_'],1),pos(45,31)),type([untyped],0),pos(41,31)),type([untyped],0),pos(24,31)),type([untyped],0),pos(5,31)),(1,30))]),commands([check(and([no(join(identifier('Root_',type(['Root_'],1),pos(9,13)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(5,4)),type(['Dir_'],1),pos(12,16)),type([untyped],0),pos(5,16)), equal(identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(5,4)),inverse(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['FSObject->this_''Dir_'],2),pos(14,17)),type([untyped],0),pos(12,17)), equal(plus(identifier('File_',type(['File_'],1),pos(5,11)),identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['File_', 'Dir_'],1),pos(10,18)),identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),type([untyped],0),pos(16,18)), in(identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),join(identifier('Root_',type(['Root_'],1),pos(9,13)),closure(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['univ->univ_'],2),pos(22,19)),type(['univ_'],1),pos(21,19)),type([untyped],0),pos(14,19)), not(no(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['Dir_'],1),pos(12,23)),type(['Dir_'],1),pos(8,23))],in(identifier('d_',type(['Dir_'],1),pos(8,23)),join(identifier('d_',type(['Dir_'],1),pos(8,23)),closure1(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['Dir->this_''FSObject_'],2),pos(25,23)),type(['FSObject_'],1),pos(24,23)),type([untyped],0),pos(20,23)),type([untyped],0),pos(5,23)),type([untyped],0),pos(1,22))],pos(5,16)),global_scope(5),exact_scopes([]),bitwidth(-1),pos(1,34)),check(and([no(join(identifier('Root_',type(['Root_'],1),pos(9,13)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(5,4)),type(['Dir_'],1),pos(12,16)),type([untyped],0),pos(5,16)), equal(identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(5,4)),inverse(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['FSObject->this_''Dir_'],2),pos(14,17)),type([untyped],0),pos(12,17)), equal(plus(identifier('File_',type(['File_'],1),pos(5,11)),identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['File_', 'Dir_'],1),pos(10,18)),identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),type([untyped],0),pos(16,18)), in(identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),join(identifier('Root_',type(['Root_'],1),pos(9,13)),closure(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['univ->univ_'],2),pos(22,19)),type(['univ_'],1),pos(21,19)),type([untyped],0),pos(14,19)), not(one(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['Dir_'],1),pos(13,27)),type(['Dir_'],1),pos(9,27))],no(join(identifier('d_',type(['Dir_'],1),pos(9,27)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(5,4)),type(['Dir_'],1),pos(23,27)),type([untyped],0),pos(19,27)),type([untyped],0),pos(5,27)),type([untyped],0),pos(1,26))],pos(5,16)),global_scope(5),exact_scopes([]),bitwidth(-1),pos(1,36)),check(and([no(join(identifier('Root_',type(['Root_'],1),pos(9,13)),identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(5,4)),type(['Dir_'],1),pos(12,16)),type([untyped],0),pos(5,16)), equal(identifier('parent_',type(['FSObject->this_''Dir_'],2),pos(5,4)),inverse(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['FSObject->this_''Dir_'],2),pos(14,17)),type([untyped],0),pos(12,17)), equal(plus(identifier('File_',type(['File_'],1),pos(5,11)),identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['File_', 'Dir_'],1),pos(10,18)),identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),type([untyped],0),pos(16,18)), in(identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),join(identifier('Root_',type(['Root_'],1),pos(9,13)),closure(identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['univ->univ_'],2),pos(22,19)),type(['univ_'],1),pos(21,19)),type([untyped],0),pos(14,19)), not(all(['f_'],[field('f_',oneof(identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),type(['FSObject_'],1),pos(13,31)),type(['FSObject_'],1),pos(9,31))],lone(['d_'],[field('d_',oneof(identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['Dir_'],1),pos(33,31)),type(['Dir_'],1),pos(29,31))],in(identifier('f_',type(['FSObject_'],1),pos(9,31)),join(identifier('d_',type(['Dir_'],1),pos(29,31)),identifier('contents_',type(['Dir->this_''FSObject_'],2),pos(5,8)),type(['FSObject_'],1),pos(45,31)),type([untyped],0),pos(41,31)),type([untyped],0),pos(24,31)),type([untyped],0),pos(5,31)),type([untyped],0),pos(1,30))],pos(5,16)),global_scope(5),exact_scopes([]),bitwidth(-1),pos(1,38))]),functions([]),signatures([signature('FSObject_',[field('parent_',loneof(identifier('Dir_',type(['Dir_'],1),pos(5,7)),type(['Dir_'],1),pos(14,4)),type(['Dir_'],1),pos(5,4))],[],[],pos(5,3)),signature('Dir_',[field('contents_',setof(identifier('FSObject_',type(['FSObject_'],1),pos(5,3)),type(['FSObject_'],1),pos(16,8)),type(['FSObject_'],1),pos(5,8))],[],[subsig('FSObject_')],pos(5,7)),signature('File_',[],[],[subsig('FSObject_')],pos(5,11)),signature('Root_',[],[],[one, subsig('Dir_')],pos(9,13))])) , 
    translate_model(AlloyTerm,Translated) , ! , 
    Translated == machine(generated(none,abstract_machine(none,machine(none),machine_header(none,alloytranslation,[]),[properties(none,conjunct(none,member(pos(0,0,4,5,0,0),identifier(none,parent_),partial_function(none,identifier(none,'FSObject_'),identifier(pos(0,0,7,5,0,0),'Dir_'))),conjunct(none,subset(pos(0,0,7,5,0,0),identifier(none,'Dir_'),identifier(none,'FSObject_')),conjunct(none,member(pos(0,0,8,5,0,0),identifier(none,contents_),relations(none,identifier(none,'Dir_'),identifier(pos(0,0,3,5,0,0),'FSObject_'))),conjunct(none,subset(pos(0,0,11,5,0,0),identifier(none,'File_'),identifier(none,'FSObject_')),conjunct(none,subset(pos(0,0,13,9,0,0),set_extension(none,[identifier(none,'Root_')]),identifier(none,'Dir_')),conjunct(none,negation(pos(0,0,23,5,0,0),exists(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,23,8,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,7,5,0,0),'Dir_')),subset(pos(0,0,23,20,0,0),set_extension(pos(0,0,23,8,0,0),[identifier(none,d_)]),image(pos(0,0,23,24,0,0),reflexive_closure(pos(0,0,23,25,0,0),identifier(pos(0,0,8,5,0,0),contents_)),set_extension(pos(0,0,23,8,0,0),[identifier(none,d_)])))))),conjunct(none,equal(pos(0,0,27,5,0,0),card(none,comprehension_set(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,27,9,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,7,5,0,0),'Dir_')),equal(pos(0,0,27,19,0,0),image(pos(0,0,27,23,0,0),identifier(pos(0,0,4,5,0,0),parent_),set_extension(pos(0,0,27,9,0,0),[identifier(none,d_)])),empty_set(none))))),integer(none,1)),conjunct(none,forall(pos(0,0,31,5,0,0),[identifier(none,f_)],implication(none,subset(pos(0,0,31,9,0,0),set_extension(none,[identifier(none,f_)]),identifier(pos(0,0,3,5,0,0),'FSObject_')),less_equal(pos(0,0,31,24,0,0),card(none,comprehension_set(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,31,29,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,7,5,0,0),'Dir_')),subset(pos(0,0,31,41,0,0),set_extension(pos(0,0,31,9,0,0),[identifier(none,f_)]),image(pos(0,0,31,45,0,0),identifier(pos(0,0,8,5,0,0),contents_),set_extension(pos(0,0,31,29,0,0),[identifier(none,d_)])))))),integer(none,1)))),conjunct(none,subset(pos(0,0,19,14,0,0),identifier(pos(0,0,3,5,0,0),'FSObject_'),image(pos(0,0,19,21,0,0),closure(pos(0,0,19,22,0,0),identifier(pos(0,0,8,5,0,0),contents_)),set_extension(pos(0,0,13,9,0,0),[identifier(none,'Root_')]))),conjunct(none,equal(pos(0,0,18,16,0,0),union(pos(0,0,18,10,0,0),identifier(pos(0,0,11,5,0,0),'File_'),identifier(pos(0,0,7,5,0,0),'Dir_')),identifier(pos(0,0,3,5,0,0),'FSObject_')),conjunct(none,equal(pos(0,0,17,12,0,0),identifier(pos(0,0,4,5,0,0),parent_),reverse(pos(0,0,17,14,0,0),identifier(pos(0,0,8,5,0,0),contents_))),equal(pos(0,0,16,5,0,0),image(pos(0,0,16,12,0,0),identifier(pos(0,0,4,5,0,0),parent_),set_extension(pos(0,0,13,9,0,0),[identifier(none,'Root_')])),empty_set(none)))))))))))))),operations(none,[operation(pos(0,0,38,1,0,0),identifier(pos(0,0,38,1,0,0),check2),[],[],precondition(pos(0,0,38,1,0,0),conjunct(pos(0,0,38,1,0,0),conjunct(none,truth(none),truth(none)),conjunct(none,negation(pos(0,0,30,1,0,0),forall(pos(0,0,31,5,0,0),[identifier(none,f_)],implication(none,subset(pos(0,0,31,9,0,0),set_extension(none,[identifier(none,f_)]),identifier(pos(0,0,3,5,0,0),'FSObject_')),less_equal(pos(0,0,31,24,0,0),card(none,comprehension_set(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,31,29,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,7,5,0,0),'Dir_')),subset(pos(0,0,31,41,0,0),set_extension(pos(0,0,31,9,0,0),[identifier(none,f_)]),image(pos(0,0,31,45,0,0),identifier(pos(0,0,8,5,0,0),contents_),set_extension(pos(0,0,31,29,0,0),[identifier(none,d_)])))))),integer(none,1))))),conjunct(none,subset(pos(0,0,19,14,0,0),identifier(pos(0,0,3,5,0,0),'FSObject_'),image(pos(0,0,19,21,0,0),closure(pos(0,0,19,22,0,0),identifier(pos(0,0,8,5,0,0),contents_)),set_extension(pos(0,0,13,9,0,0),[identifier(none,'Root_')]))),conjunct(none,equal(pos(0,0,18,16,0,0),union(pos(0,0,18,10,0,0),identifier(pos(0,0,11,5,0,0),'File_'),identifier(pos(0,0,7,5,0,0),'Dir_')),identifier(pos(0,0,3,5,0,0),'FSObject_')),conjunct(none,equal(pos(0,0,17,12,0,0),identifier(pos(0,0,4,5,0,0),parent_),reverse(pos(0,0,17,14,0,0),identifier(pos(0,0,8,5,0,0),contents_))),equal(pos(0,0,16,5,0,0),image(pos(0,0,16,12,0,0),identifier(pos(0,0,4,5,0,0),parent_),set_extension(pos(0,0,13,9,0,0),[identifier(none,'Root_')])),empty_set(none))))))),skip(none))),operation(pos(0,0,36,1,0,0),identifier(pos(0,0,36,1,0,0),check1),[],[],precondition(pos(0,0,36,1,0,0),conjunct(pos(0,0,36,1,0,0),conjunct(none,truth(none),truth(none)),conjunct(none,negation(pos(0,0,26,1,0,0),equal(pos(0,0,27,5,0,0),card(none,comprehension_set(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,27,9,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,7,5,0,0),'Dir_')),equal(pos(0,0,27,19,0,0),image(pos(0,0,27,23,0,0),identifier(pos(0,0,4,5,0,0),parent_),set_extension(pos(0,0,27,9,0,0),[identifier(none,d_)])),empty_set(none))))),integer(none,1))),conjunct(none,subset(pos(0,0,19,14,0,0),identifier(pos(0,0,3,5,0,0),'FSObject_'),image(pos(0,0,19,21,0,0),closure(pos(0,0,19,22,0,0),identifier(pos(0,0,8,5,0,0),contents_)),set_extension(pos(0,0,13,9,0,0),[identifier(none,'Root_')]))),conjunct(none,equal(pos(0,0,18,16,0,0),union(pos(0,0,18,10,0,0),identifier(pos(0,0,11,5,0,0),'File_'),identifier(pos(0,0,7,5,0,0),'Dir_')),identifier(pos(0,0,3,5,0,0),'FSObject_')),conjunct(none,equal(pos(0,0,17,12,0,0),identifier(pos(0,0,4,5,0,0),parent_),reverse(pos(0,0,17,14,0,0),identifier(pos(0,0,8,5,0,0),contents_))),equal(pos(0,0,16,5,0,0),image(pos(0,0,16,12,0,0),identifier(pos(0,0,4,5,0,0),parent_),set_extension(pos(0,0,13,9,0,0),[identifier(none,'Root_')])),empty_set(none))))))),skip(none))),operation(pos(0,0,34,1,0,0),identifier(pos(0,0,34,1,0,0),check0),[],[],precondition(pos(0,0,34,1,0,0),conjunct(pos(0,0,34,1,0,0),conjunct(none,truth(none),truth(none)),conjunct(none,negation(pos(0,0,22,1,0,0),negation(pos(0,0,23,5,0,0),exists(none,[identifier(none,d_)],conjunct(none,subset(pos(0,0,23,8,0,0),set_extension(none,[identifier(none,d_)]),identifier(pos(0,0,7,5,0,0),'Dir_')),subset(pos(0,0,23,20,0,0),set_extension(pos(0,0,23,8,0,0),[identifier(none,d_)]),image(pos(0,0,23,24,0,0),reflexive_closure(pos(0,0,23,25,0,0),identifier(pos(0,0,8,5,0,0),contents_)),set_extension(pos(0,0,23,8,0,0),[identifier(none,d_)]))))))),conjunct(none,subset(pos(0,0,19,14,0,0),identifier(pos(0,0,3,5,0,0),'FSObject_'),image(pos(0,0,19,21,0,0),closure(pos(0,0,19,22,0,0),identifier(pos(0,0,8,5,0,0),contents_)),set_extension(pos(0,0,13,9,0,0),[identifier(none,'Root_')]))),conjunct(none,equal(pos(0,0,18,16,0,0),union(pos(0,0,18,10,0,0),identifier(pos(0,0,11,5,0,0),'File_'),identifier(pos(0,0,7,5,0,0),'Dir_')),identifier(pos(0,0,3,5,0,0),'FSObject_')),conjunct(none,equal(pos(0,0,17,12,0,0),identifier(pos(0,0,4,5,0,0),parent_),reverse(pos(0,0,17,14,0,0),identifier(pos(0,0,8,5,0,0),contents_))),equal(pos(0,0,16,5,0,0),image(pos(0,0,16,12,0,0),identifier(pos(0,0,4,5,0,0),parent_),set_extension(pos(0,0,13,9,0,0),[identifier(none,'Root_')])),empty_set(none))))))),skip(none)))]),constants(none,[identifier(pos(0,0,13,9,0,0),'Root_'),identifier(pos(0,0,11,5,0,0),'File_'),identifier(none,contents_),identifier(pos(0,0,7,5,0,0),'Dir_'),identifier(none,parent_)]),sets(none,[deferred_set(pos(0,0,3,5,0,0),'FSObject_')])]))) , 
    load_model(AlloyTerm).

test(self_grandpas,[]) :- 
    AlloyTerm = alloy_model(facts([fact(no(['p_'],[field('p_',oneof(identifier('Person_',type(['Person_'],1),pos(14,3)),type(['Person_'],1),pos(12,17)),type(['Person_'],1),pos(8,17))],in(identifier('p_',type(['Person_'],1),pos(8,17)),join(identifier('p_',type(['Person_'],1),pos(8,17)),closure1(plus(identifier('mother_',type(['Person->this_''Woman_'],2),pos(5,5)),identifier('father_',type(['Person->this_''Man_'],2),pos(5,4)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(37,17)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(28,17)),type(['Woman_', 'Man_'],1),pos(27,17)),type([untyped],0),pos(23,17)),type([untyped],0),pos(5,17)),(1,16)),fact(equal(identifier('wife_',type(['Man->this_''Woman_'],2),pos(5,9)),inverse(identifier('husband_',type(['Woman->this_''Man_'],2),pos(5,13)),type(['Man->this_''Woman_'],2),pos(12,21)),type([untyped],0),pos(10,21)),(1,20)),fact(and([no(intersection(identifier('wife_',type(['Man->this_''Woman_'],2),pos(5,9)),join(closure(plus(identifier('mother_',type(['Person->this_''Woman_'],2),pos(5,5)),identifier('father_',type(['Person->this_''Man_'],2),pos(5,4)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(24,25)),type(['univ->univ_'],2),pos(15,25)),identifier('mother_',type(['Person->this_''Woman_'],2),pos(5,5)),type(['univ->this_''Woman_'],2),pos(33,25)),type(['Man->this_''Woman_'],2),pos(13,25)),type([untyped],0),pos(5,25)), no(intersection(identifier('husband_',type(['Woman->this_''Man_'],2),pos(5,13)),join(closure(plus(identifier('mother_',type(['Person->this_''Woman_'],2),pos(5,5)),identifier('father_',type(['Person->this_''Man_'],2),pos(5,4)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(27,26)),type(['univ->univ_'],2),pos(18,26)),identifier('father_',type(['Person->this_''Man_'],2),pos(5,4)),type(['univ->this_''Man_'],2),pos(36,26)),type(['Woman->this_''Man_'],2),pos(16,26)),type([untyped],0),pos(5,26))],pos(1,1)),(1,24))]),assertions([]),commands([run(and([no(['p_'],[field('p_',oneof(identifier('Person_',type(['Person_'],1),pos(14,3)),type(['Person_'],1),pos(12,17)),type(['Person_'],1),pos(8,17))],in(identifier('p_',type(['Person_'],1),pos(8,17)),join(identifier('p_',type(['Person_'],1),pos(8,17)),closure1(plus(identifier('mother_',type(['Person->this_''Woman_'],2),pos(5,5)),identifier('father_',type(['Person->this_''Man_'],2),pos(5,4)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(37,17)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(28,17)),type(['Woman_', 'Man_'],1),pos(27,17)),type([untyped],0),pos(23,17)),type([untyped],0),pos(5,17)), equal(identifier('wife_',type(['Man->this_''Woman_'],2),pos(5,9)),inverse(identifier('husband_',type(['Woman->this_''Man_'],2),pos(5,13)),type(['Man->this_''Woman_'],2),pos(12,21)),type([untyped],0),pos(10,21)), no(intersection(identifier('wife_',type(['Man->this_''Woman_'],2),pos(5,9)),join(closure(plus(identifier('mother_',type(['Person->this_''Woman_'],2),pos(5,5)),identifier('father_',type(['Person->this_''Man_'],2),pos(5,4)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(24,25)),type(['univ->univ_'],2),pos(15,25)),identifier('mother_',type(['Person->this_''Woman_'],2),pos(5,5)),type(['univ->this_''Woman_'],2),pos(33,25)),type(['Man->this_''Woman_'],2),pos(13,25)),type([untyped],0),pos(5,25)), no(intersection(identifier('husband_',type(['Woman->this_''Man_'],2),pos(5,13)),join(closure(plus(identifier('mother_',type(['Person->this_''Woman_'],2),pos(5,5)),identifier('father_',type(['Person->this_''Man_'],2),pos(5,4)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(27,26)),type(['univ->univ_'],2),pos(18,26)),identifier('father_',type(['Person->this_''Man_'],2),pos(5,4)),type(['univ->this_''Man_'],2),pos(36,26)),type(['Woman->this_''Man_'],2),pos(16,26)),type([untyped],0),pos(5,26)), some(['m_'],[field('m_',identifier('Man_',type(['Man_'],1),pos(5,8)),type(['Man_'],1),pos(17,34))],in(identifier('m_',type(['Man_'],1),pos(17,34)),fun_call('grandpas_',[identifier('m_',type(['Man_'],1),pos(17,34))],type(['Person_'],1),pos(10,35)),type([untyped],0),pos(7,35)),type([untyped],0),pos(1,1))],pos(1,16)),global_scope(-1),exact_scopes([]),bitwidth(-1),pos(1,38))]),functions([function('grandpas_',[identifier('p_',type(['Person_'],1),pos(14,29))],[field('p_',identifier('Person_',type(['Person_'],1),pos(14,3)),type(['Person_'],1),pos(14,29))],let('parent_',plus(plus(plus(identifier('mother_',type(['Person->this_''Woman_'],2),pos(5,5)),identifier('father_',type(['Person->this_''Man_'],2),pos(5,4)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(25,30)),join(identifier('father_',type(['Person->this_''Man_'],2),pos(5,4)),identifier('wife_',type(['Man->this_''Woman_'],2),pos(5,9)),type(['Person->this_''Woman_'],2),pos(42,30)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(34,30)),join(identifier('mother_',type(['Person->this_''Woman_'],2),pos(5,5)),identifier('husband_',type(['Woman->this_''Man_'],2),pos(5,13)),type(['Person->this_''Man_'],2),pos(56,30)),type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(48,30)),intersection(join(join(identifier('p_',type(['Person_'],1),pos(14,29)),identifier('parent_',type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(9,30)),type(['Woman_', 'Man_'],1),pos(8,31)),identifier('parent_',type(['Person->this_''Woman_', 'Person->this_''Man_'],2),pos(9,30)),type(['Woman_', 'Man_'],1),pos(15,31)),identifier('Man_',type(['Man_'],1),pos(5,8)),type(['Man_'],1),pos(23,31)),type(['Man_'],1),pos(16,30)),pos(1,29)),predicate('ownGrandpa_',[identifier('m_',type(['Man_'],1),pos(17,34))],[field('m_',identifier('Man_',type(['Man_'],1),pos(5,8)),type(['Man_'],1),pos(17,34))],in(identifier('m_',type(['Man_'],1),pos(17,34)),fun_call('grandpas_',[identifier('m_',type(['Man_'],1),pos(17,34))],type(['Person_'],1),pos(10,35)),type([untyped],0),pos(7,35)),pos(1,34))]),signatures([signature('Person_',[field('father_',loneof(identifier('Man_',type(['Man_'],1),pos(5,8)),type(['Man_'],1),pos(14,4)),type(['Man_'],1),pos(5,4)),field('mother_',loneof(identifier('Woman_',type(['Woman_'],1),pos(5,12)),type(['Woman_'],1),pos(14,5)),type(['Woman_'],1),pos(5,5))],[],[abstract],pos(14,3)),signature('Man_',[field('wife_',loneof(identifier('Woman_',type(['Woman_'],1),pos(5,12)),type(['Woman_'],1),pos(12,9)),type(['Woman_'],1),pos(5,9))],[],[subsig('Person_')],pos(5,8)),signature('Woman_',[field('husband_',loneof(identifier('Man_',type(['Man_'],1),pos(5,8)),type(['Man_'],1),pos(15,13)),type(['Man_'],1),pos(5,13))],[],[subsig('Person_')],pos(5,12))])) , 
    translate_model(AlloyTerm,Translated) , ! , 
    Translated == machine(generated(none,abstract_machine(none,machine(none),machine_header(none,alloytranslation,[]),[properties(none,conjunct(none,member(pos(0,0,4,5,0,0),identifier(none,father_),partial_function(none,identifier(none,'Person_'),identifier(pos(0,0,8,5,0,0),'Man_'))),conjunct(none,member(pos(0,0,5,5,0,0),identifier(none,mother_),partial_function(none,identifier(none,'Person_'),identifier(pos(0,0,12,5,0,0),'Woman_'))),conjunct(none,subset(pos(0,0,8,5,0,0),identifier(none,'Man_'),identifier(none,'Person_')),conjunct(none,member(pos(0,0,9,5,0,0),identifier(none,wife_),partial_function(none,identifier(none,'Man_'),identifier(pos(0,0,12,5,0,0),'Woman_'))),conjunct(none,subset(pos(0,0,12,5,0,0),identifier(none,'Woman_'),identifier(none,'Person_')),conjunct(none,member(pos(0,0,13,5,0,0),identifier(none,husband_),partial_function(none,identifier(none,'Woman_'),identifier(pos(0,0,8,5,0,0),'Man_'))),conjunct(none,negation(pos(0,0,17,5,0,0),exists(none,[identifier(none,p_)],conjunct(none,subset(pos(0,0,17,8,0,0),set_extension(none,[identifier(none,p_)]),identifier(pos(0,0,3,14,0,0),'Person_')),subset(pos(0,0,17,23,0,0),set_extension(pos(0,0,17,8,0,0),[identifier(none,p_)]),image(pos(0,0,17,27,0,0),reflexive_closure(pos(0,0,17,28,0,0),union(pos(0,0,17,37,0,0),identifier(pos(0,0,5,5,0,0),mother_),identifier(pos(0,0,4,5,0,0),father_))),set_extension(pos(0,0,17,8,0,0),[identifier(none,p_)])))))),conjunct(none,equal(pos(0,0,21,10,0,0),identifier(pos(0,0,9,5,0,0),wife_),reverse(pos(0,0,21,12,0,0),identifier(pos(0,0,13,5,0,0),husband_))),conjunct(none,equal(pos(0,0,26,5,0,0),intersection(pos(0,0,26,16,0,0),identifier(pos(0,0,13,5,0,0),husband_),composition(pos(0,0,26,36,0,0),closure(pos(0,0,26,18,0,0),union(pos(0,0,26,27,0,0),identifier(pos(0,0,5,5,0,0),mother_),identifier(pos(0,0,4,5,0,0),father_))),identifier(pos(0,0,4,5,0,0),father_))),empty_set(none)),equal(pos(0,0,25,5,0,0),intersection(pos(0,0,25,13,0,0),identifier(pos(0,0,9,5,0,0),wife_),composition(pos(0,0,25,33,0,0),closure(pos(0,0,25,15,0,0),union(pos(0,0,25,24,0,0),identifier(pos(0,0,5,5,0,0),mother_),identifier(pos(0,0,4,5,0,0),father_))),identifier(pos(0,0,5,5,0,0),mother_))),empty_set(none)))))))))))),definitions(none,[predicate_definition(pos(0,0,34,1,0,0),ownGrandpa_,[identifier(pos(0,0,34,17,0,0),m_)],conjunct(none,subset(pos(0,0,35,7,0,0),set_extension(pos(0,0,34,17,0,0),[identifier(none,m_)]),definition(pos(0,0,35,10,0,0),grandpas_,[identifier(none,m_)])),subset(pos(0,0,34,17,0,0),set_extension(none,[identifier(none,m_)]),identifier(pos(0,0,8,5,0,0),'Man_')))),expression_definition(pos(0,0,29,1,0,0),grandpas_,[identifier(pos(0,0,29,14,0,0),p_)],comprehension_set(pos(0,0,29,1,0,0),[identifier(none,temp)],conjunct(none,member(none,identifier(none,temp),let_expression(pos(0,0,30,16,0,0),[identifier(none,parent_)],equal(pos(0,0,30,16,0,0),identifier(none,parent_),union(pos(0,0,30,48,0,0),union(pos(0,0,30,34,0,0),union(pos(0,0,30,25,0,0),identifier(pos(0,0,5,5,0,0),mother_),identifier(pos(0,0,4,5,0,0),father_)),composition(pos(0,0,30,42,0,0),identifier(pos(0,0,4,5,0,0),father_),identifier(pos(0,0,9,5,0,0),wife_))),composition(pos(0,0,30,56,0,0),identifier(pos(0,0,5,5,0,0),mother_),identifier(pos(0,0,13,5,0,0),husband_)))),intersection(pos(0,0,31,23,0,0),image(pos(0,0,31,15,0,0),identifier(pos(0,0,30,9,0,0),parent_),image(pos(0,0,31,8,0,0),identifier(pos(0,0,30,9,0,0),parent_),set_extension(pos(0,0,29,14,0,0),[identifier(none,p_)]))),identifier(pos(0,0,8,5,0,0),'Man_')))),subset(pos(0,0,29,14,0,0),set_extension(none,[identifier(none,p_)]),identifier(pos(0,0,3,14,0,0),'Person_')))))]),operations(none,[operation(pos(0,0,38,1,0,0),identifier(pos(0,0,38,1,0,0),run0),[],[],precondition(pos(0,0,38,1,0,0),conjunct(pos(0,0,38,1,0,0),conjunct(none,truth(none),truth(none)),conjunct(none,exists(pos(0,0,1,1,0,0),[identifier(none,m_)],conjunct(none,subset(pos(0,0,34,17,0,0),set_extension(none,[identifier(none,m_)]),identifier(pos(0,0,8,5,0,0),'Man_')),subset(pos(0,0,35,7,0,0),set_extension(pos(0,0,34,17,0,0),[identifier(none,m_)]),definition(pos(0,0,35,10,0,0),grandpas_,[identifier(none,m_)])))),conjunct(none,equal(pos(0,0,26,5,0,0),intersection(pos(0,0,26,16,0,0),identifier(pos(0,0,13,5,0,0),husband_),composition(pos(0,0,26,36,0,0),closure(pos(0,0,26,18,0,0),union(pos(0,0,26,27,0,0),identifier(pos(0,0,5,5,0,0),mother_),identifier(pos(0,0,4,5,0,0),father_))),identifier(pos(0,0,4,5,0,0),father_))),empty_set(none)),conjunct(none,equal(pos(0,0,25,5,0,0),intersection(pos(0,0,25,13,0,0),identifier(pos(0,0,9,5,0,0),wife_),composition(pos(0,0,25,33,0,0),closure(pos(0,0,25,15,0,0),union(pos(0,0,25,24,0,0),identifier(pos(0,0,5,5,0,0),mother_),identifier(pos(0,0,4,5,0,0),father_))),identifier(pos(0,0,5,5,0,0),mother_))),empty_set(none)),conjunct(none,equal(pos(0,0,21,10,0,0),identifier(pos(0,0,9,5,0,0),wife_),reverse(pos(0,0,21,12,0,0),identifier(pos(0,0,13,5,0,0),husband_))),negation(pos(0,0,17,5,0,0),exists(none,[identifier(none,p_)],conjunct(none,subset(pos(0,0,17,8,0,0),set_extension(none,[identifier(none,p_)]),identifier(pos(0,0,3,14,0,0),'Person_')),subset(pos(0,0,17,23,0,0),set_extension(pos(0,0,17,8,0,0),[identifier(none,p_)]),image(pos(0,0,17,27,0,0),reflexive_closure(pos(0,0,17,28,0,0),union(pos(0,0,17,37,0,0),identifier(pos(0,0,5,5,0,0),mother_),identifier(pos(0,0,4,5,0,0),father_))),set_extension(pos(0,0,17,8,0,0),[identifier(none,p_)]))))))))))),skip(none)))]),constants(none,[identifier(none,husband_),identifier(pos(0,0,12,5,0,0),'Woman_'),identifier(none,wife_),identifier(pos(0,0,8,5,0,0),'Man_'),identifier(none,mother_),identifier(none,father_)]),sets(none,[deferred_set(pos(0,0,3,14,0,0),'Person_')])]))) , 
    load_model(AlloyTerm).

test(queens2,[]) :- 
    AlloyTerm = alloy_model(facts([fact(all(['q_', 'q__'],[field('q_',oneof(identifier('queen_',type(['queen_'],1),pos(5,7)),type(['queen_'],1),pos(14,17)),type(['queen_'],1),pos(12,17)), field('q__',oneof(minus(identifier('queen_',type(['queen_'],1),pos(5,7)),identifier('q_',type(['queen_'],1),pos(12,17)),type(['queen_'],1),pos(30,17)),type(['queen_'],1),pos(25,17)),type(['queen_'],1),pos(21,17))],and([not(equal(join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(8,18)),join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(15,18)),type([untyped],0),pos(11,18)),type([untyped],0),pos(5,18)), not(equal(join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(8,19)),join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(15,19)),type([untyped],0),pos(11,19)),type([untyped],0),pos(5,19)), not(equal(fun_call('integer_''plus_',[join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(13,20)), join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(17,20))],type(['Int_'],1),pos(7,20)),fun_call('integer_''plus_',[join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(30,20)), join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(35,20))],type(['Int_'],1),pos(23,20)),type([untyped],0),pos(21,20)),type([untyped],0),pos(5,20)), not(equal(fun_call('integer_''plus_',[join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('x__',type(['queen->Int_'],2),pos(20,7)),type(['Int_'],1),pos(13,21)), join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(18,21))],type(['Int_'],1),pos(7,21)),fun_call('integer_''plus_',[join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('x__',type(['queen->Int_'],2),pos(20,7)),type(['Int_'],1),pos(31,21)), join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(37,21))],type(['Int_'],1),pos(24,21)),type([untyped],0),pos(22,21)),type([untyped],0),pos(5,21))],pos(1,1)),type([untyped],0),pos(8,17)),(1,17))]),assertions([]),commands([run(and([all(['q_', 'q__'],[field('q_',oneof(identifier('queen_',type(['queen_'],1),pos(5,7)),type(['queen_'],1),pos(14,17)),type(['queen_'],1),pos(12,17)), field('q__',oneof(minus(identifier('queen_',type(['queen_'],1),pos(5,7)),identifier('q_',type(['queen_'],1),pos(12,17)),type(['queen_'],1),pos(30,17)),type(['queen_'],1),pos(25,17)),type(['queen_'],1),pos(21,17))],and([not(equal(join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(8,18)),join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(15,18)),type([untyped],0),pos(11,18)),type([untyped],0),pos(5,18)), not(equal(join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(8,19)),join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(15,19)),type([untyped],0),pos(11,19)),type([untyped],0),pos(5,19)), not(equal(fun_call('integer_''plus_',[join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(13,20)), join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(17,20))],type(['Int_'],1),pos(7,20)),fun_call('integer_''plus_',[join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(30,20)), join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(35,20))],type(['Int_'],1),pos(23,20)),type([untyped],0),pos(21,20)),type([untyped],0),pos(5,20)), not(equal(fun_call('integer_''plus_',[join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('x__',type(['queen->Int_'],2),pos(20,7)),type(['Int_'],1),pos(13,21)), join(identifier('q_',type(['queen_'],1),pos(12,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(18,21))],type(['Int_'],1),pos(7,21)),fun_call('integer_''plus_',[join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('x__',type(['queen->Int_'],2),pos(20,7)),type(['Int_'],1),pos(31,21)), join(identifier('q__',type(['queen_'],1),pos(21,17)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(37,21))],type(['Int_'],1),pos(24,21)),type([untyped],0),pos(22,21)),type([untyped],0),pos(5,21))],pos(1,1)),type([untyped],0),pos(8,17))],pos(1,17)),global_scope(-1),exact_scopes([('queen_',4)]),bitwidth(5),pos(1,26))]),functions([predicate('show_',[],[],boolean(true,pos(11,24)),pos(1,24))]),signatures([signature('queen_',[field('x_',oneof(identifier('Int_',type(['Int_'],1),pos(1,1)),type(['Int_'],1),pos(15,7)),type(['Int_'],1),pos(13,7)),field('x__',oneof(identifier('Int_',type(['Int_'],1),pos(1,1)),type(['Int_'],1),pos(23,7)),type(['Int_'],1),pos(20,7)),field('y_',oneof(identifier('Int_',type(['Int_'],1),pos(1,1)),type(['Int_'],1),pos(30,7)),type(['Int_'],1),pos(28,7))],[and([greater_equal(join(identifier(this,type(['queen_'],1),pos(1,1)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(5,8)),integer(1,pos(10,8)),type([untyped],0),pos(7,8)), greater_equal(join(identifier(this,type(['queen_'],1),pos(1,1)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(5,9)),integer(1,pos(10,9)),type([untyped],0),pos(7,9)), less_equal(join(identifier(this,type(['queen_'],1),pos(1,1)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(5,10)),card(identifier('queen_',type(['queen_'],1),pos(5,7)),type(['Int_'],1),pos(10,10)),type([untyped],0),pos(7,10)), less_equal(join(identifier(this,type(['queen_'],1),pos(1,1)),identifier('y_',type(['queen->Int_'],2),pos(28,7)),type(['Int_'],1),pos(5,11)),card(identifier('queen_',type(['queen_'],1),pos(5,7)),type(['Int_'],1),pos(10,11)),type([untyped],0),pos(7,11)), greater_equal(join(identifier(this,type(['queen_'],1),pos(1,1)),identifier('x__',type(['queen->Int_'],2),pos(20,7)),type(['Int_'],1),pos(5,12)),integer(1,pos(10,12)),type([untyped],0),pos(8,12)), less_equal(join(identifier(this,type(['queen_'],1),pos(1,1)),identifier('x__',type(['queen->Int_'],2),pos(20,7)),type(['Int_'],1),pos(5,13)),card(identifier('queen_',type(['queen_'],1),pos(5,7)),type(['Int_'],1),pos(11,13)),type([untyped],0),pos(8,13)), equal(join(identifier(this,type(['queen_'],1),pos(1,1)),identifier('x__',type(['queen->Int_'],2),pos(20,7)),type(['Int_'],1),pos(5,14)),fun_call('integer_''minus_',[fun_call('integer_''plus_',[card(identifier('queen_',type(['queen_'],1),pos(5,7)),type(['Int_'],1),pos(21,14)), integer(1,pos(28,14))],type(['Int_'],1),pos(16,14)), join(identifier(this,type(['queen_'],1),pos(1,1)),identifier('x_',type(['queen->Int_'],2),pos(13,7)),type(['Int_'],1),pos(31,14))],type(['Int_'],1),pos(10,14)),type([untyped],0),pos(8,14))],pos(1,1))],[],pos(5,7))])) , 
    translate_model(AlloyTerm,Translated) , ! , 
    Translated == machine(generated(none,abstract_machine(none,machine(none),machine_header(none,alloytranslation,[]),[properties(none,conjunct(none,member(pos(0,0,7,13,0,0),identifier(none,x_),total_function(none,identifier(none,queen_),integer_set(pos(0,0,1,1,0,0)))),conjunct(none,member(pos(0,0,7,20,0,0),identifier(none,x__),total_function(none,identifier(none,queen_),integer_set(pos(0,0,1,1,0,0)))),conjunct(none,member(pos(0,0,7,28,0,0),identifier(none,y_),total_function(none,identifier(none,queen_),integer_set(pos(0,0,1,1,0,0)))),conjunct(none,forall(none,[identifier(none,this)],implication(none,subset(none,set_extension(none,[identifier(none,this)]),identifier(none,queen_)),conjunct(none,equal(pos(0,0,14,8,0,0),function(pos(0,0,14,5,0,0),identifier(pos(0,0,7,20,0,0),x__),[identifier(pos(0,0,1,1,0,0),this)]),minus(pos(0,0,14,10,0,0),add(pos(0,0,14,16,0,0),card(pos(0,0,14,21,0,0),identifier(pos(0,0,7,5,0,0),queen_)),integer(pos(0,0,14,28,0,0),1)),function(pos(0,0,14,31,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(pos(0,0,1,1,0,0),this)]))),conjunct(none,less_equal(pos(0,0,13,8,0,0),function(pos(0,0,13,5,0,0),identifier(pos(0,0,7,20,0,0),x__),[identifier(pos(0,0,1,1,0,0),this)]),card(pos(0,0,13,11,0,0),identifier(pos(0,0,7,5,0,0),queen_))),conjunct(none,greater_equal(pos(0,0,12,8,0,0),function(pos(0,0,12,5,0,0),identifier(pos(0,0,7,20,0,0),x__),[identifier(pos(0,0,1,1,0,0),this)]),integer(pos(0,0,12,10,0,0),1)),conjunct(none,less_equal(pos(0,0,11,7,0,0),function(pos(0,0,11,5,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(pos(0,0,1,1,0,0),this)]),card(pos(0,0,11,10,0,0),identifier(pos(0,0,7,5,0,0),queen_))),conjunct(none,less_equal(pos(0,0,10,7,0,0),function(pos(0,0,10,5,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(pos(0,0,1,1,0,0),this)]),card(pos(0,0,10,10,0,0),identifier(pos(0,0,7,5,0,0),queen_))),conjunct(none,greater_equal(pos(0,0,9,7,0,0),function(pos(0,0,9,5,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(pos(0,0,1,1,0,0),this)]),integer(pos(0,0,9,10,0,0),1)),greater_equal(pos(0,0,8,7,0,0),function(pos(0,0,8,5,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(pos(0,0,1,1,0,0),this)]),integer(pos(0,0,8,10,0,0),1)))))))))),forall(pos(0,0,17,8,0,0),[identifier(none,q_),identifier(none,q__)],implication(none,conjunct(none,subset(pos(0,0,17,21,0,0),set_extension(none,[identifier(none,q__)]),set_subtraction(pos(0,0,17,30,0,0),identifier(pos(0,0,7,5,0,0),queen_),set_extension(pos(0,0,17,12,0,0),[identifier(none,q_)]))),subset(pos(0,0,17,12,0,0),set_extension(none,[identifier(none,q_)]),identifier(pos(0,0,7,5,0,0),queen_))),conjunct(none,negation(pos(0,0,21,5,0,0),equal(pos(0,0,21,22,0,0),add(pos(0,0,21,7,0,0),function(pos(0,0,21,13,0,0),identifier(pos(0,0,7,20,0,0),x__),[identifier(none,q_)]),function(pos(0,0,21,18,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q_)])),add(pos(0,0,21,24,0,0),function(pos(0,0,21,31,0,0),identifier(pos(0,0,7,20,0,0),x__),[identifier(none,q__)]),function(pos(0,0,21,37,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q__)])))),conjunct(none,negation(pos(0,0,20,5,0,0),equal(pos(0,0,20,21,0,0),add(pos(0,0,20,7,0,0),function(pos(0,0,20,13,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(none,q_)]),function(pos(0,0,20,17,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q_)])),add(pos(0,0,20,23,0,0),function(pos(0,0,20,30,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(none,q__)]),function(pos(0,0,20,35,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q__)])))),conjunct(none,negation(pos(0,0,19,5,0,0),equal(pos(0,0,19,11,0,0),function(pos(0,0,19,8,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q_)]),function(pos(0,0,19,15,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q__)]))),negation(pos(0,0,18,5,0,0),equal(pos(0,0,18,11,0,0),function(pos(0,0,18,8,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(none,q_)]),function(pos(0,0,18,15,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(none,q__)]))))))))))))),definitions(none,[predicate_definition(pos(0,0,24,1,0,0),show_,[],boolean_true(pos(0,0,24,11,0,0)))]),operations(none,[operation(pos(0,0,26,1,0,0),identifier(pos(0,0,26,1,0,0),run0),[],[],precondition(pos(0,0,26,1,0,0),conjunct(pos(0,0,26,1,0,0),conjunct(none,equal(none,card(none,identifier(none,queen_)),integer(none,4)),truth(none)),forall(pos(0,0,17,8,0,0),[identifier(none,q_),identifier(none,q__)],implication(none,conjunct(none,subset(pos(0,0,17,21,0,0),set_extension(none,[identifier(none,q__)]),set_subtraction(pos(0,0,17,30,0,0),identifier(pos(0,0,7,5,0,0),queen_),set_extension(pos(0,0,17,12,0,0),[identifier(none,q_)]))),subset(pos(0,0,17,12,0,0),set_extension(none,[identifier(none,q_)]),identifier(pos(0,0,7,5,0,0),queen_))),conjunct(none,negation(pos(0,0,21,5,0,0),equal(pos(0,0,21,22,0,0),add(pos(0,0,21,7,0,0),function(pos(0,0,21,13,0,0),identifier(pos(0,0,7,20,0,0),x__),[identifier(none,q_)]),function(pos(0,0,21,18,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q_)])),add(pos(0,0,21,24,0,0),function(pos(0,0,21,31,0,0),identifier(pos(0,0,7,20,0,0),x__),[identifier(none,q__)]),function(pos(0,0,21,37,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q__)])))),conjunct(none,negation(pos(0,0,20,5,0,0),equal(pos(0,0,20,21,0,0),add(pos(0,0,20,7,0,0),function(pos(0,0,20,13,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(none,q_)]),function(pos(0,0,20,17,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q_)])),add(pos(0,0,20,23,0,0),function(pos(0,0,20,30,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(none,q__)]),function(pos(0,0,20,35,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q__)])))),conjunct(none,negation(pos(0,0,19,5,0,0),equal(pos(0,0,19,11,0,0),function(pos(0,0,19,8,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q_)]),function(pos(0,0,19,15,0,0),identifier(pos(0,0,7,28,0,0),y_),[identifier(none,q__)]))),negation(pos(0,0,18,5,0,0),equal(pos(0,0,18,11,0,0),function(pos(0,0,18,8,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(none,q_)]),function(pos(0,0,18,15,0,0),identifier(pos(0,0,7,13,0,0),x_),[identifier(none,q__)]))))))))),skip(none)))]),constants(none,[identifier(none,y_),identifier(none,x__),identifier(none,x_)]),sets(none,[deferred_set(pos(0,0,7,5,0,0),queen_)])]))) , 
    load_model(AlloyTerm).

test(river_crossing,[]) :- 
    AlloyTerm = alloy_model(facts([fact(equal(identifier('eats_',type([['Object_', 'Object_']],2),pos(23,5)),plus(cartesian(identifier('Fox_',type([['Fox_']],1),pos(17,6)),identifier('Chicken_',type([['Chicken_']],1),pos(22,6)),type([['Fox_', 'Chicken_']],2),pos(18,9)),cartesian(identifier('Chicken_',type([['Chicken_']],1),pos(22,6)),identifier('Grain_',type([['Grain_']],1),pos(31,6)),type([['Chicken_', 'Grain_']],2),pos(37,9)),type([['Fox_', 'Chicken_'], ['Chicken_', 'Grain_']],2),pos(28,9)),type([untyped],0),pos(13,9)),(1,9)),fact(and([equal(join(fun_call('ordering_''first_',[],type(['State_'],1),pos(8,15)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(13,15)),identifier('Object_',type([['Object_']],1),pos(14,5)),type([untyped],0),pos(19,15)), no(join(fun_call('ordering_''first_',[],type(['State_'],1),pos(34,15)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(39,15)),type([untyped],0),pos(31,15))],pos(28,15)),(1,15)),fact(all(['s_', 's__'],[field('s_',oneof(identifier('State_',type([['State_']],1),pos(5,12)),type([['State_']],1),pos(10,27)),type([['State_']],1),pos(7,27)), field('s__',oneof(join(identifier('s_',type([['State_']],1),pos(7,27)),fun_call('ordering_''next_',[],type(['State_'],2),pos(23,27)),type([['State_']],1),pos(22,27)),type([['State_']],1),pos(21,27)),type([['State_']],1),pos(17,27))],if_then_else(in(identifier('Farmer_',type([['Farmer_']],1),pos(9,6)),join(identifier('s_',type([['State_']],1),pos(7,27)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(16,28)),type([untyped],0),pos(12,28)),pred_call('crossRiver_',[join(identifier('s_',type([['State_']],1),pos(7,27)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(20,29)), join(identifier('s__',type([['State_']],1),pos(17,27)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(29,29)), join(identifier('s_',type([['State_']],1),pos(7,27)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(37,29)), join(identifier('s__',type([['State_']],1),pos(17,27)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(45,29))],type([untyped],0),pos(7,29)),pred_call('crossRiver_',[join(identifier('s_',type([['State_']],1),pos(7,27)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(20,31)), join(identifier('s__',type([['State_']],1),pos(17,27)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(28,31)), join(identifier('s_',type([['State_']],1),pos(7,27)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(35,31)), join(identifier('s__',type([['State_']],1),pos(17,27)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(44,31))],type([untyped],0),pos(7,31)),type([untyped],0),pos(22,28)),type([untyped],0),pos(3,27)),(1,26))]),assertions([]),commands([run(and([equal(identifier('eats_',type([['Object_', 'Object_']],2),pos(23,5)),plus(cartesian(identifier('Fox_',type([['Fox_']],1),pos(17,6)),identifier('Chicken_',type([['Chicken_']],1),pos(22,6)),type([['Fox_', 'Chicken_']],2),pos(18,9)),cartesian(identifier('Chicken_',type([['Chicken_']],1),pos(22,6)),identifier('Grain_',type([['Grain_']],1),pos(31,6)),type([['Chicken_', 'Grain_']],2),pos(37,9)),type([['Fox_', 'Chicken_'], ['Chicken_', 'Grain_']],2),pos(28,9)),type([untyped],0),pos(13,9)), equal(join(fun_call('ordering_''first_',[],type(['State_'],1),pos(8,15)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(13,15)),identifier('Object_',type([['Object_']],1),pos(14,5)),type([untyped],0),pos(19,15)), no(join(fun_call('ordering_''first_',[],type(['State_'],1),pos(34,15)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(39,15)),type([untyped],0),pos(31,15)), all(['s_', 's__'],[field('s_',oneof(identifier('State_',type([['State_']],1),pos(5,12)),type([['State_']],1),pos(10,27)),type([['State_']],1),pos(7,27)), field('s__',oneof(join(identifier('s_',type([['State_']],1),pos(7,27)),fun_call('ordering_''next_',[],type(['State_'],2),pos(23,27)),type([['State_']],1),pos(22,27)),type([['State_']],1),pos(21,27)),type([['State_']],1),pos(17,27))],if_then_else(in(identifier('Farmer_',type([['Farmer_']],1),pos(9,6)),join(identifier('s_',type([['State_']],1),pos(7,27)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(16,28)),type([untyped],0),pos(12,28)),pred_call('crossRiver_',[join(identifier('s_',type([['State_']],1),pos(7,27)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(20,29)), join(identifier('s__',type([['State_']],1),pos(17,27)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(29,29)), join(identifier('s_',type([['State_']],1),pos(7,27)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(37,29)), join(identifier('s__',type([['State_']],1),pos(17,27)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(45,29))],type([untyped],0),pos(7,29)),pred_call('crossRiver_',[join(identifier('s_',type([['State_']],1),pos(7,27)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(20,31)), join(identifier('s__',type([['State_']],1),pos(17,27)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(28,31)), join(identifier('s_',type([['State_']],1),pos(7,27)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(35,31)), join(identifier('s__',type([['State_']],1),pos(17,27)),identifier('near_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(44,31))],type([untyped],0),pos(7,31)),type([untyped],0),pos(22,28)),type([untyped],0),pos(3,27)), equal(join(fun_call('ordering_''last_',[],type(['State_'],1),pos(7,36)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(11,36)),identifier('Object_',type([['Object_']],1),pos(14,5)),type([untyped],0),pos(15,36))],pos(1,9)),global_scope(-1),exact_scopes([('State_',8)]),bitwidth(-1),pos(1,36))]),functions([predicate('crossRiver_',[identifier('from_',type([['Object_']],1),pos(18,18)), identifier('from__',type([['Object_']],1),pos(24,18)), identifier('to_',type([['Object_']],1),pos(31,18)), identifier('to__',type([['Object_']],1),pos(35,18))],[field('from_',setof(identifier('Object_',type([['Object_']],1),pos(14,5)),type([['Object_']],1),pos(40,18)),type([['Object_']],1),pos(18,18)),field('from__',setof(identifier('Object_',type([['Object_']],1),pos(14,5)),type([['Object_']],1),pos(40,18)),type([['Object_']],1),pos(18,18)),field('to_',setof(identifier('Object_',type([['Object_']],1),pos(14,5)),type([['Object_']],1),pos(40,18)),type([['Object_']],1),pos(18,18)),field('to__',setof(identifier('Object_',type([['Object_']],1),pos(14,5)),type([['Object_']],1),pos(40,18)),type([['Object_']],1),pos(18,18))],one(['x_'],[field('x_',oneof(identifier('from_',type([['Object_']],1),pos(18,18)),type([['Object_']],1),pos(10,19)),type([['Object_']],1),pos(7,19))],and([equal(identifier('from__',type([['Object_']],1),pos(24,18)),minus(minus(minus(identifier('from_',type([['Object_']],1),pos(18,18)),identifier('x_',type([['Object_']],1),pos(7,19)),type([['Object_']],1),pos(18,20)),identifier('Farmer_',type([['Farmer_']],1),pos(9,6)),type([['Object_']],1),pos(22,20)),join(identifier('from__',type([['Object_']],1),pos(24,18)),identifier('eats_',type([['Object_', 'Object_']],2),pos(23,5)),type([['Object_']],1),pos(38,20)),type([['Object_']],1),pos(31,20)),type([untyped],0),pos(11,20)), equal(identifier('to__',type([['Object_']],1),pos(35,18)),plus(plus(identifier('to_',type([['Object_']],1),pos(31,18)),identifier('x_',type([['Object_']],1),pos(7,19)),type([['Object_']],1),pos(14,21)),identifier('Farmer_',type([['Farmer_']],1),pos(9,6)),type([['Object_']],1),pos(18,21)),type([untyped],0),pos(9,21))],pos(1,1)),type([untyped],0),pos(3,19)),pos(1,18)),predicate('run$1_',[],[],equal(join(fun_call('ordering_''last_',[],type(['State_'],1),pos(7,36)),identifier('far_',type([['State_', 'Object_']],2),pos(13,12)),type([['Object_']],1),pos(11,36)),identifier('Object_',type([['Object_']],1),pos(14,5)),type([untyped],0),pos(15,36)),pos(1,36))]),signatures([signature('Object_',[field('eats_',setof(identifier('Object_',type([['Object_']],1),pos(14,5)),type([['Object_']],1),pos(29,5)),type([['Object_']],1),pos(23,5))],[],[abstract],pos(14,5)),signature('Farmer_',[],[],[one, subsig('Object_')],pos(9,6)),signature('Fox_',[],[],[one, subsig('Object_')],pos(17,6)),signature('Chicken_',[],[],[one, subsig('Object_')],pos(22,6)),signature('Grain_',[],[],[one, subsig('Object_')],pos(31,6)),signature('State_',[field('near_',setof(identifier('Object_',type([['Object_']],1),pos(14,5)),type([['Object_']],1),pos(24,12)),type([['Object_']],1),pos(13,12)),field('far_',setof(identifier('Object_',type([['Object_']],1),pos(14,5)),type([['Object_']],1),pos(24,12)),type([['Object_']],1),pos(13,12))],[],[ordered],pos(5,12))])) , 
    translate_model(AlloyTerm,_Translated) , ! , 
    load_model(AlloyTerm).

/*
test(crewalloc,[]) :- 
    translate_model(alloy_model(facts([fact(equal(identifier('male_',type(['Personell_'],1),pos(5,7)),plus(plus(identifier('tom_',type(['tom_'],1),pos(9,6)),identifier('david_',type(['david_'],1),pos(14,6)),type(['tom_', 'david_'],1),pos(13,13)),identifier('jeremy_',type(['jeremy_'],1),pos(21,6)),type(['tom_', 'david_', 'jeremy_'],1),pos(19,13)),type([untyped],0),pos(8,13)),(1,12)),fact(equal(identifier('speaks_',type(['Personell->this_''Language_'],2),pos(26,5)),plus(plus(plus(plus(plus(cartesian(identifier('tom_',type(['tom_'],1),pos(9,6)),identifier('german_',type(['german_'],1),pos(17,4)),type(['tom->this_''german_'],2),pos(14,16)),cartesian(identifier('david_',type(['david_'],1),pos(14,6)),identifier('french_',type(['french_'],1),pos(9,4)),type(['david->this_''french_'],2),pos(30,16)),type(['tom->this_''german_', 'david->this_''french_'],2),pos(23,16)),cartesian(identifier('jeremy_',type(['jeremy_'],1),pos(21,6)),identifier('german_',type(['german_'],1),pos(17,4)),type(['jeremy->this_''german_'],2),pos(47,16)),type(['tom->this_''german_', 'david->this_''french_', 'jeremy->this_''german_'],2),pos(39,16)),cartesian(identifier('carol_',type(['carol_'],1),pos(29,6)),identifier('spanish_',type(['spanish_'],1),pos(25,4)),type(['carol->this_''spanish_'],2),pos(63,16)),type(['tom->this_''german_', 'david->this_''french_', 'jeremy->this_''german_', 'carol->this_''spanish_'],2),pos(56,16)),cartesian(identifier('janet_',type(['janet_'],1),pos(36,6)),identifier('french_',type(['french_'],1),pos(9,4)),type(['janet->this_''french_'],2),pos(80,16)),type(['tom->this_''german_', 'david->this_''french_', 'jeremy->this_''german_', 'carol->this_''spanish_', 'janet->this_''french_'],2),pos(73,16)),cartesian(identifier('tracy_',type(['tracy_'],1),pos(43,6)),identifier('spanish_',type(['spanish_'],1),pos(25,4)),type(['tracy->this_''spanish_'],2),pos(96,16)),type(['tom->this_''german_', 'david->this_''french_', 'jeremy->this_''german_', 'carol->this_''spanish_', 'janet->this_''french_', 'tracy->this_''spanish_'],2),pos(89,16)),type([untyped],0),pos(10,16)),(1,15))]),assertions([]),commands([run(and([equal(identifier('male_',type(['Personell_'],1),pos(5,7)),plus(plus(identifier('tom_',type(['tom_'],1),pos(9,6)),identifier('david_',type(['david_'],1),pos(14,6)),type(['tom_', 'david_'],1),pos(13,13)),identifier('jeremy_',type(['jeremy_'],1),pos(21,6)),type(['tom_', 'david_', 'jeremy_'],1),pos(19,13)),type([untyped],0),pos(8,13)), equal(identifier('speaks_',type(['Personell->this_''Language_'],2),pos(26,5)),plus(plus(plus(plus(plus(cartesian(identifier('tom_',type(['tom_'],1),pos(9,6)),identifier('german_',type(['german_'],1),pos(17,4)),type(['tom->this_''german_'],2),pos(14,16)),cartesian(identifier('david_',type(['david_'],1),pos(14,6)),identifier('french_',type(['french_'],1),pos(9,4)),type(['david->this_''french_'],2),pos(30,16)),type(['tom->this_''german_', 'david->this_''french_'],2),pos(23,16)),cartesian(identifier('jeremy_',type(['jeremy_'],1),pos(21,6)),identifier('german_',type(['german_'],1),pos(17,4)),type(['jeremy->this_''german_'],2),pos(47,16)),type(['tom->this_''german_', 'david->this_''french_', 'jeremy->this_''german_'],2),pos(39,16)),cartesian(identifier('carol_',type(['carol_'],1),pos(29,6)),identifier('spanish_',type(['spanish_'],1),pos(25,4)),type(['carol->this_''spanish_'],2),pos(63,16)),type(['tom->this_''german_', 'david->this_''french_', 'jeremy->this_''german_', 'carol->this_''spanish_'],2),pos(56,16)),cartesian(identifier('janet_',type(['janet_'],1),pos(36,6)),identifier('french_',type(['french_'],1),pos(9,4)),type(['janet->this_''french_'],2),pos(80,16)),type(['tom->this_''german_', 'david->this_''french_', 'jeremy->this_''german_', 'carol->this_''spanish_', 'janet->this_''french_'],2),pos(73,16)),cartesian(identifier('tracy_',type(['tracy_'],1),pos(43,6)),identifier('spanish_',type(['spanish_'],1),pos(25,4)),type(['tracy->this_''spanish_'],2),pos(96,16)),type(['tom->this_''german_', 'david->this_''french_', 'jeremy->this_''german_', 'carol->this_''spanish_', 'janet->this_''french_', 'tracy->this_''spanish_'],2),pos(89,16)),type([untyped],0),pos(10,16)), pred_call('allLanguages_',[],type([untyped],0),pos(3,31)), pred_call('allSexes_',[],type([untyped],0),pos(20,31)), pred_call('scheduleOk_',[],type([untyped],0),pos(33,31)), pred_call('everybodyInSchedule_',[],type([untyped],0),pos(48,31))],pos(1,12)),global_scope(-1),exact_scopes([]),bitwidth(-1),pos(1,34))]),functions([predicate('allLanguages_',[],[],all(['f_'],[field('f_',oneof(identifier('Flight_',type(['Flight_'],1),pos(5,8)),type(['Flight_'],1),pos(9,19)),type(['Flight_'],1),pos(7,19))],equal(join(join(identifier('f_',type(['Flight_'],1),pos(7,19)),identifier('assign_',type(['Flight->this_''Personell_'],2),pos(3,9)),type(['Personell_'],1),pos(19,19)),identifier('speaks_',type(['Personell->this_''Language_'],2),pos(26,5)),type(['Language_'],1),pos(26,19)),identifier('Language_',type(['Language_'],1),pos(14,3)),type([untyped],0),pos(34,19)),type([untyped],0),pos(3,19)),pos(1,18)),predicate('allSexes_',[],[],all(['f_'],[field('f_',oneof(identifier('Flight_',type(['Flight_'],1),pos(5,8)),type(['Flight_'],1),pos(9,22)),type(['Flight_'],1),pos(7,22))],and([some(minus(join(identifier('f_',type(['Flight_'],1),pos(7,22)),identifier('assign_',type(['Flight->this_''Personell_'],2),pos(3,9)),type(['Personell_'],1),pos(25,22)),identifier('male_',type(['Personell_'],1),pos(5,7)),type(['Personell_'],1),pos(32,22)),type([untyped],0),pos(18,22)), some(intersection(join(identifier('f_',type(['Flight_'],1),pos(7,22)),identifier('assign_',type(['Flight->this_''Personell_'],2),pos(3,9)),type(['Personell_'],1),pos(50,22)),identifier('male_',type(['Personell_'],1),pos(5,7)),type(['Personell_'],1),pos(58,22)),type([untyped],0),pos(43,22))],pos(39,22)),type([untyped],0),pos(3,22)),pos(1,21)),predicate('scheduleOk_',[],[],all(['p_', 'f_'],[field('p_',oneof(identifier('Personell_',type(['Personell_'],1),pos(14,5)),type(['Personell_'],1),pos(9,25)),type(['Personell_'],1),pos(7,25)), field('f_',oneof(identifier('Flight_',type(['Flight_'],1),pos(5,8)),type(['Flight_'],1),pos(24,25)),type(['Flight_'],1),pos(20,25))],implication(and([in(identifier('p_',type(['Personell_'],1),pos(7,25)),join(identifier('f_',type(['Flight_'],1),pos(20,25)),identifier('assign_',type(['Flight->this_''Personell_'],2),pos(3,9)),type(['Personell_'],1),pos(39,25)),type([untyped],0),pos(35,25)), in(identifier('p_',type(['Personell_'],1),pos(7,25)),join(join(identifier('f_',type(['Flight_'],1),pos(20,25)),fun_call('ordering_''next_',[],type(['Flight->this_''Flight_'],2),pos(56,25)),type(['Flight_'],1),pos(56,25)),identifier('assign_',type(['Flight->this_''Personell_'],2),pos(3,9)),type(['Personell_'],1),pos(63,25)),type([untyped],0),pos(53,25))],pos(47,25)),not(in(identifier('p_',type(['Personell_'],1),pos(7,25)),join(join(join(identifier('f_',type(['Flight_'],1),pos(20,25)),fun_call('ordering_''next_',[],type(['Flight->this_''Flight_'],2),pos(89,25)),type(['Flight_'],1),pos(89,25)),fun_call('ordering_''next_',[],type(['Flight->this_''Flight_'],2),pos(84,25)),type(['Flight_'],1),pos(84,25)),identifier('assign_',type(['Flight->this_''Personell_'],2),pos(3,9)),type(['Personell_'],1),pos(97,25)),type([untyped],0),pos(81,25)),type([untyped],0),pos(74,25)),type([untyped],0),pos(71,25)),type([untyped],0),pos(3,25)),pos(1,24)),predicate('everybodyInSchedule_',[],[],equal(join(identifier('univ_',type(['univ_'],1),pos(1,1)),identifier('assign_',type(['Flight->this_''Personell_'],2),pos(3,9)),type(['Personell_'],1),pos(7,28)),identifier('Personell_',type(['Personell_'],1),pos(14,5)),type([untyped],0),pos(15,28)),pos(1,27)),predicate('crewAlloc_',[],[],and([pred_call('allLanguages_',[],type([untyped],0),pos(3,31)), pred_call('allSexes_',[],type([untyped],0),pos(20,31)), pred_call('scheduleOk_',[],type([untyped],0),pos(33,31)), pred_call('everybodyInSchedule_',[],type([untyped],0),pos(48,31))],pos(44,31)),pos(1,30))]),signatures([signature('Language_',[],[],[abstract],pos(14,3)),signature('french_',[],[],[one, subsig('Language_')],pos(9,4)),signature('german_',[],[],[one, subsig('Language_')],pos(17,4)),signature('spanish_',[],[],[one, subsig('Language_')],pos(25,4)),signature('Personell_',[field('speaks_',setof(identifier('Language_',type(['Language_'],1),pos(14,3)),type(['Language_'],1),pos(35,5)),type(['Language_'],1),pos(26,5))],[],[abstract],pos(14,5)),signature('tom_',[],[],[one, subsig('Personell_')],pos(9,6)),signature('david_',[],[],[one, subsig('Personell_')],pos(14,6)),signature('jeremy_',[],[],[one, subsig('Personell_')],pos(21,6)),signature('carol_',[],[],[one, subsig('Personell_')],pos(29,6)),signature('janet_',[],[],[one, subsig('Personell_')],pos(36,6)),signature('tracy_',[],[],[one, subsig('Personell_')],pos(43,6)),signature('male_',[],[],[subset(['Personell_'])],pos(5,7)),signature('Flight_',[field('assign_',setof(identifier('Personell_',type(['Personell_'],1),pos(14,5)),type(['Personell_'],1),pos(12,9)),type(['Personell_'],1),pos(3,9))],[],[ordered],pos(5,8))])),Translated) , ! , 
    Translated == machine(generated(none,abstract_machine(none,machine(none),machine_header(none,alloytranslation,[]),[properties(none,conjunct(none,member(none,identifier(none,french_),identifier(none,'Language_')),conjunct(none,member(none,identifier(none,german_),identifier(none,'Language_')),conjunct(none,member(none,identifier(none,spanish_),identifier(none,'Language_')),conjunct(none,member(none,identifier(none,speaks_),relations(none,identifier(none,'Personell_'),identifier(none,'Language_'))),conjunct(none,member(none,identifier(none,tom_),identifier(none,'Personell_')),conjunct(none,member(none,identifier(none,david_),identifier(none,'Personell_')),conjunct(none,member(none,identifier(none,jeremy_),identifier(none,'Personell_')),conjunct(none,member(none,identifier(none,carol_),identifier(none,'Personell_')),conjunct(none,member(none,identifier(none,janet_),identifier(none,'Personell_')),conjunct(none,member(none,identifier(none,tracy_),identifier(none,'Personell_')),conjunct(none,member(none,identifier(none,male_),identifier(none,'Personell_')),conjunct(none,member(none,identifier(none,assign_),relations(none,identifier(none,'Flight_'),identifier(none,'Personell_'))),conjunct(none,equal(none,identifier(none,male_),union(none,union(none,identifier(none,tom_),identifier(none,david_)),identifier(none,jeremy_))),equal(none,identifier(none,speaks_),union(none,union(none,union(none,union(none,union(none,cartesian(none,identifier(none,tom_),identifier(none,german_)),cartesian(none,identifier(none,david_),identifier(none,french_))),cartesian(none,identifier(none,jeremy_),identifier(none,german_))),cartesian(none,identifier(none,carol_),identifier(none,spanish_))),cartesian(none,identifier(none,janet_),identifier(none,french_))),cartesian(none,identifier(none,tracy_),identifier(none,spanish_)))))))))))))))))),definitions(none,[predicate_definition(none,crewAlloc_,[],conjunct(none,definition(everybodyInSchedule_,[],type([untyped],0),pos(48,31)),conjunct(none,definition(scheduleOk_,[],type([untyped],0),pos(33,31)),conjunct(none,definition(allSexes_,[],type([untyped],0),pos(20,31)),definition(allLanguages_,[],type([untyped],0),pos(3,31)))))),predicate_definition(none,everybodyInSchedule_,[],equal(none,ran(none,identifier(none,assign_)),identifier(none,'Personell_'))),predicate_definition(none,scheduleOk_,[],forall(none,[identifier(none,p_),identifier(none,f_)],implication(none,conjunct(none,subset(none,set_extension(none,[identifier(none,f_)]),identifier(none,'Flight_')),subset(none,set_extension(none,[identifier(none,p_)]),identifier(none,'Personell_'))),implication(none,conjunct(none,subset(none,identifier(none,p_),image(none,identifier(none,assign_),image(none,definition('ordering_\'next_',[],type(['Flight->this_\'Flight_'],2),pos(56,25)),identifier(none,f_)))),subset(none,identifier(none,p_),image(none,identifier(none,assign_),identifier(none,f_)))),not(none,subset(none,identifier(none,p_),image(none,identifier(none,assign_),image(none,definition('ordering_\'next_',[],type(['Flight->this_\'Flight_'],2),pos(84,25)),image(none,definition('ordering_\'next_',[],type(['Flight->this_\'Flight_'],2),pos(89,25)),identifier(none,f_)))))))))),predicate_definition(none,allSexes_,[],forall(none,[identifier(none,f_)],implication(none,subset(none,set_extension(none,[identifier(none,f_)]),identifier(none,'Flight_')),conjunct(none,greater(card(none,intersection(none,image(none,identifier(none,assign_),identifier(none,f_)),identifier(none,male_))),integer(none,0)),greater(card(none,set_subtraction(none,image(none,identifier(none,assign_),identifier(none,f_)),identifier(none,male_))),integer(none,0)))))),predicate_definition(none,allLanguages_,[],forall(none,[identifier(none,f_)],implication(none,subset(none,set_extension(none,[identifier(none,f_)]),identifier(none,'Flight_')),equal(none,image(none,identifier(none,speaks_),image(none,identifier(none,assign_),identifier(none,f_))),identifier(none,'Language_')))))]),operations(none,[operation(none,identifier(none,run),[],[],precondition(none,conjunct(none,conjunct(none,truth(none),truth(none)),conjunct(none,definition(everybodyInSchedule_,[],type([untyped],0),pos(48,31)),conjunct(none,definition(scheduleOk_,[],type([untyped],0),pos(33,31)),conjunct(none,definition(allSexes_,[],type([untyped],0),pos(20,31)),conjunct(none,definition(allLanguages_,[],type([untyped],0),pos(3,31)),conjunct(none,equal(none,identifier(none,speaks_),union(none,union(none,union(none,union(none,union(none,cartesian(none,identifier(none,tom_),identifier(none,german_)),cartesian(none,identifier(none,david_),identifier(none,french_))),cartesian(none,identifier(none,jeremy_),identifier(none,german_))),cartesian(none,identifier(none,carol_),identifier(none,spanish_))),cartesian(none,identifier(none,janet_),identifier(none,french_))),cartesian(none,identifier(none,tracy_),identifier(none,spanish_)))),equal(none,identifier(none,male_),union(none,union(none,identifier(none,tom_),identifier(none,david_)),identifier(none,jeremy_))))))))),skip(none)))]),constants(none,[identifier(none,assign_),identifier(none,male_),identifier(none,tracy_),identifier(none,janet_),identifier(none,carol_),identifier(none,jeremy_),identifier(none,david_),identifier(none,tom_),identifier(none,speaks_),identifier(none,spanish_),identifier(none,german_),identifier(none,french_)]),sets(none,[deferred_set(none,'Flight_'),deferred_set(none,'Personell_'),deferred_set(none,'Language_')])]))).
*/
:- end_tests(full_machine_translation).

