#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "queue.h"
#include "hash.h"
#include "magic_constants.h"
#include "verbose.h"

#include "../libs/leveldb-1.20/include/leveldb/c.h"

#include "assert_helper.h"

static void enqueue_fifo(wQueue * const q, char * const term, const size_t len, const char digest[HASH_LEN]);
static void enqueue_lifo(wQueue * const q, char * const term, const size_t len, const char digest[HASH_LEN]);

static void ensure_errptr(char const * const errptr, char const * const loc) {
    if (errptr != NULL) {
        printf("levelDB error during %s: %s\n", loc, errptr);
        assert(errptr == NULL);
    }
}


/* TODO: disregard the below, we always to BFS right now. (pk, 12.01.2018)
 * initialises a new queue; 
 * if strategy is 0, it is a queue;
 * if strategy is 100, it is a stack;
 * any value in between results in a mixed data structure */
wQueue *init_queue(const int strategy, const size_t thresh, char const * const tmpdir) {
    struct db db;
    if (thresh > 0) {
        leveldb_options_t *ldb_opts = leveldb_options_create();
        leveldb_options_set_create_if_missing(ldb_opts, 1);
        leveldb_options_set_error_if_exists(ldb_opts, 1);

        leveldb_readoptions_t *ldb_readopts = leveldb_readoptions_create();
        leveldb_readoptions_set_fill_cache(ldb_readopts, 0);

        leveldb_writeoptions_t *ldb_writeopts = leveldb_writeoptions_create();

        char *errptr = NULL;

        char *file = strdup(tmpdir);
        printf("opening db in %s\n", file);
        leveldb_t *ldb = leveldb_open(ldb_opts, file, &errptr);
        ensure_errptr(errptr, "open");

        db = (struct db) {ldb, ldb_opts, ldb_readopts, ldb_writeopts, file};
    }

    wQueue *q = (wQueue*) malloc(sizeof(wQueue));
    q->terms_available = 0;
    q->first = NULL;
    q->last = NULL;
    q->c = 0;
    q->thresh = thresh;
    q->strategy = strategy;
    q->db = db;

    return q;
}

/* enqueue a term with length len and hash code digest into the given queue;
 * function takes ownership of the term */
void enqueue(wQueue *q, char * const term, const size_t len, const char digest[HASH_LEN]) {
      enqueue_fifo(q, term, len, digest);
 //   int r = rand() % 100;
 //   if (r < q->strategy) {
 //       enqueue_lifo(q, term, len, digest);
 //   } else {
 //       enqueue_fifo(q, term, len, digest);
 //   }
}

void ensure_at_least_terms_available(wQueue *q) {
    assert(0); /* this is a paranoia check and should not be in production at any time */
    int i;
    tCell *cell = q->first;
    for (i = 0; i < q->terms_available; i++) {
        assert(cell->term != NULL);
        cell = cell->next;
    }
}

static void enqueue_fifo(wQueue * const q, char * const term, const size_t len, const char digest[HASH_LEN]) {
    tCell *new = (tCell*) malloc(sizeof(tCell));

    int to_disk = (q->thresh > 0 && /* only write to disk if threshold is sane */
                   (q->c >= q->thresh /* write to disk if over threshold */
                    || q->terms_available != q->c)); /* or if written to disk before */
    if (to_disk) {
        char *errptr = NULL;
        if (VERY_VERBOSE_QUEUE) {
            char buf[READABLE_HASH_LEN];
            print_readable_digest((unsigned char*) digest, buf);
            printf("LEVELDB: putting %s for cell %p\n", buf, new);
        }
        leveldb_put(q->db.db, q->db.write_opts,
                    digest, HASH_LEN, /* key */
                    term, len, /* val */
                    &errptr);
        ensure_errptr(errptr, "put");
    } else {
        q->terms_available++;
    }

    if (VERY_VERBOSE_QUEUE) {
        printf("enqueued cell %p, available: %ld, count: %ld\n", new, q->terms_available, q->c);
    }
    q->c++;

    /* keep mini-states:
     * a mini-state is like a state, but the term is stored on disk;
     * the term size and its hash code are still kept in memory 
     * (pk, 08.01.2018) */
    if (to_disk) {
        new->term = NULL;
        free(term);
    } else {
        new->term = term;
    }
    new->term_length = len;
    memcpy(new->digest, digest, HASH_LEN);
    new->next = NULL;

        
    if (q->first == NULL) {
        q->first = new; 
        q->last = new;
        q->first->next = NULL;
    } else { 
        tCell *temp = q->last; // for multithreading
        q->last = new;
        temp->next = new;
    }

    //ensure_at_least_terms_available(q);
    
}

static void enqueue_lifo(wQueue * const q, char * const term, const size_t len, const char digest[HASH_LEN]) {
    q->c++;

    tCell *new = (tCell*) malloc(sizeof(tCell));
    new->term = term;
    new->term_length = len;
    memcpy(new->digest, digest, HASH_LEN);
    new->next = q->first;

    q->first = new;
    if (q->last == NULL) {
        q->last = new;
    }
}

size_t load_terms(wQueue *q) {
    size_t terms_to_load = q->thresh > q->c ? q->c : q->thresh;
    int i;
    tCell *cur = q->first;
    assert(cur);

    for (i = 0; i < terms_to_load; i++) {
        assert(cur->term == NULL);
        size_t term_length = -1;
        char *errptr = NULL;
        if (VERY_VERBOSE_QUEUE) {
            char buf[READABLE_HASH_LEN];
            print_readable_digest((unsigned char*) cur->digest, buf);
            printf("LEVELDB: reading %s\n", buf);
        }
        cur->term = leveldb_get(q->db.db, q->db.read_opts,
                                cur->digest, HASH_LEN,
                                &term_length,
                                &errptr);
        ensure_errptr(errptr, "load");
        assert(cur->term);
        if (VERY_VERBOSE_QUEUE) {
            printf("loading into cell %p\n", cur);
        }
        assert_es(term_length == cur->term_length, "term length of term gotten from DB did not match");

        leveldb_delete(q->db.db, q->db.write_opts,
                       cur->digest, HASH_LEN, /* key */
                       &errptr);
        ensure_errptr(errptr, "delete");


        /* FIXME: add paranoid hash check if things seem to go wrong (pk, (08.01.2018) */
        cur = cur->next;
    }
    return terms_to_load;
}

/* dequeues a cell from given queue;
 * returns NULL, if queue was empty.
 * note that the next pointer will be NULL. */
tCell *dequeue(wQueue * const q) {
    if (q->c == 0) {
        return NULL;
    }
    if (q->thresh > 0 && q->terms_available == 0) {
        q->terms_available += load_terms(q);
    }

    q->c--;
    q->terms_available--;
    //ensure_at_least_terms_available(q);

    tCell *res = q->first;
    if (VERY_VERBOSE_QUEUE) {
        printf("dequeuing cell %p, available: %ld, count: %ld\n", res, q->terms_available, q->c);
    }
    assert_es(res->term, "The head of the queue had a NULL term. Aborting.");
    
    q->first = q->first->next;
    
    if (q->first == NULL) {
        q->last = NULL;
    }
    
    res->next = NULL;
    return res;
}

/* returns 1 if the queue is empty,
 * 0 otherwise */
int is_empty(wQueue const * const q) {
    return (q->first == NULL);
}

/* returns the current queue size */
size_t q_size(wQueue const * const q) {
    return q->c;
}

/* tears down a queue and all its queued items */
void teardown_queue(wQueue **q) {
    while((*q)->c > 0) {
        tCell *cell = dequeue(*q);
        free(cell->term);
        free(cell);
    }

    if ((*q)->thresh > 0) {
        leveldb_close((*q)->db.db);
        char *errptr = NULL;
        leveldb_destroy_db((*q)->db.opts, (*q)->db.name, &errptr);
        ensure_errptr(errptr, "close");
        leveldb_writeoptions_destroy((*q)->db.write_opts);
        leveldb_readoptions_destroy((*q)->db.read_opts);
        leveldb_options_destroy((*q)->db.opts);
        free((*q)->db.name);
    }

    free(*q);
    *q = NULL;
}

/* tears down a cell and its term */
void teardown_cell(tCell **c) {
    assert(c);
    assert(*c);
    assert((*c)->term);
    free((*c)->term);
    free(*c);
    *c = NULL;
}
