#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/queue.h"
#include "../common/hash.h"
#include "../common/msg_headers.h"

#include "mstats.h"
#include "mid.h"

#include "mwork.h"

int has_empty_worker(mStats const * const stats) {
    pDigest *pdigests = stats->pdigests;
    int i;
    for (i = 1; i <= stats->len; i++) {
        if (pdigests[i].header == PSTATS_HAS_EMPTY
                || pdigests[i].header == PSTATS_COMPLETELY_EMPTY) {
            return i;
        }
    }
    return 0;
}

/* TODO: termination for assertion checking */

int work_sender(zloop_t * const loop,
                const int timer_id,
                mStats * const stats,
                struct id_map const * const ips,
                wQueue * const q,
                int fd_logfile) {
    if (is_empty(q)) {
        dprintf(fd_logfile, "work sender stop\n");
        zloop_timer_end(loop, timer_id);
        return 0;
    }

    int empty;
    while ((empty = has_empty_worker(stats)) != 0) {
        if (is_empty(q)) {
            dprintf(fd_logfile, "work sender stop\n");
            zloop_timer_end(loop, timer_id);
            return 0;
        }
        dprintf(fd_logfile, "work sender active\n");
        stats->pdigests[empty].empty--;
        if (stats->pdigests[empty].empty == 0) {
            stats->pdigests[empty].header = PSTATS_HAS_PMS;
            //printf("expecting an ACK on proxy%d\n", empty);
            stats->pdigests[empty].acks_expected++;
        } else if (stats->pdigests[empty].header == PSTATS_COMPLETELY_EMPTY) {
            stats->pdigests[empty].header = PSTATS_HAS_EMPTY;
            stats->pdigests[empty].acks_expected++;
            //printf("expecting an ACK on proxy%d\n", empty);
        } else {
            assert(0);
        }
        char *port = zmsg_popstr(stats->pdigests[empty].need_msg);

        char *endpoint;
        int rc = asprintf(&endpoint, "tcp://%s:%s", ips->id_map[empty], port);
        assert(rc != -1);

        dprintf(fd_logfile, "binding to %s\n", endpoint);
        zsock_t *req = zsock_new_req(endpoint);
        tCell *item = dequeue(q);

        zmsg_t *msg = zmsg_new();
        zmsg_addmem(msg, item->digest, HASH_LEN);
        zmsg_addmem(msg, item->term, item->term_length);

        dprintf(fd_logfile, "sending workpackage to a worker\n");
        uint64_t timestamp = time(NULL);
        rc = zsock_send(req, "i8m", CHECK_THIS, timestamp, msg);
        assert(rc == 0);
        dprintf(fd_logfile, "sent!\n");

        int header;
        rc = zsock_recv(req, "im", &header, NULL);
        assert(rc == 0);
        assert(header == TRANSFER_ACK);
        dprintf(fd_logfile, "received ACK!\n");

        zmsg_destroy(&msg);

        teardown_cell(&item);
        zsock_disconnect(req, "%s", endpoint);

        zsock_destroy(&req);
        free(endpoint);
        free(port);
    }
    return 0;
}


/* a version of the above function that does not require a zloop
 * and returns whether it should be called again */
int work_sender_no_loop(mStats * const stats,
                        struct id_map const * const ips,
                        wQueue * const q,
                        int fd_logfile) {
    if (is_empty(q)) {
        return 0;
    }

    int empty;
    while ((empty = has_empty_worker(stats)) != 0) {
        if (is_empty(q)) {
            return 0;
        }
        dprintf(fd_logfile, "work sender active\n");
        stats->pdigests[empty].empty--;
        if (stats->pdigests[empty].empty == 0) {
            stats->pdigests[empty].header = PSTATS_HAS_PMS;
            //printf("expecting an ACK on proxy%d\n", empty);
            stats->pdigests[empty].acks_expected++;
        }
        if (stats->pdigests[empty].header == PSTATS_COMPLETELY_EMPTY) {
            stats->pdigests[empty].header = PSTATS_HAS_EMPTY;
            //printf("expecting an ACK on proxy%d\n", empty);
            stats->pdigests[empty].acks_expected++;
        }
        char *port = zmsg_popstr(stats->pdigests[empty].need_msg);

        char *endpoint;
        int rc = asprintf(&endpoint, "tcp://%s:%s", ips->id_map[empty], port);
        assert(rc != -1);

        dprintf(fd_logfile, "binding to %s\n", endpoint);
        zsock_t *req = zsock_new_req(endpoint);
        tCell *item = dequeue(q);

        zmsg_t *msg = zmsg_new();
        zmsg_addmem(msg, item->digest, HASH_LEN);
        zmsg_addmem(msg, item->term, item->term_length);

        dprintf(fd_logfile, "sending workpackage to a worker\n");
        uint64_t timestamp = time(NULL);
        rc = zsock_send(req, "i8m", CHECK_THIS, timestamp, msg);
        assert(rc == 0);
        dprintf(fd_logfile, "sent!\n");

        int header;
        rc = zsock_recv(req, "im", &header, NULL);
        assert(rc == 0);
        assert(header == TRANSFER_ACK);
        dprintf(fd_logfile, "received ACK!\n");

        zmsg_destroy(&msg);

        teardown_cell(&item);
        zsock_disconnect(req, "%s", endpoint);

        zsock_destroy(&req);
        free(endpoint);
        free(port);
    }
    return 1;
}


int send_work_timer_fn(zloop_t *loop, int timer_id, void *args) {
    struct msend_work_args *arg = (struct msend_work_args*) args;
    return work_sender(loop, timer_id, arg->stats, arg->ips, arg->q, arg->fd_logfile);
}
