#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <zmq.h>
#include <czmq.h>

#include <stdio.h>
#include <stdlib.h>

#include "../common/msg_headers.h"
#include "pstats.h"
#include "p_pull_push.h"
#include "p_pub_sub.h"
#include "../common/verbose.h"


extern size_t sub_message_count;
extern size_t subhashes_message_count;
static size_t push_message_count = 0;
static size_t pull_message_count = 0;

void psend_stats(zsock_t * const zocket,
                 const int id,
                 pStats const * const wqs,
                 const int port_start,
                 const int proxynumber,
                 const int fd_logfile) {
    zmsg_t *msg = create_stats_msg(id, wqs, port_start, proxynumber, fd_logfile);
    int rc = zsock_send(zocket, "m", msg);
    assert(rc == 0);
    zmsg_destroy(&msg);
}

void p_send_transfer_done(zsock_t * const push,
                          const int id,
                          pStats const * const wqs,
                          const int port_start,
                          const int proxynumber,
                          const int fd_logfile,
                          const uint64_t timestamp) {
    zmsg_t *msg = create_stats_msg(id, wqs, port_start, proxynumber, fd_logfile);
    ddprintf(VERBOSE_TRANSFER, fd_logfile, "{:type :P_TRANSFER_DONE, id %d, @time %lu}\n", id, timestamp);

    zsock_send(push, "ii8m", P_TRANSFER_DONE, id, timestamp, msg);
    zmsg_destroy(&msg);
}

void proxy_h_pull(zsock_t * const pull,
                  zsock_t * const push,
                  zsock_t * const pub,
                  pStats * const stats,
                  const int id,
                  const int thresh,
                  const int port_start,
                  const int fd_logfile,
                  const hashtrie trie,
                  const int proxynumber) {
    int rc;

    int dirty = 0;
    zmsg_t *msg;
    int header;

    rc = zsock_recv(pull, "im", &header, &msg);
    assert(rc == 0);
    pull_message_count++;

    switch (header) {
        case IS_HASH_MSG: {
            // TODO: test this
            dirty |= handle_stats(msg, header, thresh, stats, fd_logfile);

            zmsg_t *copy = zmsg_dup(msg);
            put_hashes(copy, trie, fd_logfile);
            rc = zsock_send(push, "im", P_IS_HASH_MSG, msg);
            zmsg_destroy(&copy);

            push_message_count++;
            assert(rc == 0);
            local_balance(pub, stats, thresh, port_start, fd_logfile, proxynumber);
            break;
        }
        case IS_RESULT_MSG:
            rc = zsock_send(push, "im", P_RESULT_MSG, msg);
            assert(rc == 0);
            break;
        case G_TRANSFER_DONE: {
            char *timestamp_s = zmsg_popstr(msg);
            uint64_t timestamp;
            rc = sscanf(timestamp_s, "%lu", &timestamp);
            assert(rc == 1);
            free(timestamp_s);
            dirty |= handle_stats(msg, header, thresh, stats, fd_logfile);
            local_balance(pub, stats, thresh, port_start, fd_logfile, proxynumber);
            p_send_transfer_done(push, id, stats, port_start, proxynumber, fd_logfile, timestamp);
            break;
        }
        case L_TRANSFER_DONE: {
            char *timestamp_s = zmsg_popstr(msg);
            ddprintf(VERBOSE_TRANSFER, fd_logfile, "{:type :local-transfer-done, :time %s}\n", timestamp_s);
            free(timestamp_s);
            // drop in next case
        }
        case IS_SMALL_STATS:
            dirty |= handle_stats(msg, header, thresh, stats, fd_logfile);
            local_balance(pub, stats, thresh, port_start, fd_logfile, proxynumber);
            break;
        default:
            assert(0);
    }
    zmsg_destroy(&msg);

    if (dirty) {
        zmsg_t *update = create_stats_msg(id, stats, port_start, proxynumber, fd_logfile);
        rc = zsock_send(push, "m", update);
        assert(rc == 0);
        push_message_count++;
        zmsg_destroy(&update);
    }
}



zsock_t *proxy_setup_pull(const int proxynumber) {
    char *endpoint = NULL;
    int rc = asprintf(&endpoint, "ipc:///tmp/zmq-proxy-%d-pushpull", proxynumber);
    assert(rc != -1);

    zsock_t *zocket = zsock_new_pull(endpoint);
    free(endpoint);
    return zocket;
}

zsock_t *proxy_setup_push(char const * const master_ip,
                          const int port_number) {
    char *endpoint = NULL;
    int rc = asprintf(&endpoint, "tcp://%s:%d", master_ip, port_number);
    assert(rc != -1);

    zsock_t *zocket = zsock_new_push(endpoint);
    free(endpoint);
    return zocket;
}

int proxy_pull_reader(zloop_t *loop, zsock_t *reader, void *args) {
    struct p_pull_args *arg = (struct p_pull_args*) args;
    
    proxy_h_pull(reader, arg->push, arg->pub, arg->stats,
                         arg->proxy_id, arg->thresh, arg->port_start,
                         arg->fd_logfile, arg->trie, arg->proxynumber);
    return 0;
}

void push_all_queuesizes(zsock_t * const push,
                         const int proxy_id,
                         pStats const * const stats,
                         const size_t iterations,
                         const int fd_logfile) {
    zmsg_t *msg = zmsg_new();
    int i;

    dprintf(fd_logfile, "{:submsgs %ld, :subhashes %ld, :iterations %ld, :pulled %ld, :pushed %ld, :qsizes [", sub_message_count, subhashes_message_count, iterations, pull_message_count, push_message_count);
    zmsg_addstrf(msg, "%lu", sub_message_count);
    zmsg_addstrf(msg, "%lu", subhashes_message_count);
    zmsg_addstrf(msg, "%lu", pull_message_count);
    zmsg_addstrf(msg, "%lu", iterations);
    zmsg_addstrf(msg, "%lu", push_message_count);
    for (i = 1; i <= stats->len; i++) {
        zmsg_addstrf(msg, "%lu", stats->queuesizes[i] & ~((size_t) 1 << 63));
        dprintf(fd_logfile, " %lu", stats->queuesizes[i] & ~((size_t) 1 << 63));
    }
    dprintf(fd_logfile, " ]}\n");


    int rc = zsock_send(push, "iim", PSTATS_QUEUES, proxy_id, msg);
    assert(rc == 0);
    zmsg_destroy(&msg);
}

int print_queuesize_timer(zloop_t *loop, int timer_id, void *args) {
    struct queue_printer_arg *arg = (struct queue_printer_arg*) args;
    push_all_queuesizes(arg->push, arg->proxy_id, arg->stats, arg->iterations, arg->fd_logfile);
    return 0;
}
