#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <zmq.h>
#include <czmq.h>

#include <stdlib.h>
#include <stdio.h>

#include "../common/assert_helper.h"
#include "../common/msg_headers.h"
#include "../common/magic_constants.h"
#include "pid.h"

static unsigned int id_counter = 0;

/* sets up an IPC socket to serve id requests from workers; 
 * its endpoint will be ipc:///tmp/zmq-proxy-proxynumber-id;
 * returns said socket */
zsock_t *proxy_setup_id_rep_zocket(const int proxynumber) {
    char *endpoint = NULL;
    int rc = asprintf(&endpoint, "ipc:///tmp/zmq-proxy-%d-id", proxynumber);
    assert(rc != -1);

    zsock_t *zocket = zsock_new_rep(endpoint);
    free(endpoint);
    return zocket;
}

/* tears down the entire id module */
void proxy_teardown_id(void) {
    id_counter = 0;
}

/* receives exactly one id request and replies with an id and a model */
void proxy_h_id(zsock_t * const rep, unsigned int proxy_id, zchunk_t *model) {
    (void) proxy_id;
    int cmd = 0;
    zmsg_t *msg = NULL;
    int rc = zsock_recv(rep, "im", &cmd, &msg);
    assert(rc == 0);

    switch (cmd) {
        case W_ID_REQUEST: {
            unsigned int id = ++id_counter;
            char *unique_template = NULL;

            rc = zsock_send(rep, "ic", id, model);
            free(unique_template);
            assert(id_counter < WORKERS_PER_PROXY);
            assert(rc == 0);
            break;
        }
        case PING:
            assert(msg);
            assert(zmsg_size(msg) == 1);
            rc = zsock_send(rep, "im", PONG, msg);
            break;
        default:
            assert_esf(0, "unknown header %d in proxy_h_id", cmd);
    }
    zmsg_destroy(&msg);
}

int proxy_id_reader(zloop_t *reactor, zsock_t *reader, void *arg) {
    struct proxy_h_id_args *args = (struct proxy_h_id_args*) arg;
    proxy_h_id(reader, args->proxy_id, args->model);
    return 0;
}


/* requests an id from tcp://master_ip:port_req,
 * transmits the own IP address,
 * returns the received id and the model
 * as well as known hashcodes */
struct pid_response proxy_request_id(char const * const master_ip,
                                     const int port_req,
                                     const int port_pub,
                                     char const * const own_ip,
                                     const int proxynumber) {
    char *endpoint = NULL;
    int rc = asprintf(&endpoint, "tcp://%s:%d", master_ip, port_req);
    assert(rc != -1);

    zsock_t *id_req = zsock_new_req(endpoint);
    free(endpoint);

    int req_header = (0 == strcmp(master_ip, own_ip)) && proxynumber == 0 ?
                     P_ID_REQ_TRIE_UP : P_ID_REQ;

    rc = zsock_send(id_req, "isi", req_header, own_ip, port_pub);
    assert(rc == 0);

    struct pid_response res = {0};
    rc = zsock_recv(id_req, "icbs", &(res.id), &(res.model), &(res.digest_block), &(res.blocksize), &(res.subscribe_hash_recv_to));
    assert(rc == 0);


    zsock_destroy(&id_req);
    return res;
}


