#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/shm_srv.h"
#include "../common/shm_common.h"
#include "../common/hashtrie.h"
#include "../common/magic_constants.h"
#include "../common/helpers.h"

#include "p_pull_push.h"
#include "p_pub_sub.h"
#include "pstats.h"
#include "pid.h"


// TCP port list:
// PORT              Proxy    Master  Content
// port_start + 0    PUSH     PULL    hashes (upstream)
// port_start + 1    SUB      PUB     commands
// port_start + 2    REQ      REP     IDs
// port_start + 3    -        PUB     hashes (downstream)
// port_start + 4    PUB      -       hashes (downstream)

size_t sub_message_count = 0;
size_t subhashes_message_count = 0;


int main(int argc, char *argv[]) {
    if (argc < 6 || argc > 7) {
        fprintf(stderr, "Usage: ./proxy <master IP> <port start> <own IP> <queue threshold> <logfile> [<proxynumber>]\n");
        return 1;
    }
    char *master_ip = argv[1];
    int port_start = atoi(argv[2]);
    char *own_ip = argv[3];
    int threshold = atoi(argv[4]);
    char *logfile_name = argv[5];
    int proxynumber = (argc == 7) ? atoi(argv[6]) : 0;

    zsys_set_io_threads(2);
    const int PORT_PUSH_PULL = port_start + 0;    
    const int PORT_SUB       = port_start + 1;
    const int PORT_REQ       = port_start + 2;
    const int PORT_PUB       = ppub_port(port_start, proxynumber);
                               
                               

    int fd_logfile = open(logfile_name, O_CREAT | O_RDWR | O_TRUNC, 0666);
    assert(fd_logfile != -1);
    dprintf(fd_logfile, "{:status :online}\n");

    struct semap_env senv = create_semaphores(proxynumber);
    hashtrie trie = get_trie(senv, HASHTRIE_KEY_START + 40000 * proxynumber);
    init_trie_meta(trie.shm);
    dprintf(fd_logfile, "{:status :trie-online}\n");

    dprintf(fd_logfile, "{:status :request-id}\n");
    struct pid_response resp = proxy_request_id(master_ip, PORT_REQ, PORT_PUB, own_ip, proxynumber);
    dprintf(fd_logfile, "{:status :got-id, :id %d, :subscribing-to: %s}\n", resp.id, resp.subscribe_hash_recv_to);

    size_t i;
    for (i = 0; i < resp.blocksize; i += HASH_INFO) {
        trie_put(trie,
                 resp.digest_block + i,
                 resp.digest_block + i + HASH_LEN);
    }
    assert(i == resp.blocksize);
    free(resp.digest_block);
    dprintf(fd_logfile, "{:status :hashes-recvd}\n");


//    zsys_set_sndhwm(0); // let's not
    zsys_set_rcvhwm(0);

    zsock_t *pull = proxy_setup_pull(proxynumber);
    zsock_t *push = proxy_setup_push(master_ip, PORT_PUSH_PULL);


    zsock_t *pub_local = proxy_setup_pub(proxynumber);
    zsock_t *pub_global = proxy_setup_pub_global(PORT_PUB);

    char *subscribe_to = resp.subscribe_hash_recv_to; // fully formatted endpoint for hashes
    zsock_t *sub = proxy_setup_sub(subscribe_to);
    zsock_t *sub_hashes = proxy_setup_sub2(master_ip, PORT_SUB);


    assert(zsock_rcvhwm(pull) == 0);
    //assert(zsock_sndhwm(pub_local) == 0);
    assert(zsock_rcvhwm(sub) == 0);
    assert(zsock_rcvhwm(sub_hashes) == 0);
    //assert(zsock_sndhwm(push) == 0);

    zsock_t *id_rep = proxy_setup_id_rep_zocket(proxynumber);

    //int fd_logfile = open("/dev/null", O_CREAT | O_RDWR | O_TRUNC, 0666);

    dprintf(fd_logfile, "{:type :start, :ip \"%s\", :endpoint-pub \"tcp://*:%d\", :endpoint-sub-master \"tcp://%s:%d\", :endpoint-sub-whoever \"%s\"}\n",
            own_ip, PORT_PUB, master_ip, PORT_SUB, subscribe_to);
    free(subscribe_to);

    pStats stats = {0};



    zsys_catch_interrupts(); // this is for internal use

    // my own fancy reactor
    zpoller_t *poller_pull       = zpoller_new(pull, NULL);
    zpoller_t *poller_sub        = zpoller_new(sub, NULL);
    zpoller_t *poller_sub_hashes = zpoller_new(sub_hashes, NULL);
    zpoller_t *poller_rep        = zpoller_new(id_rep, NULL);
    int64_t last_time = zclock_mono();
    size_t iterations = 0;
    while (!zsys_interrupted) {
        if (zpoller_wait(poller_pull, 0)) {
            proxy_h_pull(pull, push, pub_local, &stats, resp.id, threshold, port_start, fd_logfile, trie, proxynumber);
        }
        if (zpoller_wait(poller_sub, 0)) {
            int rc = handle_sub(sub, pub_local, pub_global, push, &stats, resp.id, threshold, port_start, fd_logfile, trie, proxynumber);
            if (rc == -1) { // this is the only socket that may make us shut down
                break;
            }
        }
        if (zpoller_wait(poller_sub_hashes, 0)) {
            int rc = handle_sub(sub_hashes, pub_local, pub_global, push, &stats, resp.id, threshold, port_start, fd_logfile, trie, proxynumber);
            if (rc == -1) { // this is the only socket that may make us shut down
                break;
            }
        }
        if (zpoller_wait(poller_rep, 0)) {
            proxy_h_id(id_rep, resp.id, resp.model);
        }

        int64_t tmp = zclock_mono();
        if (tmp - last_time >= 500) {
            last_time = tmp;
            push_all_queuesizes(push, resp.id, &stats, iterations, fd_logfile);
        }
        iterations++;
    }


    if (stats.queuesizes) {
        free(stats.queuesizes);
        stats.queuesizes = 0;
        stats.len = 0;
        stats.empty_workers = 0;
        stats.balanced_workers = 0;
        stats.full_workers = 0;
        stats.in_transfer = 0;
    }
    
    zpoller_destroy(&poller_pull);
    zpoller_destroy(&poller_sub);
    zpoller_destroy(&poller_sub_hashes);
    zpoller_destroy(&poller_rep);

    zsock_destroy(&id_rep);
    zsock_destroy(&pub_local);
    zsock_destroy(&pub_global);
    zsock_destroy(&sub);
    zsock_destroy(&sub_hashes);
    zsock_destroy(&pull);
    zsock_destroy(&push);

    zchunk_destroy(&(resp.model));
    proxy_teardown_id();
    close(fd_logfile);

    teardown_hashtrie(&trie, 1, proxynumber);

    return 0;
}
