#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/msg_headers.h"
#include "../common/magic_constants.h"
#include "../common/helpers.h"

#include "pstats.h"
#include "../common/verbose.h"


void print_wqs(pStats const * const wqs, int fd_logfile);

int mark_transfer_begin(pStats * const wqs, const int id, const size_t thresh, int fd_logfile) {
    size_t * const queuesizes = wqs->queuesizes;
    if (queuesizes[id] < ((size_t) 1 << 63)) {
        size_t old = queuesizes[id];
        queuesizes[id] = queuesizes[id] | ((size_t) 1 << 63);
        if (old == 0) {
            wqs->empty_workers--;
            ddprintf(VERY_VERBOSE_TRANSFER, fd_logfile, "{:type :worker-state-change, :id %d, :qs 0, :old-state :empty, :new-state :in-transfer}\n", id);
        } else if (old < thresh) {
            wqs->balanced_workers--;
            ddprintf(VERY_VERBOSE_TRANSFER, fd_logfile, "{:type :worker-state-change, :id %d, :qs %ld, :old-state :balanced, :new-state :in-transfer}\n", id, old);
        } else {
            wqs->full_workers--;
            ddprintf(VERY_VERBOSE_TRANSFER, fd_logfile, "{:type :worker-state-change, :id %d, :qs %ld, :old-state :full, :new-state :in-transfer}\n", id, old);
        }
        wqs->in_transfer++;
        print_wqs(wqs, fd_logfile);
        return 1;
    }
    return 0;
}

// TODO: check how this behaves with global transfer acks
int mark_transfer_end(pStats * const wqs, const int id, const size_t thresh, int fd_logfile) {
    size_t * const queuesizes = wqs->queuesizes;
    if (queuesizes[id] >= ((size_t) 1 << 63)) {
        queuesizes[id] = queuesizes[id] & ~((size_t) 1 << 63);
        size_t qs = queuesizes[id];
        if (qs == 0) {
            ddprintf(VERY_VERBOSE_TRANSFER, fd_logfile, "{:type :worker-state-change, :id %d, :qs 0, :old-state :in-transfer, :new-state :empty}\n", id);
            wqs->empty_workers++;
        } else if (qs < thresh) {
            ddprintf(VERY_VERBOSE_TRANSFER, fd_logfile, "{:type :worker-state-change, :id %d, :qs %ld, :old-state :in-transfer, :new-state :balanced}\n", id, qs);
            wqs->empty_workers++;
            wqs->balanced_workers++;
        } else {
            ddprintf(VERY_VERBOSE_TRANSFER, fd_logfile, "{:type :worker-state-change, :id %d, :qs %ld, :old-state :in-transfer, :new-state :full}\n", id, qs);
            wqs->full_workers++;
        }
        wqs->in_transfer--;
        print_wqs(wqs, fd_logfile);
        return 1;
    }
    return 0;
}

size_t is_in_transfer(pStats const * const wqs, const int id) {
    size_t const * const queuesizes = wqs->queuesizes;
    return queuesizes[id] & ((size_t) 1 << 63);
}

// TODO: special handling of workers that did not connect yet
int extend_queuesizes(pStats * const wqs, int id) {
    if (id > wqs->len) {
        size_t old_len = wqs->len;
        wqs->len = id;
        wqs->queuesizes = realloc(wqs->queuesizes, (id + 1) * sizeof(size_t));
        int i;
        for (i = old_len + 1; i <= id; i++) {
            wqs->queuesizes[i] = 0;
            wqs->empty_workers++;
        }
        return 1;
    }
    return 0;
}

void print_wqs(pStats const * const wqs, int fd_logfile) {
    ddprintf(VERBOSE_TRANSFER, fd_logfile, "{:type :queue-digest, :empty %d, :balanced %d, :full %d, :in-transfer %d}\n",
                                           wqs->empty_workers, wqs->balanced_workers, wqs->full_workers, wqs->in_transfer);
}

int update_qs(pStats * const wqs,
              const int id,
              const size_t queuesize,
              const size_t thresh,
              int fd_logfile) {
    int dirty = extend_queuesizes(wqs, id);
    if (is_in_transfer(wqs, id)) {
        ddprintf(VERY_VERBOSE_TRANSFER, fd_logfile, "{:type :update-qs, :worker-in-transfer %d, :info :dirty}\n", id);
        return dirty;
    }

    size_t old = wqs->queuesizes[id];
    // keep the in-transfer bit from the old value
    wqs->queuesizes[id] = (old & ((size_t) 1 << 63)) | queuesize;
    old = old & ~((size_t) 1 << 63);

    if (old == 0 && queuesize > 0) {
        ddprintf(VERY_VERBOSE_TRANSFER, fd_logfile, "{:type :worker-state-change, :id %d, :qs %ld, :old-state :in-transfer, :new-state %s}\n",
                                                    id, queuesize, queuesize < thresh ? ":balanced" : ":full");
        dirty = 1;
        wqs->empty_workers--;
        if (queuesize < thresh) {
            wqs->balanced_workers++;
        } else {
            wqs->full_workers++;
        }
        //print_wqs(wqs, fd_logfile);
    } else if (old >= thresh) {
        if (queuesize == 0) {
            //dprintf(fd_logfile, "old: %ld (full) --> new: 0 (balanced)\n", old);
            dirty = 1;
            wqs->full_workers--;
            wqs->empty_workers++;
            //print_wqs(wqs, fd_logfile);
        } else if (queuesize < thresh) {
            //dprintf(fd_logfile, "old: %ld (full) --> new: %ld (balanced)\n", old, queuesize);
            dirty = 1;
            wqs->full_workers--;
            wqs->balanced_workers++;
            //print_wqs(wqs, fd_logfile);
        }
    } else if (old) { // old : (0, thresh)
        if (queuesize == 0) {
            //dprintf(fd_logfile, "old: %ld (balanced) --> new: 0 (empty)\n", old);
            dirty = 1;
            wqs->balanced_workers--;
            wqs->empty_workers++;
            //print_wqs(wqs, fd_logfile);
        } else if (queuesize >= thresh) {
            //dprintf(fd_logfile, "old: %ld (balanced) --> new: %ld (full)\n", old, queuesize);
            dirty = 1;
            wqs->balanced_workers--;
            wqs->full_workers++;
            //print_wqs(wqs, fd_logfile);
        }
    }

    return dirty;
}

void dump_to_file(const int fd,
                  const int worker_id,
                  const size_t queuesize,
                  char const * const tProlog,
                  char const * const timestamp,
                  char const * const statenmbr,
                  char const * const readable_digest) {
    ddprintf(VERBOSE_PROXY_STATES, fd, "{:type :checked, :id %d, :queuesize %lu, :tProlog %s, :timestamp %s, :local-nmbr %s, :digest :%s}\n",
                                                             worker_id,     queuesize,    tProlog,       timestamp,      statenmbr,   readable_digest);
}

// don't destroy the message
int handle_stats(zmsg_t * const msg,
                 const int header,
                 const size_t thresh,
                 pStats * const queuesizes,
                 const int fd_logfile) {
    char *worker_id = zmsg_popstr(msg);
    int id = -1;
    int rc = sscanf(worker_id, "%d", &id);
    assert(rc == 1);
    free(worker_id);

    // FIXME: this should be brought into standard network byte order
    size_t *qs = (size_t*) zmsg_popstr(msg);
    int dirty = 0;

    if (header == L_TRANSFER_DONE) {
        dirty |= mark_transfer_end(queuesizes, id, thresh, fd_logfile);
    }

    dirty |= update_qs(queuesizes, id, *qs, thresh, fd_logfile);

    if (header == IS_HASH_MSG) {
        assert(zmsg_size(msg) > 4);
        char * const tProlog = zmsg_popstr(msg);
        char * const timestamp = zmsg_popstr(msg);
        char * const statenmbr = zmsg_popstr(msg);
        char * const readable_digest = zmsg_popstr(msg);

        dump_to_file(fd_logfile, id, *qs, tProlog, timestamp, statenmbr, readable_digest);

        free(tProlog);
        free(timestamp);
        free(statenmbr);
        free(readable_digest);
    }
    free(qs);

    return dirty;
}


/* returns the index of the greatest entry in wqs
 * which is not marked as in transfer,
 * or -1 if all are in transfer or empty */
int get_max_not_transfer_queue(pStats const * const wqs) {
    size_t max = 0;
    int i;
    int res = -1;
    for (i = 1; i <= wqs->len; i++) {
        if (!is_in_transfer(wqs, i) && wqs->queuesizes[i] > max) {
            max = wqs->queuesizes[i];
            res = i;
        }
    }
    return res;
}

/* returns the index of the greatest queuesize in wqs,
 * or -1 if all are empty */
int get_max_queue(pStats const * const wqs) {
    size_t max = 0;
    int i;
    int res = -1;
    for (i = 1; i <= wqs->len; i++) {
        if ((wqs->queuesizes[i] & ~((size_t) 1 << 63)) > max) {
            max = wqs->queuesizes[i];
            res = i;
        }
    }
    return res;
}

/* requires at least one index in wqs where the value in zero.
 * returns the first index of a zero in wqs */
int get_empty_queue(pStats const * const wqs) {
    assert(wqs->empty_workers > 0);
    int i;
    for (i = 1; i <= wqs->len; i++) {
        if (!wqs->queuesizes[i]) {
            return i;
        }
    }
    assert(0);
}


zmsg_t *create_stats_msg(const int id,
                         pStats const * const wqs,
                         const int port_start,
                         const int proxynumber,
                         const int fd_logfile) {
    zmsg_t *msg = zmsg_new();
    int header = PSTATS_HAS_PMS;
    if (wqs->empty_workers && !wqs->full_workers
            && !wqs->balanced_workers && !wqs->in_transfer) {
        header = PSTATS_COMPLETELY_EMPTY;
    } else if (wqs->empty_workers && !wqs->full_workers) {
        header = PSTATS_HAS_EMPTY;
    } else if (wqs->full_workers && !wqs->empty_workers) {
        header = PSTATS_HAS_FULL;
    }

    zmsg_addstrf(msg, "%d", header);
    zmsg_addstrf(msg, "%d", id);
    ddprintf(VERBOSE_TRANSFER, fd_logfile, "{:type :state-msg, :header %d, :id %d" , header, id);

    switch(header) {
        case PSTATS_HAS_PMS:
            break;
        case PSTATS_HAS_FULL:
            zmsg_addstrf(msg, "%d", wqs->full_workers);
            ddprintf(VERBOSE_TRANSFER, fd_logfile, ", :full %d", wqs->full_workers);
            break;
        case PSTATS_COMPLETELY_EMPTY:
        case PSTATS_HAS_EMPTY: {
            zmsg_addstrf(msg, "%d", wqs->empty_workers);
            ddprintf(VERBOSE_TRANSFER, fd_logfile, ", :empty %d, :ports [", wqs->empty_workers);
            int i;
            for (i = 1; i <= wqs->len; i++) {
                // FIXME: the second part of the condition is implicit in the first one
                if (wqs->queuesizes[i] == 0 && !is_in_transfer(wqs, i)) { 
                    zmsg_addstrf(msg, "%d", wsink_port(port_start, proxynumber, i));
                    ddprintf(VERBOSE_TRANSFER, fd_logfile, "%d ", wsink_port(port_start, proxynumber, i));
                }
            }
            ddprintf(VERBOSE_TRANSFER, fd_logfile, "]");
            break;
        }
        default:
            assert(0);
    }
    ddprintf(VERBOSE_TRANSFER, fd_logfile, "}\n");
    return msg;
}



/* returns the amount of possible work transfers based on wqs and thresh */
int proxy_can_balance(pStats const * const wqs, const size_t thresh) {
    int count_has = 0;
    int count_need = 0;
    int i;
    for (i = 1; i <= wqs->len; i++) {
        if (!is_in_transfer(wqs, i) && wqs->queuesizes[i] >= thresh) {
            count_has++;
        }

        if (!is_in_transfer(wqs, i) && wqs->queuesizes[i] == 0) {
            count_need++;
        }
    }

    if (count_need < count_has) {
        return count_need;
    }
    return count_has;
}

