#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/queue.h"
#include "../common/msg_headers.h"
#include "../common/verbose.h"

#include "wpush.h"
#include "wctrl.h"

#define MAXTERMS 100 // TODO: set this from somewhere else

extern int g_generate_successors;

/* adds every other work-item from the queue to the message provided,
 * but no more than max;
 * the format will be
 *     <digest>
 *     <term>
 *     <digest>
 *     <term>
 *     ...
 */
void add_wps(zmsg_t * const msg, wQueue * const q, const size_t max, const int fd_logfile) {
    const int qlen = q_size(q);
    int i;
    for (i = 0; i < qlen; i++) {
        tCell *c = dequeue(q);
        assert(c);

        if (i % 2 == 0 && i/2 <= max) {
            if (VERBOSE_WORKER_STATES) {
                char buf[READABLE_HASH_LEN];
                print_readable_digest((unsigned char*) c->digest, buf);
                dprintf(fd_logfile, "{:type :dequeue-for-transfer, :hash :%s}\n", buf);
            }
            zmsg_addmem(msg, c->digest, HASH_LEN);
            zmsg_addmem(msg, c->term, c->term_length);
            teardown_cell(&c);
        } else {
            enqueue(q, c->term, c->term_length, c->digest);
            free(c);
        }
    }

}

/* appends the entire queue to the message provided;
 * format is the same as described for add_wps.
 * if the queue is empty, only one frame is added 
 * (containing the string "retry") */
void add_entire_queue(zmsg_t * const msg, wQueue * const q, const int fd_logfile) {
    const int qlen = q_size(q);

    int i;
    for (i = 0; i < qlen; i++) {
        tCell *c = dequeue(q);
        assert(c);
        if (VERBOSE_WORKER_STATES) {
            char buf[READABLE_HASH_LEN];
            print_readable_digest((unsigned char*) c->digest, buf);
            dprintf(fd_logfile, "{:type :dequeue-for-transfer, :hash :%s}\n", buf);
        }
        zmsg_addmem(msg, c->digest, HASH_LEN);
        zmsg_addmem(msg, c->term, c->term_length);
        teardown_cell(&c);
    }
}

/* connects via a ZMQ_REQ zocket to endpoint and
 * sends up to max work items from the queue, if header is a TRANSFER_TO one
 * or sends all items from the queue, if header is P_TRANSFER_ALL */
/* TODO: waits for acknowledgement */
void send_work(char const * const endpoint,
               wQueue * const q,
               const size_t max,
               const int header,
               const uint64_t timestamp,
               const int fd_logfile) {
    zsock_t *zock = zsock_new_req(endpoint);
    assert(zock);
    
    zmsg_t *work = zmsg_new();
    if (header == L_TRANSFER_TO || header == G_TRANSFER_TO) {
        add_wps(work, q, max, fd_logfile);
    } else if (header == P_TRANSFER_ALL) {
        add_entire_queue(work, q, fd_logfile);
    } else {
        assert(0);
    }

    int transfer_header;
    switch (header) {
        case L_TRANSFER_TO:
            transfer_header = LOCAL_TRANSFER;
            break;
        case G_TRANSFER_TO:
        case P_TRANSFER_ALL:
            transfer_header = GLOBAL_TRANSFER;
            break;
        default:
            assert(0);

    }

    dprintf(fd_logfile, "{:type :transfer, :endpoint %s}\n", endpoint);

    int rc = zsock_send(zock, "i8m", transfer_header, timestamp, work);
    assert(rc == 0);
    dprintf(fd_logfile, "{:type :transfer, :status :done}\n");

    int peer_ack_header;
    uint64_t timestamp_echo;
    rc = zsock_recv(zock, "i8m", &peer_ack_header, &timestamp_echo, NULL);
    assert(rc == 0);
    assert(peer_ack_header == TRANSFER_ACK);
    assert(timestamp_echo == timestamp);


    zmsg_destroy(&work);
    rc = zsock_disconnect(zock, "%s", endpoint);
    assert(rc == 0);
    zsock_destroy(&zock);
    dprintf(fd_logfile, "{:type :transfer, :status :disconnected, :endpoint %s}\n", endpoint);
}


void do_transfer(zsock_t * const push,
                 char const * const endpoint,
                 wQueue * const q,
                 const size_t max,
                 const int header,
                 const int worker_id,
                 const uint64_t timestamp,
                 const int fd_logfile) {
    send_work(endpoint, q, max, header, timestamp, fd_logfile);
    int ack_header;
    switch (header) {
        case L_TRANSFER_TO:
            ack_header = L_TRANSFER_DONE;
            break;
        case G_TRANSFER_TO:
            ack_header = G_TRANSFER_DONE;
            break;
        default:
            assert(0);
    }
    signal_transfer_done(push, ack_header, worker_id, q, timestamp);
}

/* receives a control message on zocket_ctrl;
 * additional parameters are the zocket used for push-pull communication,
 * the worker's queue and its ID */
int wrecv_ctrl(zsock_t * const zocket_ctrl,
               zsock_t * const zocket_push,
               wQueue * const q,
               const int worker_id,
               const int fd_logfile) {
    int header;
    zmsg_t *cmd;

    int rc = zsock_recv(zocket_ctrl, "im", &header, &cmd);
    assert(rc == 0);

    switch (header) {
        case P_TERM:
            dprintf(fd_logfile, "{:type :info, :recv-ctrl :P_TERM}\n");
            zmsg_destroy(&cmd);
            return -1;
        case G_TRANSFER_TO:
        case L_TRANSFER_TO: {
            char *id_s = zmsg_popstr(cmd);
            int id = atoi(id_s);
            char *timestamp_s = zmsg_popstr(cmd);
            uint64_t timestamp;
            rc = sscanf(timestamp_s, "%lu", &timestamp);
            assert(rc == 1);
            free(timestamp_s);
            if (id == worker_id) {
                char *endpoint = zmsg_popstr(cmd);
                dprintf(fd_logfile, "{:type :info, :recv-ctrl :TRANSFER_TO, :endpoint %s}\n", endpoint);
                // TODO: should this message include a maximum amount of packages?
                do_transfer(zocket_push, endpoint, q, MAXTERMS, header, worker_id, timestamp, fd_logfile);
                free(endpoint);
            }
            free(id_s);
            zmsg_destroy(&cmd);

            break;
        }
        case P_STOP_SUCC:
            dprintf(fd_logfile, "{:type :info, :recv-ctrl :P_STOP_SUCC}\n");
            g_generate_successors = 0;
            break;
        default:
            assert(0);
    }

    return 0;
}


int wrecv_ctrl_reader(zloop_t *loop,
                      zsock_t *reader,
                      void *args) {
    struct wrecv_ctrl_arg *arg = (struct wrecv_ctrl_arg*) args;
    return wrecv_ctrl(reader, arg->push, arg->q, arg->worker_id, arg->fd_logfile);
}


zsock_t *init_wsub(const int proxynmbr) {
    char *endpoint_sub;
    int rc = asprintf(&endpoint_sub,"ipc:///tmp/zmq-proxy-%d-pubsub", proxynmbr);
    assert(rc != -1);

    zsock_t *res = zsock_new_sub(endpoint_sub, "");
    free(endpoint_sub);
    return res;
}

