% (c) 2014-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(atelierb_provers_interface, [prove_predicate/3,
                                       disprove_predicate/3,disprove_predicate/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,smt_solvers).
:- module_info(description,'This module provides an interface to the Atelier-B provers.').

:- use_module(bsyntaxtree,[create_negation/2,
                           conjunct_predicates/2,
                           create_implication/3,
                           conjunction_to_list/2]).
:- use_module(translate, [translate_bexpression/2,
                          translate_subst_or_bexpr_in_mode/3,
                          set_print_type_infos/1]).
:- use_module(system_call, [system_call/5]).
:- use_module(preferences, [get_preference/2,
                            temporary_set_preference/3,
                            reset_temporary_preference/2]).
:- use_module(tools, [read_string_from_file/2]).
:- use_module(error_manager, [get_error/2, add_error/2, add_error/3]).
:- use_module(debug, [debug_println/2]).

prove_predicate(L,P,R) :- %debug_println(9,prove_predicate(L,P,R)),
    (prove_predicate_aux(L,P,Res)
      -> (get_error(system_call,_) -> R = error ; R = Res)
      ;  add_error(atelierb_provers_interface,
         'Call to Atelier-B Provers failed. Be sure to set ATELIERB_KRT_PATH preference correctly.'),
         fail).
prove_predicate_aux(ListOfHypotheses,Predicate,Result) :-
    set_print_type_infos(all),

    conjunct_predicates(ListOfHypotheses,ConjunctOfHypotheses),
    create_implication(ConjunctOfHypotheses,Predicate,Sequent),

    get_preference(path_to_atb_krt,PathToKrt),
    debug_println(9,path_to_atb_krt(PathToKrt)),
    create_temp_file(ml,Sequent,TempFilePathML,ResultFilePathML),
    debug_println(9,created_temp_file(ResultFilePathML)),
    call_ml(PathToKrt,TempFilePathML,ResultFilePathML,MLResult),
    (MLResult = proved
     -> Result = MLResult
     ;  create_temp_file(pp,Sequent,TempFilePathPP,ResultFilePathPP),
        call_pp(PathToKrt,TempFilePathPP,ResultFilePathPP,Result)),
    set_print_type_infos(none).

disprove_predicate(Predicate,Result) :-
    conjunction_to_list(Predicate,Hyps),
    Falsity = b(equal(b(boolean_true,boolean,[]),b(boolean_false,boolean,[])),pred,[]),
    prove_predicate(Hyps,Falsity,Result).
disprove_predicate(ListOfHypotheses,Predicate,Result) :-
    create_negation(Predicate,NegPredicate),
    conjunct_predicates(NegPredicate,Conjunct),
    prove_predicate(ListOfHypotheses,Conjunct,Result).

create_temp_file(ml,Sequent,FileName,OutputFileName) :-
    % generate identifiers to remove primes, etc.
    temporary_set_preference(bugly_pp_scrambling,true,Chng),
    temp_file(S,FileName),
    format(S,'THEORY Lemma;Unproved IS\n',[]),
    translate_bexpression(Sequent,PPSequent),
    format(S,'~w\n',PPSequent),
    temp_file(STemp,OutputFileName),
    format(S, 'WHEN Force IS (0;1;2;3) WHEN FileOut IS "~w"\n', [OutputFileName]),
    format(S, 'WHEN Options IS ? & ? & ? & OK & "" & dummy & KO\nEND\n', []),
    close(S),
    close(STemp),
    reset_temporary_preference(bugly_pp_scrambling,Chng).

create_temp_file(pp,Sequent,FileName,OutputFileName) :-
    % generate identifiers to remove primes, etc.
    temporary_set_preference(bugly_pp_scrambling,true,Chng),
    temp_file(S,FileName),
    temp_file(STemp,OutputFileName),
    format(S,'Flag(FileOn("~w")) & Set(toto | ',[OutputFileName]),
    translate_subst_or_bexpr_in_mode(atelierb_pp,Sequent,PPSequent),
    format(S, '~w )\n', [PPSequent]),
    close(S),
    close(STemp),
    reset_temporary_preference(bugly_pp_scrambling,Chng).

temp_file(S,Filename) :-
    open(temp('atbprovers_temp_file'),write,S,[if_exists(generate_unique_name)]),
    stream_property(S, file_name(Filename)).

call_ml(Krt,TempFilePath,ResultFilePath,Result) :-
    ml_path(Krt,MLKin),
    debug_println(9,calling_ml(Krt,MLKin)),
    system_call(Krt,['-a','m1500000','-p','rmcomm','-b',MLKin,TempFilePath],_,ErrorTextAsCodeList,_),
    debug_println(9,result_err(ErrorTextAsCodeList)),
    (ErrorTextAsCodeList = []
     -> read_string_from_file(ResultFilePath,String),
        ml_result(String,Result)
     ;  format('ML Error Result: ~s~n.',[ErrorTextAsCodeList]),
        Result = error).

call_pp(Krt,TempFilePath,ResultFilePath,Result) :-
    pp_path(Krt,PPKin),
    debug_println(9,calling_pp(Krt,PPKin)),
    system_call(Krt,['-p','rmcomm','-b',PPKin,TempFilePath],_,ErrorTextAsCodeList,_),
    (ErrorTextAsCodeList = []
     -> read_string_from_file(ResultFilePath,String),
        pp_result(String,Result)
     ;  format('PP Error Result: ~s~n',[ErrorTextAsCodeList]),
        Result = error).

ml_result(String,proved) :-
    append("THEORY Etat IS Proved",_,String), !.
ml_result(_,unproved).
pp_result(String,proved) :-
    append("SUCCES",_,String), !.
pp_result(_,unproved).

ml_path(PathToKrt,PathToML) :- krt_path_to_kin(PathToKrt,"ML.kin",PathToML).
pp_path(PathToKrt,PathToML) :- krt_path_to_kin(PathToKrt,"PP.kin",PathToML).

krt_path_to_kin(PathToKrt,KINEXT,PathToML) :-
    atom_codes(PathToKrt,PathString),
    (append(BasePath,"krt",PathString)
     -> append(BasePath,KINEXT,MLString),
        atom_codes(PathToML,MLString)
     ; add_error(atelierb_provers_interface,'ATELIERB_KRT_PATH does not point to a krt binary: ',PathToKrt),
       fail).
