% (c) 2013-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(b_expression_sharing,
          [ apply_expression_sharing_to_machine/2
          , create_initial_expression_cache/1
          , find_common_subexpressions/4
          , cse_optimize_predicate/2
          , cse_optimize_substitution/2
          , expand_cse_lazy_lets/2 ,expand_cse_lazy_lets_if_cse_enabled/2
          , is_lazy_let_identifier/1
          ]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(avl)).

:- use_module(tools).
:- use_module(self_check).
:- use_module(error_manager).
:- use_module(debug).

:- use_module(bsyntaxtree).
:- use_module(bmachine_structure).

:- use_module(preferences,[get_preference/2]).

expression_sharing_enabled :-
    get_preference(use_common_subexpression_elimination,true).

create_subcache(Id,cache(Id,Avl)) :- empty_avl(Avl).

create_initial_expression_cache(A) :-
    expression_sharing_enabled,!,
    A=[Cache],create_subcache(top,Cache).
create_initial_expression_cache(disabled).


apply_expression_sharing_to_machine(InMachine,OutMachine) :-
    change_sections(InMachine,InExprs,OutExprs,OutMachine),
    create_stores(0,EmptyStores),
    remove_implicit_negation_l(InExprs,NormedExprs),
    find_common_subexpressions_l(NormedExprs,EmptyStores,true,machine,1,OutExprs,Stores,_Stripped,_UsedIds,_UsedCSEIds),
    ground_reference_counters(Stores),!.

change_sections(InMachine,InExprs,OutExprs,OutMachine) :-
    change_section2([constraints,properties,invariant,
                     assertions,operation_bodies],
                    InMachine,InExprs,OutExprs,OutMachine).
change_section2([],Machine,[],[],Machine).
change_section2([Sec|Rest],InMachine,InExprs,OutExprs,OutMachine) :-
    select_section_texprs(Sec,LInExprs,LOutExprs,InMachine,InterMachine),
    append(LInExprs,RInExprs,InExprs),
    append(LOutExprs,ROutExprs,OutExprs),
    change_section2(Rest,InterMachine,RInExprs,ROutExprs,OutMachine).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(bsyntaxtree).

% TESTS: x+y >100 & x+y <99 now fails deterministically in REPL with CSE

% (x+y)+(x+y)=(x+y)

% Note: we should detect if we are at the invariant level before calling cse_optimize_predicate: do not mess up Rodin rodinpos proof information or syntactically_relevance info

% new predicate to optimize an operation body; not yet used
:- public cse_optimize_operation/1.
:- use_module(bmachine,[b_top_level_operation/1,b_get_machine_operation/4]).
cse_optimize_operation(OpName) :-
   b_top_level_operation(OpName),
   b_get_machine_operation(OpName,_Results,_Parameters,Body),
   print(optimizing(OpName)),nl,
   cse_optimize_substitution(Body,_Res).

% no difference thus far with cse_optimize_predicate
cse_optimize_substitution(Subst,Res) :- cse_optimize_predicate(Subst,Res),
   nl,print('FINISHED CSE: '),nl, translate:print_subst(Res),nl.

cse_optimize_predicate(Pred,Res) :- already_cse_optimized(Pred),
   !,
   Res=Pred.
cse_optimize_predicate(Pred,Res) :-
   cse_print_check('Running Common Sub Expressions (CSE) Optimization on ', Pred),
   expand_cse_lazy_lets(Pred,ExpandedPred), % TO DO: see if we can do this more efficiently; do we need it at all ??
   cse_print_check('Expanded Expression: ', ExpandedPred),
   % Note: we can still have unexpanded lazy-lets hanging around in ExpandedPred from outer scopes
   find_maximum_lazy_let_id(ExpandedPred,MaxID), % find maximum ID
   StartNr is MaxID+1, % we will start numbering from there to avoid clashes with existing lazy-lets
   find_common_subexpressions(ExpandedPred,StartNr,true,R,_Refs),
   cse_print_check('Found common sub expressions: ',R),

   % print('Common: '),nl, translate:nested_print_bexpr_or_subst(R),nl,nl,nl,
   bup_init_acc(InitialAcc),
   reset_let_dependency_info,
   transform_bexpr_with_bup_accs(b_expression_sharing:bup_insert_lazy_lookups,R,R2,InitialAcc,_BUPAcc),
   %portray_avl(BUPAcc),nl,
   %print_bexpr_or_subst(R2),nl,nl,
    %  tools_printing:trace_print(R2,6),nl,nl,
   empty_avl(Empty),
   avl_store('$maximum_unregistered_lazy_let_id',Empty,MaxID,TDAcc),
   top_down_gen_lazy_lets(R2,TDAcc,OptimizedPred),

   %cse_print_check('Optimized Expression Result: ',OptimizedPred), % info fields not yet set
   transform_bexpr(b_expression_sharing:cse_cleanup_infos,OptimizedPred,COPred),
   cse_print_check('Cleaned up Optimized Expression Result: ',COPred),

    % nl,print('FINISHED CSE: '),nl, translate:nested_print_bexpr_or_subst(COPred),nl, %portray_avl(LastAcc),nl,nl,
   !,
   mark_as_optimized(COPred,Res).
cse_optimize_predicate(P,Res) :-
   add_internal_error('CSE FAILED: ',cse_optimize_predicate(P,Res)),
   Res=P.

already_cse_optimized(b(_E,_T,I)) :- member(cse_optimized,I).
mark_as_optimized(b(E,T,I),b(E,T,I2)) :-
   append(I,[cse_optimized],I2).

:- use_module(translate,[print_bexpr_or_subst/1]).
cse_print_check(Msg,Expr) :- debug:debug_mode(on), %%
   !,
   print(Msg), print_functor(Expr),
   print_bexpr_or_subst(Expr), nl,
   cse_check_ast(Msg,Expr).
cse_print_check(_,_).

print_functor(b(E,T,_)) :- functor(E,F,A), format('Type: ~w, Functor: ~w/~w~n',[T,F,A]).
print_functor(E) :- functor(E,F,A), format('Unwrapped, Functor: ~w/~w~n',[F,A]).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_safe_mode,true)).
% check the AST (should be removed in production version of ProB):
cse_check_ast(PP,Pred) :-
   debug_print(4,'Checking AST : '), debug_println(4,PP), % translate:print_bexpr_or_subst(Pred),nl,
   check_ast(true,Pred).

%cse_check_ast_complete(PP,Pred) :- cse_check_ast(PP,Pred),
%   debug_println(20,checking_ast_complete(PP)),
%   % currently will generate spurious warnings when the term to start with has @identifiers introduced by LETs higher up
%   map_over_bexpr_top_down_acc(b_expression_sharing:check_lazy_let_vars,Pred,[]),fail.
%cse_check_ast_complete(_,_).

%check_lazy_let_vars(X,Acc,R) :- print(t(X,Acc,R)),nl,fail.
check_lazy_let_vars(lazy_let_expr(TID,_,_),Acc,[ID|Acc]) :- !, get_texpr_id(TID,ID),
   (member(ID,Acc) -> print('multiply defined id: '), print(ID),nl ; true).
check_lazy_let_vars(lazy_let_pred(TID,_,_),Acc,[ID|Acc]) :- !, get_texpr_id(TID,ID),
   (member(ID,Acc) -> print('multiply defined id: '), print(ID),nl ; true).
check_lazy_let_vars(lazy_lookup_pred(ID),Acc,Acc) :- !,
   (member(ID,Acc) -> true ; add_internal_error('Lazy lookup pred id not defined: ',ID)).
check_lazy_let_vars(lazy_lookup_expr(ID),Acc,Acc) :- !,
   (member(ID,Acc) -> true ; add_internal_error('Lazy lookup expr id not defined: ',ID)).
%check_occurs_in_expr(ID,E) :-
%   (occurs_in_expr(ID,E) -> true
%    ; add_internal_error('Id does not occur in expression: ',ID),
%      print('Expression: '), translate:print_bexpr_or_subst(E),nl).
:- else.
cse_check_ast(_PP,_Pred).
%cse_check_ast_complete(_,_).
%check_occurs_in_expr(_,_).
:- endif.


% -----------------------------------------------

% utility to expand lazy let expressions for better CSE if repeatedly applied
% Note: lazy_lookups could remain if only a partial predicate is passed

expand_cse_lazy_lets_if_cse_enabled(E,EE) :-
   (expression_sharing_enabled -> expand_cse_lazy_lets(E,EE) ; E=EE).

expand_cse_lazy_lets(Expr,ExpandedExpr) :-
   transform_bexpr(b_expression_sharing:expand_lazy_let,Expr,ExpandedExpr).

:- public expand_lazy_let/2.
%expand_lazy_let(B,R) :- tools_printing:print_term_summary(B),nl,fail.
expand_lazy_let(b(E,Type,Info),Result) :- expand_lazy_let_aux(E,Type,Info,Result).
expand_lazy_let_aux(lazy_let_expr(TID,SharedExpr,MainExpr),_,_,Result) :-
   get_texpr_id(TID,ID), debug_println(4,expanding(ID)),
   expand_lazy_lookup(ID,SharedExpr,MainExpr,Result).
expand_lazy_let_aux(lazy_let_pred(TID,SharedExpr,MainExpr),_,_,Result) :-
   get_texpr_id(TID,ID), debug_println(4,expanding(ID)),
   expand_lazy_lookup(ID,SharedExpr,MainExpr,Result).
expand_lazy_let_aux(lazy_let_subst(TID,SharedExpr,MainExpr),_,_,Result) :-
   get_texpr_id(TID,ID), debug_println(4,expanding(ID)),
   expand_lazy_lookup(ID,SharedExpr,MainExpr,Result).
expand_lazy_let_aux(E,Type,Info,Result) :- E \= value(_), E \= integer(_),
   compute_static_expression(E,Val),
   %print(computed(E,Val)),nl,
   Result=b(value(Val),Type,Info).

expand_lazy_lookup(ID,SharedExpr,MainExpr,Result) :- \+ atom(ID),!,
   add_internal_error('Illegal call: ',expand_lazy_lookup(ID,SharedExpr,MainExpr,Result)),
   Result=MainExpr.
expand_lazy_lookup(ID,SharedExpr,MainExpr,Result) :-
   transform_bexpr(b_expression_sharing:replace_lazy_lookup(ID,SharedExpr),MainExpr,Result).

% TO DO: unify with b_compile and ast_cleanup !
%compute_static_texpression(b(E,_,_),Result) :- compute_static_expression(E,Result).
compute_static_int_texpression(b(E,_,_),Result) :- compute_static_expression(E,int(Result)).

:- use_module(kernel_objects,[safe_pown/3]).
compute_static_expression(add(A,B),int(Result)) :- % plus
    compute_static_int_texpression(A,IA), compute_static_int_texpression(B,IB),
    Result is IA+IB.
compute_static_expression(minus(A,B),int(Result)) :- % plus
    compute_static_int_texpression(A,IA), compute_static_int_texpression(B,IB),
    Result is IA-IB.
compute_static_expression(unary_minus(A),int(Result)) :- % plus
    compute_static_int_texpression(A,IA),
    Result is -IA.
compute_static_expression(multiplication(A,B),int(Result)) :-
    compute_static_int_texpression(A,IA), compute_static_int_texpression(B,IB),
    Result is IA*IB.
compute_static_expression(power_of(A,B),int(Result)) :-
    compute_static_int_texpression(A,IA), compute_static_int_texpression(B,IB), IB >= 0,
    kernel_objects:safe_pown(IA,IB,Result), number(Result).
% TO DO: modulo, division, set_extension
compute_static_expression(integer(I),int(I)).
compute_static_expression(value(I),I).
compute_static_expression(equal(A,B),I) :- compute_static_int_texpression(A,SA), compute_static_int_texpression(B,SB),
    (SA==SB -> I = pred_true ; I = pred_false).
compute_static_expression(not_equal(A,B),I) :- compute_static_int_texpression(A,SA), compute_static_int_texpression(B,SB),
    (SA==SB -> I = pred_false ; I = pred_true).


% -----------------------------------------------

% find maximum id used in lazy lookups; useful to generate new unused identifiers
find_maximum_lazy_let_id(BExpr,MaxID) :-
   reduce_over_bexpr(b_expression_sharing:find_maximum_lazy_let_id_aux,BExpr,-1,MaxID).

:- public find_maximum_lazy_let_id_aux/3.
find_maximum_lazy_let_id_aux(lazy_lookup_expr(ID),MaxSoFar,NewMax) :- !,
   convert_atom_to_nr(ID,IDNr),
   (IDNr @> MaxSoFar -> NewMax=IDNr ; NewMax=MaxSoFar).
find_maximum_lazy_let_id_aux(lazy_lookup_pred(ID),MaxSoFar,NewMax) :-
   convert_atom_to_nr(ID,IDNr),
   (IDNr @> MaxSoFar -> NewMax=IDNr ; NewMax=MaxSoFar).
find_maximum_lazy_let_id_aux(_,V,V).

% -----------------------------------------------

:- use_module(library(ordsets)).
:- use_module(library(avl)).

% BUP traversal; replacing shared expressions by lazy_lookup and propagate usage information upwards
% no introduction of lazy_let yet
% Accumulator contains list of introduced lazy_lets

% example call ((x+y)+z)*(x+y)+((x+z)+y) + (x+z) = 0

% the starting accumulator for this traversal:
% '$used_lazy_lookups': special key to store which lazy_lookups occur below
bup_init_acc(Acc3) :- empty_avl(E),
    avl_store('$used_lazy_lookups',E,[]-[],Acc1), % used directly and indirectly
    avl_store('$identifier_uses',Acc1,[],Acc2),
    avl_store('$lazy_lookups_directly_used',Acc2,[],Acc3).

:- public bup_insert_lazy_lookups/4.
bup_insert_lazy_lookups(BE,Res,SubAccs,NewAcc) :-
   BE2=BE, %update_info_field(BE,BE2), % we have just treated sub-expressions and we may insert BE into avl tree of shared expressions; info field is now updated in top-down pass
   % combine all infos about lazy_lets from subexpressions:
   merge_bup_accs(SubAccs,Acc),
   bup_update_identifier_uses(BE,Acc,Acc1),
   bup_insert_lazy_lookup_if_necessary(BE2,Res,Acc1,NewAcc).

% also compute used identifiers bup (ignoring @LazyLookup replacements)
% a slightly rewritten version of find_typed_identifier_uses2 without recursive calls
bup_update_identifier_uses(TExpr,Acc,NewAcc) :-
    avl_fetch('$identifier_uses',Acc,UsedIds),
    %bsyntaxtree:syntaxtraversion_fixed(TExpr,Expr,_QSubs,_Subs,TNames),
    get_texpr_expr(TExpr,Expr),safe_syntaxelement(Expr,_Subs,TNames,_,_),
    ( bsyntaxtree:uses_an_identifier(Expr,Id)
      -> ord_add_element(UsedIds,Id,New),avl_store('$identifier_uses',Acc,New,NewAcc)
       ; TNames=[] -> NewAcc=Acc
       ; get_texpr_ids(TNames,Names), % these names are introduced and no longer visible higher up
         sort(Names,SNames),
         ord_subtract(UsedIds,SNames,New),
         avl_store('$identifier_uses',Acc,New,Acc1),
         % now invalidate all lazy_lookups which use an introduced variable name in SNames
         avl_fetch('$used_lazy_lookups',Acc1,All-Intro),
         exclude(filter_lazy_lookup(SNames),All,FAll),
         exclude(filter_lazy_lookup(SNames),Intro,FIntro),
         avl_store('$used_lazy_lookups',Acc1,FAll-FIntro,Acc2),
         avl_fetch('$lazy_lookups_directly_used',Acc2,Direct),
         exclude(filter_lazy_lookup(SNames),Direct,FDirect),
         avl_store('$lazy_lookups_directly_used',Acc2,FDirect,NewAcc)
         %,print(filter(SNames,Intro,FIntro)),nl
    ).

% detect which lazy lookups are no longer valid higher up the AST
filter_lazy_lookup(SortedNames,LazyLookupID) :-
   lazy_let_to_be_introduced(LazyLookupID,_,_,UsedIds,SharedExpr),
   (ord_intersect(SortedNames,UsedIds) % if any of the introduced names is used: filter out lazy lookup
     -> true
     ;  get_texpr_info(SharedExpr,Infos),
        memberchk(contains_wd_condition,Infos)
        %,print(wd_filter(LazyLookupID)),nl, translate:print_bexpr(SharedExpr),nl
    ).
% TO DO: also filter Expressions with WD-condition attached

% merge accumulators from subexpressions:
merge_bup_accs([],Res) :- bup_init_acc(Res).
merge_bup_accs([H],Res) :- !,  % only one accumulator/argument: remove all ids marked for introduction
   % they can be introduced in the level below
   avl_delete('$used_lazy_lookups',H,All-_Intro,Acc1),
   avl_store('$used_lazy_lookups',Acc1,All-[],Acc2), % set introduce field to []
   Res=Acc2.
merge_bup_accs([H|T],Res) :- avl_fetch('$used_lazy_lookups',H,All1-_),
  avl_store('$used_lazy_lookups',H,All1-[],InitAcc), % set introduction set to empty
  merge_bup_accs_aux(T,InitAcc,Res).
merge_bup_accs_aux([],R,R).
merge_bup_accs_aux([H|T],A,R) :- combine_acc(H,A,A2), merge_bup_accs_aux(T,A2,R).


combine_acc(Acc1,Acc2,NewAcc) :-
    avl_to_list(Acc1,Acc1List),
    empty_avl(Empty),
    combine_acc_aux(Acc1List,Acc2,Empty,NewAcc).

combine_acc_aux([],_Acc2,Res,Res).
combine_acc_aux([Field-Value1|T],Acc2,RAcc,Res) :-
   avl_delete(Field,Acc2,Value2,Acc3),
   (field_combinator(Field,Value1,Value2,NewValue)
     -> avl_store(Field,RAcc,NewValue,RAcc2)
     ;  RAcc2=Acc2, Acc3=Acc2),
   combine_acc_aux(T,Acc3,RAcc2,Res).

% rules on how to combine the different fields of the accumulators:
field_combinator('$identifier_uses',Used1,Used2,Used) :- ord_union(Used1,Used2,Used).
field_combinator('$lazy_lookups_directly_used',Used1,Used2,Used) :- ord_union(Used1,Used2,Used).
field_combinator('$used_lazy_lookups',All1-_,AccAll-AccIntro,All-Intro) :-
   ord_union(All1,AccAll,All),
   ord_intersection(All1,AccAll,I1), % an @ID that is in this branch and some other branch
   ord_union([I1,AccIntro],Intro).




% check if we need to introduce a lazy_lookup at the current node
bup_insert_lazy_lookup_if_necessary(BExpr,LAZYLOOKUP,Acc,NewAcc) :-
   BExpr = b(Expr,Type,Info),
   select(sharing(ID,TotalCount,_,_),Info,Info2), % this expressions is marked as being shared
   TotalCount > 1, % with at least two occurrences
   !,
   %cse_check_ast(avl_store(ID),BExpr), % infos only re-established in TD phase
   bup_update_info(Info2,Acc,Info3),
   (lazy_let_to_be_introduced(ID,UsedLL,IntroLL,_,_)
     -> true /* let already introduced; ensure we use same Ids/version everywhere */
     ; get_used_lazy_lookups(Acc,UsedLL,IntroLL),
       avl_fetch('$identifier_uses',Acc,UsedIds),
       % nl, print(lookup(ID,UsedLL,IntroLL)),nl, translate:print_bexpr_or_subst(BExpr), % portray_avl(Acc),nl,
       register_let_id(ID,UsedLL,IntroLL,UsedIds,b(Expr,Type,Info3)) % store that we need to introduce a lazy_let
    ),
   (UsedLL=[] -> true
    ; last(UsedLL,LastLL),
      (ID > LastLL -> true
       ; format('*** LET ~w depends on later id ~w in ~w~n',[ID,LastLL,UsedLL]))
   ),
   ord_union([ID],UsedLL,NewUsedLL),
   ord_union([ID],IntroLL,NewIntroLL),
   avl_store('$lazy_lookups_directly_used',Acc,[ID],Acc1),
   avl_store('$used_lazy_lookups',Acc1,NewUsedLL-NewIntroLL,NewAcc), % to the outside (above) we only see usage of @ID
   bup_update_info(Info2,NewAcc,LazyLookupInfo),
   gen_lazy_lookup(Type,ID,Info2,LazyLookupInfo,LAZYLOOKUP).
bup_insert_lazy_lookup_if_necessary(b(Expr,Type,Info1),b(Expr,Type,Info2),Acc,Acc) :-
    %nl, print('BUP: '),translate:print_bexpr_or_subst(b(Expr,Type,Info1)), portray_avl(Acc),nl,
   bup_update_info(Info1,Acc,Info2).

bup_update_info(Info0,Acc,Info4) :-
   delete(Info0,sharing(_,_,_,_),Info1),
   delete(Info1,cse_used_ids(_),Info2),
   get_used_lazy_lookups(Acc,_,InterList),
   avl_fetch('$lazy_lookups_directly_used',Acc,DirectlyUsed),
   Info3=Info2, %Info3 = [cse_used_ids(DirectlyUsed)|Info2],
   ord_intersection(InterList,DirectlyUsed,IntroduceHere), % those not directly used will be introduced automatically inside the shared expressions
   (IntroduceHere=[] -> Info4=Info3
      %,(DirectlyUsed=[] -> true ; print('*** WARNING: directly used not empty: '),print(DirectlyUsed),nl)
   ; Info4 = ['$introduce_lazy_lookups'/IntroduceHere|Info3]). % store info in node for later top-down traversal).

get_used_lazy_lookups(Acc,All,Introduce) :-
   (avl_fetch('$used_lazy_lookups',Acc,A-I) -> All=A,Introduce=I
     ; add_internal_error('Failed: ',get_used_lazy_lookups(Acc,All,Introduce)),
       All=[], Introduce=[]).


% ----------------------------------

% PHASE 2: Top-Down Phase to introduce LETs at latest possible moment:

% top-down traversal to introduce lazy_lets for lazy_lookups we have introduced in bottom-up phase
top_down_gen_lazy_lets(InExpr,Acc,ResExpr) :-
   top_down_gen_lazy_lets1(InExpr,Acc,IntExpr),
   update_info_field(IntExpr,ResExpr). % maybe we should do this only if there was a change

top_down_gen_lazy_lets1(b(LazyLookup,Type,_Info),Acc,NewExpr1) :-
   directly_uses_cse_id(LazyLookup,LetID), % check if we have lazy_let_expr/_pred
   (lazy_let_to_be_introduced(LetID,UsedInside,IntroduceHere,_,SharedExpression) -> true
     ; avl_fetch('$maximum_unregistered_lazy_let_id',Acc,MaxId), % all ids up until MaxId come from a previous run of CSE
       LetID > MaxId,
       add_internal_error('**** Unregistered let: ',LetID),fail),
   \+ avl_fetch(lazy_let_introduced(LetID),Acc,true),
   !,
   debug_println(9,inlining_single_usage(LetID,UsedInside,IntroduceHere)),
   avl_store(lazy_let_inlined(LetID),Acc,true,Acc0), % NOT USED !
   top_down_gen_lazy_lets(SharedExpression,Acc0,NewSharedExpression),
   top_down_gen_lazy_lets_for_list(IntroduceHere,Type,Acc0,_,NewExpr1,NewSharedExpression).
top_down_gen_lazy_lets1(b(Expr,Type,Info),Acc,Res) :-
  %syntaxtraversion(BExpr,Expr,_,_,Subs,_TNames),
  select('$introduce_lazy_lookups'/IntroduceHere,Info,Info2),
  % functor(Expr,Fun,Arity),print(intro(IntroduceHere,Fun,Arity)),nl, translate:print_bexpr_or_subst(b(Expr,Type,Info)),nl,
  !,
  % the lazy lookups in Inter must now be introduced, if not already done so
  top_down_gen_lazy_lets_for_list(IntroduceHere,Type,Acc,NewAcc,Res,InnerRes),
  syntaxtransformation(Expr,Subs,_Names,NSubs,NewExpr1),
  InnerRes=b(NewExpr1,Type,Info2),
  td_treat_sub_args(Subs,NewAcc,NSubs).
top_down_gen_lazy_lets1(b(Expr,Type,Info),Acc,b(NewExpr1,Type,Info)) :-
   syntaxtransformation(Expr,Subs,_Names,NSubs,NewExpr1),
   td_treat_sub_args(Subs,Acc,NSubs).

% treat sub-arguments of an expression for which we have introduced lets
td_treat_sub_args([],_Acc,[]).
td_treat_sub_args([TE|T],Acc,[NTE|NT]) :-
    (top_down_gen_lazy_lets(TE,Acc,NTE)
     -> true
     ;  add_internal_error('Call failed: ',top_down_gen_lazy_lets(TE,Acc,NTE)),
        NTE=TE ),
    td_treat_sub_args(T,Acc,NT).


top_down_gen_lazy_lets_for_list(IntroHere,Type,Acc,NewAcc,Res,InnerRes) :-
    % first collect all lazy_lets we need here:
    %td_collect_lazy_lets_to_introduce_here(Inter,Acc,IntroHere,[]),
    % now update which lazy_lets are now visible
    topsort(IntroHere,SortedIntroHere),
    %print(sorted_intro(IntroHere,SortedIntroHere)),nl,
    td_introduce_lazy_lets(SortedIntroHere,Type,Acc,NewAcc,Res,InnerRes).

% topological sorting using a naive quadratic algorithm; TO DO: improve
topsort([],[]).
topsort([H|T],Res) :- select(X,T,TT), let_depends_upon(H,X), !,topsort([X,H|TT],Res).
topsort([H|T],[H|ST]) :- topsort(T,ST).

:- dynamic let_depends_upon/2, lazy_let_to_be_introduced/5.
% a fact to store which let depends directly upon which other let
% UsedLL: Used LazyLet @IDs, IntroduceLL: LazyLet @IDs to be introduced, UsedIds: regular identifiers referenced directly or via other lazy_lookups
register_let_id(ID,UsedLL,IntroduceLL,UsedIds,SharedExpr) :-
    assert(lazy_let_to_be_introduced(ID,UsedLL,IntroduceLL,UsedIds,SharedExpr)),
    %format('LAZY LET ~w (depends on ~w) [uses ~w ids]~n',[ID,UsedLL,UsedIds]),
    member(ID2,UsedLL),
    assert(let_depends_upon(ID,ID2)),
    % now do transitive closure in one direction
    %let_depends_upon(ID0,ID),assert(let_depends_upon(ID0,ID2)),
    fail.
register_let_id(ID,UsedLL,_,_,_) :- fail,
    % now do transitive closue in other direction
    member(ID2,UsedLL),
    let_depends_upon(ID2,ID3),
    assert(let_depends_upon(ID,ID3)),fail.
register_let_id(_,_,_,_,_).

reset_let_dependency_info :- retractall(let_depends_upon(_,_)),
   retractall(lazy_let_to_be_introduced(_,_,_,_,_)).


% introduce lets in  order
td_introduce_lazy_lets([],_,Acc,Acc,R,R).
td_introduce_lazy_lets([],Ty,Acc,AccR,T,RT) :-
    add_internal_error('Call failed: ',td_introduce_lazy_lets([],Ty,Acc,AccR,T,RT)),fail.
td_introduce_lazy_lets([LetID|T],Type,Acc,Acc1,OuterTerm,InnerTerm) :-
    avl_fetch(lazy_let_introduced(LetID),Acc,true),!,
    %print(already_introduced(LetID)),nl,
    td_introduce_lazy_lets(T,Type,Acc,Acc1,OuterTerm,InnerTerm).
td_introduce_lazy_lets([LetID|T],Type,Acc,Acc1,OuterTerm,InnerTerm) :-
    avl_store(lazy_let_introduced(LetID),Acc,true,Acc0),
    td_introduce_lazy_lets(T,Type,Acc0,Acc1,MidTerm,InnerTerm),
    % now introduce let for ID
    lazy_let_to_be_introduced(LetID,_,_,UsedIds,SharedExpression),
    top_down_gen_lazy_lets(SharedExpression,Acc0,NewSharedExpression), % the lazy_lets introduced inside are not visible
    %tools_printing:print_term_summary(gen_lazy_let(Type,LetID,NewSharedExpression,MidTerm,OuterTerm)),nl,
    gen_lazy_let(Type,LetID,UsedIds,NewSharedExpression,MidTerm,OuterTerm).



% -----------------------------------------------


% copy position information; important e.g. for discharged checks for theorems and invariants
/*
copy_position_info(From,b(To,Type,Info1),b(To,Type,Info3)) :-
   get_texpr_pos(From,PositionInfo),!,
   delete(Info1,nodeid(_),Info2),
   Info3 = [nodeid(PositionInfo)|Info2].
copy_position_info(_,E,E).
*/

:- use_module(b_ast_cleanup,[recompute_used_ids_info/2]).
% update info field of a typed expression after a modification of the expression
update_info_field(TE,NewTE) :-
   update_wd_condition(TE,TE2),
   recompute_used_ids_info(TE2,NewTE). % inserting lazy_lookups changes used ids for exists/forall

% check if we need to add contains_wd_condition info field
update_wd_condition(b(E,Type,Info1),Res) :-
   nonmember(contains_wd_condition,Info1), % inlined update_wd_condition_info to just copy E in clause below; not sure this helps memory performance
   sub_expression_contains_wd_condition(E),
   !,
   Res = b(E,Type,[contains_wd_condition|Info1]).
update_wd_condition(E,E).

update_wd_condition_info(E,Info1,Info2) :-
   nonmember(contains_wd_condition,Info1),
   sub_expression_contains_wd_condition(E),
   !,
   Info2 = [contains_wd_condition|Info1].
update_wd_condition_info(_,I,I).



% Remove sharing/4 info from SharedExpression + avoid reusing it for CSE if algorithm applied second time
%delete_sharing_info(b(E,T,Info),b(E,T,NewInfo)) :-  delete(Info,sharing(_,_,_,_),NewInfo).

:- public cse_cleanup_infos/2.
% Remove info that is no longer necessary once CSE has run to completion:
cse_cleanup_infos(b(E,T,Info),Res) :-
   cse_cleanup_infos1(E,T,Info,Res).

cse_cleanup_infos1(lazy_let_pred(ID,SharedExpr,Body),T,_,Res) :-
   collect_important_info(SharedExpr,Body,Info),
   cse_cleanup_aux(lazy_let_pred(ID,SharedExpr,Body),T,Info,Res).
cse_cleanup_infos1(lazy_let_expr(ID,SharedExpr,Body),T,_,Res) :-
   collect_important_info(SharedExpr,Body,Info),
   cse_cleanup_aux(lazy_let_expr(ID,SharedExpr,Body),T,Info,Res).
cse_cleanup_infos1(lazy_let_subst(ID,SharedExpr,Body),T,_,Res) :-
   collect_important_info(SharedExpr,Body,Info),
   cse_cleanup_aux(lazy_let_subst(ID,SharedExpr,Body),T,Info,Res).
cse_cleanup_infos1(sequence(Seq),T,Info,Res) :- !, % re-flatten sequences that were binarized for CSE sharing analysis
    flatten_sequence(Seq,FSeq,WD),
    ((WD==contains_wd_condition, nonmember(contains_wd_condition,Info)) ->
      Info1 = [contains_wd_condition|Info] ; Info1=Info),
    cse_cleanup_aux(sequence(FSeq),T,Info1,Res).
cse_cleanup_infos1(E,T,Info,Res) :- cse_cleanup_aux(E,T,Info,Res).

flatten_sequence([H,b(sequence(T),subst,Info)],[H|FT],WD) :- member(cse_generated,Info),
    !,
    % check if this sequence contains a WD condition
    (var(WD),member(contains_wd_condition,Info) -> WD=contains_wd_condition ; true),
    flatten_sequence(T,FT,WD).
flatten_sequence(X,X,_WD).


cse_cleanup_aux(E,T,Info,b(NewE,T,NewInfo)) :-
   delete(Info,cse_used_ids(_),Info2),
   ((nonmember(contains_wd_condition,Info2), % TO DO: allow in DISPROVER_MODE
     post_cse_simplifier(E,SE)) -> NewE = SE, NewInfo=Info2 % Note: maybe some lets no longer necessary now
   ; reorder_lazy_let_expr(E,T,Info2,NewE,NewInfo) -> true
   ; NewE=E,NewInfo=Info2).

% reorder/move lazy lets outside if this leads to better optimizations being applicable
reorder_lazy_let_expr(Expr,Type,Info,NewExpr,NewInfo) :-
    reorder_argument(Expr,FunctorID,Match,TArgToReorder,OtherArgs,_),
    get_texpr_expr(TArgToReorder,ArgToReorder),
    %print(try_reorder(FunctorID,TArgToReorder)),nl,
    is_lazy_let(ArgToReorder,LLType,ID,ShExpr,BodyA),
    get_texpr_expr(BodyA,Match),
    print(reorder_lazy_let_expr(ID,FunctorID)),nl,
    !,
    reorder_argument(NewBinExpr,FunctorID,_,BodyA,OtherArgs,Recur),
    !, % FunctorID ensures only one possible solution for reorder_argument
    ((Recur=recur, % we may have to re-order again
      reorder_lazy_let_expr(NewBinExpr,Type,Info,NewBinExpr2,Info2))
       -> NewInnerExpr = b(NewBinExpr2,Type,Info2)
       ;  NewInnerExpr = b(NewBinExpr,Type,Info)),
    is_lazy_let(NewExpr,LLType,ID,ShExpr,NewInnerExpr), % TYPE MAY HAVE TO CHANGE for CASE ?? TO DO: Check with appropriate B Machine
     % copy important info from InfoA ??
    update_wd_condition_info(NewExpr,Info,NewInfo).
% TO DO: add more cases + what if both arguments need reordering --> fixpoint computation ?

is_lazy_let(lazy_let_expr(ID,ShExpr,BodyA),lazy_let_expr,ID,ShExpr,BodyA).
is_lazy_let(lazy_let_subst(ID,ShExpr,BodyA),lazy_let_subst,ID,ShExpr,BodyA).
is_lazy_let(lazy_let_pred(ID,ShExpr,BodyA),lazy_let_pred,ID,ShExpr,BodyA).

% reorder argument 1 (i.e., move lazy_let out) in case we match a certain pattern
% reorder_argument(Expression, ID, MATCH,  ARG_TO_REORDER,  OtherArgs)
reorder_argument(function(A,B),function_1,set_extension(_),A,B,do_not_recur). % there is optimized code: b_apply_function_set_extension
reorder_argument(max(A),max,set_extension(_),A,nil,do_not_recur). % we have optimized code for min({a,b,...}) and max
reorder_argument(min(A),min,set_extension(_),A,nil,do_not_recur).
reorder_argument(card(A),card,set_extension(_),A,nil,do_not_recur).
reorder_argument(image(A,B),image_1,closure(_),A,B,recur).
reorder_argument(image(B,A),image_2,set_extension(_),A,B,recur).
% avoid that Lazy Lets stand at illegal places
% avoid IF THEN LET(... ELSIF P THEN ....)  if_elsif term
reorder_argument(if(IfList),if1,_,A,Rest,recur) :- select(A,IfList,Rest).
% avoid lazy let in place of a CASE case_or term
reorder_argument(case(Expr,Cases,Else),case_element,_,A,[Expr,Else|Rest],recur) :- select(A,Cases,Rest).
% avoid lazy let for a select_when term:
reorder_argument(select(SelWhenList),select,_,A,Rest,recur) :- select(A,SelWhenList,Rest). % SelWhenList: a list containing select_when/2 terms
reorder_argument(select(SelWhenList,Else),select_else,_,A,(Rest,Else),recur) :- select(A,SelWhenList,Rest).
% CSE will sometimes insert a lazy_let_subst where an if_elsif should stand; this not supported by the interpreter
%reorder_argument(while(COND,TSubst,INV,VARIANT),while_subst,_,TSubst,[COND,INV,VARIANT]). % TO DO: check if lazy_let does not depend on variables modified in loop

% TO DO: bring this in accordance with do_not_share_this rules:

% detect certain obvious inconsistencies / tautologies introduced by CSE
post_cse_simplifier(equal(A,B),truth) :- same_lazy_lookup(A,B).
post_cse_simplifier(not_equal(A,B),falsity) :- same_lazy_lookup(A,B).
post_cse_simplifier(greater(A,B),falsity) :- same_lazy_lookup(A,B).
post_cse_simplifier(greater_equal(A,B),truth) :- same_lazy_lookup(A,B).
post_cse_simplifier(less(A,B),falsity) :- same_lazy_lookup(A,B).
post_cse_simplifier(less_equal(A,B),truth) :- same_lazy_lookup(A,B).
post_cse_simplifier(subset(A,B),truth) :- same_lazy_lookup(A,B).
post_cse_simplifier(not_subset(A,B),falsity) :- same_lazy_lookup(A,B).
post_cse_simplifier(strict_subset(A,B),falsity) :- same_lazy_lookup(A,B).
post_cse_simplifier(not_strict_subset(A,B),truth) :- same_lazy_lookup(A,B).

/*
TO DO:
split up lazy_let_pred; as previously done in gen_lazy_let
gen_lazy_let(pred,CSEID,_UsedIds,SharedExpr,Body,Result) :-
    true, % TO DO <------------- RENABLE
    % check if we can split off Prefix and/or Suffix of Body not using CSEID
    cse_split_conjunct(Body, CSEID, Prefix, MainBody, Suffix),
    !,
    print(optimizing_let_with_conjunction_body(CSEID)),nl,
    gen_id(CSEID,SharedExpr,ID),
    cse_conjunct_predicates(MainBody,NewBody),
    %get_texpr_info(NewBody,NewInfo),
    collect_important_info(SharedExpr,NewBody,NewInfo), % we could remove ID from cse_used_ids
    LET = b(lazy_let_pred(ID,SharedExpr,NewBody),pred,NewInfo),
    append(Prefix,[LET|Suffix],ConjunctionList),
    cse_conjunct_predicates(ConjunctionList,Result).
*/

same_lazy_lookup(A,B) :- get_lazy_lookup_id(A,ID), get_lazy_lookup_id(B,ID).

get_lazy_lookup_id(b(E,Type,_),Type,ID) :- get_lazy_lookup_id_aux(E,ID).
get_lazy_lookup_id(b(E,_,_),ID) :- get_lazy_lookup_id_aux(E,ID).
get_lazy_lookup_id_aux(lazy_lookup_expr(ID),ID).
get_lazy_lookup_id_aux(lazy_lookup_pred(ID),ID).
% there is no lazy_lookup_subst(ID)

%gen_lazy_let_with_check(Type, Identifier, UsedIds,SharedExpression,  SubExpressionInsideLet,  Result) :-
%    cse_check_ast(before_gen_lazy_let_shared,SharedExpression),
%    cse_check_ast(before_gen_lazy_let_body,SubExpressionInsideLet),
%    (gen_lazy_let(Type, Identifier, UsedIds,SharedExpression,  SubExpressionInsideLet,  Result)
%      -> cse_check_ast(gen_lazy_let,Result)
%      ; add_internal_error('Call failed: ',gen_lazy_let_with_check(Type, Identifier, UsedIds,SharedExpression,  SubExpressionInsideLet,  Result)),
%        fail).

% TO DO: maybe check that we are not in a bad context to prevent optimized treatment (e.g. {1|->a,2|->b}(x) ), rather than delegating this to reorder lazy_let code above ?
% gen_lazy_let(Type, Identifier, UsedIds,SharedExpression,  SubExpressionInsideLet,  Result)
gen_lazy_let(pred,A,_UsedIds,SharedExpr,Body,b(R,pred,[])) :-
   % Info field will be set later in cse_cleanup_infos
    !,
    gen_id(A,SharedExpr,ID), debug_println(4,generating_lazy_let_pred(A,ID)),
    R = lazy_let_pred(ID,SharedExpr,Body).
gen_lazy_let(subst,A,_UsedIds,SharedExpr,Body,b(R,subst,[])) :-
    !,
    gen_id(A,SharedExpr,ID), debug_println(4,generating_lazy_let_subst(A,ID)),
    R = lazy_let_subst(ID,SharedExpr,Body).
gen_lazy_let(Type,A,_UsedIds,SharedExpr,C,b(lazy_let_expr(ID,SharedExpr,C),Type,[])) :-
    gen_id(A,SharedExpr,ID).

% ----------------------


collect_important_info(SubA,SubB,Info) :- (var(SubA) ; var(SubB)), !,
    add_internal_error('Illegal call: ',collect_important_info(SubA,SubB,Info)),fail.
collect_important_info(SubA,SubB,Info) :-
    extract_info(SubA,SubB,Info1), % WD info (bsyntaxtree utility)
    % was collect_uses_cse_id([SubA,SubB],IDS), % CSE usage info
    exclude(is_cse_used_ids,Info1,Info2), % is this necessary ??
    Info=Info2. %Info = [cse_used_ids(IDS)|Info2].

% ----------------------

gen_lazy_lookup(pred,A,Infos,LazyLookupInfo,R) :- !, convert_to_atom(A,ID),
   (member(negated_cse,Infos)
    -> R = b(negation(b(lazy_lookup_pred(ID),pred,LazyLookupInfo)),pred,OuterInfos),
       delete(LazyLookupInfo,'$introduce_lazy_lookups'/_,OuterInfos) % we do not need to generate the lazy_lookups here; they can be created inside the negation
       %,print(create_neg_lookup(ID,LazyLookupInfo,OuterInfos)),nl
     ; R = b(lazy_lookup_pred(ID),pred,LazyLookupInfo)).
gen_lazy_lookup(Type,A,_,LazyLookupInfo,
                b(lazy_lookup_expr(ID),Type,LazyLookupInfo)) :- convert_to_atom(A,ID).

% generate atoms: many identifier treating predicates expect atoms
gen_id(A,B,b(identifier(ID),Type,[])) :- get_texpr_type(B,Type),convert_to_atom(A,ID).

convert_to_atom(A,ID) :- var(A),!, add_internal_error('Illegal call: ', convert_to_atom(A,ID)),fail.
convert_to_atom(A,ID) :- atom(A),!, ID=A.
convert_to_atom(Nr,ID) :- number_codes(Nr,C), atom_codes(ID,[64|C]). % 64 = @

convert_atom_to_nr(Atom,Nr) :- var(Atom),!, add_internal_error('Illegal call: ',convert_atom_to_nr(Atom,Nr)),fail.
convert_atom_to_nr(Atom,Nr) :- atom_codes(Atom,[64|C]), number_codes(Nr,C).

% check if an identifier is (potentially) a lazy_let ID
is_lazy_let_identifier(Atom) :- atom(Atom), atom_codes(Atom,[64|_]).

is_cse_used_ids(cse_used_ids(_)).

/*
collect_uses_cse_id([],[]).
collect_uses_cse_id([H|T],Res) :-
    get_cse_used_ids_info(H,CSE),!,
    collect_uses_cse_id(T,CSET),
    ord_union(CSE,CSET,Res).

uses_cse_id(ID,Expr) :- get_cse_used_ids_info(Expr,List),!,
   member(ID,List).
uses_cse_id(_ID,Expr) :- print('Could not extract cse_used_ids info: '), print(Expr),nl,
   true.

% CSE Identifier usage infor either form info field or from direct usage by lazy_lookups
get_cse_used_ids_info(TExpr,IDList) :- get_texpr_expr(TExpr,E),
   directly_uses_cse_id(E,UsedID),!,
   IDList = [UsedID].
get_cse_used_ids_info(TExpr,IDList) :- get_texpr_info(TExpr,Info),
   member(cse_used_ids(IDList),Info),!.
get_cse_used_ids_info(TE,L) :- print('Could not get cse_used_ids info: '), print(TE),
   L=[].
*/

directly_uses_cse_id(E,ID) :- var(E),!, add_internal_error('Illegal call: ',directly_uses_cse_id(E,ID)),fail.
directly_uses_cse_id(lazy_lookup_pred(AtomID),ID) :- convert_atom_to_nr(AtomID,ID).
directly_uses_cse_id(lazy_lookup_expr(AtomID),ID) :- convert_atom_to_nr(AtomID,ID).
% there is no lazy_lookup_subst

:- public print_cse_id_used/1.
print_cse_id_used(TypedExpr) :- get_texpr_info(TypedExpr,Info),
   member(cse_used_ids(L),Info),!,
   print(L).
print_cse_id_used(_) :- print(' no cse used id info ').

:- public replace_lazy_lookup/4.
:- use_module(btypechecker, [unify_types_strict/2]).
% TO DO: maybe remove cse_used_ids ??
%replace_lazy_lookup(_,_,T,_) :- print(T),nl,fail.
replace_lazy_lookup(ID,SharedExpr,BExpr,SharedExpr) :-
   get_lazy_lookup_id(BExpr,Type,TID),
    !,
   TID=ID,
   get_texpr_type(SharedExpr,SharedType),
   (my_unify_types(SharedType,Type) -> true
    ; add_internal_error('Incompatible type: ',replace_lazy_lookup(ID,'\\='(SharedType,Type),BExpr,SharedExpr))).
replace_lazy_lookup(_,_,E,Res) :- update_info_field(E,Res). % probably not necessary ?!

:- if(environ(prob_safe_mode,true)).
my_unify_types(T1,T2) :- (\+ ground(T1) ; \+ ground(T2)),
   format('*** Non-ground type unification: ~w = ~w~n~n',[T1,T2]),
   fail.
:- endif.
my_unify_types(T1,T2) :- unify_types_strict(T1,T2).

% -----------------


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% common subexpression detection

:- assert_must_succeed(
   (   find_common_subexpressions(b(identifier(i),integer,[]),0,false,R),
       R == b(identifier(i),integer,[sharing(0,1,top,[]),cse_used_ids([])])
   )).
:- assert_must_succeed(
   (   find_common_subexpressions(b(identifier(i),integer,[]),0,true,R),
       R == b(identifier(i),integer,[cse_used_ids([])])
   )).
:- assert_must_succeed(
   (   I = b(identifier(i),integer,[]),
       O = b(identifier(i),integer,[sharing(0,2,top,[]),cse_used_ids([])]),
       find_common_subexpressions(b(add(I,I),integer,[]),0,false,R),
       R == b(add(O,O),integer,[sharing(1,1,top,[]),cse_used_ids([0])])
   )).
:- assert_must_succeed(
   (   I = b(identifier(i),integer,[]),
       O = b(identifier(i),integer,[cse_used_ids([])]),
       find_common_subexpressions(b(add(I,I),integer,[]),0,true,R),
       R == b(add(O,O),integer,[sharing(0,1,top,[]),cse_used_ids([])])
   )).
% two quantifiers with the same variable name must not share
% expressions that contain the variable, other expressions should be
% shared
:- assert_must_succeed(
   (   C = b(identifier(c),integer,[]),
       X = b(identifier(x),integer,[]),
       I = b(identifier(i),set(integer),[]),
       O = b(integer(1),integer,[]),
       %  !x.(x:I => x<c+1)
       F = b(forall([X],
                    b(member(X,I),pred,[]),
                    b(less(X,b(add(C,O),integer,[])),pred,[])),pred,[]),
       %  !x.(x:I => x>c+1)
       Fx = b(forall([X],
                    b(member(X,I),pred,[]),
                    b(greater(X,b(add(C,O),integer,[])),pred,[])),pred,[]),
       C1 = b(identifier(c),integer,[sharing(3,2,top,[]),cse_used_ids([])]),
       X1 = b(identifier(x),integer,[sharing(0,3,7,[]),cse_used_ids([])]),
       I1 = b(identifier(i),set(integer),[sharing(1,2,top,[]),cse_used_ids([])]),
       O1 = b(integer(1),integer,[sharing(4,2,top,[]),cse_used_ids([])]),
       F1 = b(forall([X1],
                     b(member(X1,I1),pred,[sharing(2,1,7,[]),cse_used_ids([0,1])]),
                     b(less(X1,b(add(C1,O1),integer,[sharing(5,2,top,[]),cse_used_ids([3,4])])),
                           pred,[sharing(6,1,7,[]),cse_used_ids([0,3,4,5])])),
                pred,[sharing(7,1,top,[0]),cse_used_ids([0,1,2,3,4,5,6])]),
       % e.g. c+1 should be shared (ref. 5)
       % x and x:I must not be shared, they get a new reference (8 and 9)
       X2 = b(identifier(x),integer,[sharing(8,3,11,[]),cse_used_ids([])]),
       F2 = b(forall([X2],
                     b(member(X2,I1),pred,[sharing(9,1,11,[]),cse_used_ids([1,8])]),
                     b(greater(X2,b(add(C1,O1),integer,[sharing(5,2,top,[]),cse_used_ids([3,4])])),
                                      pred,[sharing(10,1,11,[]),cse_used_ids([3,4,5,8])])),
                pred,[sharing(11,1,top,[8]),cse_used_ids([1,3,4,5,8,9,10])]),
       find_common_subexpressions(b(conjunct(F,Fx),pred,[]),0,false,R),
       R == b(conjunct(F1,F2),pred,[sharing(12,1,top,[]),cse_used_ids([0,1,2,3,4,5,6,7,8,9,10,11])])
   )).

% find_common_subexpressions/4 searches for common subexpressions and adds an
%   information field sharing/4 to each AST node.
% find_common_subexpressions(+TExpr,+Start,+SkipSimple,-NTExpr):
%   TExpr: the original AST
%   Start: at which number the IDs should start (usually 0)
%   NTExpr: the AST, each node annotated with an sharing(ID,Counter,SubId,References) information field
%      two syntax nodes share the same ID when they should yield always the same
%      value. Counter counts the number of occurrences of the same expression in
%      the given AST. SubId is the unique reference of the quantified expression
%      where a used ID is introduced, top if there is no such quantifier.
%      References is a list of ids that are are sub-expressions of the node which
%        occur more than once and which depend on the newly introduced identifiers
%      (Please note: The declaration of a variable in a quantified expression
%       is an occurrence of its own.)
%   SkipSimple: set to true if expression sharing should not be applied to simple
%      expressions like identifier(_), integer(_), etc.
find_common_subexpressions(TExpr,Start,SkipSimple,NTExpr) :-
    find_common_subexpressions(TExpr,Start,SkipSimple,NTExpr,_).
find_common_subexpressions(TExpr,Start,SkipSimple,NTExpr,References) :-
    create_stores(Start,EmptyStores),
    find_common_subexpressions2(TExpr,EmptyStores,SkipSimple,root/1,NTExpr,Stores,_Stripped,_UsedIds,_UsedCSEIds),
    %print(grounding(Stores)),nl,
    ground_reference_counters(Stores), %print(stores(Stores)),nl, print(NTExpr),nl,
    get_current_references(Stores,References). % needed ??

:- if(environ(prob_safe_mode,true)).
add_cse_used_ids_info(Ids,OldInfos,NewInfos) :- (var(Ids) ; \+ ground(OldInfos)),!,
   add_internal_error('Illegal Call: ',add_cse_used_ids_info(Ids,OldInfos,NewInfos)),
   NewInfos = OldInfos.
:- endif.
add_cse_used_ids_info(Ids,OldInfos,NewInfos) :-
    exclude(is_cse_used_ids,OldInfos,OldInfos1),
    NewInfos=[cse_used_ids(Ids)|OldInfos1].

% find_common_subexpressions2(+TExpr,+InStores,+SkipSimple,+OuterFunctor,-NTExpr,-OutStores,-Stripped,-UsedIds,-UsedCSEIds):
find_common_subexpressions2(TExpr,InStores,SkipSimple,OuterFunctor,NTExpr,OutStores,Stripped,UsedIds,UsedCSEIds) :-
    %print(find(TExpr,InStores,outer(OuterFunctor))),nl,
    % decompose the typed expression into its components (Subs), look up the newly introduced
    % identifiers (NewIds) and create a new typed expression (NTExpr) with the modified components
    % (NSubs) and the new Info field (Info)
    fc_parts(TExpr,Expr,Type,Subs,NewQuantifiedIds,InvalidatedIds,NSubs,OldInfos,NewInfos,NTExpr),
    % introduce a map from expression to reference that is only visible to this node
    % and it subnodes if new identifiers (NewIds) are introduced
    % OutStores1 and OutStores contain a pattern such that if OutStores1 is set,
    % OutStores is exactly like OutStores1 but without visibility of the newly introduced
    % identifiers.
    append(InvalidatedIds,NewQuantifiedIds,NewIds),
    create_substore(NewIds,ID,InStores,SubStores),
    % apply subexpression detection on the subexpressions
    functor(Expr,OuterExprFunctor,_), % memorise functor for sub-expressions (allows deciding if sharing makes sense)
    find_common_subexpressions_l(Subs,SubStores,SkipSimple,OuterExprFunctor,1,
                                 NSubs,SubStores1,StrippedSubs,UsedSubIds,SubUsedCSEIds),
    % create a stripped (i.e. no info field, no types) version of the syntax node for look ups
    (Type==subst -> Stripped=not_relevant % substitions are never shared (and everything containing a subst is a subst itself)
      ; create_stripped(Expr,StrippedSubs,Stripped)),
    %print(stripped(Stripped)),nl,
    LookupKey = cse_lookup_key(Stripped,Type), % we also store the type along with the stripped AST (avoid sharing, e.g., two prj1/2 functions with different types, see test 255 !)
    % merge the used ids of the subnodes with the used ids in this node,
    % subtract newly introduced quantified identifiers because they are not visible from outside
    update_used_ids(Expr,NewQuantifiedIds,UsedSubIds,UsedIds),
    % if a substore was introduced due to newly introduced identifiers, remove it
    %%(OuterExprFunctor=while -> portray_stores(SubStores1) ;  true),
    remove_last_substore(NewIds,SubStores1,Stores1,References),
    % try to find the expression in the stores
    (do_not_share_this(SkipSimple,Stripped,UsedIds,OuterFunctor,Type,OldInfos) ->
        OutStores = Stores1,
        add_cse_used_ids_info(SubUsedCSEIds,OldInfos,NewInfos),
        UsedCSEIds = SubUsedCSEIds,
        store_expression(LookupKey,UsedIds,Stores1,_,_,ID) /* just ground ID; used for Sub expressions */
        %, print(not_sharing(ID))
    ;  lookup_cse(Type,LookupKey,UsedIds,Stores1,OutStores,SharingInfo,Negated,ID) ->
        /* it is shared */
         % print(looked_up(ID,Negated,used(UsedIds,sub(UsedSubIds)),Stores1)),nl,
        (Negated=negated_cse -> NewInfos=[Negated,SharingInfo|OldInfos1]
         ; NewInfos=[SharingInfo|OldInfos1]),
        add_cse_used_ids_info(SubUsedCSEIds,OldInfos,OldInfos1),
        ord_union([ID],SubUsedCSEIds,UsedCSEIds),
        set_references(References,SharingInfo)
        % In this case we do not need to look for sharing in the sub-expressions !!
        %,check_sharing_info(lookup,SharingInfo)
    ; otherwise ->
        % if not found, store the expression: it will be shared if we use it again later
        store_expression(LookupKey,UsedIds,Stores1,OutStores,SharingInfo,ID),
        ord_union([ID],SubUsedCSEIds,UsedCSEIds),
        % print(new_id(ID,UsedIds,SubUsedCSEIds,OutStores)),nl,
        % TO DO: what if later we find out that it is shared after all -> we should remove sub-expression counts
        add_cse_used_ids_info(SubUsedCSEIds,OldInfos,OldInfos1),
        NewInfos=[SharingInfo|OldInfos1],
        set_references(References,SharingInfo)
        %,check_sharing_info(new,SharingInfo)
    ). %  print(done_find(NTExpr,OutStores,UsedIds,UsedCSEIds)),nl.

lookup_cse(Type,LookupKey,UsedIds,Stores1,OutStores,SharingInfo,Negated,ID) :-
    (lookup_expression(LookupKey,UsedIds,Stores1,OutStores,SharingInfo,ID)
      -> Negated = normal_cse
      ; Type=pred, negate_lookup_key(LookupKey,NegKey), % try looking up negation
        lookup_expression(NegKey,UsedIds,Stores1,OutStores,SharingInfo,ID),
        Negated = negated_cse).

negate_lookup_key(cse_lookup_key(P,pred),cse_lookup_key(NP,pred)) :-
    negate_norm_strip(P,NP).

%check_sharing_info(_,sharing(Id,_RefCounter,Level,Refs)) :- number(Id), ground(Level), ground(Refs),!.
%check_sharing_info(PP,SI) :- print(illegal_sharing_info(PP,SI)),nl.

% NewIds = introduced IDs or modified Ids
fc_parts(TExpr,Expr,Type,Subs,QuantifiedNewIds,InvalidatedIds,NSubs,OldInfos2,NewInfos,NTExpr) :-
    my_decompose_texpr(TExpr,Expr,Type,OldInfos),
    cse_syntaxtransformation(Expr,Subs,Names,NSubs,NExpr),
    get_texpr_ids(Names,QuantifiedNewIds),
    (select(invalidate_ids_for_cse(ModifiedIDs),OldInfos,OldInfos2)
      % we have a stored invalidation info for CSE
      -> InvalidatedIds=ModifiedIDs
      ;  InvalidatedIds= [],OldInfos=OldInfos2),
     %functor(Expr,F,A), format('-> ~w/~w : ~w + ~w~n',[F,A,QuantifiedNewIds,InvalidatedIds]),
    create_texpr(NExpr,Type,NewInfos,NTExpr).
% TO DO: re-flatten the sequential compositions later

% CSE traversal which ignores certain sub-arguments (mainly LHS of assignments)
cse_syntaxtransformation(assign(LHS,RHS),RHS,[],NRHS,NExpr) :- !, % print(assign(LHS)),nl,
    same_length(RHS,NRHS),
    NExpr=assign(LHS,NRHS).
cse_syntaxtransformation(becomes_element_of(LHS,RHS),[RHS],[],[NRHS],NExpr) :- !,
    NExpr=becomes_element_of(LHS,NRHS).
cse_syntaxtransformation(becomes_such(LHS,RHS),[RHS],[],[NRHS],NExpr) :- !,
    NExpr=becomes_such(LHS,NRHS).
cse_syntaxtransformation(evb2_becomes_such(LHS,RHS),[RHS],[],[NRHS],NExpr) :- !,
    NExpr=evb2_becomes_such(LHS,NRHS).
cse_syntaxtransformation(assign_single_id(Id,RHS),[RHS],[],[NRHS],NExpr) :- !,
    NExpr=assign_single_id(Id,NRHS).
cse_syntaxtransformation(Expr,Subs,Names,NSubs,NExpr) :- syntaxtransformation(Expr,Subs,Names,NSubs,NExpr).
% TO DO: rlevent(Id,Section,Status,Parameters,Guard,Theorems,Actions,VWitnesses,PWitnesses,_Unmod,Refines)

:- use_module(b_read_write_info, [get_accessed_vars/4]).
my_decompose_texpr(b(Expr,Type,Infos),Expr2,Type,Infos2) :- my_decompose2(Expr,Expr2,ModifiedIDs),
   add_invalidated_ids_to_infos(Infos,ModifiedIDs,Infos2).
% decompose sequence into left and right part and store invalidated parts on the right
my_decompose2(sequence([TSubst|T]),Res,[]) :- !,
   Res=sequence([TSubst|NT]),
   get_accessed_vars(TSubst,[],ModifiedIDs,_),
   print(modified_sequence(ModifiedIDs)),nl,
   construct_seq_tail(T,ModifiedIDs,NT).
% annotate parts of while loop with variables which are invalidated by loop iteration
% note the While loop itself also invalidates the same variables, ensures that a substore
% is already set-up when entering the sub-nodes; TO DO: examine whether we can simplify this by improving the way shared expressions are stored above
my_decompose2(while(COND,TSubst,INV,VARIANT),while(COND2,TSubst2,INV2,VARIANT2),ModifiedIDs) :-
   get_accessed_vars(TSubst,[],ModifiedIDs,_),
   print(modified_while(ModifiedIDs)),nl,
   add_invalidated_ids(TSubst,ModifiedIDs,TSubst2),
   add_invalidated_ids(COND,ModifiedIDs,COND2),
   add_invalidated_ids(INV,ModifiedIDs,INV2),
   add_invalidated_ids(VARIANT,ModifiedIDs,VARIANT2).
my_decompose2(X,X,[]).

add_invalidated_ids_to_infos(L,[],R) :- !, R=L.
add_invalidated_ids_to_infos(L,Ids,[invalidate_ids_for_cse(Ids)|L]).

add_invalidated_ids(b(E,T,I),Ids,b(E,T,[invalidate_ids_for_cse(Ids)|I])).

construct_seq_tail([],_,[]) :- !.
%construct_seq_tail([X],[X]) :- !.
construct_seq_tail([H|T],ModifiedIDs,[b(sequence([H|T]),subst,[cse_generated,invalidate_ids_for_cse(ModifiedIDs)])]). % store that modified vars are invalid for CSE in tail



find_common_subexpressions_l([],Stores,_SkipSimple,_OuterFunctor,_,[],Stores,[],[],[]).
find_common_subexpressions_l([Expr|Erest],InStores,SkipSimple,OuterFunctor,ArgNr,
                             [NExpr|Nrest],OutStores,[Stripped|Srest],UsedIds,UsedCSEIds) :-
    find_common_subexpressions2(Expr,InStores,SkipSimple,outer(OuterFunctor,ArgNr),NExpr,InterStores,Stripped,UsedIds1,UsedCSEIds1),
    A1 is ArgNr+1,
    find_common_subexpressions_l(Erest,InterStores,SkipSimple,OuterFunctor,A1,Nrest,OutStores,Srest,UsedIds2,UsedCSEIds2),
    ord_union(UsedIds1,UsedIds2,UsedIds), % the B identifiers used
    ord_union(UsedCSEIds1,UsedCSEIds2,UsedCSEIds). % the identifiers to CSE used

boolean(boolean_true).
boolean(boolean_false).


do_not_share_this(_,_,_,_,Type,_) :- \+ ground(Type),!. % it could be that the Type is ground, but the sub-expressions contain non-ground types; see test 1090 size(arr)
  % nl,print(not_sharing_non_ground_type(Type)),nl,nl.
% TO DO: can we maybe still share non-ground type terms if we are careful ? see also tests 402, 403 ?
do_not_share_this(_,_,_,_,Type,_) :- do_not_share_type(Type),!.
do_not_share_this(_,_Stripped,_,_Functor,_Type,Infos) :-
  memberchk(contains_wd_condition,Infos),
  get_preference(use_common_subexpression_wd_only,true),
  !. % also examine DISPROVER MODE ?
do_not_share_this(_,_Stripped,_,_Functor,Type,_Infos) :- Type==pred,
  get_preference(use_common_subexpression_also_for_predicates,false),
  !.
do_not_share_this(SkipSimple,Stripped,UsedIds,OuterFunctor,_Type,_Infos) :-
  SkipSimple==true, % print(chk(Stripped,OuterFunctor)),nl,
  do_not_share_this_3(Stripped,UsedIds,OuterFunctor).

do_not_share_type(subst).
do_not_share_type(status). % Event-B Event Status

:- use_module(external_functions,[not_declarative/1]).

do_not_share_this_3(case_or(_,_),_,_) :- !. % why does this have a pred type ?
do_not_share_this_3(if_elsif(_,_),_,_) :- !.
do_not_share_this_3(select_when(_,_),_,_) :- !. % has subst type; so will not be shared
do_not_share_this_3(set_extension(X),_,outer(OuterFunctor,ArgNr)) :- !,
   %nl,print(sx(X,OuterFunctor/ArgNr)),nl,
   (X = [V], simple_explicit_value(V) ; % singleton set with simple value
   set_extension_optimized_treatment_available(OuterFunctor,ArgNr)).
do_not_share_this_3(comprehension_set(_Id,_),_,M) :-
     % optimized treatment available; important for test 1079
    (nmem(M) -> true ; symbolic_set_op(M)). % -> true ; print(share(_Id,M)),nl,nl,fail).
do_not_share_this_3(union(_,_),_,M) :-
     % union often used for composing functions
    (nmem(M) -> true ; symbolic_set_op(M)). % -> true ; print(share_union(M)),nl,nl,fail).
% lambda is already translated into comprehension_set here
%do_not_share_this_3(closure(_L),_,outer(OuterFunctor,ArgNr)) :- !,
%   closure1_optimized_treatment_available(OuterFunctor,ArgNr).
%do_not_share_this_3(sequence_extension(_L),_,outer(OuterFunctor,ArgNr)) :- !,  % we need to add b_compute_card2 for seq ?
/*
do_not_share_this_3(X,_,M) :- nmem(M), functor(X,F),
   (kernel_mappings:symbolic_closure_unary_operator(F) ;
    kernel_mappings:symbolic_closure_binary_operator(F)).

% case for unary_in_boolean_type solutions:
*/
do_not_share_this_3(pow_subset(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(fin_subset(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(pow1_subset(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(fin1_subset(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(struct(_),_,M) :- nmem(M).
do_not_share_this_3(seq(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(iseq(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(perm(_),_,M) :- nmem(M).
do_not_share_this_3(domain(_),_,M) :- nmem(M).
do_not_share_this_3(seq1(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(iseq1(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(identity(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(closure(_),_,M) :- nmem(M).
do_not_share_this_3(external_pred_call(FunName,_Args),_,_) :- not_declarative(FunName). % in principle any element containing this should not be shared (but anyway there are little gurantees about preservation and order of side-effects)
do_not_share_this_3(external_function_call(FunName,_Args),_,_) :- not_declarative(FunName).
% case for binary_not_in_boolean_type solutions:
/*
do_not_share_this_3(partial_bijection(_,_),_,M) :- nmem(M).
do_not_share_this_3(total_relation(_,_),_,M) :- nmem(M).
do_not_share_this_3(surjection_relation(_,_),_,M) :- nmem(M).
do_not_share_this_3(total_surjection_relation(_,_),_,M) :- nmem(M).
do_not_share_this_3(parallel_product(_,_),_,M) :- nmem(M).
% set_subtraction, intersection not covered
*/
do_not_share_this_3(interval(A,B),_,M) :- nmem(M) ; small_interval(A,B).
do_not_share_this_3(total_function(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(total_injection(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(total_surjection(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(total_bijection(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(partial_function(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(partial_injection(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).  % >+>
do_not_share_this_3(partial_surjection(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(relations(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(cartesian_product(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
%do_not_share_this_3(div(A,B),_,_) :- A=identifier(xx), B=integer(1000000000). % xx / 1000000000 ; purpose: test vesg performance issue
%do_not_share_this_3(range(identifier('Block_Id')),_,_). % for vesg performance test in INVARIANT

do_not_share_this_3(Stripped,UsedIds,_OuterFunctor) :- too_simple_for_sharing2(Stripped,UsedIds).

% check if expression used as second arg of member check:
nmem(outer(M,2)) :- mem_op(M).
nmem(outer(M,_)) :- (M=partial_function ; M=total_function ; M=relations).
mem_op(member).
mem_op(not_member).
symbolic_set_op(outer(M,P)) :- (M=union; M=function,P=1).

small_interval(integer(A),integer(B)) :- number(A),number(B), B < A+10000000.
infinite_set(integer_set(_)). % :- (X='INTEGER' ; X = 'NATURAL' ; X = 'NATURAL1').
infinite_set(string_set).
% TO DO : couple
simple_explicit_value(integer(_)).
simple_explicit_value(string(_)).
simple_explicit_value(boolean_true).
simple_explicit_value(boolean_false).
simple_explicit_value(identifier(_Id)). % :- b_global_sets:is_b_global_constant(_,_,Id). % no guarantee that this is really an enumerated set element because a local variable could overide this; but
% TO DO: should any identifier not be ok here: basically, we just gain the lookup when sharing ??
%simple_explicit_value(V) :- print(sev(V)),nl,fail.

set_extension_optimized_treatment_available(card,1).
set_extension_optimized_treatment_available(min,1). % we have optimized code for min({a,b,...}) and max
set_extension_optimized_treatment_available(max,1).
set_extension_optimized_treatment_available(function,1).
set_extension_optimized_treatment_available(subset,2). % there is a rule in cleanup_post which will rewrite this to member checks {x1,x2,...} <: B  <=>  x1:B & x2:B & ...; we could also try and ensure that this runs first !
set_extension_optimized_treatment_available(image,2). % ??

%closure1_optimized_treatment_available(image,1). % there is some custom code for: image_for_closure1_wf; but it is also called if we do CSE ??!!

:- use_module(external_functions,[performs_io/1]).
too_simple_for_sharing2(equal(A,B),_) :-
 (A = identifier(_Id), boolean(B) ;
  B = identifier(_Id), boolean(A) ).
too_simple_for_sharing2(identifier(Id),[Id]).
too_simple_for_sharing2(unary_minus(Int),[]) :- Int = integer(_).
too_simple_for_sharing2(integer(_),[]).
too_simple_for_sharing2(value(_),[]).
too_simple_for_sharing2(boolean_true,[]).
too_simple_for_sharing2(boolean_false,[]).
too_simple_for_sharing2(integer_set(_),[]).
too_simple_for_sharing2(bool_set,[]).
too_simple_for_sharing2(string(_),[]).
too_simple_for_sharing2(string_set,[]).
too_simple_for_sharing2(empty_set,[]).
too_simple_for_sharing2(empty_sequence,[]).
too_simple_for_sharing2(lazy_lookup_pred(_),_).
too_simple_for_sharing2(lazy_lookup_expr(_),_).
too_simple_for_sharing2(lazy_let_expr(_,_,_),_).
too_simple_for_sharing2(lazy_let_pred(_,_,_),_).
too_simple_for_sharing2(lazy_let_subst(_,_,_),_).
too_simple_for_sharing2(couple(_,_),_). % maybe could be useful nonetheless ??

%too_simple_for_sharing2(interval(integer(_),integer(_)),_).
% TO DO: do not share arithmetic expressions which can be constant folded ??
% or maybe perform constant folding at the same time ??

too_simple_for_sharing2(truth,[]).
too_simple_for_sharing2(falsity,[]).

% TO DO: completely avoid sharing those expressions ? but usually they should not occur inside predicates/expressions anyway
too_simple_for_sharing2(external_pred_call(FunName,_Args),_) :-
        external_functions:performs_io(FunName).
too_simple_for_sharing2(external_function_call(FunName,_Args),_) :-
        external_functions:performs_io(FunName).



set_references(References,sharing(_Id,_RefCounter,_Level,References)).

% the stores data type has the following form:
% stores(Counter,References,Top,Substores)
% - Counter is the next ID for new expressions, the same expressions
%   share the same id
% - References is a mapping from an ID to a term rc(C,R) where C
%   contains the number of occurrences of the expressions so far and R the
%   final number of occurrences of this expression. Until the end, R is a variable
%   that will be unified with the last value of C
% - Top is a mapping from a stripped expression to an ID. This mapping is for all
%   expressions that do not contain references to quantified variables
% - Substores is a list of terms substore(Ids,SubStore) where Ids is a list of
%   introduced identifiers in a qualified expression and Substore is a mapping from
%   a stripped expression to its ID. The expression contains references to identifiers
%   in Ids. In case of nested quantified expressions, the inner expression has
%   a substore entry before the outer expression in the list.
create_stores(Start,stores(Start,Empty,Empty,[])) :-
    empty_avl(Empty).
create_substore([],_SubId,Stores,Stores) :- !.
create_substore(NewIds,ID,InStores,SubStores) :-
    force_create_substore(NewIds,ID,InStores,SubStores).
force_create_substore(Ids,SubId,
                stores(Counter,IR,Top, InSubs),
                stores(Counter,IR,Top, [substore(OrdIds,Empty,SubId,[])|InSubs])) :-
    list_to_ord_set(Ids,OrdIds),
    empty_avl(Empty).

% remove_last_substore(+Names,+InSubStores,-OutSubStores,-References)
remove_last_substore([],Stores,Stores,[]).
remove_last_substore([_|_], % new ids have been introduced; local scope / substore was added
                     stores(Counter,IR,Top,[Substore|Subs]),
                     stores(Counter,IR,Top,Subs),
                     References) :-
    Substore = substore(_Names,_Store,_SubId,References).

% try and get references to common sub expression ids + the corresponding expressions
get_current_references(stores(_Counter,IR,TopStore,_Subs),References) :-
   avl_to_list(IR,Refs),
   convlist(reused_expression(TopStore),Refs,References).
reused_expression(TopStore,ID-rc(Count,Count),cse(ID,Expr,Count)) :- Count>1,
   avl_member(Expr,TopStore,ID). % reverse lookup; not efficient !

lookup_expression(Stripped,UsedIds,InStores,OutStores,sharing(Id,RefCounter,Level,_References),Id) :-
    InStores =  stores(Counter,InRefCounter, TopStore,OldSubStores),
    OutStores = stores(Counter,OutRefCounter,TopStore,NewSubStores),
    lookup_expression2(OldSubStores,TopStore,Stripped,UsedIds,Id,Level,NewSubStores),!,
    avl_fetch(Id,InRefCounter,rc(OC,RefCounter)),
    NC is OC+1,
    avl_store(Id,InRefCounter,rc(NC,RefCounter),OutRefCounter).
lookup_expression2([],TopStore,Stripped,_UsedIds,Id,top,[]) :-
    avl_fetch(Stripped,TopStore,Id).
lookup_expression2([substore(Names,Store,SubId,OldRefs)|Rest],_TopStore,Stripped,_UsedIds,Id,SubId,
                   [substore(Names,Store,SubId,NewRefs)|Rest]) :-
    avl_fetch(Stripped,Store,Id),!,
    ord_add_element(OldRefs,Id,NewRefs).
lookup_expression2([substore(Names,Store,SubId,Refs)|Rest],TopStore,Stripped,UsedIds,Id,Level,
                   [substore(Names,Store,SubId,Refs)|NRest]) :-
    ord_disjoint(UsedIds,Names),
    !,
    lookup_expression2(Rest,TopStore,Stripped,UsedIds,Id,Level,NRest).

%lookup_next_id(stores(Id,_,_,_),Id).

store_expression(Stripped,UsedIds,InStores,OutStores,sharing(Id,RefCounter,Level,_References),Id) :-
    InStores = stores(Id,InRefCounter,InTopStore,InSubStores),
    Counter is Id+1, % Generate a new CSE Identifier
    OutStores = stores(Counter,OutRefCounter,OutTopStore,OutSubStores),
    store_expression2(InSubStores,InTopStore,Stripped,UsedIds,Id,OutSubStores,OutTopStore,Level),
    avl_store(Id,InRefCounter,rc(1,RefCounter),OutRefCounter).
store_expression2([],InTopStore,Stripped,_UsedIds,Id,[],OutTopStore,top) :-
    avl_store(Stripped,InTopStore,Id,OutTopStore).
store_expression2([substore(Names,InStore,SubId,Refs) |Srest],TopStore,Stripped,UsedIds,Id,
                  [substore(Names,OutStore,SubId,Refs)|Srest],TopStore,SubId) :-
    ord_intersect(Names,UsedIds),!, % the Shared Expression makes use of identifiers/names introduced at that level; it cannot be re-used higher up
    avl_store(Stripped,InStore,Id,OutStore).
store_expression2([Substore|ISrest],InTopStore,Stripped,UsedIds,Id,
                  [Substore|OSrest],OutTopStore,Level) :-
    store_expression2(ISrest,InTopStore,Stripped,UsedIds,Id,OSrest,OutTopStore,Level).

:- public portray_stores/1. % for debugging
portray_stores(stores(Counter,RefCounter,TopStore,SubStores)) :-
   format('Store Counter:~w ~n Reference Counts: ',[Counter]),
   avl:portray_avl(RefCounter),nl,
   print('Top = '), avl:portray_avl(TopStore),nl,
   (maplist(portray_substore,SubStores) -> true ; print('Illegal Substores: '),nl).
portray_substore(substore(Names,OutStore,_SubId,Refs)) :-
   format('--> Substore [Names=~w, Refs=~w]~n    Store = ',[Names,Refs]),
   avl:portray_avl(OutStore),nl.

update_used_ids(Expr,NewlyIntroducedIds,UsedSubIds,UsedIds) :-
    update_used_ids2(Expr,UsedSubIds,UsedIds1),
    ord_subtract(UsedIds1,NewlyIntroducedIds,UsedIds). % remove NewlyIntroducedIds; they are not visible outside
update_used_ids2(identifier(Id),UsedSubIds,UsedIds) :- !, % we directly use an identifier
    ord_add_element(UsedSubIds,Id,UsedIds).
update_used_ids2(_Expr,UsedSubIds,UsedIds) :- !, UsedSubIds=UsedIds.

% We do not have to use strip_and_norm_ast/2 from bsyntaxtree though
% the result should be the same. We already have the stripped versions
% of the subexpressions, so we can re-use that information
create_stripped(Expr,StrippedSubs,NormStripped) :-
    (cse_syntaxtransformation(Expr,_Subs,_Names,StrippedSubs,Stripped)
     -> norm_strip(Stripped,NormStripped)
     ;  add_internal_error('Illegal Expression: ',create_stripped(Expr,StrippedSubs,NormStripped)),
        NormStripped = Expr).

ground_reference_counters(stores(_Counter,RefCounter,_TopStore,_SubStores)) :-
    avl_to_list(RefCounter,List),
    maplist(ground_reference_counter,List).
ground_reference_counter(_Id-rc(C,C)).

:- use_module(library(samsort)).
norm_strip(empty_sequence,R) :- !, R = empty_set.
norm_strip(greater_equal(A,B),R) :- !, R = less_equal(B,A).
norm_strip(greater(A,B),R) :- !, R = less(B,A).
norm_strip(set_extension(SX),R) :- !,
    sort(SX,SSX),
    R = set_extension(SSX).
norm_strip(negation(A),R) :- negate_norm_strip(A,R).
norm_strip(Old,New) :-
    functor(Old,Functor,2),
    is_commutative(Functor),!,
    norm_commutative(Functor,Old,New).
norm_strip(Old,Old).

negate_norm_strip(equal(A,B),R) :- !, R=not_equal(A,B).
negate_norm_strip(not_equal(A,B),R) :- !, R=equal(A,B).
negate_norm_strip(less_equal(A,B),R) :- !, R=less(B,A).
negate_norm_strip(less(A,B),R) :- !, R=less_equal(B,A).
negate_norm_strip(member(A,B),R) :- !, R=not_member(A,B).
negate_norm_strip(not_member(A,B),R) :- !, R=member(A,B).
negate_norm_strip(subset(A,B),R) :- !, R=not_subset(A,B).
negate_norm_strip(not_subset(A,B),R) :- !, R=subset(A,B).
negate_norm_strip(subset_strict(A,B),R) :- !, R=not_subset_strict(A,B).
negate_norm_strip(not_subset_strict(A,B),R) :- !, R=subset_strict(A,B).
negate_norm_strip(conjunct(C),R) :- !,
    maplist(negate_norm_strip,C,NC),
    R=disjunct(NC).
negate_norm_strip(disjunct(C),R) :- !,
    maplist(negate_norm_strip,C,NC),
    R=conjunct(NC).
negate_norm_strip(negation(A),R) :- !, R=A.
% what about implication, equivalence, exists, forall
negate_norm_strip(X,negation(X)).

norm_commutative(Functor,Old,New) :- is_associative(Functor), !,
    Old =.. [Functor,OA,OB],
    % now collect all args with same functor in left and right:
    get_top_level_functor_arguments(OA,Functor,LeftArgs),
    get_top_level_functor_arguments(OB,Functor,RightArgs),
    append(LeftArgs,RightArgs,Args),
    samsort(Args,SortedArgs),
    New =.. [Functor,SortedArgs].
norm_commutative(Functor,Old,New) :- % the functor is not associative (or cannot appear at top-level in OA,OB)
    Old =.. [Functor,OA,OB],
    New =.. [Functor,NA,NB],
    ( OA @>= OB -> OA=NB,OB=NA
    ; otherwise -> OA=NA,OB=NB).

% we could use this instead of samsort
% but we get error message e.g. for % (x\/y\/v\/x) /= (v\/x\/y) & x<:BOOL  as an expression appears within itself ?!
%sort_args(Functor,Args,SortedArgs) :-
% (idempotent(Functor) -> sort(Args,SortedArgs) % we can remove duplicates
%      ; samsort(Args,SortedArgs)).


get_top_level_functor_arguments(Top,Functor,Args) :- % must also be associative
    functor(Top,Functor,N), !,
    (N=2 -> Top =.. [Functor|Args] % no normalisation applied
       ; arg(1,Top,Args)). % Note: we do not recurse: we assume normalisation has already been applied fully below
get_top_level_functor_arguments(Top,_,[Top]).

%is_commutative(_) :- !,fail.
is_commutative(conjunct). % commutative only if there are no WD issues !
is_commutative(disjunct). % commutative only if there are no WD issues !
is_commutative(equivalence).
is_commutative(equal).
is_commutative(not_equal).
is_commutative(add).
is_commutative(multiplication).
is_commutative(union).
is_commutative(intersection).


%is_associative(_) :- !,fail.
is_associative(conjunct). % even if there are WD issues !
is_associative(disjunct). % even if there are WD issues !
is_associative(add).
is_associative(multiplication).
is_associative(union).
is_associative(intersection).
is_associative(concat).
is_associative(composition).

% not used at the moment:
:- public idempotent/1.
idempotent(conjunct).
idempotent(disjunct).
idempotent(union).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% removing implicit negation:
%  predicates of the form a/=b (resp. a/:b, ...) are
%  rewritten to not(a=b), (resp. not(a:b), ...) to enable sharing of the
%  predicate a=b.

remove_implicit_negation(TExpr,NTExpr) :-
    remove_bt(TExpr,Expr,NewExpr,TExpr2),
    syntaxtransformation(Expr,Subs,_Names,NSubs,NewExpr),
    remove_implicit_negation_l(Subs,NSubs),
    remove_implicit_negation2(TExpr2,NTExpr).

remove_implicit_negation_l([],[]).
remove_implicit_negation_l([OExpr|ORest],[NExpr|NRest]) :-
    remove_implicit_negation(OExpr,NExpr),
    remove_implicit_negation_l(ORest,NRest).

remove_implicit_negation2(TExpr,NTExpr) :-
    get_texpr_expr(TExpr,Expr),
    remove_implicit_negation3(Expr,NTExpr),!.
remove_implicit_negation2(TExpr,TExpr).

remove_implicit_negation3(negation(P),R) :-
    get_texpr_expr(P,negation(Q)),!,R=Q. % eliminate double negation
remove_implicit_negation3(not_equal(A,B),R) :- negate_bexpr(equal(A,B),A,B,R).
remove_implicit_negation3(not_member(A,B),R) :- negate_bexpr(member(A,B),A,B,R).
remove_implicit_negation3(not_subset(A,B),R) :- negate_bexpr(subset(A,B),A,B,R).
remove_implicit_negation3(not_subset_strict(A,B),R) :- negate_bexpr(subset_strict(A,B),A,B,R).

negate_bexpr(Expr,SubA,SubB,R) :- extract_info(SubA,SubB,Info),
    create_texpr(Expr,pred,Info,P),
    create_negation(P,R).


% TO DO: optimize away unnecessary nested LETs as below:
% Optimized pred:
% !(sddb1,sddb2).((sddb1 : INTEGER & sddb2 : INTEGER) & (sddb1 |-> sddb2 : dom(SDDBs_Signal) & (Signals_Cap__Name~)(Routes_Cap__Opposite_Signal_ID(route)) : ran(SDDBs_Signal(sddb1 |-> sddb2))) => (LET (@9)==ran(tses) IN (LET (@10)==dom(@9) IN sddb1 /: @10 & sddb2 /: @10)))


/*

Code currently not used; but possibly to be re-included :


% a version of conjunction which recomputes the cse_used_ids INFO
cse_conjunct_predicates(L,Res) :- cse_conjunct_predicates2(L,b(E,pred,I)),
    (collect_uses_cse_id(L,CSEINFO)
      -> Res = b(E,pred,[cse_used_ids(CSEINFO)|I1]),
         exclude(is_cse_used_ids,I,I1) %delete(I,cse_used_ids(_),I1)
      ;  print('*** Could not collect uses_cse_id INFO :'), print(L),nl,
         Res = b(E,pred,I)).

% sometimes we get identical conjuncts; TO DO: generalize to longer lists,...  (test 351 still leaves redundant conjuncts inside let body)
cse_conjunct_predicates2([A,B],R) :- A==B, !, R=A.
cse_conjunct_predicates2(L,R) :- conjunct_predicates(L,R).


% find prefix list of conjuncts which do not use CSEID:
find_prefix_not_using_cse_id([],_CSEID,[],[]).
find_prefix_not_using_cse_id([H|T],CSEID,Prefix,Rest) :-
   (uses_cse_id(CSEID,H)
    -> Prefix=[], Rest= [H|T]
     ; % print('NOT USING '), print(CSEID),nl, translate:print_bexpr_or_subst(H),nl,
       Prefix = [H|TP],
       find_prefix_not_using_cse_id(T,CSEID,TP,Rest)).

% split a conjunct into  PREFIX &  MAIN  & SUFFIX   such that PREFIX/SUFFIX do not use CSEID identifier
cse_split_conjunct(Conjunct,CSEID,  PREFIX,MAIN,SUFFIX) :-
    is_a_conjunct(Conjunct,_,_),
    % try removing conjuncts which do not use CSEID by flattening first:
    conjunction_to_list(Conjunct,List),
    find_prefix_not_using_cse_id(List,CSEID,PREFIX,Rest1),
    reverse(Rest1,RevRest1),
    find_prefix_not_using_cse_id(RevRest1,CSEID,RevSuffix,RevMain),
    (PREFIX \= [] ; RevSuffix \= []), % only succeed if there is a non-trivial split
    reverse(RevMain,MAIN),
    reverse(RevSuffix,SUFFIX).


% binders which pose problem for lazy_let introduction in case SharedExpression depends on bound variables
non_unary_binder(b(P,_Type,_I),Ids) :-
  non_unary_binder_aux(P,Ids,_Left,_Right).
non_unary_binder_aux(forall(Ids,A,B),Ids,A,B).
non_unary_binder_aux(let_predicate(Ids,A,B),Ids,A,B).
% expressions
non_unary_binder_aux(general_sum(Ids,A,B),Ids,A,B).
non_unary_binder_aux(general_product(Ids,A,B),Ids,A,B).
non_unary_binder_aux(quantified_union(Ids,A,B),Ids,A,B).
non_unary_binder_aux(quantified_intersection(Ids,A,B),Ids,A,B).
non_unary_binder_aux(general_product(Ids,A,B),Ids,A,B).
non_unary_binder_aux(event_b_comprehension_set(Ids,A,B),Ids,A,B).
non_unary_binder_aux(lambda(Ids,A,B),Ids,A,B).
non_unary_binder_aux(let_expression(Ids,A,B),Ids,A,B).
%subst
non_unary_binder_aux(let(Ids,A,B),Ids,A,B).
non_unary_binder_aux(any(Ids,A,B),Ids,A,B).
% TO DO: add more subst

% binders which only have a single sub-expression; here we just have to decide whether to introduce the let inside or outside
unary_binder(b(P,_Type,_I),Ids) :-
  unary_binder_aux(P,Ids,_Left).
unary_binder_aux(exists(Ids,A),Ids,A).
unary_binder_aux(comprehension_set(Ids,A),Ids,A).


*/
