% (c) 2016-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(b_operation_cache, [ %project_in_state/3,
                               %operation_computed/4,
                               compute_operation_on_expanded_store_cache/5,
                               print_op_cache/0
                    ]).


:- use_module(module_information).
:- module_info(group,animator).
:- module_info(description,'This module caches B operation results on projected states.').


:- use_module(extension('counter/counter')). % op_cache_id counter

b_operation_cache_startup :- % call once at startup to ensure all counters exist
      counter_init,
      new_counter(op_cache_id).

:- use_module(eventhandling,[register_event_listener/3]).

:- register_event_listener(startup_prob,b_operation_cache_startup,
                           'Initialise Statespace Counters.').
:- register_event_listener(clear_specification,reset_b_operation_cache,
                           'Reset Operation Cache.').
:- register_event_listener(change_of_animation_mode,reset_b_operation_cache,
                           'Reset Operation Cache.').
% TO DO: also reset if maxNrOfEnablingsPerOperation changed ?

:- use_module(error_manager).
:- use_module(bmachine,[b_get_operation_normalized_read_write_info/3]).

:- dynamic operation_read_projection_cache/2.
reset_b_operation_cache :-
   retractall(operation_read_projection_cache(_,_)),
   retractall(operation_computed(_,_,_,_)),
   retractall(operation_cached_results(_,_,_,_)),
   reset_counter(op_cache_id).

% project the in-state for an operation onto the relevant variables
project_in_state(InState,OpName,ProjectedInState ) :-
   % TO DO: statically pre-compute which operations are worthwhile for this treatment
   operation_read_projection_cache(OpName,ProjVars),
   !,
   ProjVars \= no_caching,
   (re_project_state(ProjVars,InState,Res) -> ProjectedInState=Res
    ; add_internal_error('Re-Projection failed: ',re_project_state(ProjVars,InState,_)),
      ProjectedInState = InState).
%project_in_state(_,OpName,_ ) :- \+ worthwhile_to_cache(OpName),!,
%   assert(operation_read_projection_cache(OpName,no_caching)),
%   fail.
project_in_state(InState,OpName,ProjectedInState ) :-
   b_get_operation_normalized_read_write_info(OpName,ReadVariables,_WrittenVariables),
   project_state_onto_vars(InState,ReadVariables,ProjVarsInOrder,RelevantState,0,NrDeleted),
   (NrDeleted>0
    -> assert(operation_read_projection_cache(OpName,ProjVarsInOrder)),
      %print(proj(OpName,PredResult,RelevantState)),nl,
      ProjectedInState = RelevantState
    ; assert(operation_read_projection_cache(OpName,no_caching)),
      print(not_caching(OpName)),nl,
      fail).


% check if caching is worthwhile for this operation
%worthwhile_to_cache(OpName) :-
%   b_get_operation_normalized_read_write_info(OpName,ReadVariables,WrittenVariables),
%   length(ReadVariables,NrRV),
%   bmachine:b_machine_statistics(variables,NrVars),
%   bmachine:b_machine_statistics(constants,NrConstants),
%   Perc is (NrRV*100) // (NrVars+NrConstants),
%   length(WrittenVariables,NrWV),
%   format('Analyzing ~w, ~w %, read: ~w, written: ~w, tot vars: ~w, tot cst: ~w~n  (read: ~w)~n',[OpName,Perc,NrRV,NrWV,NrVars,NrConstants,ReadVariables]),
%   Perc < 95. % maybe provide as preference

:- use_module(library(ordsets),[ord_member/2]).
%is_read(ReadVariables,bind(Var,_)) :- ord_member(Var,ReadVariables).

% a variation of split_list which also returns a list of predicate results
% with re_project_state(L,ProjVarsInOrder,A) : we can split another list using the same pattern

project_state_onto_vars([],_,[],[],Acc,Acc).
project_state_onto_vars([Elem|Rest],ReadVariables,ProjVars,A,NrDelAcc,NrDel) :- Elem = bind(Var,_),
    (ord_member(Var,ReadVariables)
      -> ProjVars=[Var|PT],  A=[Elem|AR], ND1=NrDelAcc
     ;   ProjVars=PT,        A=AR, ND1 is NrDelAcc+1),
    project_state_onto_vars(Rest,ReadVariables,PT,AR,ND1,NrDel).

% project state(VarsInOrder,State,ProjectedState)
% faster than project_state_onto_vars
re_project_state([],_,[]).
re_project_state([V|PT],[Elem|Rest],A) :- Elem = bind(Var,_),
    (V==Var -> A=[Elem|AR],re_project_state(PT,Rest,AR)
    ; re_project_state2(V,PT,Rest,A)).

re_project_state2(V,PT,[Elem|Rest],A) :- Elem = bind(Var,_),
    (V==Var -> A=[Elem|AR],re_project_state(PT,Rest,AR)
    ; re_project_state2(V,PT,Rest,A)).

% -------------------------------

:- use_module(hashing).
:- use_module(state_packing).

:- dynamic operation_computed/4, operation_cached_results/4.

% return Hash and check if operation computed
% TO DO: do not hash constants if single value exists ?
operation_was_computed(OpName,State,PackedState,Hash,Res) :-
     pack_values(State,PackedState),
     hashing:my_term_hash(op(OpName,PackedState),Hash),
     %terms:term_hash(op(OpName,PackedState),[range(smallint),algorithm(sdbm),depth(infinite),if_var(ignore)],Hash),
     (operation_computed(Hash,OpName,PackedState,ID) -> Res = ID
      ; Res= false).

get_new_operation_computed_id(OpName,PackedState,Hash,ID) :-
  inc_counter(op_cache_id,ID),
  assert(operation_computed(Hash,OpName,PackedState,ID)).

:- use_module(probsrc(b_interpreter),[b_execute_top_level_operation_update/5]).
compute_operation_on_expanded_store_cache(OpName,Operation,InState,NewState,PathInfo) :-
     %print(comp(OpName,InState)),nl,
     project_in_state(InState,OpName,ProjInState),
     !,
     %print(proj(OpName,ProjInState)),nl,
     operation_was_computed(OpName,ProjInState,PackedState,Hash,ID),
     %print(o(Hash,ID,OpName)),nl,
     (ID \= false
       -> %print(cache(OpName,ID)),nl,
          %print('+'),
          operation_cached_results(ID,Operation,PackedNewState,PathInfo),
          unpack_values(PackedNewState,NewState)
       ;  get_new_operation_computed_id(OpName,PackedState,Hash,NewID),
          %print('-'),
          b_interpreter:b_execute_top_level_operation_update(OpName,Operation,ProjInState,NewState,PathInfo),
          pack_values(NewState,PackedNewState),
          assert(operation_cached_results(NewID,Operation,PackedNewState,PathInfo))
    ).
compute_operation_on_expanded_store_cache(OpName,Operation,InState,NewState,PathInfo) :-
    %print(not_caching(OpName)),nl,
    b_interpreter:b_execute_top_level_operation_update(OpName,Operation,InState,NewState,PathInfo).


% small utility code to analyze hash collisions:
% b_operation_cache:ahc.
:- public ahc/0.
ahc :- operation_computed(Hash,OpName,State,ID),
       operation_computed(Hash,OpName2,State2,ID2),
       ID2 > ID,
       format('Hash collision: ~w~n ~w:~w:~w~n ~w:~w:~w~n~n',[Hash,OpName,State,ID,OpName2,State2,ID2]),
       fail.
ahc.

:- use_module(translate,[translate_bstate_limited/2, translate_event_with_limit/3]).
% b_operation_cache:print_op_cache
print_op_cache :- \+ operation_computed(_,_,_,_),!. % print nothing
print_op_cache :- operation_read_projection_cache(OpName,ProjVars),
      format('Operation ~w cached onto: ~w~n',[OpName,ProjVars]),fail.
print_op_cache :-
      print('Local Operation Cache Info:'),nl,
      operation_computed(_Hash,OpName,PackedState,ID),
      format('~nNode ID = ~w, Operation = ~w~n',[ID,OpName]),
      unpack_values(PackedState,State), translate:translate_bstate_limited(State,NodeDesc),
      format(' projected state : ~w~n',[NodeDesc]),
      operation_cached_results(ID,Operation,PackedNewState,_PathInfo),
      translate:translate_event_with_limit(Operation,100,OpStr),
      unpack_values(PackedNewState,NewState), translate:translate_bstate_limited(NewState,UpdateStr),
      format('  ~w -upd-> ~w~n',[OpStr,UpdateStr]),
      fail.
print_op_cache :- ahc,print('----------'),nl.

% ----------------------
% DOT RENDERING (not yet finished)

%op_cache_node(ID,OpName,NodeDesc,Shape,Style,Color) :- Shape=box, Style=solid, Color=blue,
%    operation_computed(_Hash,OpName,PackedState,ID),
%    unpack_values(PackedState,State), translate_bstate_limited(State,NodeDesc).
%:- use_module(probsrc(dot_graph_generator),[gen_dot_graph/6]).
% Predicates for creating a dependency graph
%tcltk_operation_cache_graph(File) :- gen_dot_graph(File,b_operation_cache,op_cache_node,cbc_dependence_trans_predicate,none,none).

