% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(b_read_write_info,[get_accessed_vars/4,
                             recompute_while_read_write_info/2,
                             get_nondet_modified_vars/3,
                             tcltk_read_write_matrix/1,
                             tcltk_get_guards_for_ops/1,
                             tcltk_get_parameter_free_guards_for_ops/1,
                       b_get_read_write/3,
                       b_get_operation_read_guard_vars/3,
                       b_get_read_write_vars/5,
                       %,b_get_read_may_must_write/7  % currently not used outside

                       get_lhs_assigned_identifier/2
                       ]).


:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'ProBs type checker and type inference module.').

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(self_check).
:- use_module(error_manager).
:- use_module(bsyntaxtree).


recompute_while_read_write_info(TExpr,NewTExpr) :-
   (transform_bexpr(b_read_write_info:update_while_info,TExpr,NewTExpr)
     -> true
     ;  add_internal_error('Call failed: ',transform_bexpr(b_read_write_info:update_while_info,TExpr,NewTExpr)),
        NewTExpr=TExpr).

:- use_module(translate,[print_bexpr/1]).
:- public update_while_info/2.
update_while_info(b(WHILE,subst,OldInfo),b(WHILE,subst,NewInfo)) :-
    WHILE = while(_TPred,_TBody,_TInv,TVar),
    get_accessed_vars(b(WHILE,subst,OldInfo),[],Modifies,Reads),
    (debug:debug_mode(off) -> true
     ; nl,print('UPDATING WHILE with VARIANT: '),translate:print_bexpr(TVar),nl,
       format(' WHILE Reads : ~w~n   Modifies : ~w~n',[Reads,Modifies])
    ),
    (delete(OldInfo,reads(OldReads),OI1) -> true ; OldReads=[], OI1=OldInfo),
    (delete(OI1,modifies(OldModifies),OI2) -> true ; OldModifies=[], OI2=OI1),
    ((Reads,Modifies)=(OldReads,OldModifies)
      -> NewInfo=OldInfo
      ;  format('***** different from Reads : ~w~n   Modifies : ~w~n',[OldReads,OldModifies]),
         ord_subtract(Reads,OldReads,NewR),
         ord_subtract(OldReads,Reads,DelR),
         format(' added: ~w~n deleted: ~w~n****~n',[NewR,DelR]),
         NewInfo = [modifies(Modifies),reads(Reads)|OI2]
    ).

% TO DO: also update info for operation call !

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(bsyntaxtree,[get_global_identifiers/1]).
:- use_module(library(ordsets)).
/* get_accessed_vars/4 computes the variables that are read or written to
   by a substitution.
   get_accessed_vars(+Subst,+LocalVars,-Modified,-Reads):
    Subst is the substitution (as a typed syntax tree)
    LocalVars is a list of (untyped) ids that are local variables.
       Local variables are ignored and not included in the list of read or written
       variables.
    Modified is the list of (untyped) identifiers that are written by the substitution and
    Reads is the list of (untyped) identifiers that are read. */
get_accessed_vars(TSubst,LocalVars,Modified,Reads) :-
    get_global_identifiers(IS), % TO DO: currently this information is set up only *after* type checking !!
    % this means the accessed vars will also contain SETS and enumerated set elements !
    % either we need to filter them out later, or change the startup / precompilation process
    list_to_ord_set(LocalVars,SLV),
    ord_union(IS,SLV,LocalVarsToIgnore),
    get_accessed_vars1(TSubst,LocalVarsToIgnore,In,Out),
    extract_reads_and_modifies(In,Out,Modified,Reads).
extract_reads_and_modifies(mrinfo(UnsortedMods,_,UnsortedReads),mrinfo([],[],[]),Modified,Reads) :-
    sort(UnsortedMods,Modified),
    sort(UnsortedReads,Reads).

% get variables which are modified non-deterministically; does not traverse operation calls at the moment
get_nondet_modified_vars(TSubst,LocalVars,NonDetModified) :-
    get_accessed_vars1(TSubst,LocalVars,In,Out),
    extract_non_det_modifies(In,Out,NonDetModified).
extract_non_det_modifies(mrinfo(_,UnsortedMods,_),mrinfo([],[],[]),NonDetModified) :-
    sort(UnsortedMods,NonDetModified).

get_accessed_vars1(TSubst,LocalVars,In,Out) :-
    get_texpr_expr(TSubst,Subst),
    ( Subst = operation_call(Operation,Results,Arguments) ->
        get_texpr_info(TSubst,Info),
        def_get_texpr_id(Operation,op(OperationName)), %print(info(OperationName,Info)),nl,
        memberchk(modifies(MVars),Info),
        memberchk(reads(RVars),Info),
        %op_modifies_reads_cache(OperationName,MVarsC,RVarsC) %this is no longer necessary
        ((ground(MVars), ground(RVars)) -> true
          ; add_error(b_read_write_info, 'Cannot obtain modifies/reads info for called operation (probably illegal local call): ',OperationName),
             % TEST 655 used to reach this branch
            MVars=[], RVars=[]
           % this should only happen when local operation called !?!
           % or ensure that operations are processed before we reach operation calls
           % this may be problematic for local operations
        ),
        % print(compute_acc_vars_for_op_call(modifies(MVars),Subst)),nl, %trace,
        accessed_vars_for_operation_call(MVars,RVars,LocalVars,Results,Arguments,In,Out)
    ; get_accessed_vars2(Subst,LocalVars,In,Out) ->
        true
    ; Subst = identifier(ID) -> add_error(b_read_write_info,'Identifier not a valid B Operation: ',ID,TSubst),In=Out
    ; otherwise ->
        add_internal_error('Could not extract information about accessed variables for substitution',
                           get_accessed_vars2(Subst,LocalVars,In,Out)),
                           %trace, get_accessed_vars2(Subst,LocalVars,In,Out),
                           fail).
accessed_vars_for_operation_call(MVars,RVars,LocalVars,Results,Arguments) -->
    add_non_local_modified_vars(MVars,LocalVars),
    add_non_local_read_vars(RVars,LocalVars),
    add_modified_vars(Results,LocalVars),
    add_read_vars_l(Arguments,LocalVars).

get_accessed_vars2(skip,_) --> !.
get_accessed_vars2(block(Subst),LocalVars) --> get_accessed_vars1(Subst,LocalVars).
get_accessed_vars2(init_statement(Subst),LocalVars) --> get_accessed_vars1(Subst,LocalVars).
get_accessed_vars2(assertion(A,Subst),R) -->   add_read_vars(A,R),get_accessed_vars1(Subst,R).
get_accessed_vars2(assign(LHS,E),R) -->        %Note: LHS can be nested function application f(a)(b) :=
      {maplist(get_lhs_assigned_identifier,LHS,Ids)}, add_modified_vars(Ids,R),
      {convlist(get_lhs_read_expression,LHS,LHS_R), append(LHS_R,E,RE)}, add_read_vars_l(RE,R).
get_accessed_vars2(assign_single_id(Id,E),R) -->
      add_read_vars_l([E],R), % Note: LHS is a single Identifier: nothing is ever read there
      add_modified_vars([Id],R).
get_accessed_vars2(precondition(P,Subst),R) -->add_read_vars(P,R),get_accessed_vars1(Subst,R).
get_accessed_vars2(assertion(Subst),R) -->     get_accessed_vars1(Subst,R).
get_accessed_vars2(choice(Substs),R) -->       get_accessed_vars_for_choice(Substs,R).
get_accessed_vars2(if(Substs),R) -->           get_accessed_vars_for_if(Substs,R).
get_accessed_vars2(select(Whens,Else),R) -->   get_accessed_vars_for_select(Whens,Else,R).
get_accessed_vars2(select(Whens),R) -->        get_accessed_vars_for_select(Whens,none,R).
get_accessed_vars2(case(Expr,Substs,Else),R) -->  get_accessed_vars_for_case(Expr,Substs,Else,R).
get_accessed_vars2(case_or(Exprs,Subst),R) --> add_read_vars_l(Exprs,R),get_accessed_vars1(Subst,R).
get_accessed_vars2(any(Ids,G,Subst),R) -->     get_mrinfo(ModIn,_,_),
    append_vars(R,Ids,L),add_read_vars(G,L),
    get_accessed_vars1(Subst,L),
    add_delta_modified_to_non_det_modified(ModIn).
get_accessed_vars2(let(Ids,G,Subst),R) -->     append_vars(R,Ids,L),add_read_vars(G,L),get_accessed_vars1(Subst,L).
get_accessed_vars2(lazy_let_subst(ID,SharedExpr,Subst),R) -->     append_vars(R,[ID],L),
     add_read_vars(SharedExpr,L),get_accessed_vars1(Subst,L).
get_accessed_vars2(becomes_element_of(Ids,E),R) --> add_non_det_modified_vars(Ids,R),add_read_vars(E,R).
get_accessed_vars2(becomes_such(Ids,E),LocalVars0) -->  add_non_det_modified_vars(Ids,LocalVars0),
    {append_primed_vars(Ids,LocalVars0,LocalVars1)}, % all Ids are already marked as accessed, any reference to Id$0 is a read of the old value; do not add Id$0 to the accessed variables
    add_read_vars(E,LocalVars1).
%get_accessed_vars2(evb2_becomes_such(Ids,E),R) --> % now translated to becomes_such
% add_non_det_modified_vars(Ids,R), add_read_vars(E,R).
get_accessed_vars2(sequence(Substs),R) -->     get_accessed_vars_l(Substs,R).
get_accessed_vars2(parallel(Substs),R) -->     get_accessed_vars_l(Substs,R).
get_accessed_vars2(while(Cond,Subst,Inv,Variant),R) -->
    get_accessed_vars_for_while(Cond,Subst,Inv,Variant,R).
get_accessed_vars2(var(Ids,Subst),R)         --> append_vars(R,Ids,L),get_accessed_vars1(Subst,L).
% TODO: we need information about which variables are modified by an external function
get_accessed_vars2(external_subst_call(_Name,Exprs),R) --> add_read_vars_l(Exprs,R).
get_accessed_vars2(rlevent(_Name,_Section,_Status,Params,Guard,Theorems,Actions,VWitnesses,PWitnesses,_Unmod,AbstractEvents),R) --> append_vars(R,Params,L),
    add_read_vars(Guard,L),
    add_read_vars_l(Theorems,L),
    add_witness_read_vars_l(VWitnesses,L),
    add_witness_read_vars_l(PWitnesses,L),
    get_accessed_vars_l(Actions,L),
    get_accessed_vars_l(AbstractEvents,L).

get_accessed_vars_for_while(Cond,Subst,Inv,Variant,Locals) -->
    get_accessed_vars_for_while2(Inv,Locals),
    add_read_vars(Cond,Locals),
    add_read_vars(Variant,Locals),
    get_accessed_vars1(Subst,Locals).
get_accessed_vars_for_while2(Inv,Locals,In,Out) :-
    add_read_vars(Inv,Locals,LI,LO),
    extract_reads_and_modifies(LI,LO,Modified,Reads0),
    maplist(remove_optional_prime,Reads0,Reads),
    add_non_local_read_vars(Reads,Locals,In,Inter),
    add_non_local_modified_vars(Modified,Locals,Inter,Out).
remove_optional_prime(Id,Orig) :-  atom_concat(Orig,'$0',Id),!.
remove_optional_prime(Id,Id).

get_accessed_vars_for_if(ElseIfs,Locals) -->
    { maplist(split_elseif,ElseIfs,Conds,Subst),
      ( if_has_else_block(Conds) -> HasElse = has_else
      ; otherwise                -> HasElse = no_else) },
    % mark all used variables in the conditions as read
    add_read_vars_l(Conds,Locals),
    get_accessed_vars_for_if_when(HasElse,Subst,Locals).
split_elseif(Expr,Cond,Subst) :- get_texpr_expr(Expr,if_elsif(Cond,Subst)).

get_accessed_vars_for_select(Whens,Else,Locals) -->
    { maplist(split_when,Whens,Conds,Subst1),
      ( Else = none ->
          HasElse=no_else, Subst = Subst1
      ; otherwise ->
          HasElse=has_else, Subst = [Else|Subst1]) },
    % mark all used variables in the conditions as read
    add_read_vars_l(Conds,Locals),
    get_accessed_vars_for_if_when(HasElse,Subst,Locals).
split_when(Expr,Cond,Subst) :- get_texpr_expr(Expr,select_when(Cond,Subst)).

get_accessed_vars_for_if_when(HasElse,Subst,Locals) -->
    { get_accessed_vars_for_if_when2(HasElse,Subst,Locals,Modified,Reads) },
    add_read_vars2(Reads),
    %add_modified_vars2(Modified)
    add_non_det_modified_vars2(Modified).

get_accessed_vars_for_if_when2(has_else,Subst,Locals,ModifiedBySome,ReadBySome) :-
    maplist(get_accessed_vars_x(Locals),Subst,Modifies,Reads),
    % We want to calculate those elements that are modified by some (in the union)
    % but not by all (in the intersection)
    ord_union(Modifies,ModifiedBySome),
    ord_intersection(Modifies,ModifiedByAll),
    ord_subtract(ModifiedBySome,ModifiedByAll,ModifiedByNotAll),
    % we add those values that are not modified by all to the set of read variables
    % because e.g. "IF P THEN x:=1 END" is equivalent to "IF P THEN x:=1 ELSE x:=x END"
    ord_union([ModifiedByNotAll|Reads],ReadBySome).
get_accessed_vars_for_if_when2(no_else,Subst,Locals,ModifiedBySome,ReadBySome) :-
    % no else block present, we must assume that each written variable
    % can also be treated like a read variable (see example in the other condition)
    maplist(get_accessed_vars_x(Locals),Subst,Modifies,Reads),
    ord_union(Modifies,ModifiedBySome),
    ord_union([ModifiedBySome|Reads],ReadBySome).

get_accessed_vars_for_choice(Substs,Locals,In,mrinfo(ModOut,NDOut2,Read)) :-
    get_accessed_vars_for_if_when(has_else,Substs,Locals,In,Out),
    In = mrinfo(ModIn,_,_),
    Out = mrinfo(ModOut,NDOut,Read), % these three fields will be all variables
     %  mark all assignments as non-deterministically chosen:
    add_modified_to_non_det_modified(ModIn,ModOut,NDOut,NDOut2).

add_modified_to_non_det_modified(ModIn,ModOut,NDIn,NDOut) :- ModIn==ModOut,!,NDIn=NDOut.
add_modified_to_non_det_modified([H|ModIn],ModOut,[H|NDIn],NDOut) :-
    add_modified_to_non_det_modified(ModIn,ModOut,NDIn,NDOut). % duplicates in NDOut not an issue !?!

get_mrinfo(Modified,NonDetModified,InRead,I,I) :-
   I = mrinfo(Modified,NonDetModified,InRead).

add_delta_modified_to_non_det_modified(ModIn,mrinfo(ModOut,NonDet,Read),mrinfo(ModOut,NonDet2,Read)) :-
   add_modified_to_non_det_modified(ModIn,ModOut,NonDet,NonDet2).

get_accessed_vars_for_case(Expr,Eithers,Else,Locals) -->
    % we can treat a case statement analogously to an IF-THEN-ELSE statement
    % where an ELSE is present
    % additionally, we have to add all expressions to the read set
    add_read_vars(Expr,Locals),
    { maplist(split_case_or,Eithers,ExprsList,Substs), append(ExprsList,Exprs) },
    add_read_vars_l(Exprs,Locals),
    get_accessed_vars_for_if_when(has_else,[Else|Substs],Locals).
split_case_or(CaseOr,Exprs,Subst) :- get_texpr_expr(CaseOr,case_or(Exprs,Subst)).

% just the same as get_accessed_vars, but with different order of parameters
% to allow call with maplist
get_accessed_vars_x(Locals,Subst,Modifies,Reads) :-
    get_accessed_vars(Subst,Locals,Modifies,Reads).

if_has_else_block(Conds) :-
    last(Conds,Cond),is_truth(Cond).

append_vars(Vars,TIds,NewVars,X,X) :-
    get_texpr_ids(TIds,Ids),
    append(Vars,Ids,NewVars).

add_read_vars_l([],_LocalVars) --> !.
add_read_vars_l([Expr|Rest],LocalVars) --> !,
    add_read_vars(Expr,LocalVars),
    add_read_vars_l(Rest,LocalVars).
add_read_vars_l(E,LV) --> !, {add_internal_error('Not a list: ',add_read_vars_l(E,LV,_,_))}.

add_witness_read_vars_l([],_LocalVars) --> !.
add_witness_read_vars_l([Expr|Rest],LocalVars) -->
    add_witness_read_vars(Expr,LocalVars), !,
    add_witness_read_vars_l(Rest,LocalVars).
add_witness_read_vars_l(E,LV) --> !, {add_internal_error('Not a witness list: ',add_witness_read_vars_l(E,LV,_,_))}.

% extract modified identifier from LHS of assign:
get_lhs_assigned_identifier(b(function(LHS,_),_,_),Res) :- !, get_lhs_assigned_identifier(LHS,Res).
get_lhs_assigned_identifier(ID,ID).

% extract part of LHS of assign which reads information:
get_lhs_read_expression(b(function(LHS,Read1),T,I),Res) :- !,
    Res = b(function(LHS,Read1),T,I).
% old version which for f(a) := b only  marked a as read in LHS f(a); now we also mark f
% indeed f(a) := b is equivalent to f := f <+ {a|->b} ; so the old value of f is implicitly read (important for ltsmin symbolic model checking)
%get_lhs_read_expression(b(function(LHS,Read1),_,_),Res) :- !,
%   (get_lhs_read_expression(LHS,Read2) -> create_couple(Read1,Read2,Res)
%     ; Res = Read1).

add_witness_read_vars(b(witness(ID,Predicate),witness,_),LocalVars) --> !,
   % a parameter witness talks about a parameter no longer there: it is not really read here, but the predicate is used to set the any local parameter in another context
   % a variable witness alks about a variable non-deterministically assigned in the abstraction; we also pretend it is not read here (but assigned somewhere else)
   append_vars(LocalVars,[ID],LocalVars2), % we add ID to local variables not read "really" read here
   %{print(witness_add_read_vars(ID)),nl},
   add_read_vars(Predicate,LocalVars2).

add_read_vars(Expression,LocalVars) -->
    {find_identifier_uses(Expression,LocalVars,ReadVars)},
    add_read_vars2(ReadVars).
add_read_vars2(Vars,mrinfo(Modified,NonDetModified,InRead),
                    mrinfo(Modified,NonDetModified,OutRead)) :- append(Vars,OutRead,InRead).

add_non_local_read_vars(ReadVars,LocalVars) -->
    {ord_subtract(ReadVars,LocalVars,AddVars)},
    add_read_vars2(AddVars).
add_non_local_modified_vars(WriteVars,LocalVars) -->
    {ord_subtract(WriteVars,LocalVars,AddVars)},
    add_modified_vars2(AddVars).

add_modified_vars(Ids,LocalVars) -->
    {filter_result_vars(Ids,LocalVars,ModifiedVars)},
    add_modified_vars2(ModifiedVars).
add_modified_vars2(Vars,In,Res) :-
    (var(Vars)
     -> add_internal_error('Modified Vars variable: ',add_modified_vars2(Vars,In,Res))
      ; true),
    In = mrinfo(InModified,NonDetModified,Read),
    Res = mrinfo(OutModified,NonDetModified,Read),
    append(Vars,OutModified,InModified).

add_non_det_modified_vars(Ids,LocalVars) -->
    {filter_result_vars(Ids,LocalVars,ModifiedVars)},
    add_non_det_modified_vars2(ModifiedVars).
add_non_det_modified_vars2(Vars,In,Res) :-
    (var(Vars)
     -> add_internal_error('Modified Vars variable: ',add_non_det_modified_vars2(Vars,In,Res))
      ; true),
    In = mrinfo(InModified,InNonDetModified,Read),
    Res = mrinfo(OutModified,OutNonDetModified,Read),
    append(Vars,OutModified,InModified),
    append(Vars,OutNonDetModified,InNonDetModified).


:- use_module(tools,[remove_all/3]).
filter_result_vars(Ids,LocalVars,ModifiedVars) :-
    % extract_changed_id might fail if the user entered
    % a wrong LHS of an assignment. This should lead to an
    % type error, handled by the assign type checking rule.
    convlist(extract_changed_id,Ids,Names),
    remove_all(Names,LocalVars,ModifiedVars).
extract_changed_id(Id,Name) :-
    get_texpr_id(Id,Name),!.
extract_changed_id(Func,Name) :-
    get_texpr_expr(Func,function(TId,_)), !,
    extract_changed_id(TId,Name). % we can have nested function calls f(x)(y) := z
extract_changed_id(X,Name) :-
    add_internal_error('Could not extract changed id: ',extract_changed_id(X,Name)),fail.

get_accessed_vars_l([],_) --> [].
get_accessed_vars_l([S|Rest],R) --> get_accessed_vars1(S,R),get_accessed_vars_l(Rest,R).



:- use_module(tools,[number_suffix/3]).
:- assert_must_succeed(( b_read_write_info:append_primed_vars([b(identifier(x),integer,[])],[],R), R = ['x$0'])).
append_primed_vars(TIds,R,L) :-
    maplist(add_dollar_zero,TIds,Primed),
    append(Primed,R,L).
:- assert_must_succeed(( b_read_write_info:add_dollar_zero(b(identifier(x),integer,[]),R), R = 'x$0')).
add_dollar_zero(TId,P) :-
    get_texpr_id(TId,Id),
    number_suffix(Id,0,P).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% READ-WRITE-INFO for OPERATIONS

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


:- use_module(bmachine,[b_get_operation_normalized_read_write_info/3, b_get_machine_operation/4,
                        b_get_machine_variables/1, b_get_machine_constants/1]).

b_get_read_write(O,R,W) :-
  if(b_get_operation_normalized_read_write_info(O,R,W),true,
      (R,W)=('$all','$all')). % it used to be that in Event-B no read information was stored; all calls to b_get_operation_normalized_read_write_info failed; has been fixed

b_get_machine_operation_with_local_paras(OpName,AllLocalParas,Body) :-
   b_get_machine_operation(OpName,Results,Parameters,Body),
   append(Parameters,Results,TP),
   get_texpr_ids(TP,TPIds),
   list_to_ord_set(TPIds,AllLocalParas).

/* We distinguish between read variables in the guard and read variables in the event statement */
b_get_read_write_guard_action(Operation,ReadGuard,ReadAction,Read,Write) :-
    b_get_read_write(Operation,Read,Write),
    b_get_operation_read_guard_vars(Operation,false,ReadGuard),
    b_get_machine_operation_with_local_paras(Operation,LocalParas,Body),
    get_texpr_expr(Body,RLEVENT),
    (RLEVENT = rlevent(_Name,_Section,_Stat,_Parameters,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,_AbsEvents) ->
     % special code to deal with *all* guards/actions (also in abstractions) + not extract guards from actions
     % TO DO: maybe move this to get_guard/get_action (see below) ; or maybe provide parameter to choose
     % in some circumstances (Deadlock checking) it is sufficient to look at lowest refinement level
     % for LTSMin,... : we need info from all refinement levels and from theorems
     findall(A1,(get_full_eventb_action(RLEVENT,AA),member(A1,AA)),A)
  ; get_action(Operation,A) -> true
  ; otherwise -> add_error(b_get_read_write_guard_action,'Could not get action for operation: ',Operation), fail
  ),
    (is_list(A) % in case of Event-B events the action block is a list of assignments
        -> find_identifier_uses_l(A,LocalParas,RActionVars) % TO DO: use get_accessed_vars ?
        ;  get_accessed_vars(A,LocalParas,_,RActionVars)
    ),
  list_to_ord_set(RActionVars,RActionVars1),
  rw_ord_intersection(RActionVars1,Read,ReadAction).
  %,debug_println(9,b_get_read_write(O,RGuard,R,W)).

% this gets the variables/identifiers read in the guard of an operation
% Note: this part will include parts of the guard that depend on the parameters
b_get_operation_read_guard_vars(Operation,JustVariables,Result) :-
  b_get_machine_operation_with_local_paras(Operation,Paras,_),
  get_operation_guard(Operation,G),
  % RGuardVars can comprise also local variables by operations with parameters
  find_identifier_uses(G,Paras,RGuardVars),
  list_to_ord_set(RGuardVars,ReadGuard),
  (JustVariables=false
    -> Result = ReadGuard
    ;  b_get_machine_variables(TVars),get_texpr_ids(TVars,Vars),
       list_to_ord_set(Vars,Variables),
       ord_intersection(ReadGuard,Variables,Result)
   ).

% show full guards
tcltk_get_guards_for_ops(list(Res)) :-
   findall([Name,' GUARD ',GS],
            (get_operation_guard(Name,G),
              translate:translate_bexpression(G,GS)), Res).

% show guards which do not depend on parameters and the rest (the Action)
tcltk_get_parameter_free_guards_for_ops(list(Res)) :-
   findall([Name,' LTSMIN-GUARD ',GS, 'ACTION ',AS],
            (get_operation_propositional_guards(Name,Guards,A),conjunct_predicates(Guards,G),
              translate:translate_bexpression(G,GS),
              translate:translate_substitution(A,AS)), Res).

get_operation_guard(Operation,G) :-
    b_get_machine_operation(Operation,_RR,_P,TBody),
    get_texpr_expr(TBody,Body),
    get_operation_guard_aux(Body,Operation,G).

:- use_module(probsrc(b_state_model_check), [get_guard/2]).
get_operation_guard_aux(RLEVENT,_,G) :-
    RLEVENT = rlevent(_Name,_Section,_Stat,_Parameters,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,_AbsEvents),
    !,
    findall(G1,get_full_eventb_guard(RLEVENT,G1),Gs),
    conjunct_predicates(Gs,G).
get_operation_guard_aux(_,Operation,G) :- get_guard(Operation,G).


% TO DO: we should move this code to b_state_model_check/predicate_debugger:get_operation_enabling_condition2 ?
get_full_eventb_guard(rlevent(_Name,_Section,_Stat,_Parameters,Grd,Thms,_Act,_VWit,_PWit,_Unmod,AbsEvents),G) :-
  (G=Grd
   ; member(G,Thms) % this is relevant for LTS Min; but not for normal enabling analysis
   ; member(b(E,_,_),AbsEvents),get_full_eventb_guard(E,G)).
get_full_eventb_action(rlevent(_Name,_Section,_Stat,_Parameters,_Grd,_Thm,Act,_VWit,_PWit,_Unmod,AbsEvents),A) :-
  (A=Act ; member(b(E,_,_),AbsEvents),get_full_eventb_action(E,A)).


% Note: this is the LTSMin meaning of action (for LTSMin a guard depending on the parameter is part of the action)
% Action will also contain parts of the guard that depend on parameters !
% This is important for independence
% Example: CopyYY2XX(p) = PRE v>1  & p = yy THEN xx := p END   -->  Guard is #p.(p=yy) --> v>1 ;  LTS Min Action is  PRE p=yy THEN x := yy END;
get_action(OpName,Action) :- get_operation_propositional_guards(OpName,_,Action).


% we need just the variables that are read by the operation (constants will not be considered)
b_get_read_write_vars(Operation,ReadGuardV,ReadActionV,ReadV,Write) :-
  b_get_read_write_vars_and_consts(false,Operation,ReadGuardV,ReadActionV,ReadV,Write).
b_get_read_write_vars_and_consts(IncludeConstants,Operation,ReadGuardV,ReadActionV,ReadV,Write) :-
  b_get_read_write_guard_action(Operation,ReadGuard,ReadAction,Read,Write),
  b_get_machine_variables(TVars),
  (IncludeConstants==true
    -> b_get_machine_constants(TConsts), append(TConsts,TVars,TCV) % should we include get_global_identifiers ?
    ;  TCV = TVars),
  get_texpr_ids(TCV,Vars),
  list_to_ord_set(Vars,Variables),
  ord_intersection(ReadGuard,Variables,ReadGuardV),
  ord_intersection(ReadAction,Variables,ReadActionV),
  ord_intersection(Read,Variables,ReadV).


% -----------------------------------------------
% compute MAY/MUST write information (for LTSMin)
% -----------------------------------------------

:- use_module(bsyntaxtree, [get_texpr_id/2, def_get_texpr_ids/2, is_truth/1, def_get_texpr_id/2]).

% get read/write with may/must write info
b_get_read_may_must_write(Operation,ReadGuardV,ReadActionV,ReadV,MustWriteV,MayWriteV,WriteV) :-
  b_get_read_write_vars_and_consts(true,Operation,ReadGuardV,ReadActionV,ReadV,WriteV),
  b_get_machine_operation(Operation,ReturnVars,_P,Body),
  (is_eventb_event(Body) -> MayWriteV = [], MustWriteV=WriteV % in Event-B there are no conditions on the actions
    ; extract_must_write(Body,MustWriteV1) ->
      % remove return values: they are not global variables written to
      def_get_texpr_ids(ReturnVars,RV),list_to_ord_set(RV,SRV), %print(removing(SRV,Operation)),nl,
      ord_subtract(MustWriteV1,SRV,MustWriteV),
      ord_subtract(WriteV,MustWriteV,MayWriteV)
    ; add_internal_error('Extraction of Must-Write failed: ',extract_must_write(Body,_)),
      MustWriteV=[], MayWriteV=WriteV).

is_eventb_event(b(E,_,_)) :-
  E = rlevent(_Name,_Section,_Status,_Params,_Guard,_Theorems,_Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents).

% compute the variables that must be written by an operation if it is enabled
extract_must_write(b(Sub,_,_),MV) :- extract_must_write_aux(Sub,MV),!.
extract_must_write(O,MV) :- add_internal_error('Call failed: ',extract_must_write(O,MV)),MV=[].


%extract_must_write_aux(P,V) :- functor(P,F,N), print(ex_aux(F,N)),nl,fail.
extract_must_write_aux(assign(LHS,_),V) :- maplist(get_id,LHS,Vars), list_to_ord_set(Vars,V).
extract_must_write_aux(assign_single_id(TID,_),[ID]) :- get_id(TID,ID).
extract_must_write_aux(becomes_such(LHS,_),V) :- maplist(get_id,LHS,Vars), list_to_ord_set(Vars,V).
extract_must_write_aux(becomes_element_of(LHS,_),V) :- maplist(get_id,LHS,Vars), list_to_ord_set(Vars,V).
extract_must_write_aux(precondition(_,TBody),MV) :- extract_must_write(TBody,MV).
extract_must_write_aux(assertion(_,TBody),MV) :- extract_must_write(TBody,MV).
extract_must_write_aux(any(_,_,TBody),MV) :- extract_must_write(TBody,MV).
extract_must_write_aux(let(_,_,TBody),MV) :- extract_must_write(TBody,MV).
extract_must_write_aux(lazy_let_subst(_,_,TBody),MV) :- extract_must_write(TBody,MV).
extract_must_write_aux(block(TBody),MV) :- extract_must_write(TBody,MV).
extract_must_write_aux(choice([A]),MV) :- !, extract_must_write(A,MV).
extract_must_write_aux(choice([A|T]),MV) :-
    extract_must_write(A,MV1), extract_must_write_aux(choice(T),MV2), ord_intersection(MV1,MV2,MV).
extract_must_write_aux(if([I]),MV) :- !, get_texpr_expr(I,if_elsif(Cond,TBody)),
    (is_truth(Cond) -> extract_must_write(TBody,MV) ; MV=[]). % we have ELSE skip and then write nothing
extract_must_write_aux(if([I|T]),MV) :- get_texpr_expr(I,if_elsif(_,TBody)),
    extract_must_write(TBody,MV1), extract_must_write_aux(if(T),MV2), ord_intersection(MV1,MV2,MV).
extract_must_write_aux(choice([A|T]),MV) :-
    extract_must_write(A,MV1), extract_must_write_aux(choice(T),MV2), ord_intersection(MV1,MV2,MV).
extract_must_write_aux(select([SW]),MV) :- !, get_texpr_expr(SW,select_when(_,TBody)),extract_must_write(TBody,MV).
extract_must_write_aux(select([SW|T]),MV) :-
    get_texpr_expr(SW,select_when(_,TBody)), extract_must_write(TBody,MV1),
    extract_must_write_aux(select(T),MV2),ord_intersection(MV1,MV2,MV).
extract_must_write_aux(select(T,Else),MV) :-
    extract_must_write(Else,MV1),
    extract_must_write_aux(select(T),MV2),ord_intersection(MV1,MV2,MV).
extract_must_write_aux(case(_,[],Else),MV) :- extract_must_write(Else,MV).
extract_must_write_aux(case(E,[C|T],Else),MV) :-
    get_texpr_expr(C,case_or(_,TBody)),
    extract_must_write(TBody,MV1),
    extract_must_write_aux(case(E,T,Else),MV2),ord_intersection(MV1,MV2,MV).
extract_must_write_aux(parallel([A]),MV) :- !, extract_must_write(A,MV).
extract_must_write_aux(parallel([A|T]),MV) :-
    extract_must_write(A,MV1), extract_must_write_aux(parallel(T),MV2), ord_union(MV1,MV2,MV).
extract_must_write_aux(sequence(S),MV) :- !, extract_must_write_aux(parallel(S),MV).
extract_must_write_aux(skip,[]).
extract_must_write_aux(external_subst_call(_,_),[]). % assume there is no guarantee that something is written
extract_must_write_aux(while(_,_,_,_),[]).
extract_must_write_aux(var(Parameters,TBody),MV) :-
   def_get_texpr_ids(Parameters,IDs),
   list_to_ord_set(IDs,MV2),
   extract_must_write(TBody,MV1),
   ord_subtract(MV1,MV2,MV).
extract_must_write_aux(operation_call(TOperation,_,_),MustWriteV) :-
   def_get_texpr_id(TOperation,op(Operation)),
   b_get_read_may_must_write(Operation,_,_,_,MustWriteV,_,_). %and ideally cache values
extract_must_write_aux(Other,[]) :- print(uncovered_subst_for_must_write(Other)),nl.

get_id(b(function(A,_),_,_),ID) :- !, get_id(A,ID).
get_id(TID,ID) :- get_texpr_id(TID,ID).

% mainly used for LTS Min (uses LTSMin's definition of Action and (parameter-free) Guard):
tcltk_read_write_matrix(list([list(['Operation','ReadsGuard','ReadsAction','MustWrite','MayWrite'])|Result])) :-
   findall(list([Operation,RG,RA,MustW,MayW]),b_get_read_may_must_write_for_ltsmin(Operation,RG,RA,MustW,MayW),Result).

:- use_module(bsyntaxtree,[get_global_identifiers/1]).
% obtains the read and written variables (constants, SETS and enumerated elements not included)
% seperates the operations into parameter-independent guards and the rest, the Action
:- use_module(predicate_debugger,[get_operation_propositional_guards/3]).
:- use_module(probsrc(b_interpreter),[b_get_machine_operation_for_animation/6]).
:- use_module(bmachine,[b_top_level_operation/1]).


get_operation(Operation,Results,Paras) :-
   b_top_level_operation(Operation),
   b_get_machine_operation_for_animation(Operation,Results,Paras,_TB,_OT,true).

b_get_read_may_must_write_for_ltsmin(Operation,ReadsGuard,ReadsAction,MustW,MayW) :-
   get_operation(Operation,Results,Paras),
   % LTSmin needs the constants in the read matrices;
   % if they are not included, it will transfer the uninitialised constants from the root state
   %b_get_machine_constants(TConsts), append(Paras,TConsts,TP),
   append(Results, Paras, ResultsAndParas),
   get_texpr_ids(ResultsAndParas,TPIds),
   get_global_identifiers(Global),
   append(TPIds,Global,PIDs),list_to_ord_set(PIDs,VarsToIgnore), % the VarsToIgnore should not be included in the matrix
   b_get_read_may_must_write(Operation,_RG,_RA,_FullReads,MustW,MayW,_W), % we ignore the info about guards and actions: LTS Min has another definition of what a guard is than b_get_read_may_must_write
   get_operation_propositional_guards(Operation,Guards,RestBody), % obtain LTSMin guards, which are independent of the parameters
   get_accessed_vars(RestBody,VarsToIgnore,_Modified,ReadsAction), % this computes the ReadsAction the LTSMin way
   %format('READS ACTION for ~w ~n  ~w~n',[Operation,ReadsAction]),
   conjunct_predicates(Guards,G),
   find_identifier_uses(G,VarsToIgnore,ReadsGuard).
%   (op_has_parameters(Operation)
%    -> ReadsAction = FullReads % guards on parameters may be projected away; leading to implicit reads in actions
%       % TO DO: convert operations into normal form; until this is done: simply return full reads
%     ; ReadsAction = RA).

%op_has_parameters(OpName) :-
%    b_get_machine_operation_for_animation(OpName,_Results,Paras,_TB,_OT,true),
%    Paras = [_|_].







rw_ord_intersection('$all',B,R) :- !,R=B.
rw_ord_intersection(A,'$all',R) :- !,R=A.
rw_ord_intersection(A,B,R) :- ord_intersection(A,B,R).



