% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(b_state_model_check,
   [b_set_up_valid_state/1,
    b_set_up_valid_state_with_pred/2, b_set_up_valid_state_with_pred/3,
    b_check_valid_state/1,

    state_model_check/4,

    cbc_deadlock_freedom_check/3, get_all_guards_false_pred/1,
    get_unsorted_all_guards_false_pred/1, get_negated_guard/3, get_negated_guard/4,
    get_guard/2,

    cbc_static_assertions_check/1, cbc_static_assertions_check/2,

    tcltk_perform_cbc_check/4,
    cbc_find_redundant_invariants/2,

    set_up_transition/7,
    set_up_initialisation/5
    ]).



:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,cbc).
:- module_info(description,'This module provides various tools for constraint-based checking of B machines.').

:- use_module(self_check).
:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(store).
:- use_module(translate).
:- use_module(b_enumerate,[  b_tighter_enumerate_values/2]).
:- use_module(b_interpreter,
            [b_test_boolean_expression/4, b_test_boolean_expression/6,
             %b_det_test_boolean_expression/6,
             b_not_test_boolean_expression/6,
             b_execute_top_level_operation_wf/6,
             b_execute_top_level_statement/7,
             set_up_typed_localstate/6]).
:- use_module(b_interpreter_eventb, [b_event_constraints/7]).
:- use_module(b_global_sets).
:- use_module(bmachine).
:- use_module(bsyntaxtree).

:- use_module(kernel_objects).
:- use_module(bsets_clp).

:- use_module(solver_interface, [set_up_typed_localstate_for_pred/4]).

:- use_module(debug).
:- use_module(typechecker).
:- use_module(bmachine).
:- use_module(kernel_waitflags).
:- use_module(error_manager).
:- use_module('kodkod/kodkod',[apply_kodkod/3]).

/* --------------------------------------------------------- */

:- use_module(tools,[start_ms_timer/1, stop_ms_timer/1]).

state_model_check(OpName,State,Operation,NormalisedNewState) :-
  print_message('===> Starting Constraint-Based Checking'),
  start_ms_timer(Timer),
  if(cbc_model_check(OpName,State,Operation,NewState),
     stop_ms_timer(Timer),
     (stop_ms_timer(Timer),print('===> no counter example found for '),print(OpName),nl,fail)),
  nl,
     print_message('===> found counter example for'), print(OpName),nl,
     print('% ===> Operation: '), translate_event(Operation,OpStr),print(OpStr),nl,
     print('% ===> State: '),translate_bstate(State,StateStr),print(StateStr),nl,
     normalise_store(NewState,NormalisedNewState),
     print('% ===> NewState: '),translate_bstate(NormalisedNewState,NStateStr),print(NStateStr),nl.

:- use_module(library(avl)).
cbc_model_check(OpName,State,Operation,NewState) :-
   (nonvar(OpName)
     -> (b_get_machine_operation(OpName,_,_,_)
          -> print_message(operation_name(OpName))
          ;  add_error(constraint_based_check,'Operation does not exist: ',OpName),
             fail
         )
      ; true
   ),
   b_get_properties_from_machine(Properties),
   b_get_invariant_from_machine(Invariant),
   b_get_machine_variables(Variables),
   b_get_machine_constants(Constants),
  % print_message(extracted_invariant(Variables,Invariant)),

   empty_state(EmptyState),
   set_up_typed_localstate(Constants,_FreshVars1,TypedVals1,EmptyState,State1,positive),
   b_global_sets:static_symmetry_reduction_for_global_sets(State1),
   set_up_typed_localstate(Variables,_FreshVars2,TypedVals2,State1,State,positive),
   append(TypedVals1,TypedVals2,TypedVals),

   init_wait_flags(WF),

   print_message(' --------------- '),
   % print_message(' set up start state:'),  print_message(State),
   % translate:print_bexpr(Properties),
   empty_avl(Ai),
   b_test_boolean_expression(Properties,[],State,WF,Ai,A2), % A2 is used below for after state
   % print_message(' set up properties for constants'),
   b_test_boolean_expression(Invariant,[],State,WF,A2,_A3),
   % print_message(' set up invariant for start state'),

   set_up_transition(OpName,Operation,State1,State,NewState,_TransInfo,WF),
   print_message(' finished setting up target state for operation:'),  print_message(OpName),
   %print_message(' target state:'), print_message(NewState),
   % Note: we are not setting up an enumerator for NewState; here we would need to be careful as we cannot use tight enumeration for NewState

   % Retrieve Information on proof status of the invariant
   (get_proven_invariant(OpName,ProvenInvariant) -> true
    ;   create_texpr(truth,pred,[],ProvenInvariant), print(no_proven_invariant_info),nl),
   (get_preference(use_po,false) -> UncheckedInvariant=Invariant
     ; b_specialized_invariant_for_op(OpName,UncheckedInvariant) -> true
     ; otherwise -> UncheckedInvariant=Invariant, print(no_specialized_invariant_info_for_op(OpName)),nl
    ),

   %print_message('proven invariant:'),
   %print(proven(ProvenInvariant)),
   create_negation(UncheckedInvariant,NegUncheckedInvariant),
   % we conjoin the predicates so that predicate-level optimizations can be applied (TO DO: also allow Properties to be used)
   conjunct_predicates([ProvenInvariant,NegUncheckedInvariant],TargetPred),
   predicate_level_optimizations(TargetPred,OptTargetPred),

   translate:print_bexpr(OptTargetPred),nl,
   b_test_boolean_expression(OptTargetPred,[],NewState,WF,A2,_Ao),  % not A3 but A2; as different State
   print_message(' finished setting up invariant of proven invariants and negation of invariant for target state'),

  % print(enum(TypedVals)),nl,
   b_tighter_enumerate_values(TypedVals,WF),

   (debug_mode(on) -> ground_wait_flag0(WF),portray_waitflags(WF) ; true),
   print_bt_message(grounding_wait_flags),
   ground_wait_flags(WF).

set_up_transition(OpName,Operation,ConstantsState,State,NewState,TransInfo,WF) :-
    is_event_b_transition(OpName),!,
    b_get_machine_variables(Variables),
    create_target_state(Variables,_Values,ConstantsState,NewState,WF),
    copy_unmodified_variables(Variables,OpName,State,NewState),
    event_b_transition(OpName,Operation,State,NewState,TransInfo,WF).
set_up_transition(OpName,Operation,ConstantsState,State,NewState,TransInfo,WF) :-
    % classical B
    append(ConstantsState,State,InState),
    append(ConstantsState,NewState,OutState),
    b_get_machine_variables(Variables),
    if(set_up_target_store(InState,OutState),true,
        (add_internal_error('Call failed: ',set_up_target_store(InState,OutState)),fail)),
    copy_unmodified_variables(Variables,OpName,InState,OutState),
    b_execute_top_level_operation_wf(OpName,Operation,InState,OutState,TransInfo,WF).
is_event_b_transition(OpName) :-
    b_get_machine_operation(OpName,[],_Parameters,_TEvent,eventb_operation(_ChangeSet,_Values,_Operation),_),!.
event_b_transition(OpName,Operation,State,NewState,TransInfo,WF) :-
    b_get_machine_operation(OpName,[],Parameters,TEvent,eventb_operation(ChangeSet,Values,Operation),_),
    copy_variable_store(State,ChangeSet,_,NewState),
    b_event_constraints(TEvent,Parameters,Values,State,NewState,TransInfo,WF).

create_target_state(Variables,Values,ConstantsState,NewState,WF) :-
    set_up_typed_localstate(Variables,Values,TypedVals,ConstantsState,NewState,positive),
    b_tighter_enumerate_values(TypedVals,WF).

% TODO[DP, 7.4.2011]: refactor, see b_interpreter:b_execute_operation3/13
set_up_initialisation(ConstantsState,NewState,Operation,TransInfo,WF) :-
    b_get_machine_variables(Variables),
    create_target_state(Variables,Values,ConstantsState,NewState,WF),
    store:l_expand_and_normalise_values(Values,NormalisedValues),
    safe_univ(Operation,['$initialise_machine'|NormalisedValues]),
    b_get_initialisation_from_machine(Body,OType),
    set_up_initialisation2(OType,Body,Variables,ConstantsState,NewState,TransInfo,WF).
set_up_initialisation2(classic,Body,Variables,ConstantsState,NewState,[path(Path)],WF) :-
    set_up_undefined_localstate(Variables,ConstantsState,PreInitState),
    empty_state(EmptyState),
    b_execute_top_level_statement(Body,EmptyState,PreInitState,Updates,WF,Path,output_required),
    save_updates_in_existing_store(Updates,NewState).
set_up_initialisation2(eventb_operation(_,_,_),Body,_Variables,ConstantsState,NewState,TransInfo,WF) :-
    b_event_constraints(Body,[],_Values,ConstantsState,NewState,TransInfo,WF).

set_up_target_store([],Out) :-
  (var(Out) -> Out=[]
   ; true % OutStore already set up by somebody else
  ).
set_up_target_store([bind(V,_)|In],OutStore) :- %print(add(V)),nl,
    nonvar(OutStore),!, % OutStore already set up by somebody else
    ((OutStore=[bind(VV,_)|T], V==VV)
      -> set_up_target_store(In,T)
       ; add_binding(OutStore,V), set_up_target_store(In,OutStore)).
set_up_target_store([bind(V,_)|In],[bind(V,_)|Out]) :-
    set_up_target_store(In,Out).

add_binding(List,V) :- var(List), !, List = [bind(V,_)|_].
add_binding([],V) :- add_internal_error(add_binding,'Variable does not appear in store: ',V,unknown).
add_binding([bind(VV,_)|T],V) :-
  (V==VV -> true /* ok, found the variable */
    ; add_binding(T,V)).

% copy over the variables that are not changed by the operation from OldStore to NewStore
copy_unmodified_variables(Variables,OpName,OldStore,NewStore) :-
    ( get_operation_info(OpName,SubstitutionInfo) ->
        ( memberchk(modifies(ModIDs),SubstitutionInfo) ->
            get_texpr_ids(Variables,Ids),
            exclude(is_modified_var(ModIDs),Ids,UnmodifiedIds),
            maplist(copy_var(OldStore,NewStore),UnmodifiedIds)
        ; otherwise ->
            add_internal_error('Operation/Event has no modifies info: ',
                               copy_unmodified_variables(Variables,OpName,OldStore,NewStore))
            % assume that all variables modified: no copy_var calls
        )
    ; otherwise ->
        add_internal_error('Unknown Operation/Event:', copy_unmodified_variables(Variables,OpName,OldStore,NewStore)),
        fail
    ).
is_modified_var(ModIDs,Id) :-
    memberchk(Id,ModIDs).
copy_var(OldStore,NewStore,Id) :-
    lookup_value(Id,OldStore,V),lookup_value(Id,NewStore,VNew),
    equal_object(V,VNew).


/* --------------------------------------------------------- */
% a database of available CBC checks:
cbc_check('INV_NO_RED',
          'INV_NO_RED: Find invariants which are redundant',
          Call,Res,Ok) :- Call = (tcltk_find_redundant_invariants(Res,Ok)).
cbc_check('INV_AXM_SAT',
          'INV_AXM_SAT: Check if there exists a state that satisfies the INVARIANT and the PROPERTIES (AXIOMS)',
          Call,Res,Ok) :-
          Call = (b_set_up_valid_state(State) -> Ok=true,translate_bstate(State,Res) ; Ok=false, Res='NO_VALID_STATE_FOUND').
cbc_check('DLK',
          'DLK: Check that INVARIANT does not allow a deadlocking state',
          Call,Res,Ok) :-
          Call = (cbc_deadlock_freedom_check(State) -> Ok=false,translate_bstate(State,Res) ; Ok=true, Res='NO_DEADLOCK_FOUND').
cbc_check('THM_STATIC',
          'THM_STATIC: Check that static ASSERTIONS (THEOREMS) follow from PROPERTIES (AXIOMS)',
          Call,TclRes,Ok) :-
          Call = (cbc_static_assertions_check(Res),translate_result(Res,Ok),functor(Res,TclRes,_)).
%cbc_check('INV/INITIALISATION',
%          'INV/INITIALISATION: Check if INITIALISATION can violate INVARIANT',
%          Call,TclRes,Ok) :-
%          Call = (user:tcltk_cbc_find_invariant_violation([],Res) -> translate_result(Res,Ok),functor(Res,TclRes,_) ; Ok=true).
cbc_check(INVO,
          Label,
          Call,Res,Ok) :-
          b_top_level_operation(OpName),
          ajoin(['INV/',OpName],INVO),
          ajoin(['INV/',OpName,': Check that ',OpName,' preserves INVARIANT'],Label),
          Call = (state_model_check(OpName,_State,_Operation,NewState)
           -> Ok=false,translate_bstate(NewState,Res) ; Ok=true, Res='NO_INVARIANT_violation_found').
% TO DO: maybe dynamically generate cbc_checks (inv preservation per event); disable certain checks

:- public translate_result/2. % used above
translate_result(no_counterexample_found,Ok) :- !, Ok=true.
translate_result(no_counterexample_exists,Ok) :- !, Ok=true.
translate_result(no_counterexample_exists(_,_,_),Ok) :- !, Ok=true.
translate_result(time_out,Ok) :- !, Ok=time_out.
translate_result(_,false).

tcltk_perform_cbc_check(ID,Text,Result,Ok) :-
   %print(tcltk_perform_cbc_check(ID,Text,Result,Ok)),nl,
   if(cbc_check(ID,Text,Call,Res,Ok),
      (print(Text),nl,
       call(Call),
       Result=Res,
       print(result(Ok,Result)),nl
      ),
      (add_internal_error('Illegal CBC check: ',tcltk_perform_cbc_check(ID,Text,Result,Ok)),
       fail)
      ).
% enum_warning:enter_new_error_scope(ScopeID), event_occurred_in_error_scope(enumeration_warning(_,_,_,_)), exit_error_scope(ScopeID,ErrOcc),ErrOcc=true


:- use_module(probsrc(bsyntaxtree),[conjunction_to_list/2]).
:- public tcltk_find_redundant_invariants/2.
tcltk_find_redundant_invariants(list(Res),Ok) :-
   cbc_find_redundant_invariants(Invs,Timeout_occured),
    (Invs=[] -> Res = ['NO REDUNDANT INVARIANTS'], Ok=true
       ; Timeout_occured=true -> Res = ['REDUNDANT INVARIANTS (OR WITH TIME-OUT):'|Invs], Ok=false
       ; Res = ['REDUNDANT INVARIANTS:'|Invs], Ok=false).


:- dynamic timeout_occured/0.
cbc_find_redundant_invariants(RedundantInvs,Timeout_occured) :-
   retractall(timeout_occured),
   findall(RI,
           (b_find_redundant_invariant(RedInv,_,R,EnumWarning,ErrOcc),
            translate:translate_bexpression(RedInv,InvString),
            generate_explanation(R,EnumWarning,ErrOcc,InvString,RI),
            (R==time_out -> assert(timeout_occured) ; true)
            ),
            RedundantInvs),
    (timeout_occured -> Timeout_occured = true ; Timeout_occured=false).

generate_explanation(R,EnumWarning,ErrOcc,InvString,Res) :-
    (EnumWarning=true -> L1=[' /* ENUM WARNING */ '|L2] ; L1=L2),
    (ErrOcc=true -> L2=[' /* REQUIRED FOR WELL-DEFINEDNESS */ '] ; L2=[]),
    (R=redundant -> ajoin([InvString|L1],Res)
     ; R=useful -> EnumWarning=true,  ajoin(['ENUMERATION WARNING (POSSIBLY USEFUL): ',InvString|L2],Res)
     ; R=time_out -> ajoin(['TIME OUT (POSSIBLY USEFUL): ',InvString|L1],Res)
    ).

:- use_module(tools_timeout, [time_out_with_factor_call/3]).

b_find_redundant_invariant(RedundantInv,State,Redundant,EnumWarning,ErrOcc) :-
   b_get_invariant_from_machine(Invariant),
   conjunction_to_list(Invariant,InvList),
   select(RedundantInv,InvList,RestInvariantList),
   print('CHECKING REDUNDANCY: '),translate:print_bexpr(RedundantInv),nl,
   create_negation(RedundantInv,NegRed),
   conjunct_predicates(RestInvariantList,AdaptedInvariant),
   %translate:print_bexpr(AdaptedInvariant),nl,
   enter_new_error_scope(ScopeID,b_find_redundant_invariant),
   ( time_out_with_factor_call(predicate_satisfiable_relevant_ids(NegRed,AdaptedInvariant,State),5,
                               TO=time_out)
      -> (TO==time_out -> Redundant=time_out ; Redundant=useful)
      ; Redundant = redundant),
    (enumeration_warning_occured_in_error_scope -> EnumWarning=true
      ; EnumWarning=false),
    exit_error_scope(ScopeID,ErrOcc,b_find_redundant_invariant),
    print('RESULT: '),print(Redundant),nl.



/* --------------------------------------------------------- */



b_set_up_valid_state(State) :- b_set_up_valid_state_with_pred(State,b(truth,pred,[]),true).

b_set_up_valid_state_with_pred(NormalisedState,Pred) :-
    b_set_up_valid_state_with_pred(NormalisedState,Pred,true).
b_set_up_valid_state_with_pred(NormalisedState,Pred,UseInvariant) :-
    predicate_satisfiable_all_ids(Pred,State,UseInvariant),
    normalise_store(State,NormalisedState).


/* ---------------------- */
/* DEADLOCK FREEDOM CHECK */
/* ---------------------- */

 :- use_module(bsyntaxtree,[predicate_identifiers/2]).
:- use_module(b_ast_cleanup,[predicate_level_optimizations/2]).

:- public cbc_deadlock_freedom_check/1. % used above
cbc_deadlock_freedom_check(State) :-
   create_texpr(truth,pred,[],True), Filter=0,
   cbc_deadlock_freedom_check(State,True,Filter).

cbc_deadlock_freedom_check(State,Goal,Filter) :-
    print(cbc_deadlock_freedom_check(State,Filter)),nl,flush_output,
    ( is_truth(Goal) ->
        get_all_guards_false_pred(Target)
    ; otherwise ->
        print('Adding additional goal predicate: '),nl,
        print_bexpr(Goal),nl,
        ( Filter=1 ->
            get_all_guards_false_pred(AllGFalse,Goal)
        ; otherwise ->
            get_all_guards_false_pred(AllGFalse)),
        conjunct_predicates([AllGFalse,Goal],Target)
    ),!,flush_output,
    statistics(runtime,[Start,_]),
    statistics(walltime,[WStart,_]),
    cbc_deadlock_freedom_check2(Target,RState,Start,WStart,false,TO), % TO DO: control SkipIrrelevantComponents by preference or GUI
    (TO==time_out -> State=time_out ; State=RState).
cbc_deadlock_freedom_check2(Target,NormalisedState,Start,WStart,SkipIrrelevantComponents,TO) :-
   get_texpr_expr(Target,T), T \= falsity,
   b_get_properties_from_machine(Properties),
   b_get_invariant_from_machine(Invariant),
   conjunct_predicates([Properties,Invariant,Target],Pred1),
   b_get_machine_variables(Variables),
   b_get_machine_constants(Constants), append(Constants,Variables,CV),

   apply_kodkod(CV,Pred1,Pred),
   %predicate_level_optimizations(Pred2,Pred), % detect set partitions, etc..., now done inside b_interpreter_components

   set_up_typed_localstate_for_pred(CV,Pred,TypedVals,State),
   b_interpreter_components:reset_component_info(false),
   (debug_mode(on) -> translate:nested_print_bexpr_as_classicalb(Pred) ; true),
   %%visualize_graph:print_predicate_dependency_as_graph_for_dot(Pred,'~/Desktop/pdg.dot'),
   (SkipIrrelevantComponents==true
     -> predicate_identifiers(Target,TargetIds)  % Warning: can skip over obvious inconsistencies TRUE:BOOL
     ; TargetIds=all),
   print(target(TargetIds)),nl,
   b_global_sets:static_symmetry_reduction_for_global_sets(State),
   time_out_with_factor_call(
        b_interpreter_components:b_trace_test_components(Pred,State,TypedVals,TargetIds),
        10, (nl,TO=time_out)),
   statistics(runtime,[Stop,_]), Time is Stop - Start,
   statistics(walltime,[WStop,_]), WTime is WStop - WStart,
   (TO==time_out -> print(time_out_during_deadlock_checking(Time)),nl, NormalisedState=State
     ; TO=no_time_out, print(deadlock_counter_example_found(Time,WTime)),nl,
       normalise_store(State,NormalisedState)
    ).
cbc_deadlock_freedom_check2(Target,_State,Start,WStart,_,no_time_out) :-
   (get_texpr_expr(Target,falsity) -> print('Disjunction of guards obviously true'),nl ; true),
   statistics(runtime,[Stop,_]), Time is Stop - Start,
   statistics(walltime,[WStop,_]), WTime is WStop - WStart,
   print(no_deadlock_counter_example_found(Time,WTime)),nl,fail.

% test if a predicate can be satisfied for some instantiation of constants and variables
predicate_satisfiable(P) :- predicate_satisfiable(P,_),!, print('SATISFIABLE'),nl,nl.
predicate_satisfiable(_) :- print('*UNSATISFIABLE*'),nl,nl,fail.
predicate_satisfiable(Predicate,State) :-
   b_get_invariant_from_machine(Invariant),
   predicate_satisfiable_relevant_ids(Predicate,Invariant,State).
predicate_satisfiable_all_ids(Predicate,State,UseInvariant) :-
   (UseInvariant=true -> b_get_invariant_from_machine(Invariant) ; create_texpr(truth,pred,[],Invariant)),
   predicate_satisfiable4(Predicate,Invariant,State,all). % look at all variables/constants: we want a complete solution

predicate_satisfiable_relevant_ids(Predicate,Invariant,State) :-
   predicate_identifiers(Predicate,RelevantIds), % only look at relevant ids for Predicate; skip over components not involving them
   predicate_satisfiable4(Predicate,Invariant,State,RelevantIds).

predicate_satisfiable4(Predicate,Invariant,State,RelevantIds) :-
   b_get_properties_from_machine(Properties),
   conjunct_predicates([Properties,Invariant,Predicate],TotPred0),
   b_get_machine_variables(Variables),
   b_get_machine_constants(Constants), append(Constants,Variables,CV),
   set_up_typed_localstate_for_pred(CV,TotPred0,TypedVals,State),
   apply_kodkod(CV,TotPred0,TotPred),
   %predicate_level_optimizations(TotPred1,TotPred), %now done inside b_interpreter_components
   b_interpreter_components:reset_component_info(false),
   translate:nested_print_bexpr_as_classicalb(Predicate),
   % TO DO: catch time-outs
   b_global_sets:static_symmetry_reduction_for_global_sets(State),
   b_interpreter_components:b_trace_test_components(TotPred,State,TypedVals,RelevantIds).


% set up a predicate which is true if all guards are false
get_all_guards_false_pred(AllGuardsFalsePred) :-
   findall(NegGuard,
          (b_get_machine_operation(OpName,_,_,_),
           get_negated_guard(OpName,_,NegGuard)), ListOfNegGuards),
   ~~mnf(conjunct_and_sort_smt_predicates(ListOfNegGuards,AllGuardsFalsePred)).
get_all_guards_false_pred(AllGuardsFalsePred,FilterPred) :-
   findall(NegGuard,
      (b_get_machine_operation(OpName,_,_,_),
       get_negated_guard(OpName,Guard,NegGuard),
       conjunct_predicates([Guard,FilterPred],FGuard),
       nl,print('---> CHECKING IF GUARD SATISFIABLE: '), print(OpName),nl,
       predicate_satisfiable(FGuard) % if Guard not satisfiable given FilterPred: remove it
       ), ListOfNegGuards),
   ~~mnf(conjunct_and_sort_smt_predicates(ListOfNegGuards,AllGuardsFalsePred)).

% the following is useful, e.g., for graphical visualization
get_unsorted_all_guards_false_pred(AllGuardsFalsePred) :-
   findall(NegGuard,
          (b_get_machine_operation(OpName,_,_,_),
           get_negated_guard(OpName,_,NegGuard)), ListOfNegGuards),
   conjunct_predicates(ListOfNegGuards,AllGuardsFalsePred).

conjunct_and_sort_smt_predicates(List,Result) :-
   print('Counting basic predicates'),nl,
   empty_avl(Ai), count(List,2,Ai,Ao),
   % portray_avl(Ao),nl,
   sort_smt_predicates(List,Ao,SList),
   %nl,print(sorted(SList)),nl,
   conjunct_smt_predicates(SList,Result) , print(finished),nl.

conjunct_smt_predicates([],b(truth,pred,[])).
conjunct_smt_predicates([P|T],Res) :- conjunct_smt_predicates(T,TRes),
   extract_info(P,TRes,Info), % extract e.g., wd condition info
   conjoin(P,TRes,[try_smt|Info],Res). % try_smt currently no longer useful ?

% Count the number of occurences of given basic predicates
count(b(Pred,pred,_Info),Rel,Ai,Ao) :- !,count2(Pred,Rel,Ai,Ao).
count([],_,Ai,Ao) :- !,Ao=Ai.
count([H|T],Rel,Ai,Ao) :- !,count(H,Rel,Ai,Aii), count(T,Rel,Aii,Ao).
count(R,_,_,_) :- add_error_fail(count,'Illegal argument to count: ',R).

count2(negation(Pred),Rel,Ai,Ao) :- !, count(Pred,Rel,Ai,Ao).
count2(conjunct(A,B),Rel,Ai,Ao) :- !, count(A,Rel,Ai,Aii), count(B,Rel,Aii,Ao).
count2(Pred,Rel,Ai,Ao) :- disjunctive_pred(Pred,A,B),!,
  (Rel>1
    -> R1 is Rel-1, count(A,R1,Ai,Aii), count(B,R1,Aii,Ao)
    ;  Ai=Ao).
count2(Pred,Rel,Ai,Ao) :- smt_predicate(Pred),!, norm_pred(Pred,NP),
  (avl_fetch(NP,Ai,C1) -> Count is C1+Rel ; Count = Rel),
  avl_store(NP,Ai,Count,Ao).
count2(_Pred,_Rel,A,A).

disjunctive_pred(disjunct(A,B),A,B).
disjunctive_pred(implication(A,B),A,B).
disjunctive_pred(equivalence(A,B),A,B).


sort_smt_predicates(b(Pred,pred,Info),Ai,b(SP,pred,Info)) :- sort_smt_predicates2(Pred,Ai,SP).
sort_smt_predicates([],_,[]).
sort_smt_predicates([H|T],Ai,[SH|ST]) :- sort_smt_predicates(H,Ai,SH), sort_smt_predicates(T,Ai,ST).

sort_smt_predicates2(negation(Pred),AVL,negation(SPred)) :- !,sort_smt_predicates(Pred,AVL,SPred).
sort_smt_predicates2(implication(A,B),AVL,implication(SA,SB)) :- !,
   sort_smt_predicates(A,AVL,SA), sort_smt_predicates(B,AVL,SB).
sort_smt_predicates2(equivalence(A,B),AVL,equivalence(SA,SB)) :- !,
   sort_smt_predicates(A,AVL,SA), sort_smt_predicates(B,AVL,SB).
sort_smt_predicates2(disjunct(A,B),AVL,disjunct(SA,SB)) :- !,  % To do: also sort disjuncts
   sort_smt_predicates(A,AVL,SA), sort_smt_predicates(B,AVL,SB).
sort_smt_predicates2(conjunct(A,B),AVL,Res) :- !,
   conjunction_to_count_list(b(conjunct(A,B),pred,[]),AVL,CountList),
   sort(CountList,SC),
   reverse(SC,SortedCList), % to do : use samsort with custom order
   %print(sorted(SortedCList)),nl,
   project_count_list(SortedCList,PR),
   conjunct_smt_predicates(PR,BRes),
   translate:print_bexpr(BRes),nl,
   BRes = b(Res,pred,_).
sort_smt_predicates2(X,_,X).

project_count_list([],[]).
project_count_list([count(_,P)|T],[P|PT]) :- project_count_list(T,PT).

conjunction_to_count_list(C,AVL,List) :- is_a_conjunct(C,LHS,RHS),!,
   conjunction_to_count_list(LHS,AVL,L1),
   conjunction_to_count_list(RHS,AVL,R1),
   append(L1,R1,List).  % TO DO: improve, use Difference Lists
conjunction_to_count_list(b(Pred,pred,Info),AVL,[count(Count,b(Pred,pred,Info))]) :- smt_predicate(Pred),!,
  norm_pred(Pred,NX),
  (avl_fetch(NX,AVL,Count) -> true ; Count=0).
conjunction_to_count_list(P,_AVL,[count(0,P)]).

smt_predicate(equal(_,_)).
smt_predicate(not_equal(_,_)).
smt_predicate(member(_,_)).
smt_predicate(not_member(_,_)).
smt_predicate(less(_,_)).
smt_predicate(less_equal(_,_)).
smt_predicate(greater(_,_)).
smt_predicate(greater_equal(_,_)).

norm_pred(equal(A,B),Res) :- !,norm_equal(A,B,Res).
norm_pred(not_equal(A,B),Res) :- !,norm_equal(A,B,Res).
norm_pred(member(A,B),member(AA,BB)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(not_member(A,B),member(AA,BB)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(less(A,B),less(AA,BB)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(greater(A,B),less(BB,AA)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(less_equal(A,B),less(BB,AA)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(greater_equal(A,B),less(AA,BB)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(b(B,_,_),Res) :- !, norm_pred(B,Res).
norm_pred(set_extension(L),Res) :- !, Res=set_extension(NL), norm_pred(L,NL).
norm_pred(integer(A),Res) :- !,Res=A.
norm_pred(identifier(A),Res) :- !,Res=A.
norm_pred([],Res) :- !, Res=[].
norm_pred([H|T],Res) :- !, Res=[NH|NT], norm_pred(H,NH), norm_pred(T,NT).
norm_pred(X,X).


norm_equal(b(identifier(ID),_,_),_B,Res) :- !,Res=ID.
norm_equal(A,B,Res) :- norm_pred(A,AA),norm_pred(B,BB),
  (AA==boolean_false -> AAA=boolean_true ; AAA=AA),
  (BB==boolean_false -> BBB=boolean_true ; BBB=BB),
  (BBB @< AAA -> Res = equal(AAA,BBB) ; Res= equal(BBB,AAA)).
/*
operation_satisfiable(OpName,FilterPred) :- get_machine_operation(OpName,_,_,_),
       get_negated_guard(OpName,Guard,_NegGuard),
       conjunct_predicates([Guard,FilterPred],FGuard),
       print('CHECKING IF GUARD SATISFIABLE: '), print(OpName),nl,
       predicate_satisfiable(FGuard).*/

% Optionally: Remove certain complicated parts from the guards
/*
simplify_guard(b(Pred,pred,Info),SPred) :- simplify_guard2(Pred,Info,SPred).
simplify_guard2(conjunct(A,B),Info,Res) :- !,
    simplify_guard(A,SA), simplify_guard(B,SB),
    conjoin(SA,SB,Info,Res).
simplify_guard2(exists(P,B),_Info,R) :- !, R=b(truth,pred,[]),
     print('REMOVED: '),translate:print_bexpr(b(exists(P,B),pred,[])),nl.
simplify_guard2(X,Info,b(X,pred,Info)).
*/

conjoin(b(truth,_,_),B,_,R) :- !,R=B.
conjoin(b(falsity,_,_),_B,_,R) :- !,R=b(falsity,pred,[]).
conjoin(A,b(truth,_,_),_,R) :- !,R=A.
conjoin(_A,b(falsity,_,_),_,R) :- !,R=b(falsity,pred,[]).
conjoin(A,B,Info,b(conjunct(A,B),pred,Info)).

get_negated_guard(OpName,Guard,NegationOfGuard) :- get_negated_guard(OpName,Guard,NegationOfGuard,_Precise).
get_negated_guard(OpName,Guard,NegationOfGuard,Precise) :-
    get_guard3(OpName,Guard,Precise),
    SGuard=Guard,
    %simplify_guard(Guard,SGuard), %% comment in to remove complicated parts
    create_negation(SGuard,NegationOfGuard). % we used to add try_smt Info; no longer relevant !?


:- use_module(predicate_debugger,[get_simplified_operation_enabling_condition/5]).
%:- use_module(b_ast_cleanup, [clean_up/3]).
:- use_module(b_interpreter_components,[construct_optimized_exists/3]).

% get a guard, by translating all parameters into existential quantifiers
get_guard(OpName,Guard) :- get_guard3(OpName,Guard,_).
get_guard3(OpName,Guard,Precise) :-
  get_simplified_operation_enabling_condition(OpName,Parameters,EnablingCondition,_BecomesSuchVars,Precise),
  % TO DO: partition; avoid lifting becomes_such_that conditions
  %   (in EventB the feasibility PO will ensure that Guard => BecomeSuchThat is ok)
  construct_optimized_exists(Parameters,EnablingCondition,Guard).

/* --------------------------------------------------------- */

:- use_module(tools_printing,[print_error/1]).
b_check_valid_state(State) :-
   (b_valid_state(State) -> true
     ;  print_error('Invalid State:'),
        print_error(State)
   ).

b_valid_state(State) :-
   ~~mnf(bmachine:b_get_machine_variables(Variables)),
  ~~pp_cll(b_state_model_check:b_check_variable_types(Variables,State)).


% TODO(DP,6.5.2008): removed type check
:- assert_pre(b_state_model_check:b_check_variable_types(V,S), (list_skeleton(V),list_skeleton(S))).
%              (list_skeleton(V),list_skeleton(VT),list_skeleton(S))).

:- assert_post(b_state_model_check:b_check_variable_types(_V,_S),true).

b_check_variable_types([],_).
b_check_variable_types([Var|VT],State) :-
   debug_println(9,checking(Var,Type)),
   def_get_texpr_id(Var,VarID), get_texpr_type(Var,Type),
   ((~~pp_mnf(store:lookup_value_for_existing_id(VarID,State,Val)),Val\==fail)
    -> ((~~pp_cll(kernel_objects:basic_type(Val,Type))
          -> debug_println(check_basic_type(Val,Type))
          ;  print_error('Type Error ! Variable,Value,Type: '),print_error(VarID),
             print_error(Val),print_error(Type)
        ),
        b_check_variable_types(VT,State))
    ;   add_error(b_check_variable_types,'Variable not defined in state: ', VarID),
        add_error(b_check_variable_types,'Be sure that INITIALISATION initialises the variable: ', VarID),
        print_error(State)
   ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% constraint based assertion check
:- use_module(preferences).

cbc_static_assertions_check(Result) :- cbc_static_assertions_check(Result,[]).
% currently supported options: tautology_check : check that assertions are tautologies
cbc_static_assertions_check(Result,Options) :-
    get_cbc_assertions_predicate_to_solve(Options,  Constants, TotalPredicate,true,_),
    enter_new_error_scope(ScopeID,cbc_static_assertions_check),
    start_ms_timer(Timer),
    (member(observe_state,Options) -> external_functions:observe_state(State) ; true),
    ( time_out_with_factor_call(cbc_static_assertions_check2(State,Constants, TotalPredicate),10,TO=time_out) ->
       ( TO==time_out -> Result = time_out
        ; otherwise    -> Result = counterexample_found(State))
    ; enumeration_warning_occured_in_error_scope -> Result = no_counterexample_found
    ; get_elapsed_walltime(Timer,WTime),
      cbc_static_assertions_contradiction_check(Constants,TotalPredicate,Result,Options,WTime)
    ),
    (debug_mode(on) -> print('cbc_static_assertions_check: '),stop_ms_timer(Timer) ; true),
    exit_error_scope(ScopeID,_ErrOcc,cbc_static_assertions_check).

% in case of proof: check if we can establish a contradiction in the hypotheses
cbc_static_assertions_contradiction_check(Constants,TotalPredicate,Result,Options,_) :-
    member(contradiction_check,Options),
    print('Looking for contradiction in hypothesis'),nl,
    get_cbc_assertions_predicate_to_solve(Options,  Constants, TotalPredicate2,false,Goal),
    start_ms_timer(Timer),
    (time_out_with_factor_call(cbc_static_assertions_check2(_,Constants, TotalPredicate2),10,TO=time_out)
      -> debug_println(no_contradiction(TO)),fail ; true),
    get_elapsed_walltime(Timer,WTime),
    (enumeration_warning_occured_in_error_scope -> print(no_contradiction(enumeration_warning)),nl,fail ; true),
    !,
    print('*** CONTRADICTION IN HYPOTHESIS FOUND !'),nl,
    translate:translate_bexpression_with_limit(Goal,GS),
    add_warning(contradiction_in_hypotheses,
          'Prover double check result: Contradiction in hypotheses (or Bug) detected for goal: ',GS),
    compute_unsat_core_if_requested(Options,TotalPredicate2,WTime),
    Result = no_counterexample_exists(Constants,TotalPredicate,[contradiction_in_hypotheses]).
cbc_static_assertions_contradiction_check(Constants,TotalPredicate,Result,Options,WTime) :-
    compute_unsat_core_if_requested(Options,TotalPredicate,WTime),
    Result = no_counterexample_exists(Constants,TotalPredicate,[]).

cbc_static_assertions_check2(concrete_constants(NormalisedState),Constants, TotalPredicate) :-
    empty_state(EmptyState),
    set_up_typed_localstate(Constants,_FreshVars1,TypedVals,EmptyState,State,positive),
    b_global_sets:static_symmetry_reduction_for_global_sets(State),

    apply_kodkod(Constants,TotalPredicate,TotPred),
    %predicate_level_optimizations(TotPred1,TotPred), % now done inside b_interpreter_components
    b_interpreter_components:reset_component_info(false),
    % TO DO: prioritise components containing parts/variables from Assertions (see set_projection_on_static_assertions/)
    b_interpreter_components:b_trace_test_components(TotPred,State,TypedVals,all),
    normalise_store(State,NormalisedState).

% input Options: output: list of identifiers over which predicate solved + predicate itself
get_cbc_assertions_predicate_to_solve(Options,  Constants, TotalPredicate,NegateGoal,NegGoal) :-
   get_cbc_assertions_sequent(Options,  Constants, Properties, GoalToProve),
   conjunct_predicates(GoalToProve,GoalC),
    (debug_mode(on) -> print('Trying to prove goal: '),nl,translate:nested_print_bexpr_as_classicalb(GoalC) ; true),
   (NegateGoal=true -> create_negation(GoalC,NegGoal) ; NegGoal=GoalC),
    %% print('Using Hypotheses: '),nl,translate:nested_print_bexpr_as_classicalb(Properties), %%
   conjunct_predicates([NegGoal,Properties],TotalPredicate).

get_cbc_assertions_sequent(Options,  Constants, Properties, GoalToProve) :-
    b_get_machine_constants(AllConstants),
    (member(specific(Label),Options) -> WhichAss=specific(Label)
     ; member(main_assertions,Options) -> WhichAss=main
     ; WhichAss=all),
    b_get_assertions(WhichAss,static,Assertions),
   ((get_preference(use_po,true), nonmember(tautology_check,Options))
      -> exclude(bmachine:is_discharged_assertion,Assertions,UnProvenAssertions)
       ; UnProvenAssertions = Assertions
    ),
    (member(tautology_check,Options) % used for ProB as an Atelier-B disprover/prover: the ASSERTIONS are an implication
      -> find_identifier_uses_l(UnProvenAssertions,[],Ids),
         include(id_mem(Ids),AllConstants,Constants), % only include constants used in tautology
         construct_sequent(UnProvenAssertions,Properties,GoalToProve)
      ; % Sequent: Properties |- UnProvenAssertions
        b_get_properties_from_machine(Properties),
        Constants = AllConstants,
        GoalToProve = UnProvenAssertions
    ).

construct_sequent([IMPLICATION],HYP,[RHS]) :- is_an_implication(IMPLICATION,HYP,RHS),!.
construct_sequent(UPAssertions,b(truth,pred,[]),UPAssertions).

id_mem(IDList,TID) :- get_texpr_id(TID,ID), member(ID,IDList).

:- use_module(unsat_cores,[unsat_core_with_time_limit/3]).
compute_unsat_core_if_requested(Options,Predicate,WTime) :-
    member(unsat_core,Options),!,
    format('Computing UNSAT CORE (~w ms to find initial contradiction)~n',[WTime]),
    unsat_core_with_time_limit(Predicate,CorePredicate,WTime),
    nl,print('UNSAT CORE: '),nl,
    print('--------'),nl,
    translate:nested_print_bexpr_as_classicalb(CorePredicate),
    print('--------'),nl.
compute_unsat_core_if_requested(_,_,_).
