% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(bmachine_construction,[reset_bmachine_construction/0,
                                 check_machine/4,
                                 type_in_machine_l/6,
                                 type_open_predicate_with_quantifier/6,
                                 get_constructed_machine_name/2, get_constructed_machine_name_and_filenumber/3,
                                 type_open_formula/8,
                                 filter_linking_invariant/3,
                                 external_procedure_used/1]).

:- use_module(library(lists)).
:- use_module(library(avl)).

:- use_module(self_check).
:- use_module(tools).
:- use_module(error_manager).
:- use_module(debug).
:- use_module(preferences).

:- use_module(btypechecker).
:- use_module(b_ast_cleanup).
:- use_module(bsyntaxtree).
:- use_module(bmachine_structure).
:- use_module(pragmas).

:- use_module(bmachine_static_checks).

:- use_module(translate,[print_machine/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module contains the rules for loading, including, seeing, etc. B machines, scope of constants, variables, parameters, etc.').

%maximum_type_errors(100).

:- dynamic debug_machine/0.

:- use_module(pref_definitions,[b_get_important_preferences_from_raw_machine/1]).
set_important_prefs_from_machine(Main,Machines) :-
    find_machine(Main,Machines,_MType,_Header,RawMachine),
    b_get_important_preferences_from_raw_machine(RawMachine).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(preferences,[temporary_set_preference/3,reset_temporary_preference/2]).

% typecheck complete machines (incl. includes)
check_machine(MainName,MachinesMayContainGenerated,Result,FinalErrors) :-
    clear_warnings,
    debug_println(9,checking_machine(MainName)),
    temporary_set_preference(perform_stricter_static_checks,true,CHNG),
    strip_global_pragmas(MachinesMayContainGenerated,Machines,_WasGenerated), % TODO: is generated still used?
    set_important_prefs_from_machine(MainName,Machines), % set important prefs before type-checking,...
    % we need to find all SEES/USES declarations, because all references
    % refer to the same machine and that has to be included exactly once.
    ~~mnf( find_uses(Machines,Uses,NotIncluded,Errors,E1) ),
    % if a seen/used machine was not included, we include it by constructing
    % a dummy machine that extends the seen/used machine and the origional
    % main machine
    ~~mnf( extend_not_included_uses(NotIncluded,MainName,NewMainName,Machines,IMachines) ),
    % figure out the right order of machines, so that each machine is
    % type-checked before a dependend machine is loaded
    ~~mnf( machine_order(IMachines,Order) ),
    % expand (= type-check + includes) machines in the given order
    ~~mnf( expand_machines(Order,IMachines,Uses,[],Results1,E1,E2) ),
    % until now, the initialisation section consists of a list of init(MachineName,Substitution)
    % fold_initialisation merges those to one substitution, respecting the dependencies
    % between the machines
    maplist(fold_initialisation(Order),Results1,Results),
    % find main machine
    memberchk(machine(NewMainName,MainMachine),Results),  %nl,print('MAIN:'),nl,print_machine(MainMachine),
    % if the main machine has parameters, we convert them to global sets resp. constants
    ~~mnf( convert_parameters_to_global_sets(E2,[],MainMachine,Result1) ),
    % add some additional informations about the machine
    add_machine_infos(MainName,Machines,Results,Result1,Result),
    ( debug_machine -> print_machine(Result); true),
    % finalize the list of errors, remove duplicates
    sort(Errors,FinalErrors),
    % output warnings
    show_warnings,
    % run static checks on the resulting machine
    static_check_machine(Result),
    reset_temporary_preference(perform_stricter_static_checks,CHNG).
check_machine(Main,_,Result,FinalErrors) :-
    add_internal_error('Internal error: Checking the machine failed: ',
                       check_machine(Main,_,Result,FinalErrors)),
%    trace,check_machine(Main,Machines,Result,FinalErrors),
    fail.

strip_global_pragmas([],[],false).
strip_global_pragmas([generated(_,M)|Ms],MachinesOut,true) :- !,
    strip_global_pragmas([M|Ms],MachinesOut,_).
strip_global_pragmas([unit_alias(_,Alias,Content,M)|Ms],MachinesOut,true) :- !,
    assert(pragmas:global_pragma(unit_alias,[Alias|Content])),
    strip_global_pragmas([M|Ms],MachinesOut,_).
strip_global_pragmas([M|Ms],[M|SMs],WasGenerated) :-
    strip_global_pragmas(Ms,SMs,WasGenerated).

fold_initialisation(Order,machine(Name,Sections),machine(Name,NewSections)) :-
    select_section(initialisation,List,Subst,Sections,NewSections),
    maplist(extract_init_substitutions(List),Order,LSubstitutions),
    append(LSubstitutions,Substitutions),
    create_sequence(Substitutions,Subst).
extract_init_substitutions(Unsorted,Name,Substitutions) :-
    convlist(unzip_init(Name),Unsorted,Substitutions).
unzip_init(Name,init(Name,Subst),Subst).
create_sequence([],Subst) :- create_texpr(skip,subst,[generated],Subst).
create_sequence([I],I) :- !.
create_sequence(L,Sequence) :- create_texpr(sequence(L),subst,[initial_sequence],Sequence).

add_machine_infos(Main,Machines,CheckedMachines,Old,New) :-
    append_to_section(meta,[hierarchy/Hierarchy],Old,NewT),
    get_refinement_hierarchy(Main,Machines,Hierarchy),
    add_refined_machine(Hierarchy,CheckedMachines,NewT,New).
get_refinement_hierarchy(Main,Machines,[Main|Abstractions]) :-
    ( find_refinement(Machines,Main,Abstract) ->
        get_refinement_hierarchy(Abstract,Machines,Abstractions)
    ; otherwise ->
        Abstractions = []).
find_refinement([M|Rest],Name,Abstract) :-
    ( get_constructed_machine_name(M,Name) ->
        refines(M,Abstract)
    ; otherwise ->
        find_refinement(Rest,Name,Abstract)).

add_refined_machine([_Main,Refined|_],Machines,Old,New) :-
    member(machine(Refined,Body),Machines),
    append_to_section(meta,[refined_machine/Body],Old,New).
add_refined_machine(_,_,M,M). % not refining

convert_parameters_to_global_sets(Ein,Eout) -->
    % extract and remove parameters and constraints
    select_section(parameters,PP,[]),
    select_section(internal_parameters,IP,[]),
    {create_texpr(truth,pred,[],Truth)},
    select_section(constraints,C,Truth),
    % split parameters into sets and scalars
    { split_list(is_set_parameter,PP,Sets,Scalars),
      foldl(type_set_parameter,Sets,Ein,Eout) },
    % put the sets to the deferred sets
    append_to_section(deferred_sets,Sets),
    % and the scalars to the constants
    append_to_section(concrete_constants,Scalars),
    append_to_section(concrete_constants,IP),
    % the scalars should be typed by constraints,
    % so move the constraints to the properties
    select_section(properties,OldProps,NewProps),
    {conjunct_predicates([C,OldProps],NewProps)}.
is_set_parameter(TExpr) :-
    % upper case identifiers denote set parameters, otherwise scalars
    get_texpr_id(TExpr,Name),is_upper_case(Name).
type_set_parameter(TExpr,Ein,Eout) :-
    get_texpr_id(TExpr,Name),
    get_texpr_type(TExpr,Type),
    get_texpr_pos(TExpr,Pos),
    % we directly type the deferred set
    unify_types_werrors(set(global(Name)),Type,Pos,Ein,Eout).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% included machines

expand_machines([],_,_,M,M,Errors,Errors).
expand_machines([M|Rest],Machines,Uses,Typed,Result,Ein,Eout) :-
    ~~mnf( expand_machine(M,Machines,Uses,Typed,New,Ein,E1) ),
    expand_machines(Rest,Machines,Uses,[machine(M,New)|Typed],Result,E1,Eout).

% resolve all includes and typecheck the machine
% expand_machine(Name,Machines,TypeChecked,Expanded) :
%   Name of the Machine to expand
%   Machines contains the list of all loaded machines
%   TypeChecked contains the list of all expanded machines so far
%   Expanded is the resulting machine
expand_machine(Name,Machines,GlobalUses,TypeChecked,Expanded,Ein,Eout) :-
    % find the machine in the list
    ~~mnf( find_machine(Name,Machines,MType,Header,RawMachineWithPragmas) ),
    % remove pragmas for later reattachment
    ~~mnf( strip_machine_section_pragmas(RawMachineWithPragmas,RawMachine,_Pragmas) ),
    % look for abstract machine
    ~~mnf( get_abstractions(Name,Machines,TypeChecked,Abstractions) ),
    % merge all used and seen machines
    ~~mnf( use_and_see_machines(RawMachine,TypeChecked,SeenRefs) ),
    % merge all included machines into one machine
    ~~mnf( include_machines(RawMachine, TypeChecked, GlobalUses, Includes, Parameters, Included, Promotes) ),
    % Parameters contains now a list of parameters for each included machine
    % Included contains now a machine that represents all included machines
    %   their parameters are renamed to internal_parameters
    append([Abstractions,Included,SeenRefs],RefMachines),
    % typecheck this machine
    debug_stats(type_machine(Name)),
    ~~mnf( type_machine(Header,Name,MType,RawMachine,RefMachines,TypedLocal,RefMachines2,Ein,Err1) ),
    % merge included and including machine
    debug_stats(merge_included_machines(Name)),
    ~~mnf( merge_included_machines(Name,TypedLocal,Included,Promotes,Expanded2,Err1,Err2) ),
    % add some predicates that state the equivalence between arguments
    %  in the include statement and the parameters of the included machine
    ~~mnf( add_link_constraints(Includes,MType,Parameters,RefMachines2,Expanded2,Expanded3,Err2,Err3) ),
    % put together refinement and abstraction.
    ~~mnf( merge_refinement_and_abstraction(Expanded3,[ref(local,TypedLocal)|RefMachines2],Expanded4,Err3,Eout)),
    % merge used machines, will also prefix
    debug_stats(merge_used_machines(Name)),
    ~~mnf( merge_used_machines(Included,Expanded4,Expanded5) ),
    % clean up the syntax tree
    debug_stats(clean_up_machine(Name)),
    ~~mnf( clean_up_machine(Expanded5,RefMachines2,Expanded) ),
    debug_stats(finished_expand_machine(Name)).

strip_machine_section_pragmas(RawMachineWithPragmas,RawMachine,Pragmas) :-
    selectchk(units(_Pos,RealVariables,Pragmas),RawMachineWithPragmas,TRawMachine), !,
    RawMachine = [RealVariables|TRawMachine].
strip_machine_section_pragmas(Machine,Machine,[]).

merge_included_machines(Name, TypedLocal, RefMachines, Promotes, Merged, Ein, Eout) :-
    include(is_included_ref,RefMachines,Included),
    create_machine(Name,Empty),
    % move the included operations into the promoted or unpromoted section
    ~~mnf( move_operations(Included,Promotes,Included2,Ein,Eout) ),
    LocalAndIncluded = [ref(local,TypedLocal)|Included2],
    % TODO: make sure that the constants of two instances of the same machine do not repeat (check what should be done for distinct machines with same identifiers)
    ~~mnf( concat_sections_of_refs([identifiers,initialisation,operation_bodies,
                                    assertions,used,values],LocalAndIncluded,Empty,Merged1) ),
    ~~mnf( conjunct_sections_of_refs([constraints,properties,invariant],LocalAndIncluded,Merged1,Merged2) ),
    ~~mnf( concat_section_of_simple_lists(freetypes,LocalAndIncluded,Merged2,Merged3) ),
    ~~mnf( get_section(definitions,TypedLocal,Defs) ),
    ~~mnf( write_section(definitions,Defs,Merged3,Merged) ).
is_included_ref(ref(included,_)).

concat_section_of_simple_lists(Sec,References,In,Out) :-
    maplist(extract_machine_from_ref,References,Machines),
    foldl(concat_section_of_simple_lists2(Sec),Machines,In,Out).
concat_section_of_simple_lists2(Sec,Machine,In,Out) :-
    get_section(Sec,In,Orig),
    get_section(Sec,Machine,List),
    append(Orig,List,NewList),
    write_section(Sec,NewList,In,Out).

merge_used_machines(RefMachines,Old,New) :-
    % get all included machines from the references
    convlist(ref_to_included_machine,RefMachines,Included),
    foldl(merge_used_machines2,Included,Old,New).
merge_used_machines2(Included,Old,New) :-
    get_section(used,Included,IncludedUse),
    ( IncludedUse = [] -> New = Old
    ; otherwise ->
        rename_used(IncludedUse,Old,New)).
ref_to_included_machine(ref(included,Inc),Inc).

% will add prefixes to identifiers
rename_used(IncludedUse,Old,New) :-
    %print(renaming(IncludedUse)),nl,
    expand_shortcuts([properties,invariant, assertions,
                      initialisation,operation_bodies],Sections), % TO DO: also traverse GOAL ? should we later apply this at the REPL level as well ? should we apply it to other DEFINITIONS ?
    foldl(rename_used_sec(IncludedUse),Sections,Old,New).
rename_used_sec(IncludedUse,Sec,Old,New) :-
    select_section_texprs(Sec,TExprs,NewTExprs,Old,New),
    rename_used2_l(TExprs,IncludedUse,NewTExprs).

rename_used2(TExpr,IncludedUse,NewTExpr) :-
    get_texpr_expr(TExpr,operation(Id,Params,Results,Subst)),!,
    rename_used2_l(Params,IncludedUse,NParams),
    rename_used2_l(Results,IncludedUse,NResults),
    rename_used2(Subst,IncludedUse,NSubst),
    get_texpr_type(TExpr,Type),
    get_texpr_info(TExpr,Info),
    selectchk(modifies(MIn),Info,Info1),selectchk(reads(RIn),Info1,RestInfo),
    rename_used_ids(MIn,IncludedUse,MOut),
    rename_used_ids(RIn,IncludedUse,ROut),
    create_texpr(operation(Id,NParams,NResults,NSubst),Type,[modifies(MOut),reads(ROut)|RestInfo],
                 NewTExpr).
rename_used2(TExpr,IncludedUse,NewTExpr) :-
    % rename identifier by adding machine prefix M.id
    get_texpr_id(TExpr,_), % result is also Id (always ??)
    get_texpr_info(TExpr,Info),
    member(usesee(Name,Id,_Mode),Info),
    member(includeduse(Name,Id,NewTExpr),IncludedUse),  % This seems to reuse position information from includeduse list for all identifiers !! TO DO: check this
    % we now rename Id to Name.Id
    !. %, print(ren_id(Name,Id)),nl.
rename_used2(TExpr,IncludedUse,NTExpr1) :-
    remove_bt(TExpr,Expr,NExpr,NTExpr),
    syntaxtransformation(Expr,Subs,_,NSubs,NExpr),!,
    rename_used2_l(Subs,IncludedUse,NSubs),
    rename_infos(NTExpr,IncludedUse,NTExpr1).
rename_used2(TExpr,IncludedUse,NTExpr) :-
    add_internal_error('Rename failed: ',rename_used2(TExpr,IncludedUse,NTExpr)),
    NTExpr = TExpr.

% update infos, e.g., read/modifies for while loops
rename_infos(b(E,T,I),List,b(E,T,NI)) :- maplist(rename_info(List),I,NI).

rename_info(IncludeUseList,Info,NewInfo) :- info_field_contains_ids(Info,I,NewInfo,SNI),!,
     maplist(rename_usage_info(IncludeUseList),I,NI), sort(NI,SNI).
rename_info(_,I,I).

info_field_contains_ids(reads(I),I,reads(SNI),SNI).
info_field_contains_ids(modifies(I),I,modifies(SNI),SNI).
info_field_contains_ids(used_ids(I),I,used_ids(SNI),SNI).

rename_usage_info(IncludeUseList,ID,NewID) :-
   (member(includeduse(_,ID,NewTExpr),IncludeUseList) -> get_texpr_id(NewTExpr,NewID) ; NewID = ID).


rename_used2_l([],_,R) :- !, R=[].
rename_used2_l([T|Rest],IncludedUse,[NT|NRest]) :-
    rename_used2(T,IncludedUse,NT), !,
    rename_used2_l(Rest,IncludedUse,NRest).
rename_used2_l(X,Y,Z) :- add_internal_error('Rename failed: ', rename_used2_l(X,Y,Z)),Z=X.

rename_used_ids(InIds,IncludedUse,OutIds) :-
    maplist(rename_used_ids2(IncludedUse),InIds,OutIds).
rename_used_ids2(IncludedUse,InId,OutId) :-
    memberchk(includeduse(_,InId,TOutId),IncludedUse),!,
    get_texpr_id(TOutId,OutId).
rename_used_ids2(_IncludedUse,Id,Id).

get_abstractions(CName,Machines,TypedMachines,[ref(abstraction,Abstraction)]) :-
    refines(M,AName),
    get_constructed_machine_name(M,CName),
    memberchk(M,Machines),!,
    memberchk(machine(AName,Abstraction),TypedMachines).
get_abstractions(_,_,_,[]).

get_includes_and_promotes(Sections,M,Includes,Promotes) :-
    optional_rawmachine_section(includes,Sections,[],Includes1),
    optional_rawmachine_section(imports,Sections,[],Imports),
    optional_rawmachine_section(extends,Sections,[],Extends),
    optional_rawmachine_section(promotes,Sections,[],Promotes1),
    append([Includes1,Extends,Imports],Includes),
    maplist(expand_extends(M),Extends,Promotes2),
    append([Promotes1|Promotes2],Promotes).
expand_extends(Machines,machine_reference(_,Ref,_),Promotes) :-
    split_prefix(Ref,Prefix,Name),
    memberchk(machine(Name,Body),Machines),
    get_section(promoted,Body,Promoted),
    prefix_identifiers(Promoted,Prefix,Renamings),
    rename_bt_l(Promoted,Renamings,TPromotes),
    maplist(add_nonpos,TPromotes,Promotes).
add_nonpos(TId,identifier(none,Id)) :-
    get_texpr_id(TId,op(Id)).

move_operations([],Promotes,[],Ein,Eout) :-
    foldl(add_promotes_not_found_error,Promotes,Ein,Eout).
move_operations([ref(_,IncMachine)|IncRest],Promotes,[ref(included,NewIncMachine)|RefRest],Ein,Eout) :-
    move_operations2(IncMachine,Promotes,NewIncMachine,RestPromotes),
    move_operations(IncRest,RestPromotes,RefRest,Ein,Eout).
move_operations2(Included,Promotes,Result,RestPromotes) :-
    select_section(promoted,IncOperations,Promoted,Included,Included1),
    select_section(unpromoted,OldUnpromoted,Unpromoted,Included1,Result),
    filter_promoted(IncOperations,Promotes,Promoted,NewUnpromoted,RestPromotes),
    append(OldUnpromoted,NewUnpromoted,Unpromoted).
filter_promoted([],Promotes,[],[],Promotes).
filter_promoted([TExpr|OpsRest],Promotes,Promoted,Unpromoted,RestPromotes) :-
    get_texpr_id(TExpr,op(P)),
    ( select(identifier(_,P),Promotes,RestPromotes1) ->
        !,Promoted = [TExpr|RestPromoted],
        Unpromoted = RestUnpromoted
    ; otherwise ->
        RestPromotes1 = Promotes,
        Promoted = RestPromoted,
        Unpromoted = [TExpr|RestUnpromoted]),
    filter_promoted(OpsRest,RestPromotes1,RestPromoted,RestUnpromoted,RestPromotes).
add_promotes_not_found_error(identifier(Pos,Id),[error(Msg,Pos)|E],E) :-
    ajoin(['Promoted operation ',Id,' not found'],Msg).

find_machine(Name,Machines,Type,Header,Sections) :-
    Header = machine_header(_,Name,_),
    ( (member(abstract_machine(_,_ModelType,Header,Sections),Machines),
          Type=machine)
    ; (member(refinement_machine(_,Header,_Abstract,Sections),Machines),
          Type=refinement)
    ; (member(implementation_machine(_,Header,_Abstract,Sections),Machines),
          Type=implementation)),
    !.

include_machines(RawMachine, TypeChecked, GlobalUses, Includes, Parameters, Included, Promotes) :-
    get_includes_and_promotes(RawMachine,TypeChecked,Includes,Promotes),
    maplist(include_machine(TypeChecked, GlobalUses), Includes, Parameters, Singles),
    remove_duplicate_set_inclusions(Singles,Singles1),
    % create refs
    maplist(create_ref,Singles1,Included).
create_ref(IncMachine,ref(included,IncMachine)).

% it is possible that a deferred or enumerated set is declared in an included machine
% that is included more than once. We remove all but one occurrences.
remove_duplicate_set_inclusions([],[]).
remove_duplicate_set_inclusions([M],[M]) :- !.
remove_duplicate_set_inclusions([M|Rest],[M|CleanedRest]) :-
    get_section(deferred_sets,M,DSets),
    get_section(enumerated_sets,M,ESets),
    get_section(enumerated_elements,M,EElements),
    append([DSets,ESets,EElements],Identifiers),
    maplist(remove_duplicate_sets2(Identifiers),Rest,CleanedRest).
remove_duplicate_sets2(Identifiers,M,Cleaned) :-
    remove_duplicate_sets_section(deferred_sets,Identifiers,M,M1),
    remove_duplicate_sets_section(enumerated_sets,Identifiers,M1,M2),
    remove_duplicate_sets_section(enumerated_elements,Identifiers,M2,Cleaned).
remove_duplicate_sets_section(Section,Identifiers,In,Out) :-
    select_section(Section,Old,New,In,Out),
    exclude(element_is_duplicate(Identifiers),Old,New).
element_is_duplicate(Identifiers,TId) :-
    get_texpr_id(TId,Name),
    get_texpr_type(TId,Type),
    get_texpr_id(ToRemove,Name),
    get_texpr_type(ToRemove,Type),
    member(ToRemove,Identifiers),
    get_texpr_info(TId,InfosA),
    get_texpr_info(ToRemove,InfosB),
    member(def(Sec,File),InfosA),
    member(def(Sec,File),InfosB),
    !.

include_machine(TypeChecked,Used,machine_reference(_Pos,FullRef,_Args),
                parameters(FullRef,Parameters), TM) :-
    split_prefix(FullRef,Prefix,Name),
    % TM1 is the already typechecked included machine:
    member(machine(Name,TM1),TypeChecked),!,
    debug_println(9,including_machine(Name)),
    % TM2 is a fresh copy, so we prevent unification of the parameter types:
    copy_term(TM1,TM2),
    % If the included machine is used somewhere, we store the identifiers
    % to enable joining the different references later:
    include_usings(Name,Used,TM2,TM3),
    % TM3 is typechecked and all internal variables are renamed with a prefix
    hide_private_information(FullRef,TM3,TM4),
    % TM4 is typechecked, and if it was referenced with a prefix (e.g. INCLUDES p.M2)
    % all variables are prefixed
    %print(prefixing(Prefix)),nl,
    prefix_machine(Prefix,TM4,TM5),
    % We need the parameters later to state their equivilance to the arguments
    get_section(parameters,TM5,Parameters),
    % We move the parameters to the internal parameters, because the resulting
    % machine has only the parameters of the including machine.
    parameters_to_internal(TM5,TM).

include_usings(Name,Uses,Old,New) :-
    ( member(Name,Uses) ->
        store_usage_info(Old,Name,UsedInfo),
        append_to_section(used,UsedInfo,Old,New)
    ; otherwise ->
        Old = New).
% returns a list of which identifiers are used in the machine
% for each identifier, we have a trible includeuse(Name,Id,TExpr)
% where Name is the name of the used machine, Id is the
% original ID and TExpr is the currently used reference to this
% identifier
store_usage_info(Machine,Name,UsedInfo) :-
    expand_shortcuts([identifiers],IdSections),
    foldl(store_usage_info2(Machine,Name),IdSections,UsedInfo,[]).
store_usage_info2(Machine,Name,Sec,I,O) :-
    get_section(Sec,Machine,Content),
    foldl(store_usage_info3(Name),Content,I,O).
store_usage_info3(Name,TId,[includeduse(Name,Id,TId)|L],L) :-
    get_texpr_id(TId,Id).

% conjunct sections that contain predicates
conjunct_sections_of_refs(Sections1,References,Old,New) :-
    expand_shortcuts(Sections1,Sections),
    maplist(extract_machine_from_ref,References,Machines),
    foldl(conjunct_sections2(Machines),Sections,Old,New).
conjunct_sections2(Machines,Section,Old,New) :-
    write_section(Section,NewContent,Old,New),
    get_section_of_machines(Machines,Section,Preds),
    conjunct_predicates(Preds,NewContent).

% merge sections that contain a list of expressions
concat_sections_of_refs(Sections1,References,Old,New) :-
    maplist(extract_machine_from_ref,References,Machines),
    % for each machine, create a tag where the expression comes from
    maplist(create_tag_by_reference,References,Tags),
    concat_sections(Sections1,Machines,Tags,Old,New).
extract_machine_from_ref(ref(_,M),M).
create_tag_by_reference(ref(local,_Machine),[]) :- !.
create_tag_by_reference(ref(RefType,Machine),[RefType/Name]) :-
    get_machine_name(Machine,Name).
concat_sections(Sections1,Machines,Tags,Old,New) :-
    expand_shortcuts(Sections1,Sections),
    foldl(concat_section2(Machines,Tags),Sections,Old,New).
concat_section2(Machines,Tags,Section,Old,New) :-
    write_section(Section,NewContent,Old,New),
    maplist(get_tagged_lsection_of_machine(Section),Machines,Tags,Contents),
    append(Contents,NewContent).
get_tagged_lsection_of_machine(Section,Machine,Tags,TaggedContent) :-
    get_section(Section,Machine,Content),
    maplist(tag_with_origin(Tags),Content,TaggedContent).
tag_with_origin(Origins,TExpr,TaggedExpr) :-
    change_info_of_expression_or_init(TExpr,Info,TaggedInfo,TaggedExpr),
    % add a new origin to the old tag or if not existent, add a new info field
    ( Origins = [] -> TaggedInfo = Info
    ; selectchk(origin(Orest),Info,origin(Onew),TaggedInfo) -> append(Origins,Orest,Onew)
    ; otherwise -> TaggedInfo = [origin(Origins)|Info]).
% the substitutions in the initialisation are additionally wrapped by an init/2 term
% a small hack to work with those to.
% TODO: this became a very ugly hack -- redo!
change_info_of_expression_or_init(init(A,OExpr),Old,New,init(A,NExpr)) :-
    !,change_info_of_expression_or_init(OExpr,Old,New,NExpr).
% ignore the info for includeduse/3 completely
change_info_of_expression_or_init(includeduse(A,B,C),[],_,includeduse(A,B,C)) :- !.
change_info_of_expression_or_init(OExpr,Old,New,NExpr) :-
    create_texpr(Expr,Type,Old,OExpr), create_texpr(Expr,Type,New,NExpr).


% adds a prefix to all variables and promoted operations
prefix_machine('',TM,TM) :- !.
prefix_machine(Prefix,Old,New) :-
    debug_println(5,prefixing_machine(Prefix)),
    expand_shortcuts([variables,promoted], RenamedIdentiferSections),
    get_all_identifiers(RenamedIdentiferSections,Old,Identifiers),
    prefix_identifiers(Identifiers,Prefix,Renamings),
    find_relevant_sections(RenamedIdentiferSections,[machine],Sections1),
    append(RenamedIdentiferSections,Sections1,Sections),
    rename_in_sections(Sections,Renamings,Old,M),
    rename_includeduse(M,Renamings,New).
rename_includeduse(Old,Renamings,New) :-
    select_section(used,OldContent,NewContent,Old,New),
    maplist(rename_includeduse2(Renamings),OldContent,NewContent).
rename_includeduse2(Renamings,includeduse(M,N,TExpr),includeduse(M,N,NExpr)) :-
    rename_bt(TExpr,Renamings,NExpr).

get_all_identifiers(Sections1,M,Identifiers) :-
    expand_shortcuts(Sections1,Sections),
    maplist(get_all_identifiers2(M),Sections,LIdentifiers),
    append(LIdentifiers,Identifiers).
get_all_identifiers2(M,Sec,Identifiers) :-
    get_section(Sec,M,Identifiers).

% hide all parameters and unpromoted operations
hide_private_information(Prefix,Machine,NewMachine) :-
    get_section(parameters,Machine,Params),
    get_section(unpromoted,Machine,UnPromoted),
    append(Params,UnPromoted,ToHide),
    %debug_println(9,hide_private(Prefix,ToHide)),
    ( ToHide = [] -> NewMachine=Machine
    ; otherwise ->
        prefix_identifiers(ToHide,Prefix,Renamings),
        % we have to do the renaming in promoted operations, too, because
        % those operations might use the renamed parameters and their reads/modifies
        % info must be updated
        rename_in_sections([parameters,promoted,unpromoted],Renamings,Machine,Machine1),
        rename_includeduse(Machine1,Renamings,Machine2),
        rename_relevant_sections([parameters,operations],Renamings,Machine2,NewMachine)).

prefix_identifiers(Identifiers,'',Identifiers) :- !.
prefix_identifiers(Old,Prefix,New) :-
    maplist(prefix_identifier(Prefix),Old,New).
prefix_identifier(Prefix,TExpr,rename(Old,New)) :-
    get_texpr_expr(TExpr,identifier(Old)),
    ( Old=op(OI) -> New=op(NI)
    ; otherwise  -> OI=Old,NI=New),
    ajoin([Prefix,'.',OI],NI). %   , print(rename(Old,New)),nl.

parameters_to_internal(M1,M2) :-
    select_section(internal_parameters,OldParams,Params,M1,M3),
    select_section(parameters,NewParams,[],M3,M2),
    append(OldParams,NewParams,Params).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% uses and sees relations
find_uses(Machines,Uses,NotIncluded,Ein,Eout) :-
    findall(U, use_usage(Machines,_,U), UnsortedUses),
    sort(UnsortedUses,Uses),
    check_include_use(Uses,Machines,NotIncluded,Ein,Eout).

% check_include_use/5 checks if the used machines are included in any machine
% check_include_use(+Uses,+Machines,-NotIncluded,Ein,Eout)
%   Uses: The list of machine names that are used
%   Machines: The list of machines
%   NotIncluded: The list of used machines (their names) that are not included
%   Ein/Eout: The errors (as difference-list)
check_include_use([],_,[],Errors,Errors).
check_include_use([U|Rest],Machines,NotIncluded,Ein,Eout) :-
    findall(I,include_usage(Machines,U,I),Inc),
    ( Inc=[]  -> NotIncluded = [U|RestNotIncluded], Ein=E1
    ; Inc=[_] -> NotIncluded = RestNotIncluded, Ein=E1
    ; otherwise ->
        NotIncluded = RestNotIncluded,
        use_usage(Machines,Pos,U),
        Ein = [error('used machine is included more than once',Pos)|E1]
    ),
    check_include_use(Rest,Machines,RestNotIncluded,E1,Eout).
% extend_not_included_uses(+Uses,+Main,-Name,+Machines,-AllMachines):
%   Create a dummy machine that extends the main machine and extends or includes
%   all seen/used machines that are not included if such machines exist.
%   In case of refinement this is done for the whole refinement chain.
%   Uses: List of machine names of the machines that are used/seen but not included
%   Main: The name of the main machine
%   Name: The name of the generated dummy machine (or Main if no dummy machine is generated)
%   Machines: List of all specified Machines
%   AllMachines: List of all Machines + the new dummy machine(s)
extend_not_included_uses([],Main,Main,Machines,Machines) :- !.
extend_not_included_uses(Uses,Main,Name,Machines,AllMachines) :-
    get_refinement_hierarchy(Main,Machines,RefChain),
    maplist(extend_not_included_uses2(Uses,Machines),RefChain,NewMachines),
    append(NewMachines,Machines,AllMachines),
    dummy_machine_name(Main,Name).
extend_not_included_uses2(Uses,Machines,Name,DummyMachine) :-
    create_dummy_machine(Name,Machines,Parameters,DummyParameters,Sections,DummyMachine),
    IncludeMain = machine_reference(none,Name,Parameters),
    ( get_preference(seen_machines_included,true) ->
        % extend just the main machine, include the used/seen machines
        References = [extends(none,[IncludeMain]),includes(none,UReferences)]
    ; otherwise ->
        % extends the main machine and all used/seen machines
        References = [extends(none,[IncludeMain|UReferences])]),
    maplist(find_using(Machines),Uses,UReferences,LUParameters),
    append([Parameters|LUParameters],DummyParameters),
    copy_raw_definitions(Name,Machines,OptDefs),
    % we store a flag "is_dummy" in the machine because we have a special case
    % later, see type_constraints/7.
    append([References,[is_dummy],OptDefs],Sections).
create_dummy_machine(Name,Machines,Parameters,DummyParameters,Sections,DummyMachine) :-
    dummy_machine_name(Name,DummyName),
    get_model_type(Name,Machines,ModelType),
    Header = machine_header(_,Name,Parameters),
    DummyHeader = machine_header(none,DummyName,DummyParameters),
    member(Machine,Machines),
    generate_raw_machine(Header,DummyHeader,ModelType,Sections,Machine,DummyMachine),!.

generate_raw_machine(OldHeader,NewHeader,_,NewSections,
                     abstract_machine(_,   ModelType,OldHeader,_),
                     abstract_machine(none,ModelType,NewHeader,NewSections)).
generate_raw_machine(OldHeader,NewHeader,ModelType,NewSections,
                     refinement_machine(_,   OldHeader,_Abstract,   _),
                     abstract_machine(none,ModelType,NewHeader,NewSections)).
generate_raw_machine(OldHeader,NewHeader,ModelType,NewSections,
                     implementation_machine(_,   OldHeader,_Abstract,   _),
                     abstract_machine(none,ModelType,NewHeader,NewSections)).
dummy_machine_name(Name,DummyName) :-
    atom_concat('MAIN_MACHINE_FOR_',Name,DummyName).

find_using(Machines,U,machine_reference(none,Name,Arguments),Arguments) :-
    % was using(U).U
    ajoin([U,'.',U],Name),
    member(M,Machines), get_machine_parameters(M,U,Params),!,
    maplist(add_use_param,Params,Arguments).
add_use_param(identifier(_,Param),identifier(none,Name)) :-
    ( is_upper_case(Param) ->
        ajoin(['Useparam(',Param,')'],Name)
    ; otherwise ->
        ajoin(['useparam(',Param,')'],Name)).

% copy_raw_definitions(+Name,+Machines,-OptDefs):
%  Get the definitions section from a raw (untyped) machine
%  Name: Name of the machine
%  Machines: List of Machines
%  OptDefs: [] if no definitions are present or [Def] if a definition section Def is present
copy_raw_definitions(Name,Machines,OptDefs) :-
    get_constructed_machine_name_and_body(M,Name,_,Sections),
    memberchk(M,Machines),!,
    Def = definitions(_,_),
    ( memberchk(Def,Sections) ->
        OptDefs = [Def]
    ; otherwise ->
        OptDefs = []).

add_def_dependency_information(DefsIn,DefsOut,Ein,Eout) :-
    extract_def_name_set(DefsIn,DefNames,DN),
    maplist(add_def_dep(DN),DefsIn,Defs1),
    check_for_cyclic_def_dependency(Defs1,DefNames,DefsOut,Ein,Eout).

extract_def_name_set(Defs,DefNames,DN) :-
    maplist(get_def_name,Defs,DefNames),
    maplist(to_mapset_entry,DefNames,DN1),
    list_to_avl(DN1,DN).

get_def_name(Def,Name) :- arg(1,Def,Name).
get_def_pos(Def,Pos) :- arg(3,Def,Pos).
get_def_dependencies(Def,Dependencies) :- arg(6,Def,Dependencies).
to_mapset_entry(Name,Name-true).

add_def_dep(DN,In,Out) :-
    analyse_definition_dependencies(In,DN,Deps),
    In  = definition_decl(Name,DefType,Pos,Args,RawExpr),
    Out = definition_decl(Name,DefType,Pos,Args,RawExpr,Deps).

check_for_cyclic_def_dependency(Defs,DefNames,DefsOut,Ein,Eout) :-
    % check if we have a cyclic definition:
    create_definitions_avl(Defs,DefsAvl),
    search_for_cyclic_definition(DefNames,DefsAvl,[],Pos,RCycle),!,
    % if we have found a cyclic definition, generate an error message, ...
    reverse(RCycle,Cycle),add_arrows(Cycle,Msg0),
    ajoin(['Found cyclic definitions: '|Msg0],Msg),
    Ein = [error(Msg,Pos)|E1],
    % ... remove the definitions in the cycle (to prevent later infinite loops) ...
    exclude(definition_in_list(Cycle),Defs,Defs1),
    % ... and check the remaining definitions.
    check_for_cyclic_def_dependency(Defs1,DefNames,DefsOut,E1,Eout).
check_for_cyclic_def_dependency(Defs,_DefNames,Defs,E,E).
add_arrows([E],[E]) :- !.
add_arrows([E|Erest],[E,'->'|Arest]) :- add_arrows(Erest,Arest).
definition_in_list(List,Def) :-
    get_def_name(Def,Name),memberchk(Name,List).

create_definitions_avl(Defs,DefsAvl) :-
    maplist(split_def_name,Defs,Entries),
    list_to_avl(Entries,DefsAvl).
split_def_name(Def,Name-Def) :- get_def_name(Def,Name).


% just a depth-first search to find a cycle
search_for_cyclic_definition(DefNames,Definitions,Visited,Pos,Cycle) :-
    member(Name,DefNames),avl_fetch(Name,Definitions,Definition),
    get_def_pos(Definition,Pos),
    ( memberchk(Name,Visited) ->
        Cycle = [Name|Visited]
    ; otherwise ->
        get_def_dependencies(Definition,Dependencies),
        search_for_cyclic_definition(Dependencies,Definitions,[Name|Visited],_,Cycle)
    ).

:- assert_must_succeed((
      list_to_avl([def1-true,def2-true,def3-true,def4-true],DefNames),
      analyse_definition_dependencies(
        conjunct(none,
                 equals(none,
                        identifier(none,x),
                        identifier(none,def1)),
                 equals(none,
                        definition(none,def4,
                                   [function(none,
                                             identifier(none,def3),
                                             integer(none,5))]),
                        identifier(y))),DefNames,Defs),
      Defs==[def1,def3,def4]
    )).
% analyse_definition_dependencies(+Expr,+DefinitionNames,Deps):
%  Expr: raw (i.e. untyped) expression to analyse
%  DN: AVL set (i.e. mapping from elements to 'true') of the names of the definitions
%      This is needed to decide if an identifier is a reference to a definition
%  Deps: A list of used definitions (a list of their names)
analyse_definition_dependencies(Expr,DN,Deps) :-
    analyse_definition_dependencies2(Expr,DN,Unsorted,[]),
    sort(Unsorted,Deps).
analyse_definition_dependencies2([Expr|Rest],DN) -->
    !, analyse_definition_dependencies2(Expr,DN),
    analyse_definition_dependencies2(Rest,DN).
analyse_definition_dependencies2(definition(_Pos,Name,Args),DN) -->
    !,[Name],analyse_definition_dependencies2(Args,DN).
analyse_definition_dependencies2(identifier(_Pos,Name),DN) -->
    {avl_fetch(Name,DN),!},[Name].
analyse_definition_dependencies2(Expr,DN) -->
    { compound(Expr),functor(Expr,_Functor,Arity),!},
    analyse_definition_dependencies_arg(2,Arity,Expr,DN).
analyse_definition_dependencies2(_Expr,_DN) --> [].

analyse_definition_dependencies_arg(I,Arity,Expr,DN) -->
    { I =< Arity,!,arg(I,Expr,Arg),I2 is I+1 },
    analyse_definition_dependencies2(Arg,DN),
    analyse_definition_dependencies_arg(I2,Arity,Expr,DN).
analyse_definition_dependencies_arg(_I,_Arity,_Expr,_DN) --> [].



get_constructed_machine_name(MachineTerm,Name) :- get_constructed_machine_name_and_body(MachineTerm,Name,_Pos,_).
% name and pos; pos can be used for filenumber
get_constructed_machine_name_and_filenumber(MachineTerm,Name,Filenumber) :-
    get_constructed_machine_name_and_body(MachineTerm,Name,Pos,_),
    (Pos = pos(_,FN,_Srow,_Scol,_Erow,_Ecol) -> Filenumber=FN ; Filenumber=unknown).
get_constructed_machine_name_and_body(abstract_machine(_,_ModelType,machine_header(Pos,Name,_),Body),Name,Pos,Body).
get_constructed_machine_name_and_body(refinement_machine(_,machine_header(Pos,Name,_),_Abstract,Body),Name,Pos,Body).
get_constructed_machine_name_and_body(implementation_machine(_,machine_header(Pos,Name,_),_Abstract,Body),Name,Pos,Body).

refines(refinement_machine(_,machine_header(_,_Name,_),Abstract,_Body),Abstract).
refines(implementation_machine(_,machine_header(_,_Name,_),Abstract,_Body),Abstract).

get_machine_parameters(abstract_machine(_,_ModelType,machine_header(_,Name,Parameters),_),Name,Parameters).
get_machine_parameters(refinement_machine(_,machine_header(_,Name,Parameters),_,_),Name,Parameters).
get_machine_parameters(implementation_machine(_,machine_header(_,Name,Parameters),_,_),Name,Parameters).

get_model_type(Main,Machines,ModelType) :-
    get_constructed_machine_name_and_body(M,Main,_,_),
    memberchk(M,Machines),
    ( refines(M,RefName) ->
        get_model_type(RefName,Machines,ModelType)
    ; otherwise ->
        M = abstract_machine(_,ModelType,_,_)).

some_machine_name_body(Machines,M,Name,Body) :-
    member(M,Machines),
    get_constructed_machine_name_and_body(M,Name,_,Body).

use_usage(Machines,Pos,Use) :-
    some_machine_name_body(Machines,_,_,Body),
    ( member(sees(Pos,R),Body)
    ; member(uses(Pos,R),Body)),
    member(identifier(_,PrefixUse),R),
    split_prefix(PrefixUse,_,Use).
% include_usage/3 checks if there is any machine in Machines that
% includes/extends/imports the used machine
% include_usage(+Machines,+Use,-Name):
%  Machines: The list of machines
%  Use: The name of the used machine
%  Name: The name of the machine that imports the used machine
% include_usage/3 fails if there is not such an import
include_usage(Machines,Use,Name) :-
    some_machine_name_body(Machines,_,Name,Body),
    ( member(includes(_,R),Body)
    ; member(extends(_,R), Body)
    ; member(imports(_,R), Body)),
    member(machine_reference(_Pos,PrefixRef,_),R),
    % The name could be prefixed, we need it without prefix:
    split_prefix(PrefixRef,_,Use).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% uses and sees clauses

% returns a list of references to used and seen machines
% the machines contain only identifier sections and
% the identifier are prefixed accordingly
% all identifier are marked as coming from a seen/used machine
use_and_see_machines(Sections,Machines,References) :-
    get_uses_and_sees(Sections,Uses,Sees),
    maplist(use_or_see_machine(used,Machines),Uses,Used),
    maplist(use_or_see_machine(seen,Machines),Sees,Seen),
    append(Used,Seen,References).

% get uses and sees sections from the untyped machines
get_uses_and_sees(Sections,Uses,Sees) :-
    get_uses_or_sees2(sees,Sections,Sees),
    get_uses_or_sees2(uses,Sections,Uses).
get_uses_or_sees2(Mode,Sections,US) :-
    optional_rawmachine_section(Mode,Sections,[],US1),
    findall(I,member(identifier(_,I),US1),US).

use_or_see_machine(Mode,TypedMachines,Ref,ref(Mode,Result)) :-
    split_prefix(Ref,Prefix,Name),
    memberchk(machine(Name,Typed),TypedMachines),
    create_machine(Name,Empty),
    use_sections([sets],Mode,'',Name,Typed,Empty,M1),
    use_sections([constants,variables,promoted],Mode,Prefix,Name,Typed,M1,Result).
use_sections(Sections,Mode,Prefix,MName,Typed,Old,New) :-
    expand_shortcuts(Sections,AllSections),
    foldl(use_section(Mode,Prefix,MName,Typed),AllSections,Old,New).
use_section(Mode,Prefix,MName,Machine,Section,OldM,NewM) :-
    get_section_texprs(Section,Machine,Identifiers),
    write_section(Section,NewIds,OldM,NewM),
    ( Prefix='' ->
        Ids1=Identifiers
    ; otherwise ->
        prefix_identifiers(Identifiers,Prefix,Renamings),
        rename_bt_l(Identifiers,Renamings,Ids1)),
    maplist(add_use_info_to_identifier(Mode,MName),Ids1,NewIds).
add_use_info_to_identifier(Mode,Name,TExpr,New) :-
    get_texpr_id(TExpr,PId), get_texpr_type(TExpr,Type),
    get_texpr_id(New,PId), get_texpr_type(New,Type),
    split_prefix(PId,_Prefix,Id),
    get_texpr_info(TExpr,Info),
    get_texpr_info(New,[usesee(Name,Id,Mode)|Info]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% add a section to the machine that describes the linking of
% parameters and arguments
add_link_constraints(Includes,MType,Parameters,RefMachines,Old,New,Ein,Eout) :-
    AllRefMachines = [ref(local,Old)|RefMachines],
    extract_parameter_types(RefMachines,NonGroundExceptions),
    foldl(add_link_section(MType,Parameters,AllRefMachines,NonGroundExceptions),
          Includes,Links/Ein,[]/Eout), % TO DO: Daniel check if [] is correct here
    %print(conjunct_predicates(Links,Link)),nl,
    conjunct_predicates(Links,Link),
    select_section(constraints,OConstraints,NConstraints,Old,New),
    conjunct_predicates([Link,OConstraints],NConstraints).
add_link_section(MType,Parameters,RefMachines,NonGroundExceptions,
                 machine_reference(Pos,Ref,Args),Links/Ein,RLinks/Eout) :-
    visible_env(MType,includes,RefMachines,Env,Ein,E1),
    memberchk(parameters(Ref,TParameters),Parameters),
    ( same_length(TParameters, Args) ->
        get_texpr_types(TParameters,Types),
        btype_ground_dl(Args,Env,NonGroundExceptions,Types,TArgs,E1,Eout),
        maplist(create_equality,TParameters,TArgs,LLinks)
    ; otherwise ->
        E1 = [error('wrong number of machine arguments',Pos)|Eout],
        LLinks = []),
    append(LLinks,RLinks,Links).
create_equality(P,A,E) :-
    create_texpr(equal(P,A),pred,[parameterlink],E).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type machines

type_machine(Header,Name,MType,RawMachine,RefMachines,TypedMachine,NewRefMachines,Ein,Eout) :-
    Header = machine_header(_,Name,_),
    create_machine(Name,Empty),
    % (optional) definitions
    extract_definitions(RawMachine,Empty,DefsOnly,Ein,E0),
    debug_stats(extracted_definitions(Name)),
    % create the identifiers that will be typed,
    % Local will contain the identifier sections
    ~~mnf( create_id_sections(Header,RawMachine,Name,DefsOnly,Local)),
    ~~mnf( create_freetypes(RawMachine,MType,RefMachines,Local,Local1,E0,E1)),
    % in case of a refinement, check if all newly defined operations are refinements
    ~~mnf( link_to_refinement(MType,RawMachine,RefMachines,Local1,Local2,E1,E2)),
    % extract types that can be variables
    ~~mnf(extract_parameter_types([ref(local,Local2)|RefMachines],NonGroundExceptions)),
    % check for a VALUES clause. They are a little bit tricky because they can replace
    % already defined deferred sets by integers or other deferred sets
    process_values_section(MType,RawMachine,NonGroundExceptions,
                           Local2/E2/RefMachines,Local3/E3/NewRefMachines),
    % type-check the other sections (properties, invariant, operation_bodies, etc)
    ~~mnf( type_sections(RawMachine,MType,[ref(local,Local3)|NewRefMachines],NonGroundExceptions,
                         Name,E3,Eout,Local3,TypedMachine)).

extract_definitions(RawMachine,In,Out,Ein,Eout) :-
    optional_rawmachine_section(definitions,RawMachine,[],AllDefinitions),
    write_section(definitions,Definitions,In,Out),
    % remove all references to definition files from definitions
    exclude(is_file_definition,AllDefinitions,Definitions1),
    % replace expression_definition(...) by defintion(expression,...), etc.
    change_definition_style(Definitions1,Definitions2),
    % analyse dependencies
    add_def_dependency_information(Definitions2,Definitions3,Ein,Eout),
    % replace external function definitions by calls to the external function
    replace_external_declarations(Definitions3,Definitions).

is_file_definition(file_definition(_Pos,_Filename)).

change_definition_style(DefsIn,DefsOut) :-
    maplist(change_definition_style2,DefsIn,DefsOut).
change_definition_style2(conversion(Pos,InnerDef),definition_decl(Name,DefType,InnerPos,Args,conversion(Pos,RawExpr))) :-
    change_definition_style2(InnerDef,definition_decl(Name,DefType,InnerPos,Args,RawExpr)).
change_definition_style2(Def,definition_decl(Name,DefType,Pos,Args,RawExpr)) :-
    Def =.. [Functor,Pos,Name,Args,RawExpr],
    maplist(check_def_argument(Name,Pos),Args),
    map_def_functor(Functor,DefType).
map_def_functor(expression_definition,expression).
map_def_functor(substitution_definition,substitution).
map_def_functor(predicate_definition,predicate).

% check formal arguments of definitions , e.g., here xx in square2 is not an identifier: xx == 20; square2(xx) == xx*xx
check_def_argument(_,_,identifier(_,_)) :- !.
check_def_argument(DefName,DefPos,definition(_,ID,_)) :- !,
   tools:ajoin(['Formal parameter ', ID, ' is a definition call in Definition: '],Msg),
   add_error(bmachine_construction,Msg,DefName,DefPos).
check_def_argument(DefName,DefPos,FP) :- !,
   tools:ajoin(['Formal parameter ', FP, ' is not an identifier in Definition: '],Msg),
   add_error(bmachine_construction,Msg,DefName,DefPos).

replace_external_declarations(Definitions,NewDefs) :-
    split_list(is_external_declaration,Definitions,ExtFunctionDecls,RestDefs),
    foldl(replace_external_declaration,ExtFunctionDecls,RestDefs,NewDefs).
is_external_declaration(definition_decl(DefName,expression,_Pos,_Params,_Def,_Dependencies)) :-
    external_name(DefName,ExtType,_ExpectedDefType,_ExtCall,FunName),
    (debug_mode(on) -> format('external ~w declared: ~w~n', [ExtType,FunName]) ; true).
external_name(DefName,function,expression,external_function_call,FunName) :-
    atom_concat('EXTERNAL_FUNCTION_',FunName,DefName).
external_name(DefName,predicate,predicate,external_pred_call,FunName) :-
    atom_concat('EXTERNAL_PREDICATE_',FunName,DefName).
external_name(DefName,substitution,substitution,external_subst_call,FunName) :-
    atom_concat('EXTERNAL_SUBSTITUTION_',FunName,DefName).
replace_external_declaration(definition_decl(DefName,expression,DefPos,TypeParams,Decl,_Deps),In,Out) :-
    OldDefinition = definition_decl(FunName,ExpectedDefType,Pos,FunParams,FunDef,Deps),
    NewDefinition = definition_decl(FunName,ExpectedDefType,Pos,FunParams,ExtCall,Deps),
    ( external_name(DefName,_ExtType,ExpectedDefType,ExtCallFunctor,FunName),
        ExtCall =.. [ExtCallFunctor,Pos,FunName,FunParams,FunDef,
                     rewrite_protected(TypeParams),rewrite_protected(Decl)],
        selectchk(OldDefinition,In,NewDefinition,Out) ->
        assert_external_procedure_used(FunName)
    ; otherwise ->
        % no definition found for external function
        add_error(replace_external_declaration,'No DEFINITION associated with: ',DefName,DefPos),
        Out=In).

reset_bmachine_construction :- reset_external_procedure_used.

% maybe this information should be stored somewhere else ??
:- dynamic external_procedure_used/1.
reset_external_procedure_used :- retractall(external_procedure_used(_)).
assert_external_procedure_used(FunName) :-
  (external_procedure_used(FunName) -> true ; assert(external_procedure_used(FunName))).

link_to_refinement(machine,_RawMachine,_RefMachines,Local,Local,Errors,Errors).
link_to_refinement(refinement,RawMachine,RefMachines,Local,NewLocal,Ein,Eout) :-
    link_to_refinement(implementation,RawMachine,RefMachines,Local,NewLocal,Ein,Eout).
link_to_refinement(implementation,RawMachine,RefMachines,Local,NewLocal,Ein,Eout) :-
    link_to_refinement2(implementation,RawMachine,RefMachines,Local,NewLocal,Ein,Eout).
link_to_refinement2(_MType,RawMachine,RefMachines,Local,NewLocal,Ein,Eout) :-
    memberchk(ref(abstraction,AbstractMachine),RefMachines),
    copy_constraints(Local,AbstractMachine,NewLocal,Ein,E1),
    type_vars_in_refinement(AbstractMachine,NewLocal),
    type_refinement_operations(RawMachine,AbstractMachine,NewLocal,E1,Eout).

copy_constraints(Local,AbstractMachine,NewLocal,Ein,Eout) :-
    get_section(parameters,Local,LocalParameters),
    get_section(parameters,AbstractMachine,AbstractParameters),
    check_if_equal_identifiers(LocalParameters,AbstractParameters,Ein,Eout,Local),
    get_section(constraints,AbstractMachine,Constraints),
    write_section(constraints,Constraints,Local,NewLocal).
check_if_equal_identifiers(Local,Abstract,Ein,Eout,LocalMachine) :-
    ( same_length(Local,Abstract) ->
        foldl(check_if_equal_identifiers2,Local,Abstract,Ein,Eout)
    ; otherwise ->
        get_texpr_ids(Abstract,AIDs),
        get_machine_name(LocalMachine,MachName),
        ajoin(['Refinement ',MachName,' must have same Parameters ', AIDs,' as abstract Machine'],Msg),
        Ein = [error(Msg,none)|Eout]
    ).
check_if_equal_identifiers2(LParam,AParam,Ein,Eout) :-
    get_texpr_id(LParam,LName),
    get_texpr_id(AParam,AName),
    ( LName = AName ->
        % type in refinement is the same as in the abstraction
        get_texpr_type(LParam,Type),
        get_texpr_type(AParam,Type),
        Ein=Eout
    ; otherwise ->
        get_texpr_pos(LParam,Pos),
        ajoin(['Parameter ',LName,' of refinement machine must be ',
               AName,' like in the abstract machine'],Msg),
        Ein = [error(Msg,Pos)|Eout]
    ).

% in case of a refinement, give variables the same type as in the abstract machine
% the same for constants
type_vars_in_refinement(AbstractMachine,ConcreteMachine) :-
    pass_type(AbstractMachine,[abstract_variables,concrete_variables],
              ConcreteMachine,[abstract_variables,concrete_variables]),
    pass_type(AbstractMachine,[abstract_constants,concrete_constants],
              ConcreteMachine,[abstract_constants,concrete_constants]).

% pass the type from variables in Sections1 of Machine1 to
% the variables of the same name in Sections2 of Machine2
% Machine1,Machine2: a machine
% Sections1, Sections2: a list of section names
pass_type(Machine1,Sections1,Machine2,Sections2) :-
    get_sections_and_append(Sections1,Machine1,Vars1),
    get_sections_and_append(Sections2,Machine2,Vars2),
    maplist(pass_type2(Vars2),Vars1).
get_sections_and_append([],_M,[]).
get_sections_and_append([Sec|RestSections],M,R) :-
    get_section(Sec,M,L), append(L,Rest,R),
    get_sections_and_append(RestSections,M,Rest).
pass_type2(DstVariables,SrcVariable) :-
    get_texpr_id(SrcVariable,Id),
    get_texpr_id(DstVariable,Id),
    ( memberchk(DstVariable,DstVariables) ->
        get_texpr_type(DstVariable,Type),
        get_texpr_type(SrcVariable,Type)
    ; otherwise ->
        true).


% in case of a refinement, check if the defined operations are already defined
% in the abstract machine and copy that type.
type_refinement_operations(RawMachine,AbstractMachine,Local,Ein,Eout) :-
    get_operation_identifiers(RawMachine,Operations),
    ~~mnf(type_refinement_operations2(Operations,Local,AbstractMachine,Ein,Eout)).
type_refinement_operations2([],_,_AbstractMachine,Errors,Errors).
type_refinement_operations2([Op|Rest],M,AbstractMachine,Ein,Eout) :-
    copy_texpr_wo_info(Op,AOp),
    get_texpr_pos(Op,Pos),
    % look up the abstract definition
    get_abstract_op(AOp,AbstractMachine,Pos,Ein,E1),
    % store the type in the identifier section
    copy_texpr_wo_info(AOp,SOp),
    get_section(promoted,M,Operations),memberchk(SOp, Operations),
    % do the rest
    type_refinement_operations2(Rest,M,AbstractMachine,E1,Eout).
% looks up the type of the operator in an abstract machine
get_abstract_op(Op,Abstraction,_,Errors,Errors) :-
    % look for the operation in promoted and unpromoted
    ( get_section(promoted,Abstraction,AbstractOps), member(Op,AbstractOps)
    ; get_section(unpromoted,Abstraction,AbstractOps), member(Op,AbstractOps) ),
    % forget alternatives
    !.
get_abstract_op(Op,_Abstraction,_Pos,Errors,Errors) :-
    % we might allow new operations
    get_preference(allow_new_ops_in_refinement,true),!,
    get_texpr_type(Op,op(_,_)).
get_abstract_op(Op,_,Pos,[error(Msg,Pos)|Eout],Eout) :-
    % in case we do not find the operation, store an error
    get_texpr_id(Op,op(Id)),
    ajoin(['operation ', Id, ' is not defined in the abstract machine'], Msg).
% copy a typed expression without the additional information (just expression and type)
copy_texpr_wo_info(A,B) :-
    % copy the expression and type, the additional information may be different
    get_texpr_expr(A,Expr),get_texpr_expr(B,Expr),
    get_texpr_type(A,Type),get_texpr_type(B,Type).

create_id_sections(Header,RawMachine,Name) -->
    create_id_sections_header(Header),
    %{print(created_header(Name)),nl},
    create_set_sections(RawMachine,Name),
    %{print(created_set(Name)),nl},
    create_constants_sections(RawMachine),
    %{print(created_constants(Name)),nl},
    create_variables_sections(RawMachine),
    %{print(created_variables(Name)),nl},
    create_operations_sections(RawMachine).

extract_parameter_types(MachineRefs,ParameterTypes) :-
    maplist(extract_parameter_types2,MachineRefs,ParameterTypesL),
    append(ParameterTypesL,ParameterTypes).
extract_parameter_types2(ref(_,Machine),ParameterTypes) :-
    get_section(parameters,Machine,VisibleParams),
    get_section(internal_parameters,Machine,InternalParams),
    append(VisibleParams,InternalParams,Params),
    include(is_a_parameter_set,Params,ParameterSets),
    maplist(get_texpr_set_type,ParameterSets,ParameterTypes).
is_a_parameter_set(TExpr) :-
    get_texpr_info(TExpr,Info),
    memberchk(parameter_set,Info).

type_sections(RawMachine,MachineType,RefMachines,NonGroundExceptions,Name,Ein,Eout) -->
    {debug_stats('TYPING CONSTRAINTS'(Name))},
    type_constraints(MachineType,Name,RawMachine,RefMachines,NonGroundExceptions,Ein,E1),
    % Maybe the VALUES section must be moved up later because it may be used to
    % substitute types (e.g. deferred sets to integers) for later use
    {debug_stats('TYPING PROPERTIES'(Name))},
    type_section_with_single_predicate(properties,Name,[constants],MachineType,RawMachine,RefMachines,NonGroundExceptions,E1,E2),
    {debug_stats('TYPING INVARIANT'(Name))},
    type_section_with_single_predicate(invariant,Name,[variables],MachineType,RawMachine,RefMachines,NonGroundExceptions,E2,E3),
    {debug_stats('TYPING ASSERTIONS'(Name))},
    type_section_with_predicate_list(assertions,[],MachineType,RawMachine,RefMachines,NonGroundExceptions,E3,E4),
    {debug_stats('TYPING INITIALISATION'(Name))},
    type_initialisation_section(RawMachine,Name,MachineType,RefMachines,NonGroundExceptions,E4,E5),
    {debug_stats('TYPING OPERATIONS'(Name))},
    type_operations_section(RawMachine,MachineType,RefMachines,NonGroundExceptions,E5,Eout),
    {debug_stats('FINISHED TYPING SECTIONS'(Name))}.

% skip this section, it is copied from the abstract machine and
% does not need to be typed again
type_constraints(refinement,_,_RawMachine,_RefMachines,_NonGroundExceptions,Errors,Errors) --> !.
type_constraints(implementation,_,_RawMachine,_RefMachines,_NonGroundExceptions,Errors,Errors) --> !.
type_constraints(machine,Name,RawMachine,RefMachines,NonGroundExceptions,Ein,Eout) -->
    % if the machine is a dummy machine (in presence of a not included seen or used
    % machine), we must omit the check that the (lower case) parameters are all typed.
    % We can assume that they are properly typed by the included machine.
    {( is_dummy_machine(RawMachine) -> IdsToType = []
     ; otherwise -> IdsToType = [parameters])},
    type_section_with_single_predicate(constraints,Name,IdsToType,machine,
                                       RawMachine,RefMachines,NonGroundExceptions,Ein,Eout).

is_dummy_machine(RawMachine) :- member(is_dummy,RawMachine),!.

% Header: Parameters
create_id_sections_header(machine_header(_,_,Parameters),Old,New) :-
    write_section(parameters,TParams,Old,New),
    maplist(create_id_section_parameter,Parameters,TParams).
create_id_section_parameter(Param,TParam) :-
    Expr=identifier(Name),
    ext2int(Param,Expr,_,Type,Expr,[ParameterType],TParam),
    ( is_upper_case(Name) ->
        ParameterType = parameter_set,
        Type = set(_)
    ; otherwise ->
        ParameterType = parameter_scalar).
is_upper_case(Name) :- atom_codes(Name,[C|_]),
    memberchk(C,"ABCDEFGHIJKLMNOPQRSTUVWXYZ").

% Body: Sets
create_set_sections(Sections,MachineName) -->
    write_section(deferred_sets,DeferredSets),
    write_section(enumerated_sets,EnumeratedSets),
    write_section(enumerated_elements,EnumeratedElements),
    {optional_rawmachine_section(sets,Sections,[],Sets),
     split_list(is_deferred_set_element,Sets,RawDeferredSets,RawEnumeratedSets),
     maplist(create_deferred_set_section(MachineName),RawDeferredSets,DeferredSets),
     maplist(create_enumerated_set_section(MachineName),
             RawEnumeratedSets,EnumeratedSets,LEnumeratedElements),
     append(LEnumeratedElements,EnumeratedElements)}.
is_deferred_set_element(AstElem) :- has_functor(AstElem,deferred_set,_).
create_deferred_set_section(MachineName,DS,TExpr) :-
    unwrap_opt_description(DS,deferred_set(Pos,I),TInfos),
    Infos = [given_set,def(deferred_set,MachineName)|TInfos],
    create_identifier(I,Pos,set(global(I)),Infos,TExpr).
create_enumerated_set_section(MachineName,ES,TExpr,Elements) :-
    unwrap_opt_description(ES,enumerated_set(Pos,I,Elems),TInfos),
    Infos = [given_set,def(enumerated_set,MachineName)|TInfos],
    create_identifier(I,Pos,set(global(I)),Infos,TExpr),
    maplist(create_element(global(I),MachineName),Elems,Elements).

% deal with optional description(Pos,Desc,A) wrapper
has_functor(description(_,_Desc,A),Functor,Arity) :- !, functor(A,Functor,Arity).
has_functor(A,Functor,Arity) :- functor(A,Functor,Arity).

% remove description wrapper and generate info field
unwrap_opt_description(description(_,Desc,A),Res,Infos) :- !, Res=A, Infos=[description(Desc)].
unwrap_opt_description(A,A,[]).

create_element(Type,MachineName,Ext,Elem) :-
    ext2int(Ext,Expr,_,Type,Expr,[enumerated_set_element,def(enumerated_element,MachineName)],Elem).
create_identifier(Id,Pos,Type,Infos,TExpr) :-
    create_texpr(identifier(Id),Type,[nodeid(Pos)|Infos],TExpr).

% Constants
create_constants_sections(RawMachine) -->
    create_section_identifiers(constants,concrete_constants,RawMachine),
    create_section_identifiers(abstract_constants,abstract_constants,RawMachine).
% Variables
create_variables_sections(RawMachine) -->
    create_section_identifiers(concrete_variables,concrete_variables,RawMachine),
    create_section_identifiers(variables,abstract_variables,RawMachine).

% Freetypes: Treat them as additional constants, plus add entries in the "freetypes"
% section of the resulting machine
create_freetypes(RawMachine,MType,RefMachines,Old,New,Ein,Eout) :-
    optional_rawmachine_section(freetypes,RawMachine,[],RawFreetypes),
    create_freetypes2(RawFreetypes,MType,[ref(local,Old)|RefMachines],Old,New,Ein,Eout).
create_freetypes2([],_MType,_RefMachines,M,M,E,E) :- !.
create_freetypes2(RawFreetypes,MType,RefMachines,Old,New,Ein,Eout) :-
    % we need the NonGroundExceptions for type checking
    extract_parameter_types(RefMachines,NonGroundExceptions),
    % create identifiers in the constants section
    maplist(create_ids_for_freetype,RawFreetypes,IdsFreetypes),
    % we just combined the results to keep the numbers of arguments low (too much for maplist)
    maplist(split_ft,IdsFreetypes,IdentifiersL,Freetypes),
    append(IdentifiersL,Identifiers),
    % create properties for each freetype
    foldl(create_properties_for_freetype(MType,RefMachines,NonGroundExceptions),
          RawFreetypes,IdsFreetypes,PropertiesL,Ein,Eout),
    conjunct_predicates(PropertiesL,Properties),
    append_to_section(abstract_constants,Identifiers,Old,Inter),
    conjunct_to_section(properties,Properties,Inter,Inter2),
    write_section(freetypes,Freetypes,Inter2,New).
split_ft(ft(Ids,Freetype),Ids,Freetype).
create_ids_for_freetype(freetype(_Pos,Id,Constructors),ft([TId|TCons],freetype(Id,Cases))) :-
    create_typed_id(Id,set(freetype(Id)),TId),
    maplist(create_ids_for_constructor(Id),Constructors,TCons,Cases).
create_ids_for_constructor(Id,constructor(_Pos,Name,_Arg),TCons,case(Name,Type)) :-
    create_typed_id(Name,set(couple(Type,freetype(Id))),TCons).
create_ids_for_constructor(Id,element(_Pos,Name),TCons,case(Name,constant([Name]))) :-
    create_typed_id(Name,freetype(Id),TCons).
create_properties_for_freetype(MType,RefMachines,NonGroundExceptions,
                               freetype(_Pos,Id,Constructors),ft(Ids,_Freetypes),Properties,Ein,Eout) :-
    % The freetype type
    FType = freetype(Id),
    % We use the standard type environment of properties...
    visible_env(MType,properties,RefMachines,CEnv,Ein,E1),
    % ...plus the identifiers of the free type (type name and constructor names)
    add_identifiers_to_environment(Ids,CEnv,FEnv),
    % We generate a comprehension set for all elements
    create_typed_id(Id,set(FType),TId),
    create_texpr(equal(TId,TComp),pred,[],FDef),
    unique_typed_id("_freetype_arg",FType,Element),
    create_recursive_compset([Element],ECond,set(FType),[],RecId,TComp),
    create_typed_id(RecId,set(FType),TRecId),
    % For each constructor, we generate a definition and a condition for the
    % comprehension set above
    foldl(create_properties_for_constructor(Id,FEnv,Element,TRecId,NonGroundExceptions),
          Constructors,Defs,Conds,E1,Eout),
    conjunct_predicates(Conds,ECond),
    conjunct_predicates([FDef|Defs],Properties).

/* create_properties_for_constructor(+Env,+Element,+RecId,+NGE,+Constructor,-Def,-Cond,Ein,Eout)
    Env:     Type environment
    Element: A typed identifier "e" that is used in the definition of the freetype set:
             ft = {e | e has_freetype_constructor x => e:...}
    RecId:   The typed identifier that can be used to refer to the freetype set (ft in the
             example above
    NGE:     "Non ground exceptions", needed for type checking when having a parametrized machine
    Constructor: The constructor expression as it comes from the parser
             (constructor(Pos,Name,ArgSet) or element(Pos,Name))
    Def:     The predicate that defines the constant for the constructor,
             e.g. "cons = {i,o | i:NAT & o = freetype(cons,i)}"
    Cond:    The predicate that checks the argument of a freetype in the freetype set
             (That would be the part "e:..." in the example above.
    Ein,Eout: Used for type checker errors
*/
create_properties_for_constructor(FID,Env,Element,RecId,NonGroundExceptions,
                                  Constructor,Def,Cond,Ein,Eout) :-
    constructor_name(Constructor,Name),
    env_lookup_type(Name,Env,CType),
    create_properties_for_constructor2(Constructor,Env,NonGroundExceptions,FID,
                                       Element,RecId,CDef,Cond,Ein,Eout),
    get_texpr_type(CDef,CType),
    create_typed_id(Name,CType,CId),
    create_texpr(equal(CId,CDef),pred,[],Def).
constructor_name(element(_Pos,Name),Name).
constructor_name(constructor(_Pos,Name,_Domain),Name).
create_properties_for_constructor2(element(_Pos,Name),_Env,_NonGroundExceptions,FID,
                                  _Element,_RecId,CDef,Cond,Ein,Ein) :-
    create_texpr(value(freeval(FID,Name,term(Name))),freetype(FID),[],CDef),
    create_texpr(truth,pred,[],Cond).
create_properties_for_constructor2(constructor(_Pos,Name,RArg),Env,NonGroundExceptions,
                                   FID,Element,RecId,CDef,Cond,Ein,Eout) :-
    % First, type check the given set of the domain:
    btype_ground_dl([RArg],Env,NonGroundExceptions,[set(DType)],[TDomain],Ein,Eout),
    % then create the RHS of "c = {i,o | i:Domain & o=freetype_constructor(Name,i)}"
    create_definition_for_constructor(Name,TDomain,FID,CDef),
    % The check in the freetype comprehension set is of the form
    %   e "of_freetype_case" Name => "content_of"(e) : Domain
    create_texpr(implication(IsCase,DomainTest),pred,[],Cond),
    create_texpr(freetype_case(FID,Name,Element),pred,[],IsCase),
    create_texpr(freetype_destructor(FID,Name,Element),DType,[],Content),
    % all references to the freetype itself are replaced by the recursive reference
    replace_id_by_expr(TDomain,FID,RecId,TDomain2),
    create_texpr(member(Content,TDomain2),pred,[],DomainTest).

:- use_module(bsyntaxtree,[get_texpr_set_type/2]).

/* The constructor is a function to the freetype, defined with a comprehension set:
   The general form is "c = {i,o | i:Domain & o=freetype_constructor(Name,i)}"
   create_definition_for_constructor(+Name,+DType,+FID,-CType,-CDef) :-
   Name:    Name of the constructor
   TDomain: The user-specified domain
   FID:     The name of the free type
   CDef:    The RHS of the definition "Name = CDef"
*/
create_definition_for_constructor(Name,TDomain,FID,CDef) :-
    % get the type of the domain:
    get_texpr_set_type(TDomain,DType),
    % create argument and result identifiers:
    unique_typed_id("_constr_arg",DType,TArgId),          % was constructor_arg and constructor_res
    unique_typed_id("_constr_res",freetype(FID),TResId),
    % The comprehension set as a whole
    CType = set(couple(DType,freetype(FID))),
    create_texpr(comprehension_set([TArgId,TResId],Pred),CType,
                 [prob_annotation('SYMBOLIC')],CDef),
    create_texpr(conjunct(DomainCheck,Construction),pred,[],Pred),
    % "i:Domain"
    create_texpr(member(TArgId,TDomain),pred,[],DomainCheck),
    % "o=freetype_constructor(i)
    create_texpr(freetype_constructor(FID,Name,TArgId),freetype(FID),[],FreetypeCons),
    create_texpr(equal(TResId,FreetypeCons),pred,[],Construction).

% Operations
create_operations_sections(RawMachine,Old,New) :-
    write_section(promoted,OperationIdentifiers,Old,New),
    get_operation_identifiers(RawMachine,OperationIdentifiers).
get_operation_identifiers(RawMachine,OperationIdentifiers) :-
    optional_rawmachine_section(operations,RawMachine,[],Ops),
    maplist(extract_operation_identifier,Ops,OperationIdentifiers).
extract_operation_identifier(Ext,TId) :-
    remove_pos(Ext, operation(ExtId,_,_,_)),
    ext2int(ExtId,identifier(I),_,op(_,_),identifier(op(I)),Infos,TId),
    operation_infos(Infos).

% VALUES section:
% process_values_section(MachineType,RawMachine,NonGroundExceptions,Min/Ein/RefMIn,Mout/Eout/RefMOut):
%   Type-check the VALUES section and change the type of valuated deferred sets, if necessary
%   MachineType, RawMachine, NonGroundExceptions: as usual, see other comments
%   Min/Mout: The currently constructed machine
%   Ein/Eout: The difference list of errors
%   RefMin/RefMout: The list of already typechecked machines. These typechecked machines can be
%     altered by this predicate because if a deferred set is valuated by an integer set or
%     other deferred set, all occurrences of the original type are replaced by the new type,
%     even for the already typed machines.
process_values_section(MachineType,RawMachine,NonGroundExceptions,Min/Ein/RefMIn,Mout/Eout/RefMOut) :-
    optional_rawmachine_section(values,RawMachine,[],RawValues),
    process_values_section_aux(RawValues,MachineType,NonGroundExceptions,
                               Min/Ein/RefMIn,Mout/Eout/RefMOut).
process_values_section_aux([],_MachineType,_NonGroundExceptions,In,Out) :- !,In=Out.
process_values_section_aux(RawValues,MachineType,NonGroundExceptions,
                           Min/Ein/RefMin,Mout/Eout/RefMout) :-
    debug_stats('TYPING VALUES'),
    type_values_section(MachineType,RawValues,RefMin,NonGroundExceptions,Min/Ein,Mout/Eout),
    % will be added as additional_property in bmachine
    RefMin=RefMout.

type_values_section(MachineType,RawValues,RefMachines,NonGroundExceptions,Min/Ein,Mout/Eout) :-
    write_section(values,Values,Min,Mout),
    visible_env(MachineType,values_expression,RefMachines,Env,Ein,E1),
    % We have to pass an environment that can be modified because in each
    % valuation the previously valuated constants can be used.
    foldl(extract_values_entry(NonGroundExceptions),RawValues,Values,Env/E1,_ResultingEnv/Eout).

extract_values_entry(NonGroundExceptions, values_entry(POS,ValueID,ValueExpr), Entry,
                     EnvIn/Ein,EnvOut/Eout) :-
    % TODO: There seem to be a lot of additional constraints for valuations that are not
    %       yet handled here
    btype_ground_dl([ValueExpr],EnvIn,NonGroundExceptions,[Type],[TExpr],Ein,Eout),
    create_identifier(ValueID,POS,Type,[valuated_constant],TypedID),
    create_texpr(values_entry(TypedID,TExpr),values_entry,[nodeid(POS)],Entry),
    debug_println(9,valued_constant(ValueID)),
    add_identifiers_to_environment([TypedID],EnvIn,EnvOut).

% type_section_with_single_predicate(+Sec,+Name,+SectionsToType,+MachineType,+Sections,
%                                   +RefMachines,+NonGroundExceptions,+Ein,-Eout,+Old,-New):
% Type a section such as INVARIANT, PROPERTIES, CONSTRAINTS with a single predicate
%   Sec:         section name in the raw (untyped) machine (e.g. invariant)
%   Name:        name of machine from which this section comes
%   SectionsToType: list of section names that contain identifiers that must be typed
%                by this predicate (e.g. all variables must be typed by the invariant)
%   MachineType: machine type (machine, refinement, ...)
%   Sections:    list of sections representing the raw (untyped) machine
%   RefMachines: list of already typed machines
%   NonGroundExceptions: list of types that may be not ground because the concrete type
%                is determinded by machine parameter
%   Ein/Eout:    difference list of errors
%   Old/New:     the new typed section is added (by conjunction) to the machine
type_section_with_single_predicate(Sec,Name,SectionsToType,MachineType,Sections,
                                   RefMachines,NonGroundExceptions,Ein,Eout,Old,New) :-
    ~~mnf(optional_rawmachine_section(Sec,Sections,truth(none),Predicate)),
    ( Predicate = truth(_) ->
        % even if there is no content, we must check if all identifiers are typed
        check_if_all_ids_are_typed(SectionsToType,RefMachines,NonGroundExceptions,Ein,Eout),
        Old=New
    ; otherwise ->
        toplevel_raw_predicate_sanity_check(Sec,Name,Predicate),
        type_predicates(Sec,SectionsToType,MachineType,[Predicate],RefMachines,NonGroundExceptions,
                        [Typed],Ein,Eout),
        conjunct_to_section(Sec,Typed,Old,New)
    ),
    !.
type_section_with_single_predicate(Sec,Name,SectsToType,MchType,_,_,_,_,_,_,_) :-
    add_internal_error('type_section_with_single_predicate failed',
                        type_section_with_single_predicate(Sec,Name,SectsToType,MchType,_,_,_,_,_,_,_)),
    fail.

% Type a section with multiple predicates, such as ASSERTIONS
type_section_with_predicate_list(Sec,SectionsToType,MachineType,Sections,
                                 RefMachines,NonGroundExceptions,Ein,Eout,Old,New) :-
    write_section(Sec,Typed,Old,New),
    ~~mnf(optional_rawmachine_section(Sec,Sections,[],Predicates)),
    type_predicates(Sec,SectionsToType,MachineType,Predicates,RefMachines,NonGroundExceptions,Typed,Ein,Eout),
    !.
type_section_with_predicate_list(Sec,SectsToType,MchType,Sects,RefMchs,NonGrndExc,Ein,Eout,Old,New) :-
    add_internal_error('type_section_with_predicate_list failed',
                        type_section_with_predicate_list(Sec,SectsToType,MchType,Sects,RefMchs,NonGrndExc,Ein,Eout,Old,New)),
     fail.


type_predicates(_Sec,SectionsToType,_MachineType,[],RefMachines,NonGroundExceptions,Typed,Ein,Eout) :-
    !,Typed=[],
    check_if_all_ids_are_typed(SectionsToType,RefMachines,NonGroundExceptions,Ein,Eout).
type_predicates(Sec,SectionsToType,MachineType,Predicates,RefMachines,NonGroundExceptions,Typed,Ein,Eout) :-
    debug:debug_println(9,typing(Sec,MachineType)),
    visible_env(MachineType, Sec, RefMachines, Env, Ein, E1),
    same_length(Predicates,Types),maplist(is_pred_type,Types),
    debug:debug_println(9,btype(Sec,MachineType)),
    btype_ground_dl_in_section(Sec,Predicates,Env,NonGroundExceptions,Types,Typed1,E1,E2),
    mark_with_section(Sec,Typed1,Typed),
    debug:debug_println(9,check_all_ids_typed),
    check_if_all_ids_are_typed(SectionsToType,RefMachines,NonGroundExceptions,E2,Eout),
    debug:debug_println(9,finished_check_all_ids_typed).

% check if the identifiers that should be typed by this section are completly typed
check_if_all_ids_are_typed([],_RefMachines,_NonGroundExceptions,Ein,Eout) :- !,Ein=Eout.
check_if_all_ids_are_typed(SectionsToType,RefMachines,NonGroundExceptions,Ein,Eout) :-
    memberchk(ref(local,Local),RefMachines),
    get_all_identifiers(SectionsToType,Local,IdentifiersToType),
    check_ground_types_dl(IdentifiersToType, NonGroundExceptions, Ein, Eout).


mark_with_section(Sec,In,Out) :-
    maplist(mark_with_section2(Sec),In,Out).
mark_with_section2(Sec,In,Out) :-
    remove_bt(In,conjunct(A,B),conjunct(NA,NB),Out),!,
    mark_with_section2(Sec,A,NA),  mark_with_section2(Sec,B,NB).
mark_with_section2(Sec,In,Out) :-
    add_texpr_infos(In,[section(Sec)],Out).

type_initialisation_section(Sections,Name,MType,RefMachines,NonGroundExceptions,Ein,Eout,Old,New) :-
    write_section(initialisation,Initialisation,Old,New),
    ( rawmachine_section(initialisation,Sections,Init) ->
        visible_env(MType, initialisation, RefMachines, InitEnv, Ein, E1),
        btype_ground_dl([Init],InitEnv,NonGroundExceptions,[subst],[TypedInit],E1,Eout),
        Initialisation=[init(Name,TypedInit)]
    ; otherwise ->
        Ein=Eout,
        Initialisation=[]).

:- use_module(library(ugraphs)).

type_operations_section(Sections,MType,RefMachines,NonGroundExceptions,Ein,Eout,Old,New) :-
    ~~mnf( write_section(operation_bodies,Operations,Old,New) ),
    ~~mnf( visible_env(MType, operation_bodies, RefMachines, Env, Ein, E1) ),
    ~~mnf( optional_rawmachine_section(operations,Sections,[],Ops) ),
    topological_sort(Ops,SOps),
    same_length(SOps,Types), maplist(is_op_type,Types),
    ~~mnf( btype_ground_dl(SOps,Env,NonGroundExceptions,Types,Operations,E1,Eout) ),!.

% perform a topological sort of the operations: treat called operation before calling operation
% only relevant when allow_local_operation_calls is set to true
topological_sort(Ops,SortedOps) :-
    get_preference(allow_local_operation_calls,true),
    findall(OtherID-ID, (member(Op,Ops),op_calls_op(Op,ID,OtherID)),Edges),
    % print(edges(Edges)),nl,
    % TO DO: maybe only store edges where OtherID also appears in Ops (i.e., call within same machine)
    Edges \= [],
    !,
    findall(ID,(member(operation(_,RawID,_,_,_),Ops),raw_id(RawID,ID)),Vertices), % print(v(Vertices)),nl,
    vertices_edges_to_ugraph(Vertices,Edges,Graph),
    (top_sort(Graph,Sorted)
     -> sort_ops(Sorted,Ops,SortedOps)
      ; add_error(topological_sort,'Cycle in the (local) operation call graph: ',Edges),
        SortedOps=Ops).
topological_sort(Ops,Ops).

sort_ops([],Ops,Ops). % Ops should be []
sort_ops([OpID|T],Ops,Res) :-
    raw_op_id(Op1,OpID),
    (select(Op1,Ops,TOps)
     -> Res = [Op1|TSOps], sort_ops(T,TOps,TSOps)
     ;   % operation from another machine
         % print(could_not_find(OpID,Ops)),nl,
         sort_ops(T,Ops,Res)
     ).

is_op_type(op(_,_)).
is_pred_type(pred).

% compute which other operations are directly called
op_calls_op(operation(_,RawID,_,_,RawBody),ID,OtherID) :- raw_id(RawID,ID),
   raw_body_calls_operation(RawBody,OtherID).

raw_body_calls_operation(RawBody,OtherID) :-
   raw_member(operation_call(_,RawOtherID,_,_),RawBody), raw_id(RawOtherID,OtherID).

raw_op_id(operation(_,RawID,_,_,_RawBody),ID) :- raw_id(RawID,ID).
raw_id(identifier(_,ID),ID).

% a utility function to work on the raw AST  Functor(POS,Arg1,...,Argn)
% this will not be able to look inside DEFINITIONS !
% TO DO: deal with more raw substitutions which have list arguments
raw_member(X,X).
raw_member(X,parallel(_,List)) :- !, member(Term,List), raw_member(X,Term).
raw_member(X,sequence(_,List)) :- !, member(Term,List), raw_member(X,Term).
raw_member(X,[H|T]) :- !, (raw_member(X,H) ; raw_member(X,T)).
raw_member(X,Term) :- compound(Term), Term =.. [_Functor,_Pos|Args],
   member(A,Args), raw_member(X,A).


create_section_identifiers(Section,DestSection,RawMachine,Old,New) :-
    ~~mnf(write_section(DestSection,Vars,Old,New)),
    ~~mnf(optional_rawmachine_section(Section,RawMachine,[],Identifiers)),
    ~~mnf(create_section_ids2(Identifiers,[],Vars,DestSection)).

create_section_ids2([],_,[],_).
create_section_ids2([Ext|Rest],Infos,Res,DestSection) :-
    create_section_id(Ext,Infos,DestSection,TId),
    ( TId = error(Msg,Term,Pos) ->
        Res = TRest, add_error(bmachine_construction,Msg,Term,Pos)
    ; otherwise ->
        Res = [TId|TRest]),
    create_section_ids2(Rest,Infos,TRest,DestSection).
create_section_id(Ext,Infos,DestSection,TId) :-
    I=identifier(_),
    ( ext2int(Ext,I,_Pos,_Type,I,Infos,TId) ->
        true
    ; unwrap_pragma(Ext,ID,PragmaInfos) -> append(PragmaInfos,Infos,FullInfos),
        ext2int(ID,I,_,_,I,FullInfos,TId)
    ; Ext = definition(POSINFO,ID,_) ->
        TId = error('Trying to use DEFINITION name as identifier: ',
                    (ID,within(DestSection)), POSINFO)
    ; otherwise ->
        TId = error('Illegal identifier: ',
                    (Ext,within(DestSection)), Ext)
    ).

unwrap_pragma(description(_,D,ID),ID,[description(D)]).
unwrap_pragma(unit(_,U,ID),ID,[unit(U)]).
unwrap_pragma(new_unit(_,U,ID),ID,[new_unit(U)]).
unwrap_pragma(inferred_unit(_,_,ID),ID,[]). % currently ignored
unwrap_pragma(inferredunit(_,_,ID),ID,[]). % currently ignored

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sort the machines topologically

machine_order(Machines,Order) :-
    machine_dependencies(Machines,Dependencies),
    topsort(Dependencies,Order).

% sort the machines topologically
topsort(Deps,Sorted) :-
    topsort2(Deps,[],Sorted).
topsort2([],_,[]) :- !.
topsort2(Deps,Known,Sorted) :-
    split_list(all_deps_known(Known),Deps,DAvailable,DNotAvailable),
    DAvailable = [_|_], % we have new machines available whose dependencies are all known
    !,
    maplist(dep_name,DAvailable,Available),
    append(Available,Known,NewKnown),
    append(Available,Rest,Sorted),
    topsort2(DNotAvailable,NewKnown,Rest).
topsort2(Deps,_Known,_) :-
    member(dep(Name,NameDeps),Deps),
    add_error(bmachine_construction,'Could not resolve machine dependencies for: ',Name:depends_on(NameDeps)),
    fail.

all_deps_known(K,dep(_Name,Deps)) :- sort(Deps,DS),sort(K,KS),subseq0(KS,DS),!.
dep_name(dep(Name,_Deps),Name).

% find dependencies between machines
machine_dependencies(Machines,Dependencies) :-
    maplist(machine_dependencies2,Machines,Deps),
    sort(Deps,Dependencies).
machine_dependencies2(M,dep(Name,Deps)) :-
    get_constructed_machine_name_and_body(M,Name,_,Body),
    findall(Ref,
            (refines(M,Ref);machine_reference(Body,Ref)),
            Deps).

machine_reference(MachineBody,Ref) :-
    ( member(sees(_,R),MachineBody)
    ; member(uses(_,R),MachineBody) ),
    member(identifier(_,PrefixRef),R),
    split_prefix(PrefixRef,_,Ref).
machine_reference(MachineBody,Ref) :-
    ( member(includes(_,R),MachineBody)
    ; member(extends(_,R),MachineBody)
    ; member(imports(_,R),MachineBody) ),
    member(machine_reference(_,PrefixRef,_),R),
    split_prefix(PrefixRef,_,Ref).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% refinements
merge_refinement_and_abstraction(Concrete,RefMachines,Result,Ein,Eout) :-
    memberchk(ref(abstraction,Abstraction),RefMachines),
    append_sections([sets,concrete_constants,concrete_variables],Abstraction,Concrete,M1),
    append_if_new(abstract_constants,Abstraction,M1,M2),
    get_section(properties,Abstraction,AbstractProperties),
    conjunct_to_section(properties,AbstractProperties,M2,M3),
    select_section(invariant,Invariant,ConcreteInvariant,M3,M4),
    write_section(linking_invariant,LinkingInvariant,M4,Merged),
    filter_linking_invariant(Invariant,LinkingInvariant,ConcreteInvariant),
    propagate_abstract_operations(Abstraction,Merged,RefMachines,Result,Ein,Eout).
merge_refinement_and_abstraction(Machine,_,Machine,Errors,Errors).

append_sections(Sections,Machine,Old,New) :-
    expand_shortcuts(Sections,AllSections),
    append_sections2(AllSections,Machine,Old,New).
append_sections2([],_,M,M).
append_sections2([Section|Rest],Machine,Old,New) :-
    get_section(Section,Machine,Content),
    append_to_section(Section,Content,Old,Inter),
    append_sections2(Rest,Machine,Inter,New).

:- assert_must_succeed((create_machine(abs,EA), create_machine(conc,EB),
                        write_section(abstract_constants,[b(identifier(x),integer,[some_info])],EA,A),
                        write_section(abstract_constants,[b(identifier(x),integer,[other_info]),
                                                          b(identifier(y),integer,[info])],EB,B),
                        append_if_new(abstract_constants,A,B,ResultM),
                        get_section(abstract_constants,ResultM,ResultConst),
                        ResultConst==[b(identifier(x),integer,[other_info]),
                                      b(identifier(y),integer,[info])]
                        )).

append_if_new(Section,Machine,In,Out) :-
    get_section(Section,Machine,Content),
    select_section(Section,Old,New,In,Out),
    get_texpr_ids(Old,ExistingIds),
    exclude(is_in_existing_ids(ExistingIds),Content,NewElements),
    append(Old,NewElements,New).
is_in_existing_ids(ExistingIds,TId) :-
    get_texpr_id(TId,Id),
    memberchk(Id,ExistingIds).

filter_linking_invariant(Invariant,Linking,Concrete) :-
    split_conjuncts(Invariant,Predicates),
    split_list(contains_abstraction_refs,Predicates,Linkings,Concretes),
    conjunct_predicates(Linkings,Linking),
    conjunct_predicates(Concretes,Concrete).
contains_abstraction_refs(TExpr) :-
    syntaxtraversion(TExpr,_,_,Infos,Subs,_),
    ( memberchk(abstraction,Infos) ->
        true
    ; otherwise ->
        member(Sub,Subs),
        contains_abstraction_refs(Sub)).


split_conjuncts(Expr,List) :-
    split_conjuncts2(Expr,List,[]).
split_conjuncts2(Expr) -->
    {get_texpr_expr(Expr,conjunct(A,B)),!},
    split_conjuncts2(A),
    split_conjuncts2(B).
split_conjuncts2(Expr) --> [Expr].

% copy the abstract operations or re-use their preconditions
propagate_abstract_operations(Abstract,Concrete,RefMachines,Result,Ein,Eout) :-
    get_section(promoted,Abstract,APromoted),
    get_section(operation_bodies,Abstract,ABodies),
    select_section(promoted,CPromotedIn,CPromotedOut,Concrete,Concrete2),
    select_section(operation_bodies,CBodiesIn,CBodiesOut,Concrete2,Result),
    propagate_aops(APromoted,ABodies,RefMachines,CPromotedIn,CBodiesIn,CPromotedOut,CBodiesOut,Ein,Eout).
propagate_aops([],_ABodies,_RefMachines,CPromoted,CBodies,CPromoted,CBodies,Errors,Errors).
propagate_aops([APromoted|ApRest],ABodies,RefMachines,CPromotedIn,CBodiesIn,CPromotedOut,CBodiesOut,Ein,Eout) :-
    get_operation(APromoted,ABodies,AbstractOp),
    copy_texpr_wo_info(APromoted,CProm),
    ( member(CProm,CPromotedIn) ->
        extract_concrete_preconditions(AbstractOp,RefMachines,Pre),
        change_operation(APromoted,ConcreteOpOld,ConcreteOpNew,CBodiesIn,CBodiesRest),
        add_precondition(Pre,ConcreteOpOld,ConcreteOpNew),
        CPromotedIn = CPromotedRest,
        Ein = E1
    ; otherwise ->
        check_copied_operation(APromoted,AbstractOp,RefMachines,Ein,E1),
        append(CPromotedIn,[APromoted],CPromotedRest),
        append(CBodiesIn,[AbstractOp],CBodiesRest)),
    propagate_aops(ApRest,ABodies,RefMachines,CPromotedRest,CBodiesRest,CPromotedOut,CBodiesOut,E1,Eout).

get_operation(TId,Bodies,Operation) :-
    select_operation(TId,Bodies,Operation,_BodiesRest).
change_operation(TId,OldOp,NewOp,OldBodies,[NewOp|NewBodies]) :-
    select_operation(TId,OldBodies,OldOp,NewBodies).
select_operation(TId,Bodies,Operation,BodiesRest) :-
    copy_texpr_wo_info(TId,OpId),
    get_texpr_expr(Operation,operation(OpId,_,_,_)),
    select(Operation,Bodies,BodiesRest),!.

extract_concrete_preconditions(Op,RefMachines,FPre) :-
    extract_preconditions_op(Op,Pre),
    extract_op_arguments(Op,Args),
    conjunction_to_list(Pre,Pres),
    % todo: check the "machine" parameter
    visible_env(machine,operation_bodies,RefMachines,Env1,_Errors,[]),
    store_variables(Args,Env1,Env),
    filter_predicates_with_unknown_identifiers(Pres,Env,FPres),
    conjunct_predicates(FPres,FPre).

extract_op_arguments(Op,Params) :-
    get_texpr_expr(Op,operation(_,_,Params,_)).

extract_preconditions_op(OpExpr,Pre) :-
    get_texpr_expr(OpExpr,operation(_,_,_,Subst)),
    extract_preconditions(Subst,Pres,_),
    conjunct_predicates(Pres,Pre).
extract_preconditions(TExpr,Pres,Inner) :-
    get_texpr_expr(TExpr,Expr),
    extract_preconditions2(Expr,TExpr,Pres,Inner).
extract_preconditions2(precondition(Pre,Subst),_,[Pre|Pres],Inner) :- !,
    extract_preconditions(Subst,Pres,Inner).
extract_preconditions2(block(Subst),_,Pres,Inner) :- !,
    extract_preconditions(Subst,Pres,Inner).
extract_preconditions2(_,Inner,[],Inner).

filter_predicates_with_unknown_identifiers([],_Env,[]).
filter_predicates_with_unknown_identifiers([Pred|Prest],Env,Result) :-
    ( find_unknown_identifier(Pred,Env,_Id) ->
        !,Result = Rrest
    ; otherwise ->
        Result = [Pred|Rrest]),
    filter_predicates_with_unknown_identifiers(Prest,Env,Rrest).
find_unknown_identifier(Pred,Env,Id) :-
    get_texpr_id(Pred,Id),!,
    \+ env_lookup_type(Id,Env,_).
find_unknown_identifier(Pred,Env,Id) :-
    syntaxtraversion(Pred,_,_,_,Subs,Names),
    store_variables(Names,Env,Subenv),
    find_unknown_identifier_l(Subs,Subenv,Id).
find_unknown_identifier_l([S|_],Env,Id) :-
    find_unknown_identifier(S,Env,Id),!.
find_unknown_identifier_l([_|Rest],Env,Id) :-
    find_unknown_identifier_l(Rest,Env,Id).

add_precondition(Pre,Old,New) :-
    remove_bt(Old,operation(Id,Res,Params,Subst),operation(Id,Res,Params,NewSubst),New),
    extract_preconditions(Subst,OldPres,Inner),
    conjunct_predicates([Pre|OldPres],NewPre),
    create_texpr(precondition(NewPre,Inner),subst,[],NewSubst).


check_copied_operation(OpRef,Op,RefMachines,Ein,Eout) :-
    % todo: check the "refinement" parameter
    visible_env(refinement,operation_bodies,RefMachines,Env1,_Errors,[]),
    get_texpr_id(OpRef,OpId),get_texpr_type(OpRef,OpType),
    env_store(OpId,OpType,[],Env1,Env),
    findall(U, find_unknown_identifier(Op,Env,U), Unknown1),
    sort(Unknown1,Unknown),
    ( Unknown=[] -> Ein=Eout
    ; otherwise ->
        op(OpName) = OpId,
        join_ids(Unknown,IdList),
        (Unknown = [_] -> Plural=[]; Plural=['s']),
        append([['Operation ',OpName,
                 ' was copied from abstract machine but the identifier'],
                Plural,
                [' '],
                IdList,
                [' cannot be seen anymore']],Msgs),
        ajoin(Msgs,Msg), Ein = [error(Msg,none)|Eout]
    ).
join_ids([I],[Msg]) :- !,opname(I,Msg).
join_ids([A|Rest],[Msg,','|Mrest]) :- opname(A,Msg),join_ids(Rest,Mrest).
opname(op(Id),Id) :- !.
opname(Id,Id).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% split an identifier into a (possible empty) prefix and the name itself
% e.g. split_prefix('path.to.machine', 'path.to', 'machine').
split_prefix(Term,Prefix,Main) :-
    one_arg_term(Functor,Arg,Term),!,
    one_arg_term(Functor,MArg,Main),
    split_prefix(Arg,Prefix,MArg).
split_prefix(PR,Prefix,Main) :- %print(split_prefix(PR,Prefix,Main)),nl,trace,
    safe_atom_chars(PR,Chars,split_prefix1),
    split_prefix2(Chars,Chars,[],CPrefix,CMain),
    safe_atom_chars(Main,CMain,split_prefix2),
    safe_atom_chars(Prefix,CPrefix,split_prefix3).
split_prefix2([],Main,_,[],Main).
split_prefix2([C|Rest],Previous,PrefBefore,Prefix,Main) :-
    ( C='.' ->
        append(PrefBefore,RestPrefix,Prefix),
        split_prefix2(Rest,Rest,[C],RestPrefix,Main)
    ; otherwise ->
        append(PrefBefore,[C],NextPref),
        split_prefix2(Rest,Previous,NextPref,Prefix,Main)).

rawmachine_section(Elem,List,Result) :- %
    functor(Pattern,Elem,2),
    arg(2,Pattern,Result),
    member(Pattern,List),!.

optional_rawmachine_section(Elem,List,Default,Result) :-
    ( rawmachine_section(Elem,List,Result1) -> true
    ; Result1=Default),
    Result1=Result.

one_arg_term(Functor,Arg,Term) :- %print(one_arg_term(Functor,Arg,Term)),nl,
    functor(Term,Functor,1),arg(1,Term,Arg).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% visibility rules

% the visibility/5 predicate declares what a part of a machine can see
% visibility(MType, Scope, Section, Access, Info) means:
%   In a machine of type MType (machine, refinement, implementation),
%   an expression in a Section (like invariant) can see
%   the identifiers in the Access sections. Access is a list of
%   sections, where shortcuts are allowed (see shortcut/2).
%   Scope defines where (in relation to the Section)
%   the section in Access are (local: same machine, included: in an
%   included machine, etc.)
%   Info is a list of additional information that is added to each
%   identifier in the environment to build up. E.g. in an operation
%   definition, constants are marked readonly.
visibility(machine, local, constraints,      [parameters],[]).
visibility(machine, local, includes,         [parameters,sets,constants],[]).
visibility(machine, local, properties,       [sets,constants],[]).
visibility(machine, local, invariant,        [parameters,sets,constants,variables],[]).
visibility(machine, local, operation_bodies, [parameters,sets,constants],[readonly]).
visibility(machine, local, operation_bodies, [operations],[readonly,dontcall]).
visibility(machine, local, operation_bodies, [variables],[]).
visibility(machine, local, initialisation,   Access,[not_initialised|Info]) :- visibility(machine,local,operation_bodies,Access,Info).

visibility(machine, included, properties,       [sets,constants],[]).
visibility(machine, included, invariant,        [sets,constants,variables],[]).
visibility(machine, included, operation_bodies, [sets,constants,variables,promoted],[readonly]).
visibility(machine, included, initialisation,   [sets,constants,variables,promoted],[readonly]).

visibility(machine, used, properties,       [sets,constants],[]).
visibility(machine, used, invariant,        [parameters,sets,constants,variables],[]).
visibility(machine, used, operation_bodies, [parameters,sets,constants,variables],[readonly]).
visibility(machine, used, initialisation,   [parameters,sets,constants,variables],[readonly]).
visibility(machine, used, operation_bodies, [operations],[inquiry]). %% added by leuschel

visibility(machine, seen, includes,         [sets,constants],[]).
visibility(machine, seen, properties,       [sets,constants],[]).
visibility(machine, seen, invariant,        [sets,constants],[]).
visibility(machine, seen, operation_bodies, [sets,constants,variables],[readonly]).
visibility(machine, seen, initialisation,   [sets,constants,variables],[readonly]).
visibility(machine, seen, operation_bodies, [operations],[inquiry]).  %% added by leuschel

visibility(refinement, local, Section, Access, Info) :-
    Section \= assertions, % assertions are handled below
    visibility(machine, local, Section, Access, Info).

visibility(refinement, abstraction, includes,         [sets,concrete_constants],[]).
visibility(refinement, abstraction, properties,       [sets,constants],[]).
visibility(refinement, abstraction, invariant,        [sets,constants,concrete_variables],[]).
visibility(refinement, abstraction, invariant,        [abstract_variables],[abstraction]).
visibility(refinement, abstraction, operation_bodies, [sets,concrete_constants],[readonly]).
visibility(refinement, abstraction, operation_bodies, [concrete_variables],[]).

visibility(refinement, included, properties,       [sets,constants],[]).
visibility(refinement, included, invariant,        [sets,constants,variables],[]).
visibility(refinement, included, operation_bodies, [sets,constants,variables,promoted],[re]).

visibility(refinement, seen, includes,         [sets,constants],[]).
visibility(refinement, seen, properties,       [sets,constants],[]).
visibility(refinement, seen, invariant,        [sets,constants],[]).
visibility(refinement, seen, operation_bodies, [sets,constants,variables],[readonly]).
visibility(refinement, seen, operation_bodies, [operations],[inquiry]).

visibility(refinement, Ref, initialisation, Access, [not_initialised|Info]) :-
    visibility(refinement,Ref,operation_bodies,Access,Info).

% assertions have same visibility as invariant
visibility(MType, Ref, assertions, Part, Access) :-
    visibility(MType,Ref,invariant,Part,Access).

visibility(implementation, Ref, Section, Part, Access) :-
    visibility(refinement, Ref, Section, Part, Access).
visibility(implementation, included, values_expression, [concrete_constants,sets],[]).
visibility(implementation, seen,     values_expression, [concrete_constants,sets],[]).

% For predicates over pre- and poststates
visibility(MType, Rev,  prepost, Access, Info) :-
    visibility(MType, Rev, invariant, Access, Info).
visibility(MType, Rev,  prepost, Access, [primed,poststate|Info]) :-
    visibility(MType, Rev, invariant, Access, Info).

% add error messages for some common mistakes (access to parameters in the properties)
visibility(machine,    local,       properties,       [parameters],         [error('a parameter cannot be accessed in the PROPERTIES section')]).
% the following rule should be helpful for the user, but additionally it also
% provides a mechanism to re-introduce abstract variables in the INVARIANT
% section of a while loop (to enable this, an identifier is also marked with "abstraction")
visibility(refinement, abstraction, operation_bodies, [abstract_variables], [error('illegal access to an abstract variable in an operation'),abstraction]).


% lookups up all identifier sections that are accessible from
% the given section, removes all shortcuts and removes
% duplicate entries
%
% Returns a list of vis(Section,Info) where Section is the
% identifier section that can be seen with Info as a list
% of additional information
expanded_visibility(MType, Ref, Part, Access) :-
    findall(vis(Section,Infos),
            ( visibility(MType,Ref,Part,Sections1,Infos),
              expand_shortcuts(Sections1,Sections),
              member(Section,Sections)),
            Access1),
    sort(Access1,Access).

% visible_env/6 creates a type environment for a certain
% part of a machine by looking up which identifier should
% be visible from there and what additional information should
% be added (e.g. to restrict access to read-only)
%
% The visibility/6 facts are used to define which identifier are visible
%
% MType: machine type (machine, refinement, ...)
% Part: part of the machine for which the environment should be created
% RefMachines: referred machines that are already typed
% Env: The created environment
% Errors: errors might be added if multiple variables with the same identifier
%  are declared
visible_env(MType, Part, RefMachines, Env, Err_in, Err_out) :-
    env_empty(Init),
    visible_env(MType, Part, RefMachines, Init, Env, Err_in, Err_out).
% visible_env/7 is like visible_env/6, but an initial environment
% can be given in "In"
visible_env(MType, Part, RefMachines, In, Out, Err_in, Err_out) :-
    foldl(visible_env2(MType,Part),RefMachines,In/Err_in,Out/Err_out).
visible_env2(MType,Part,ref(Scope,Machine),InEnvErr,OutEnvErr) :-
    get_machine_name(Machine,MName),
    debug_println(9,visible_env2(MName,Scope)),
    ( Scope == local ->
        ~~mnf(get_section(definitions,Machine,Defs)),
        %nl,print(adding_defs(Defs)),nl,nl,
        foldl(env_add_def(MName),Defs,InEnvErr,InterEnvErr)
    ; otherwise ->
        InEnvErr=InterEnvErr),
    ~~mnf(expanded_visibility(MType, Scope, Part, Access)),
    foldl(create_vis_env(Scope,MName,Machine),Access,InterEnvErr,OutEnvErr).
env_add_def(MName,definition_decl(Name,PType,Pos,Params,Body,_Deps),InEnvErr,OutEnvErr) :-
    Type = definition(PType,Params,Body),
    Info = [nodeid(Pos),loc(local,MName,definitions)],
    create_texpr(identifier(Name),Type,Info,TExpr),
    add_unique_variable(TExpr,InEnvErr,OutEnvErr),!.
env_add_def(MName,Def,EnvErr,EnvErr) :-
    add_internal_error('Cannot deal with DEFINITION: ',env_add_def(MName,Def,EnvErr,EnvErr)).

create_vis_env(Scope,MName,IDS,vis(Section,Infos),InEnvErr,OutEnvErr) :-
    ~~mnf(get_section(Section,IDS,Identifiers1)),
    maplist(add_infos_to_identifier([loc(Scope,MName,Section)|Infos]),Identifiers1,Identifiers),
    foldl(add_unique_variable,Identifiers,InEnvErr,OutEnvErr).

add_unique_variable(Var1,Old/Err_in,New/Err_out) :-
    optionally_rewrite_id(Var1,Var),
    get_texpr_id(Var,Id),
    get_texpr_type(Var,Type),
    get_texpr_info(Var,InfosOfNew),!,
    ( env_lookup_type(Id,Old,_) ->
        % we have a collision of two identifiers
        handle_collision(Var,Id,Type,Old,InfosOfNew,New,Err_in,Err_out)
    ; otherwise ->
        % no collision, just introduce the new identifier
        env_store(Id,Type,InfosOfNew,Old,New),
        Err_in=Err_out
    ).
add_unique_variable(Var,Env/Err,Env/Err) :-
    ( Var = b(definition(DEFNAME,_),_,INFO) ->
        add_error(add_unique_variable,'DEFINITION used in place of Identifier: ',DEFNAME,INFO)
    ; otherwise ->
        add_error(add_unique_variable,'Expected Identifier, but got: ',Var,Var)
    ).

add_infos_to_identifier(NewInfos,In,Out) :-
    add_texpr_infos(In,NewInfos,Out).

% get the ID of the variable, prime it if the infos contain "primed"
optionally_rewrite_id(Var,NewVar) :-
    get_texpr_info(Var,InfosIn),
    ( selectchk(primed,InfosIn,InfosOut) ->
        get_texpr_id(Var,Id1),
        get_texpr_type(Var,Type),
        atom_concat(Id1,'\'',Id),
        create_texpr(identifier(Id),Type,InfosOut,NewVar)
    ; otherwise ->
        Var = NewVar).

:- use_module(translate,[translate_span/2]).
% in case of a collision, we have three options:
%  - overwrite the old identifier,
%  - ignore the new identifier or
%  - generate an error message
handle_collision(Var,Name,Type,OldEnv,InfosOfNew,NewEnv,Ein,Eout) :-
    env_lookup_infos(Name,OldEnv,InfosOfExisting),
    ( collision_precedence(Name,InfosOfExisting,InfosOfNew) ->
        % this identifier should be ignored
        OldEnv = NewEnv,
        Ein = Eout
    ; collision_precedence(Name,InfosOfNew,InfosOfExisting) ->
        % the existing should be overwritten
        env_store(Name,Type,InfosOfNew,OldEnv,NewEnv),
        Ein = Eout
    ; otherwise ->
        % generate error and let the environment unchanged
        (Name = op(IName) -> true; Name=IName),
        get_texpr_pos(Var,Pos1),
        safe_get_info_pos(InfosOfExisting,Pos2),
        ( double_inclusion_allowed(Pos1,Pos2,InfosOfExisting)
         -> print(double_inclusion_of_id_allowed(Name,Type,Pos1,OldEnv,InfosOfExisting)),nl,
            OldEnv=NewEnv, Ein=Eout
         ; ((better_pos(Pos2,Pos1), \+ better_pos(Pos1,Pos2))
            -> Pos = Pos2
             ; Pos = Pos1),
          translate_span(Pos1,PS1), translate_span(Pos2,PS2),
          ajoin(['Identifier \'', IName, '\' declared twice at (Line:Col[:File]) ', PS1, ' and ', PS2],Msg),
          Ein = [error(Msg,Pos)|Eout],
          OldEnv = NewEnv
        )
    ).

% SEE ISSUE PROB-403
% Correct behaviour related to multiple instantiation is specified in
% B Reference Manual (AtelierB 4.2.1), 8.3 B Project/Instantiating and renaming.
% => Constants and Sets defined in machines instantiated multiple times CAN be used in the including machine with their original (non-prefixed names).
% we do not use this code yet: this leads to the constants being added multiple times
double_inclusion_allowed(Pos1,Pos2,InfosOfExisting) :- fail, % DISABLE AT THE MOMENT; TO DO: FIX
  Pos1==Pos2,
  print(chk(InfosOfExisting)),nl,
  member(loc(LOC,_,Section),InfosOfExisting),
  section_can_be_included_multiple_times_nonprefixed(Section),
  % check that we are in a context of an included machine identifier:
  (inclusion_directive(LOC)
    -> true
    ; %LOC is probably local
      member(origin([INCL/_MACHINE|_]),InfosOfExisting), inclusion_directive(INCL)
  ).

inclusion_directive(included).
inclusion_directive(used).
inclusion_directive(seen). % imports ??

section_can_be_included_multiple_times_nonprefixed(abstract_constants).
section_can_be_included_multiple_times_nonprefixed(concrete_constants).
section_can_be_included_multiple_times_nonprefixed(sets).
section_can_be_included_multiple_times_nonprefixed(enumerated_sets).
section_can_be_included_multiple_times_nonprefixed(enumerated_elements).


% decide which position info is better: prefer info in main file (highlighting)
:- use_module(bmachine,[b_get_main_filenumber/1]).
better_pos(pos(_,Filenumber,_Srow,_Scol,_Erow,_Ecol),_) :- bmachine:b_get_main_filenumber(Filenumber).
better_pos(_,none).

safe_get_info_pos(Info,Pos) :- (get_info_pos(Info,Pos) -> true ; Pos=none).

% collision_precedence/3 decides if the first variable takes
% precedence over the second in case of a collision
% the decision is made by the additional information of both
% variables
collision_precedence(_Name,PreferredVarInfos,DroppedVarInfos) :-
    % in case of a re-introduced variable from the abstraction,
    % we prefer the concrete variable to the abstract one.
    is_abstraction(DroppedVarInfos,PreferredVarInfos),!.
collision_precedence(_Name,PreferredVarInfos,DroppedVarInfos) :-
    % We are checking an Event-B model with multi-level support
    % and have the same variable in two different refinement levels.
    % Then the one in the more refined module takes precedence
    member(level(L1),PreferredVarInfos),
    member(level(L2),DroppedVarInfos),
    % Level 0 is the abstract level, level 1 the first refinement, etc.
    L1 > L2,!.
collision_precedence(Name,PreferredVarInfos,DroppedVarInfos) :-
        % A local definition takes precedence over a non-local identifier
        % TODO:
        member(loc(local,_DefMachine,definitions),PreferredVarInfos),
        member(loc(_,_VarMachine,Section),DroppedVarInfos),
        Section \= definitions,!,
        (preferences:get_preference(warn_if_definition_hides_variable,true)
          -> ((get_info_pos(PreferredVarInfos,Pos1), Pos1 \= none,
               get_info_pos(DroppedVarInfos,Pos2), Pos2 \= none)
             -> translate:translate_span(Pos1,Pos1Str), translate:translate_span(Pos2,Pos2Str),
                ajoin(['Warning: DEFINITION of ', Name, ' at (line,column) =', Pos1Str,
                       ' hides variable with same name at', Pos2Str, '.'], Msg)
             ;  ajoin(['Warning: DEFINITION of ', Name, ' hides variable with same name.'], Msg)
             ),
             add_warning(Msg)
          ;  true
        ).

% is_abstraction/2 tries to find out if (in case of a name clash of
% two variables) the second variable is just the re-introduced abstract
% variable in a refinement.
% InfosAbs is the list of information about the abstract variable
% InfosConc is the list of information about the concrete variable
is_abstraction(InfosAbs,InfosConc) :-
    % one variable is an abstract variable, introduced in the abstraction
    member(loc(abstraction,_,abstract_variables),InfosAbs),
    % the other is either an abstract or concrete variable,
    member(VarType,[abstract_variables,concrete_variables]),
    % introduced either locally or in an included machine
    member(Scope,[local,included]),
    member(loc(Scope,_,VarType),InfosConc).
% and the same for constants:
is_abstraction(InfosAbs,InfosConc) :-
    % one variable is an abstract variable, introduced in the abstraction
    member(loc(abstraction,_,abstract_constants),InfosAbs),
    % the other is either an abstract or concrete variable,
    member(VarType,[abstract_constants,concrete_constants]),
    % introduced either locally or in an included machine
    member(Scope,[local,included]),
    member(loc(Scope,_,VarType),InfosConc).

% shortcuts for sections, to ease the use of typical combinations of
% sections
shortcut(all_parameters,[parameters,internal_parameters]).
shortcut(sets,[deferred_sets,enumerated_sets,enumerated_elements]).
shortcut(constants,[abstract_constants,concrete_constants]).
shortcut(variables,[abstract_variables,concrete_variables]).
shortcut(operations,[unpromoted,promoted]).
shortcut(identifiers,[all_parameters,sets,constants,variables,operations]).

expand_shortcuts(Sections,Expanded) :-
    foldl(expand_shortcut,Sections,Expanded,[]).
expand_shortcut(Section,Sections,RSections) :-
    ( shortcut(Section,Expanded) ->
        foldl(expand_shortcut,Expanded,Sections,RSections)
    ; valid_section(Section) ->
        Sections = [Section|RSections]
    ; otherwise ->
        add_internal_error('invalid section',expand_shortcut(Section,Sections,RSections)),fail).

% find sections that can see given sections
find_relevant_sections(RSecs,MTypes,Result) :-
    expand_shortcuts(RSecs,Sections),
    % print(find_rel_sections_for(Sections)),nl,
    findall(R,
              (   member(MType,MTypes),
                  visibility(MType,local,R,SAccess,_),
                  expand_shortcuts(SAccess,Access),
                  member(S,Sections),
                  member(S,Access),
                  valid_section(R)),
              Result1),
    sort(Result1,Result).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% renaming

rename_relevant_sections(DefSecs,Renamings,Machine,New) :-
    find_relevant_sections(DefSecs,[machine],Relevant),
    rename_in_sections(Relevant,Renamings,Machine,New).
rename_in_sections([],_,M,M).
rename_in_sections([Section|Rest],Renamings,Old,New) :-
    select_section_texprs(Section,OTExprs,NTExprs,Old,Inter),
    rename_bt_l(OTExprs,Renamings,NTExprs),
    rename_in_sections(Rest,Renamings,Inter,New).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% cleaning up machine
clean_up_machine(In,RefMachines,Out) :-
    extract_parameter_types([ref(local,In)|RefMachines],NonGroundExceptions),
    clean_up_machine2(NonGroundExceptions,In,Out).
clean_up_machine2(NonGroundExceptions) -->
    clean_up_section(constraints,NonGroundExceptions),
    clean_up_section(properties,NonGroundExceptions),
    clean_up_section(invariant,NonGroundExceptions),
    clean_up_section(initialisation,NonGroundExceptions),
    clean_up_section(assertions,NonGroundExceptions),
    clean_up_section(operation_bodies,NonGroundExceptions).
:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_safe_mode,true)).
clean_up_section(Section,NonGroundExceptions,In,Out) :-
    select_section_texprs(Section,Old,New,In,Out),
    %format('Cleaning up and optimizing section ~w~n',[Section]), %maplist(check_ast,Old), % this will raise errors
    clean_up_l_with_optimizations(Old,NonGroundExceptions,New,Section),
    %format('Checking result of clean_up section ~w~n',[Section]),
    maplist(check_ast(true),New),
    format('Finished checking section ~w~n',[Section]).
:- else.
clean_up_section(Section,NonGroundExceptions,In,Out) :-
    select_section_texprs(Section,Old,New,In,Out),
    clean_up_l_with_optimizations(Old,NonGroundExceptions,New,Section).
:- endif.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type expressions in context of an already type-checked machine
type_in_machine_l(Exprs,Scope,Machine,Types,TExprs,Errors) :-
    MType = machine,
    create_scope(Scope,MType,Machine,Env,Errors,E1),
    btype_ground_dl(Exprs,Env,[],Types,TExprUncleans,E1,[]),
    maplist(clean_up_pred_or_expr([]),TExprUncleans,TExprs).

create_scope(Scope,MType,Machine,Env,Ein,Eout) :-
    env_empty(Init),
    add_theory_operators(Machine,Init,WithOperators),
    foldl(create_scope2(MType,Machine),Scope,WithOperators/Ein,Env/Eout).
add_theory_operators(Machine,Ein,Eout) :-
    get_section(operators,Machine,Operators),
    keys_and_values(Operators,Ids,Ops),
    foldl(env_store_operator,Ids,Ops,Ein,Eout).
create_scope2(MType,Machine,Scope,In/Ein,Out/Eout) :-
    ( Scope = constants ->
        ~~mnf( visible_env(MType,properties,[ref(local,Machine)],In,Out,Ein,Eout) )
    ; Scope = variables ->
        ~~mnf( visible_env(MType,invariant,[ref(local,Machine)],In,Out,Ein,Eout) )
    ; Scope = prepost ->
        ~~mnf( visible_env(MType,prepost,[ref(local,Machine)],In,Out,Ein,Eout) )
    ; Scope = operation_bodies ->
        ~~mnf( visible_env(MType,operation_bodies,[ref(local,Machine)],In,Out,Ein,Eout) )
    ; Scope = operation(Op) ->
        create_operation_scope(Op,Machine,In,Out), Ein=Eout
    ; Scope = env(ExplicitEnv) ->
        ExplicitEnv = Out, Ein=Eout
    ; Scope = identifier(Ids) ->
        ~~mnf( store_variables(Ids,In,Out) ), Ein=Eout
    ; otherwise ->
        add_error(bmachine_construction, 'invalid scope', Scope),fail).
create_operation_scope(Op,Machine,In,Out) :-
    get_section(operation_bodies,Machine,OpBodies),
    get_texpr_id(OpId,op(Op)),
    get_texpr_expr(TOp,operation(OpId,Results,Params,_)),
    ( member(TOp,OpBodies) ->
        append(Results,Params,LocalVariables),
        store_variables(LocalVariables,In,Out)
    ; otherwise ->
        ajoin(['operation \'',Op,'\' not found for building scope'], Msg),
        add_error(bmachine_construction,Msg),fail).
store_variables(Ids,In,Out) :- foldl(store_variable,Ids,In,Out).
store_variable(Id,In,Out) :-
    get_texpr_id(Id,Name),get_texpr_type(Id,Type),
    env_store(Name,Type,[],In,Out).

type_open_predicate_with_quantifier(OptionalOuterQuantifier,Predicate,Scope,Machine,Result,Errors) :-
    type_open_formula(Predicate,Scope,false,Machine,pred,Identifiers,TPred,Errors),
    ( Identifiers = [] ->
        Result = TPred
    ; OptionalOuterQuantifier=forall ->
        create_forall(Identifiers,TPred,Result)
    ; OptionalOuterQuantifier=no_quantifier ->
        Result = TPred
    ;  otherwise -> create_exists(Identifiers,TPred,Result)
    ).

type_open_formula(Raw,Scope,AllowOpenIdsinExpressions,Machine,Type,Identifiers,Result,Errors) :-
    create_scope(Scope,machine,Machine,Env1,Errors,E1),
    ( Identifiers==[] -> Mode=closed, Env1=Env
    ; otherwise       -> Mode=open,   openenv(Env1,Env)),
    btype_ground_dl([Raw],Env,[],[Type],[TExprUnclean],E1,E2),
    ( Mode=closed -> true
    ; otherwise ->
        openenv_identifiers(Env,Identifiers),
        check_ground_types_dl(Identifiers,[],E2,E3)
    ),
    %print(type_open_formula(Identifiers,TExprUnclean)),nl,
    mark_outer_quantifier_ids(TExprUnclean,TExprUnclean2),
    clean_up_pred_or_expr([],TExprUnclean2,TResult),
    ( Identifiers = [] ->       % no newly introduced identifiers, no problem
        E3 = []
    ; Type = pred ->            % newly introduced identifiers in a predicate - ok
        E3 = []
    ; AllowOpenIdsinExpressions=true -> % we explicitly allow open ids in expressions
        E3 = []
    ; otherwise ->              % newly introduced identifiers in expression make no sense
                                % (is that so?)
        foldl(add_unknown_identifier_error,Identifiers,E3,[])
    ),
    Result = TResult.
add_unknown_identifier_error(TId,[error(Msg,Pos)|E],E) :-
    get_texpr_id(TId,Id),
    ajoin(['Unknown identifier ',Id,'.'],Msg),
    get_texpr_pos(TId,Pos).

% mark outermost identfiers so that they don't get optimized away
% e.g., ensure that we print the solution for something like #(y2,x2).(x2 : 0..10 & y2 : 0..10 & x2 = 10 & y2 = 10)
mark_outer_quantifier_ids(b(exists(IDS,Pred),pred,Info),Res) :-
    maplist(mark_id,IDS,MIDS),!,
    Res = b(exists(MIDS,Pred),pred,Info).
mark_outer_quantifier_ids(b(let_predicate(IDS,Exprs,Pred),pred,Info),Res) :-
    maplist(mark_id,IDS,MIDS),!,
    Res = b(let_predicate(MIDS,Exprs,Pred),pred,Info).
mark_outer_quantifier_ids(b(forall(IDS,LHS,RHS),pred,Info),Res) :-
    maplist(mark_id,IDS,MIDS),!,
    Res = b(forall(MIDS,LHS,RHS),pred,Info).
mark_outer_quantifier_ids(X,X).

mark_id(b(identifier(ID),TYPE,INFO),b(identifier(ID),TYPE,[do_not_optimize_away|INFO])).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Warnings

:- dynamic warnings/1.
clear_warnings :-
    retractall( warnings(_) ).
show_warnings :-
    warnings(Warning),
    add_warning(bmachine_construction, Warning),
    fail.
show_warnings.


add_warning(Warning) :-
    ( warnings(Warning) -> true
    ; otherwise -> assert( warnings(Warning) ) ).
