% (c) 2016-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(clpfd_tables,[can_translate_avl_to_table/2, translate_avl_to_table/2, check_element_of_table/4,
    check_element_of_avl_with_table/4,
    clpfd_in_table/2,

    can_translate_function_to_element_constraint/2,
    check_apply_with_element_constraint/5]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'Provide interface to CLP(FD) table constraint.').

% Example usage of CLPFD table/2 :
%| ?- table([[X,Y]],[[1,2],[2,3],[3,4],[4,5]]), X in 1..2.
%X in 1..2,
%Y in 2..3 ?

:- use_module(library(clpfd)).
:- use_module(library(avl),[avl_domain/2]).
:- use_module(library(lists),[maplist/3, maplist/4, maplist/2]).
:- use_module(kernel_waitflags,[add_fd_variables_for_labeling/2]).
:- use_module(debug,[debug_println/2]).
:- use_module(error_manager,[add_internal_error/2]).

% can an AVL set be translated into a CLPFD table membership constraint
can_translate_avl_to_table(AVL,SkeletonType) :-
   AVL = node(Element,_True,_,_,_),
   translate_to_fd_values(Element,SkeletonType,ElVList,[]),
   maplist(no_overflow,ElVList), % check that the element constraint will not generate an overflow, e.g., for test 1353
   % TO DO: ideally we should check avl_min / avl_max (but with multiple FD values in an element; even that will not guarantee that there are no overflows)
   %(ElVList = [_,_|_] -> true ; print(Element),nl,fail,trace,true),
   SkeletonType \= boolean. % should we also disallow rec([field(_,boolean)]) % boolean treatment causes problems for 1624, 1627, 140, 1426

:- use_module(tools,[is_tagged_integer/1]).
no_overflow(H) :- (var(H) -> true ; is_tagged_integer(H)).

translate_avl_to_table(AVL,TableList) :-
   avl_domain(AVL,List),
   maplist(translate_avl_element_to_fd_list,List,TableList).

check_element_of_avl_with_table(X,SkeletonType,AVL,WF) :-
   translate_avl_to_table(AVL,TableList),
   check_element_of_table(X,SkeletonType,TableList,WF).

check_element_of_table(X,SkeletonType,TableList,WF) :-
   translate_to_fd_values(X,SkeletonType,XList,[]),
   clpfd_in_table(XList,TableList),
   add_fd_variables_for_labeling(XList,WF).

:- use_module(clpfd_interface,[clpfd_inlist/2]).
get_unique_el([X],X).
clpfd_in_table([X],TableList) :- !, maplist(get_unique_el,TableList,FDList),
   %print(posting_clpfd_inlist(X,FDList)),nl,
   clpfd_inlist(X,FDList).
clpfd_in_table(XList,TableList) :-
   %print(posting_clpfd_table([XList],TableList)),nl,
   clpfd:table([XList],TableList). % we could provide optional third argument, e.g., [order(id3)]
% from the SICStus Manual:
% table(+Tuples,+Extension,+Options)
% Defines an n-ary constraint by extension. Extension should be a list of lists of
% integers, each of length n. Tuples should be a list of lists of domain variables,
% each also of length n. The constraint holds if every Tuple in Tuples occurs in
% the Extension. The constraint will maintain domain consistency.
% For convenience, Extension may contain ConstantRange (see Section 10.10.13.1
% [Syntax of Indexicals], page 485) expressions in addition to integers.

% TO DO: also allow partial FD infos, in which case we would use table/2 just for quick propagation
% we could also translate variables into ConstantRange expressions (see above)
translate_avl_element_to_fd_list(Val,FDList) :- translate_to_fd_values(Val,_,FDList,[]).

translate_to_fd_values(fd(X,T),fd(T)) --> [X].
translate_to_fd_values(int(A),integer) --> [A]. % warning: could overflow, we check above !
translate_to_fd_values((A,B),(SkelA,SkelB)) --> translate_to_fd_values(A,SkelA), translate_to_fd_values(B,SkelB).
translate_to_fd_values(Var,Type) --> {var(Var),Type==boolean},!, {prop_01(Var,FD)},[FD].
translate_to_fd_values(pred_true,boolean) --> [1].
translate_to_fd_values(pred_false,boolean) --> [0].
translate_to_fd_values(rec(F),rec(SkelF)) --> l_translate_to_fd_values(F,SkelF).
% TO DO: also free-types

l_translate_to_fd_values([],[]) --> [].
l_translate_to_fd_values([field(N,F)|T],[field(N,SF)|ST]) -->
   translate_to_fd_values(F,SF),l_translate_to_fd_values(T,ST).

:- block prop_01(-,-).
prop_01(pred_true,1).
prop_01(pred_false,0).

% ----------------------

:- use_module(bsyntaxtree,[is_set_type/2]).
:- use_module(preferences,[preference/2]).

% can a list be translated into CLPFD element constraints for function application
can_translate_function_to_element_constraint(List,FunctionType) :-
    FunctionType \=  unknown,
    is_list_skeleton(List), % TO DO: allow tail AVL list + enable delaying treatment
    preference(find_abort_values,false),
    preference(use_chr_solver,false), % avoid falling into the SICStus element/3 bug prior to 4.3.6; TO DO: remove later
    is_set_type(FunctionType,couple(Domain,Range)),
    is_simple_fd_type(Domain), % Domain must be translatable to simple FD value list
    is_fd_type(Range).

is_list_skeleton(X) :- nonvar(X),is_list_skeleton_aux(X).
is_list_skeleton_aux([]).
is_list_skeleton_aux([_H|T]) :- is_list_skeleton(T).

is_simple_fd_type(global(_)).
is_simple_fd_type(integer).
is_simple_fd_type(boolean).

is_fd_type(couple(A,B)) :- !, is_fd_type(A), is_fd_type(B).
is_fd_type(record(F)) :- !, is_fd_field_types(F).
is_fd_type(X) :- is_simple_fd_type(X).

is_fd_field_types([]).
is_fd_field_types([field(_,FT)|T]) :- is_fd_type(FT), is_fd_field_types(T).

get_fd_type(couple(A,B),(VA,VB)) --> get_fd_type(A,VA), get_fd_type(B,VB).
get_fd_type(global(T),fd(X,T)) --> [X].
get_fd_type(integer,int(X)) --> [X].
get_fd_type(boolean,B) --> [FD], {prop_01(B,FD)}.
get_fd_type(record(Fields),rec(VF)) --> get_field_fd_types(Fields,VF).

get_field_fd_types([],[]) --> [].
get_field_fd_types([field(Name,Type)|FT],[field(Name,VF)|VT]) -->
     get_fd_type(Type,VF),
     get_field_fd_types(FT,VT).

get_fd_types(Domain,Range,(VD,VR),FDDom,FDRList) :- get_fd_type(Domain,VD,[FDDom],[]), get_fd_type(Range,VR,FDRList,[]).

% Examples:
% f = {x|->2, y|->3} & x:1..2 & y:4..5 & r = f(4)
% f = {x|->(2,x+1), y|->(3,y+1)} & x:1..2 & y:4..5 & r = f(v) & v:3..4  --> deterministically sets v=4, r=(3,5),y=4
% from the SICStus manual:
% element(?X,+List,?Y)
% where X and Y are domain variables and List is a list of domain variables. True
% if the X:th element of List is Y. Operationally, the domains of X and Y are
% constrained so that for every element in the domain of X, there is a compatible
% element in the domain of Y, and vice versa.
% Maintains domain consistency in X and bounds consistency in List and Y.

:- use_module(library(clpfd),[all_distinct/1]).
:- use_module(library(lists),[nth1/3]).
check_apply_with_element_constraint(List,X,Y,FunctionType,WF) :-
    is_set_type(FunctionType,couple(Domain,Range)),
    get_fd_type(Domain,X,[FDX],[]),
    if(l_get_fd_types(List,DomainList,RangeLists,Domain,Range,FDX,FoundY),true,
       (add_internal_error('Get FD Types failed: ',get_fd_types(Domain,Range)),fail)),
    debug_println(4,element_domain_list(FDX,DomainList)),
    %print(element_domain_list(FDX,DomainList,FoundY)),nl,
    (FoundY \== y_not_found
      -> FoundY=Y % no need for equal_object: we only have FD types + Bool + couples
      ; element(Idx,DomainList,FDX), % no need to call equal_object for X as we have simple fd types
        (nonvar(Idx)
         -> nth1(Idx,List,(X,Y)) %relevant for tests 387, 1579
         ; all_distinct(DomainList), % ensures that Idx will get bound eventually if we enumerate DomainList and X (if multiple entries in DomainList then Idx would not get bound)
           maplist(get_fd_types(Domain,Range),List,DomainList,RangeLists),
           get_fd_type(Range,Y,FDYList,[]),
           setup_element_domain_constraints(FDYList,Idx,RangeLists),
           add_fd_variables_for_labeling([FDX|FDYList],WF)
        )
    ).

% an unfolded version of maplist(get_fd_types(Domain,Range),List,DomainList,RangeLists)
% we also check if the lookup FD variable occurs in the list; if so we return the Y value in Found
% relevant for performance, e.g., for NQueens60; we avoid traversing the list a second time
l_get_fd_types([],[],[],_,_,_,y_not_found).
l_get_fd_types([Pair|PT],[FDDom|DT],[FDRList|RT],Domain,Range,FDX,FoundY) :-
   get_fd_types(Domain,Range,Pair,FDDom,FDRList),
   (FDX==FDDom -> Pair=(_,FoundY), DT=[],RT=[]
    ; l_get_fd_types(PT,DT,RT,Domain,Range,FDX,FoundY)).


setup_element_domain_constraints([],_,_).
setup_element_domain_constraints([FDY|YT],Idx,RangeLists) :-
    maplist(get_next_fd,RangeLists,FDYList,RestLists),
    %print(setting_up_range_element_constraint(Idx,FDYList,FDY)),nl,
    element(Idx,FDYList,FDY),
    setup_element_domain_constraints(YT,Idx,RestLists).

get_next_fd([H|T],H,T).
