% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(enabling_analysis,[reset_enabling_analysis/0,
                             tcltk_cbc_cfg_analysis/1, tcltk_dot_cfg_analysis/1,
                             tcltk_cbc_enabling_analysis/1, tcltk_cbc_enabling_analysis/2,
                             tcltk_cbc_enabling_relations_for_operation/3,
                             tcltk_cbc_enabling_relations_after_operation/3, compute_cbc_enable_rel/4,
                             tcltk_cbc_simultaneous_enabling_analysis/1,
                             tcltk_cbc_dependence_analysis/1, print_enable_table/1,
                             eop_node_predicate/6,init_or_op/1,
			                 feasible_operation/2, feasible_operation_with_timeout/3,
			                 check_if_feasible_operation/5,
			                 feasible_operation_cache/2,infeasible_operation/1, infeasible_operation_cache/1,
			                 cbc_enable_analysis/4,
                       is_timeout_enabling_result/1,
                       translate_enable_res/6,
                       operation_sequence_possible/3, operation_can_be_enabled_by/3,

                       cbc_enable_analysis_cache/4 % for prologTasks
                       ]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,cbc).
:- module_info(description,'This module computes enabling relations for B operations.').
% --------------

:- use_module(probporsrc(static_analysis),[action_dependent_to_itself/4, get_conj_inv_predicate/3, is_timeout/1]).
:- use_module(bmachine,            [b_top_level_operation/1]).
:- use_module(b_state_model_check, [get_negated_guard/3,get_negated_guard/4]).
:- use_module(sap).
:- use_module(library(ordsets)).
:- use_module(library(lists)).
:- use_module(error_manager).
:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2]).
:- use_module(probsrc(translate), [translate_bexpression/2]).
:- use_module(probsrc(b_read_write_info), [b_get_read_write/3,b_get_read_write_vars/5, b_get_operation_read_guard_vars/3]).

% -------------------- BINARY STATIC ANALYSIS -------------------------

% certain static analyses used by CBC Test case generation

% check if OpName1 can influence the truth value of the guard of OpName2 by looking at read/write info
activation_dependent(OpName1,OpName2) :-
    b_top_level_operation(OpName1),
    b_get_read_write(OpName1,_,Writes1),
    b_top_level_operation(OpName2),
    b_get_operation_read_guard_vars(OpName2,true,GuardReads2),
    (my_ord_intersect(Writes1,GuardReads2) -> true
     ; %print(indep(OpName1,Writes1,OpName2,GuardReads2)),nl,
       fail
     ).

% a version of ord_intersect(ion) which deals with the '$all' term
my_ord_intersect('$all',_) :- !.
my_ord_intersect(_,'$all') :- !.
my_ord_intersect(A,B) :- ord_intersect(A,B).


% return Res=impossible if from Invariant we can never execute OpName1 followed by OpName2
% is used by sap:check_operation_sequence_possible
operation_sequence_possible(OpName1,OpName2,Res) :-
    \+ activation_dependent(OpName1,OpName2),
    !,
    Res = activation_independent.
operation_sequence_possible(OpName1,OpName2,Res) :-
  % we assume OpName2 is feasible on its own
    get_negated_guard(OpName2,PosGuard2,_NegGuard2),
    Timeout1 = 200,
    % print('Guard: '),translate:print_bexpr(PosGuard2),nl,
    (testcase_path_timeout_catch(invariant,Timeout1,[OpName1],PosGuard2,_Csts,_Ops,_TestS,_TI,Res)
      -> true
      ;  Res = impossible
    ).


% return whether OpName1 can enable a previously disabled OpName2
operation_can_be_enabled_by(OpName1,OpName2,Res) :-
    b_top_level_operation(OpName1),
    b_get_read_write(OpName1,_Reads1,Writes1),
    b_top_level_operation(OpName2),
    format("~nCHECKING if ~w (~w) can enable ~w~n",[OpName1,Writes1,OpName2]),
    get_negated_guard(OpName2,PosGuard2,_NegGuard2),
    filter_predicate(PosGuard2,Writes1,FilteredPosGuard2),
    create_negation(FilteredPosGuard2,FilteredNegGuard2),
    TIMEOUT = 500,
   % print('Before: '), translate:print_bexpr(FilteredNegGuard2),nl,print('After: '), translate:print_bexpr(PosGuard2),nl,
    testcase_path_timeout_catch(pred(FilteredNegGuard2),TIMEOUT,[OpName1],PosGuard2,_,_,_,_,Res),
    print(result(Res)),nl.



% -------------------- CFG ANALYSIS -------------------------
:-use_module(dot_graph_generator,[gen_dot_graph/6]).
tcltk_dot_cfg_analysis(File) :-
       gen_dot_graph(File,enabling_analysis,eop_node_predicate,cfg_trans_predicate,none,none).

:- use_module(tools,[ajoin/2]).
:- use_module(preferences,[get_preference/2]).
op_node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,Color) :-
     get_preference(dot_enabling_show_readwrites,false), !,
     SubGraph=none, Shape=box, Style=none,
     b_top_level_operation(Op), NodeID = Op, NodeDesc=Op, Color=blue.
op_node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,Color) :- SubGraph=none,
     Shape=record, Style=none,
     b_top_level_operation(Op), NodeID = Op,
     %NodeDesc = NodeID,
     b_get_read_write(Op,Reads2,Writes1),
     b_get_operation_read_guard_vars(Op,false,ReadsGrd),
     insert_commas(ReadsGrd,0,R1), insert_commas(Reads2,0,R2), insert_commas(Writes1,0,W1),
     append(W1,['}|'],W2),
     append(R1,['\\n|reads: '|R2],Reads),
     append(Reads,['\\n|writes: '|W2],As),
     ajoin(['|{',Op,'\\n|reads (guard): '|As],NodeDesc),
     Color=blue.

insert_commas('$all',_,['ALL']).
insert_commas([],_,[]).
insert_commas([H],_,R) :- !, R=[H].
insert_commas([H|T],N,[H,Sep|IT]) :-
    (N>5 -> N1=0, Sep=',\\n' ; N1 is N+1, Sep=', '),
    insert_commas(T,N1,IT).

:- public cfg_trans_predicate/5.
cfg_trans_predicate(NodeID,Label,SuccID,Color,Style) :-
    cbc_quick_cfg_analysis(NodeID,SuccID,Res),
    translate_res(Res,NodeID,SuccID,Label,Color,Style).

translate_res(cannot_enable,_,_,_,_,_) :- !, fail.
translate_res(syntactic_independent,X,Y,'independent',yellow,bold) :- !,X=Y. % only show if source & dest identical
translate_res(race_dependent,X,Y,'race_independent',lightgray,bold) :- !,X=Y. % only show if source & dest identical
translate_res(possible,_,_,'',black,solid) :- !.
translate_res(timeout_possible,_,_,'',tomato,solid) :- !.
translate_res(X,_,_,X,red,dashed).

% compute a Control Flow Graph very quickly
% ideal when we have program counter style variables
tcltk_cbc_cfg_analysis(list([list(['Origin'|SOps])|Result])) :-
  findall(Op, b_top_level_operation(Op), Ops), sort(Ops,SOps),
  findall(list([OpName1|EnableList]),
     (b_get_sorted_op(OpName1),
      findall(Possible,cbc_quick_cfg_analysis(OpName1,_OpName2,Possible),EnableList)),
     Result),
  print_enable_table([list(['Origin'|Ops])|Result]).

/*  TO DO: complete into a determinacy analysis ?
cbc_quick_det_analysis(OpName1,OpName2,Res) :-
    b_top_level_operation(OpName1),  % top_level
    ((b_get_sorted_op(OpName2), OpName2 \= OpName1,
   % get_negated_guard(OpName1,PosGuard1,NegGuard1),
    get_negated_guard(OpName2,PosGuard2,NegGuard2),
    sap:testcase_path_timeout(pred(PosGuard2),1200,[OpName1],b(truth,pred,[]),_,_,_,_,Res))
     -> format(user_output,'Operation ~w can be simultaneously enabled with ~w (~w).~n',[OpName1,OpName2,Res])
     ;  format(user_output,'Operation ~w cannot be simultanously enabled with another operation.~n',[OpName1]), Res=det
     ).
*/

cbc_quick_cfg_analysis(OpName1,OpName2,Res) :-
   OpName1='INITIALISATION',
   b_top_level_operation(OpName2),
   b_get_read_write(OpName2,Reads2,Writes2),
   format(user_output,"COMPUTING CFG: INITIALISATION  -->  ~w r:~w / w:~w~n~n",
				 [OpName2,Reads2,Writes2]),
    ( testcase_path_timeout_catch(init,250,[OpName2],b(truth,pred,[]),_Constants,_Ops,_TestS,_TI,R1)
      ->  format(user_output," ~w can be enabled by ~w (~w)!~n",[OpName2,OpName1,R1]),
		  (is_timeout(R1) -> Res=timeout_possible; Res = possible)
	    ; format(user_output," ~w cannot be enabled by ~w!~n",[OpName2,OpName1]),
		  Res = cannot_enable
	).
cbc_quick_cfg_analysis(OpName1,OpName2,Res) :-
    b_top_level_operation(OpName1),  % top_level
    b_get_read_write(OpName1,Reads1,Writes1),
    %get_negated_guard(OpName1,PosGuard1,NegGuard1),
    b_get_sorted_op(OpName2),
    (b_get_read_write(OpName2,Reads2,Writes2),
		format(user_output,"COMPUTING CFG: ~w r:~w / w:~w  -->  ~w r:~w / w:~w~n~n",
				 [OpName1,Reads1,Writes1,OpName2,Reads2,Writes2]),
		(\+ my_ord_intersect(Writes1,Reads2)
		  -> (my_ord_intersect(Writes1,Writes2)
		      -> Res = race_dependent, format(user_output," ~w cannot be enabled/disabled by ~w!~n",[OpName2,OpName1])
		       ; Res = syntactic_independent, format(user_output," ~w cannot be enabled/disabled/modified by ~w!~n",[OpName2,OpName1])
		      )
		  ;  get_negated_guard(OpName2,PosGuard2,NegGuard2),
		     garbage_collect,
		     print('Guard: '), translate:print_bexpr(PosGuard2),nl,
			 ((testcase_path_timeout_catch(typing(NegGuard2),60,[OpName1],PosGuard2,_,_,_,_,_R0), % quick check without invariants & properties
			 % TO DO: project constants, variables onto the ones really needed
			   testcase_path_timeout_catch(pred(NegGuard2),250,[OpName1],PosGuard2,_Constants,_Ops,_TestS,_TI,R1))
			   ->  format(user_output," ~w can be enabled by ~w (~w)!~n",[OpName2,OpName1,R1]),
				   (is_timeout(R1) -> Res=timeout_possible; Res = possible)
			  ; format(user_output," ~w cannot be enabled by ~w!~n",[OpName2,OpName1]),
			    Res = cannot_enable)
		)
    ).

%use_module(enabling_analysis),enabling_analysis:cbc_quick_cfg_analysis('SEQUENCER2','SEQUENCER3',Res).
%use_module(sap),use_module(enabling_analysis),enabling_analysis:cbc_quick_cfg_analysis('COMPUTE_SDDBs_points','prop_COMPUTE_0',Res).
%use_module(sap),use_module(enabling_analysis),enabling_analysis:cbc_quick_cfg_analysis('COMPUTE_SDDBs_points','COMPUTE_SDDBs_length',Res).

% -------------------- DEPENDENCE ANALYSIS ----------------------
/*

Independency between two actions in a B/Event-B model can be expressed also by means of LTL-formulas:
    1. Two actions a and b are independent if the following LTL-formula is satisfied by the model:
           "G ((e(b) & [a] => X e(b)) & (e(a) & [b] => X e(a)))" (ind)
    2. If (ind) is violated by the model then a and b are dependent actions.

Note: The race_dependent condition cannot be covered by the above LTL-formula. If two actions are race dependent, but never simultaniously enabled in a state
      from the state space, then (ind) will be satisfied.

*/

tcltk_cbc_dependence_analysis(list([list(['Origin'|SOps])|Result])) :-
  %findall(Op, b_get_machine_operation(Op,_,_,_), Ops), sort(Ops,SOps),
  findall(Op, b_top_level_operation(Op), Ops), sort(Ops,SOps),
  findall(list([OpName1|EnableList]),
     (b_get_sorted_op(OpName1),
      findall(Enable,cbc_dependence_analysis(OpName1,_OpName2,Enable),EnableList)),
     Result).
  %,print_enable_table([list(['Origin'|Ops])|Result]).

b_get_sorted_op(Op) :- findall(Op, b_top_level_operation(Op), Ops),
    sort(Ops,SOps), member(Op,SOps).

cbc_dependence_analysis(OpName1,OpName2,Res) :-
    b_top_level_operation(OpName1),
    b_get_read_write_vars(OpName1,GReads1,AReads1,Reads1,Writes1),
    b_get_sorted_op(OpName2),
    (OpName1=OpName2 -> action_dependent_to_itself(OpName1,GReads1,Writes1,Res) % TO DO: check if it is possible that for two different parameter values of the same event there is dependence
      ; OpName2 @< OpName1 -> Res = '-'  % our checking is symmetric; only check one pair
      ; otherwise ->
        b_get_read_write_vars(OpName2,GReads2,AReads2,Reads2,Writes2),
		    format("CHECKING DEPENDENCE: ~w gr:~w / ar:~w / w:~w  <-->  ~w gr:~w / ar:~w / w:~w~n",[OpName1,GReads1,AReads1,Writes1,OpName2,GReads2,AReads2,Writes2]),
    		( syntactical_independence(Reads1,Writes1,Reads2,Writes2) -> Res = syntactic_independent
    		; my_ord_intersect(Writes1,Writes2) -> Res = race_dependent
        ; (my_ord_intersect(AReads1,Writes2);my_ord_intersect(AReads2,Writes1)) -> Res = race_dependent
         % TO DO: in this case check if there is indeed a race dependence (e.g. in scheduler new and ready_active are actually independent !)
         % Set up [OpName1,OpName2], [OpName2,OpName1] and see if for ord_intersect(Writes1,Writes2) we can find different values
        ; otherwise ->
            % get constraints for G_{Op_1} \leadsto{Op_2} not(G_{Op_1})
            get_negated_guard(OpName1,PosGuard1,NegGuard1),
            get_conj_inv_predicate([NegGuard1],1,NegGuard1_Inv),
            % get constraints for G_{Op_2} \leadsto{Op_1} not(G_{Op_2})
            get_negated_guard(OpName2,PosGuard2,NegGuard2),
            get_conj_inv_predicate([NegGuard2],1,NegGuard2_Inv),
            conjunct_predicates([PosGuard1,PosGuard2],GuardsConj),
            ((my_ord_intersect(GReads1,Writes2),testcase_path_timeout_catch(pred(GuardsConj),500,[OpName2],NegGuard1_Inv,_Constants,_Ops,_TestS,_TI,R1))
                ->  format(" ~w may disable ~w (~w)!~n",[OpName2,OpName1,R1]),
                    (is_timeout(R1) -> Res=timeout_dependent; Res = dependent)
            ; (my_ord_intersect(GReads2,Writes1),testcase_path_timeout_catch(pred(GuardsConj),500,[OpName1],NegGuard2_Inv,_,_,_,_,R2))
                ->  format(" ~w may disable ~w (~w)!~n",[OpName1,OpName2,R2]),
                    (is_timeout(R2) -> Res=timeout_dependent; Res = dependent)
                ; Res = independent))).

syntactical_independence(Reads1,Writes1,Reads2,Writes2) :-
    \+ my_ord_intersect(Writes1,Reads2), % op1 does not modify guard/effect of op2
    \+ my_ord_intersect(Writes2,Reads1), % op2 does not modify guard/effect of op1
    \+ my_ord_intersect(Writes1,Writes2). % no race condition

% -------------- ENABLING ANALYSIS --------------

:- dynamic cbc_enable_analysis_cache/4.

% check which operations can be enabled after executing another operation
% to do: move maybe to another module; provide proper CSV export using Sebastian's modules
cbc_enable_analysis(OpName1,OpName2,Enable,ExtraTimeout) :-
   init_or_op(OpName1),
   if(cbc_enable_analysis_cache(OpName1,OpName2,Result,ExtraTimeout),
      Result = Enable, % already computed
      (% now compute all enablings form OpName1
       cbc_enable_analysis_calc(OpName1,Op2,R,ExtraTimeout),
       assert(cbc_enable_analysis_cache(OpName1,Op2,R,ExtraTimeout)),
       fail
       ;
       % now look up:
       cbc_enable_analysis_cache(OpName1,OpName2,Enable,ExtraTimeout)
      )).


cbc_enable_analysis_calc(OpName1,OpName2,Enable,ExtraTimeout) :-
   OpName1='INITIALISATION',print('CHECKING ENABLING AFTER INITIALISATION'),nl,
   b_top_level_operation(OpName2),
    print('INITIALISATION'), print(' ---> '), print(OpName2), print('  :: '),
    add_time_outs(ExtraTimeout,200,Timeout),
    ( testcase_path_timeout_catch(init,Timeout,[OpName2],b(truth,pred,[]),_Constants,_Ops,_TestS,_TI,R)
      -> print(R), print(' : '),
         get_negated_guard(OpName2,_,NegGuard),
         (testcase_path_timeout_catch(init,Timeout,[],NegGuard,_Constants2,_Ops2,_TestS2,_TI2,R2)
           -> print(R2), print(' : '),Enable=possible
            ; Enable=guaranteed)
      ;  Enable=impossible),
      print(Enable),nl.
cbc_enable_analysis_calc(OpName1,OpName2,Enable,ExtraTimeout) :-
    b_top_level_operation(OpName1),
    b_get_operation_read_guard_vars(OpName1,true,Reads1),
    %get_negated_guard(OpName1,PosGuard1,_NegGuard1),
    format("CHECKING ENABLING AFTER: ~w r:~w / w:~w~n",[OpName1,Reads1,Writes1]),
    b_top_level_operation(OpName2),
    format('~w ---> ~w  :: ',[OpName1,OpName2]),
    cbc_enable_calc_aux(OpName1,Writes1,OpName2,Enable,ExtraTimeout),
    print('Enable='),print(Enable),nl.

:- use_module(probporsrc(static_analysis),[syntactic_independence/3]).
cbc_enable_calc_aux(OpName1,_Writes1,_OpName2,Enable,ExtraTimeout) :-
    (ExtraTimeout>500 -> infeasible_operation(OpName1)
     ; infeasible_operation_cache(OpName1)), % only check cached version
    !,
    Enable=infeasible.
cbc_enable_calc_aux(_OpName1,_Writes1,OpName2,Enable,ExtraTimeout) :-
    (ExtraTimeout>500 -> infeasible_operation(OpName2)
      ; infeasible_operation_cache(OpName2)), % only check cached version
    !,
    Enable=impossible_infeasible.
cbc_enable_calc_aux(OpName1,_Writes1,OpName2,Enable,_) :-
    % first check if we can syntactically determine independence
    syntactic_independence(OpName1,OpName2,Res),
    !,
    Enable=Res.
cbc_enable_calc_aux(OpName1,Writes1,OpName2,Enable,ExtraTimeout) :-
    % now we do the semantic checks
    add_time_outs(ExtraTimeout,200,Timeout1),
    add_time_outs(ExtraTimeout,300,Timeout2),

    get_negated_guard(OpName2,PosGuard2,NegGuard2),
    filter_predicate(PosGuard2,Writes1,FilteredPosGuard2),
    create_negation(FilteredPosGuard2,FilteredNegGuard2),
    cbc_enable_calc_aux2(OpName1,OpName2,Enable,Timeout1,Timeout2,PosGuard2,NegGuard2,FilteredNegGuard2).
    % TO DO: return timeout results and compute timeout info here


cbc_enable_calc_aux2(OpName1,OpName2,Enable,Timeout1,Timeout2,PosGuard2,NegGuard2,FilteredNegGuard2) :-
      %format('Check if ~w can be enabled after ~w ~n',[OpName2,OpName1]),
      %print('  Pos Guard: '),translate:print_bexpr(PosGuard2),nl,
      %((OpName1=winc,OpName2=winc) -> trace ; true),
      testcase_path_timeout_catch(invariant,Timeout1,[OpName1],PosGuard2,_Csts,_Ops,_TestS,_TI,R), % advantage over version with [OpName1,OpName2] : one less state to setup and enumerate; but inner guards may not be checked
      !,
      print(can_be_enabled_after(R)), print(' : '),
         % print('Neg Guard: '),translate:print_bexpr(NegGuard2),nl,
         % TO DO: first check whether OpName2 can be disabled given Invariant ?
         (testcase_path_timeout_catch(invariant,Timeout1,[OpName1],NegGuard2,_Csts2,_Ops2,_TestS2,_TI2,R2)
           -> print(can_be_disabled_after(R2)), print(' : '),
            % TO DO: test if NegGuard2 holds initially it is possible to execute [OpName1,OpName2]; if not: OpName1 cannot enable OpName2, only keep it -> Enable=keep_possible
            % then we could check if OpName1 can disable OpName2: PosGuard2,[OpName1],NegGuard2
             ((OpName1\=OpName2, % otherwise OpName2 must be enabled if [OpName1] can be executed
                %testcase_path_timeout_catch(pred(NegGuard2),Timeout2,[OpName1,OpName2],b(truth,pred,[]),_,_,_,_,R3)
                testcase_path_timeout_catch(pred(NegGuard2),Timeout2,[OpName1],PosGuard2,_,_,_,_,R3))
               -> print(can_enable(R3)), print(' : '),
                 %nl, translate:print_bexpr(PosGuard2),nl,print(OpName1),nl,translate:print_bexpr(NegGuard2),nl,
                 %print('FILTERED: '), translate:print_bexpr(FilteredNegGuard2),nl,
                  (testcase_path_timeout_catch(pred(PosGuard2),Timeout2,[OpName1],FilteredNegGuard2,_,_,_,_,R4)
                  -> print(can_disable(R4)), print(' : '),
                     (contains_timeout([R,R2,R3,R4]) -> Enable=timeout_possible ; Enable=possible)
                  ;  (contains_timeout([R,R2,R3]) -> Enable=timeout_possible_enable ; Enable=possible_enable) /* Opname cannot disable OpName2; only enable it */
                  )
               ; /* OpName1 cannot enable OpName2; only preserve it */
                  (testcase_path_timeout_catch(pred(PosGuard2),Timeout2,[OpName1],FilteredNegGuard2,_,_,_,_,R4)
                  -> print(can_disable(R4)), print(' : '),
                     (contains_timeout([R,R2,R4]) -> Enable=timeout_possible_disable; Enable=possible_disable)
                  ;  (contains_timeout([R,R2]) -> Enable=timeout_keep; Enable=keep) /* Opname cannot enable or disable */
                  )
              )
            ; (is_timeout(R) -> Enable=timeout_guaranteed; Enable=guaranteed)
        ).
cbc_enable_calc_aux2(OpName1,_OpName2,Enable,_Timeout1,Timeout2,PosGuard2,_NegGuard2,FilteredNegGuard2) :-
        % OpName2 can never be enabled after OpName1; check whether it can be enabled before
        % Note: we could replace FilteredNegGuard2 by truth
        testcase_path_timeout_catch(pred(PosGuard2),Timeout2,[OpName1],FilteredNegGuard2,_,_,_,_,R4),
        !,
        print(can_disable(R4)), print(' : '),
        (is_timeout(R4) -> Enable=timeout_impossible; Enable=impossible).
cbc_enable_calc_aux2(_OpName1,_OpName2,Enable,_,_,_PosGuard2,_NegGuard2,_FilteredNegGuard2) :-
        % OpName2 can never ben enabled after nor before
        Enable=impossible_keep.


% check if contains timeout or similar result
contains_timeout(List) :- (member(timeout,List) -> true ; member(clpfd_overflow,List) -> true).


:- use_module(bsyntaxtree,[conjunction_to_list/2,some_id_occurs_in_expr/2,create_negation/2]).
% remove all predicates which are not modified
filter_predicate(Pred,ModifiedVars,FilteredPred) :- partition_predicate(Pred,ModifiedVars,FilteredPred,_).

% partition all predicates into those which are not modified and those that can be
partition_predicate(Pred,'$all',FilteredPred,Rest) :- !, FilteredPred=Pred, Rest=b(truth,pred,[]).
partition_predicate(Pred,ModifiedVars,FilteredPred,Rest) :-
    conjunction_to_list(Pred,List),
    partition(can_be_modified(ModifiedVars),List,FilteredList,[],UnFilteredList),
    conjunct_predicates(FilteredList,FilteredPred),
    conjunct_predicates(UnFilteredList,Rest).

can_be_modified(ModifiedVars,Pred,Res) :-
   some_id_occurs_in_expr(ModifiedVars,Pred),!, Res='<'.
can_be_modified(_,_,'>').

%predicate_modified_by_write_vars(_Pred,'$all') :- !.
%predicate_modified_by_write_vars(Pred,SortedWriteVars) :- some_id_occurs_in_expr(SortedWriteVars,Pred).


:- use_module(preferences,[get_time_out_preference_with_factor/2, add_time_outs/3]).
tcltk_cbc_enabling_analysis(L) :-
   get_time_out_preference_with_factor(0.1,TOF),
   (TOF > 250 -> ExtraTO is TOF-250 ; ExtraTO = 0),
   tcltk_cbc_enabling_analysis(ExtraTO,L).
tcltk_cbc_enabling_analysis(ExtraTimeout,list([list(['Origin'|Ops])|Result])) :-
  findall(Op, b_top_level_operation(Op), Ops),
  %statistics(walltime,[WT1,_]),
  findall(list([OpName1|EnableList]),
     (init_or_op(OpName1),
      findall(Enable,cbc_enable_analysis(OpName1,_OpName2,Enable,ExtraTimeout),EnableList)),
     Result).
  %statistics(walltime,[WT2,_]), Time is WT2-WT1,
  %format('Runtime for enabling analysis: ~w ms~n',[Time])
  %,print_enable_table([list(['Origin'|Ops])|Result]).

% -----------------
% compute the four edges of the enable relation for a particular operation
tcltk_cbc_enabling_relations_for_operation(OpName2,ExtraTimeout,
      list([list(['Event','Enable','KeepEnabled','Disable','KeepDisabled'])|Results])) :-
    findall(list([OpName1|ResOp2]),
          compute_cbc_enable_rel(OpName1,OpName2,ExtraTimeout,ResOp2), Results).
tcltk_cbc_enabling_relations_after_operation(OpName1,ExtraTimeout,
      list([list(['Event','Enable','KeepEnabled','Disable','KeepDisabled'])|Results])) :-
    findall(list([OpName2|ResOp2]),
          compute_cbc_enable_rel(OpName1,OpName2,ExtraTimeout,ResOp2), Results).

% compute the four edges of the enable relation for a pair of events:
compute_cbc_enable_rel(OpName1,OpName2,ExtraTimeout,[Enable,KeepEnabled,Disable,KeepDisabled]) :-
    b_top_level_operation(OpName1),
    b_get_operation_read_guard_vars(OpName1,true,Reads1),
    format("CHECKING ENABLING AFTER: ~w r:~w / w:~w~n",[OpName1,Reads1,Writes1]),
    b_top_level_operation(OpName2),
    get_negated_guard(OpName2,PosGuard2,_NegGuard2),
    partition_predicate(PosGuard2,Writes1,FilteredPosGuard2,StaticPosGuard2),
    create_negation(FilteredPosGuard2,FilteredNegGuard2),
    Timeout1 is 300+ExtraTimeout,
    format('~w ---> ~w  :: ',[OpName1,OpName2]),
    %print('     FPOS: '),translate:print_bexpr(FilteredPosGuard2),nl,
    %print('Neg Guard: '),translate:print_bexpr(NegGuard2),nl,
    %print(' Filtered: '),translate:print_bexpr(FilteredNegGuard2),nl,
    % TO DO: use syntactic conditions
    my_testcase_path_timeout(pred(PosGuard2),Timeout1,[OpName1],FilteredPosGuard2,KeepEnabled),
    my_testcase_path_timeout(pred(PosGuard2),Timeout1,[OpName1],FilteredNegGuard2,Disable),
    (OpName1==OpName2 -> KeepDisabled=false, Enable=false
     ; conjunct_predicates([StaticPosGuard2,FilteredNegGuard2],NegEnableGuard2), % as StaticPosGuard2 must be same before after: it must be true before to be true after
       my_testcase_path_timeout(pred(NegEnableGuard2),Timeout1,[OpName1],FilteredPosGuard2,Enable),
       % this is the unoptimised call:
       %my_testcase_path_timeout(pred(NegGuard2),Timeout1,[OpName1],NegGuard2,KeepDisabled),
       % we do two calls for KeepDisabled: pred(NegStaticGuard2) true after, or StaticGuard2 & NegFiltered -> NegFiltered
       create_negation(StaticPosGuard2,StaticNegGuard2),
       my_testcase_path_timeout(pred(StaticNegGuard2),Timeout1,[OpName1],b(truth,pred,[]),KeepDisabled1),
       (KeepDisabled1 = ok -> KeepDisabled=ok
         ; my_testcase_path_timeout(pred(NegEnableGuard2),Timeout1,[OpName1],FilteredNegGuard2,KeepDisabled)
        ),
       format(' KeepEnabled=~w, Disable=~w, Enable=~w, KeepDisabled=~w~n',[KeepEnabled,Disable,Enable,KeepDisabled])
    ).

my_testcase_path_timeout(Before,Timeout,Path,After,Result) :-
    (testcase_path_timeout_catch(Before,Timeout,Path,After,_,_,_,_,Result) -> true
      ; Result = false).

% ----------------------------------------------

% copy of predicat ein static_analysis to avoid meta_predicate error in Spider:
:- meta_predicate catch_enumeration_warning(0,0).
catch_enumeration_warning(Call,Handler) :-
  % throw/1 predicate raises instantiation_error
  on_exception(enumeration_warning(enumerating(_),_Type,_,_,critical),Call,call(Handler)).

:- use_module(sap,[testcase_path_timeout/9]).
testcase_path_timeout_catch(Pred,TIMEOUT,Seq,P2,Csts,Ops,TestS,TI,Res) :-
  catch_enumeration_warning(sap:testcase_path_timeout(Pred,TIMEOUT,Seq,P2,Csts,Ops,TestS,TI,Res),
         Res=virtual_time_out).

init_or_op('INITIALISATION').
init_or_op(OpName1) :- b_top_level_operation(OpName1). %b_get_machine_operation(OpName1,_,_,_).

% print table in CSV format
print_enable_table(R) :- print_enable_table_list(R).
print_enable_table_list([]) :- nl,nl.
print_enable_table_list([list(L)|T]) :-
   print_csv_list(L), print_enable_table_list(T).
print_csv_list([]) :- !,nl.
print_csv_list([H]) :- !,print(H),nl.
print_csv_list([H|T]) :- !,print(H), print(','), print_csv_list2(T).
print_csv_list(X) :- print(illegal_list(X)),nl, add_internal_error('Illegal list: ', print_csv_list(X)).

print_csv_list2([H]) :- !,print_enable_info(H),nl.
print_csv_list2([H|T]) :- !,print_enable_info(H), print(','), print_csv_list2(T).
print_csv_list2(X) :- print(illegal_list(X)),nl, add_internal_error('Illegal list: ', print_csv_list(X)).

print_enable_info(I) :- (translate_enable_info(I,TI) -> print(TI) ; print_for_csv(I)).
% simplify result further for output to CSV; reason: also testing (e.g., test 1360 which sometimes causes overflow on Mac, and time_out on Linux)
translate_enable_info(timeout,unknown).
translate_enable_info(time_out,unknown).
translate_enable_info(overflow,unknown).
translate_enable_info(virtual_time_out,unknown).

print_for_csv([]) :- !, print('{}').
print_for_csv([H|T]) :- !, % print list without commas ,
   format('{~w',[H]), maplist(print_csv_el,T), print('}').
print_for_csv(Term) :- print(Term).
print_csv_el(H) :- format(' ~w',[H]).


% ---------- create enable graph ------------------

eop_node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,black) :-
     op_node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,_Color).
eop_node_predicate('INITIALISATION',SubGraph,'INITIALISATION',Shape,Style,Color) :- SubGraph=none,
     Shape=hexagon, Style=none, Color=olivedrab2.

translate_enable_res(Rel,_,_,'impossible',red,_) :- impossible(Rel),!, fail.
translate_enable_res(guaranteed,_,_,'guaranteed',olivedrab2,solid) :- !.
translate_enable_res(Rel,X,Y,TransRel,TCol,bold) :- independent(Rel,Col),
   !,X=Y, % only show if source & dest identical
   TransRel=Rel,TCol=Col.
translate_enable_res(possible,_,_,'possible',black,solid) :- !.
translate_enable_res(keep,_,_,'keep',lightgray,solid) :- !.
translate_enable_res(possible_enable,_,_,'(enable)',black,solid) :- !.
translate_enable_res(possible_disable,_,_,'(disable)',lightblue,solid) :- !.
translate_enable_res(timeout_keep,_,_,'keep?',tomato,dashed) :- !.
translate_enable_res(timeout_possible,_,_,'possible?',tomato,dashed) :- !.
translate_enable_res(timeout_possible_disable,_,_,'(disable?)',tomato,dashed) :- !.
translate_enable_res(timeout_possible_enable,_,_,'(enable?)',tomato,dashed) :- !.
translate_enable_res(pred(Expr),_,_,String,black,dashed) :- !,
    translate_bexpression(Expr,String).
translate_enable_res(predicate(Expr),_,_,String,black,dashed) :- !,
    translate_bexpression(Expr,String).
translate_enable_res(dependent,_,_,dependent,green,solid) :- !.
translate_enable_res(race_dependent,_,_,race_dependent,green,solid) :- !.
translate_enable_res(X,_,_,'independent',lightgray,solid) :-
    memberchk(X,[independent,syntactic_independent,syntactic_fully_independent,syntactic_unchanged]),!.
translate_enable_res(X,_,_,X,red,solid).

impossible(impossible).
impossible(impossible_keep).
impossible(impossible_disable).
impossible(impossible_infeasible).
impossible(infeasible).

independent(independent,yellow).
independent(syntactic_keep,lightgray).
independent(syntactic_independent,lightgray).
independent(syntactic_fully_independent,lightgray).
independent(syntactic_unchanged,lightgray).
independent(activation_indepdendent,lightgray).

is_timeout_enabling_result(timeout_keep).
is_timeout_enabling_result(timeout_possible).
is_timeout_enabling_result(timeout_possible_disable).
is_timeout_enabling_result(timeout_possible_enable).
is_timeout_enabling_result(timeout).
is_timeout_enabling_result(timeout_dependent).
is_timeout_enabling_result(timeout_guaranteed).
is_timeout_enabling_result(timeout_impossible).

% Simultaneous Enabling Analysis
% Find events which cannot be enabled simultaneously
% Usage:
%  a) if one event is known to be enabled we do not need to check the impossible ones
%  b) if one event is known to be disabled we do not need check events which guarantee it

% use_module(enabling_analysis), tcltk_cbc_simultaneous_enabling_analysis(R).

tcltk_cbc_simultaneous_enabling_analysis(list([list(['Origin'|SOps])|Result])) :-
  findall(Op, b_top_level_operation(Op), Ops),
  sort(Ops,SOps), % sort operations to ensure nice triangular display
  statistics(walltime,[WT1,_]),
  findall(list([OpName1|EnableList]),
     (member(OpName1,SOps),
      findall(SimulRes,tcltk_simult(OpName1,SOps,SimulRes),EnableList)),
     Result),
  statistics(walltime,[WT2,_]), Time is WT2-WT1,
  format('Runtime for simultaneous enabling analysis: ~w ms~n',[Time])
  . %,print_enable_table([list(['Origin'|Ops])|Result]).

tcltk_simult(Op1,SOps,Res) :- member(Op2,SOps),tcltk_simult2(Op1,Op2,Res).
tcltk_simult2(Op1,Op2,Res) :-
  Op1 = Op2, % Op2 @=< Op1, % the simultaneous impossible analysis is symmetric:  TO DO: do not compute/display info twice
  !,
   Res = '-'.
tcltk_simult2(Op1,Op2,TRes) :-
   simult_enable_analysis(Op1,Op2,1,fixed_time_out(150),TRes).

simult_enable_analysis(OpName1,OpName2,UseInvariant,TimeoutFactor,Res) :-
    b_top_level_operation(OpName1),
    b_top_level_operation(OpName2),
    %OpName2 @>OpName1,
	get_negated_guard(OpName1,PosGuard1,_NegGuard1),
    get_negated_guard(OpName2,PosGuard2,NegGuard2),
	get_conj_inv_predicate([PosGuard1,PosGuard2],UseInvariant,Pred),
	%print(checking_simultaneous(OpName1,OpName2)),nl,
	(solve_predicate_with_chr(Pred,_State,TimeoutFactor,[],RR)
	  -> %print(state(_State)),nl, print(result(RR)),nl,
	     ((RR = contradiction_found ; RR = no_solution_found(unfixed_deferred_sets))
	      -> Res = impossible % we do not compute PosGuard1,NegGuard2 in this case: otherwise this would mean that Operation 1 can never fire !
	      ;  get_conj_inv_predicate([PosGuard1,NegGuard2],UseInvariant,Pred2),
	         solve_predicate_with_chr(Pred2,_State2,TimeoutFactor,[],RR2),
	         combine_simult_result(RR,RR2,Res))
	   ; print('********* ERROR ***********'),nl,fail).

combine_simult_result(solution(_),solution(_),R) :- !,R=possible.
combine_simult_result(solution(_),B,R) :-
    (B = contradiction_found ; B = no_solution_found(unfixed_deferred_sets)), !,
    R=guaranteed.
combine_simult_result(A,B,Res) :- functor(A,FA,_), functor(B,FB,_),
   ajoin([FA,'_',FB],Res).

% ---------------------------------------------------

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(specification_initialised,reset_enabling_analysis,
                           'Initialise module enabling analysis.').

:- dynamic feasible_operation_cache/2.
reset_enabling_analysis :-
   %retractall(disable_graph(_)),
   retractall(feasible_operation_cache(_,_)),
   retractall(cbc_enable_analysis_cache(_,_,_,_)).

% in principle you should only call this for Op with operation_name_not_yet_covered(Op)
feasible_operation(Op,Res) :- feasible_operation_cache(Op,CR),!,Res=CR.
feasible_operation(Op,Res) :- UseInvariant=1,
   check_if_feasible_operation(Op,UseInvariant,fixed_time_out(500),CR,_),
   assert(feasible_operation_cache(Op,CR)),
   Res=CR.

feasible_operation_with_timeout(Op,TimeOut,Res) :- feasible_operation_cache(Op,CR),
   (CR \= unknown ; TimeOut =< 500), !, Res=CR.
feasible_operation_with_timeout(Op,TimeOut,Res) :- check_if_feasible_operation(Op,1,fixed_time_out(TimeOut),CR,_),
   retractall(feasible_operation_cache(Op,_)), % must be unknown if it was asserted before
   assert(feasible_operation_cache(Op,CR)),
   Res=CR.


:- use_module(state_space,[operation_not_yet_covered/1]).
infeasible_operation(Op) :- feasible_operation_cache(Op,CR),!, CR=impossible.
infeasible_operation(Op) :- operation_not_yet_covered(Op),
   feasible_operation(Op,impossible). % will also assert feasible_operation_cache

infeasible_operation_cache(Op) :- feasible_operation_cache(Op,impossible).

:- use_module(solver_interface, [solve_predicate/5]).
solve_predicate_with_chr(Pred,State,TimeoutFactor,Options,RR) :-
    solve_predicate(Pred,State,TimeoutFactor,['CHR','CLPFD','SMT'|Options],RR).

:- use_module(debug, [debug_mode/1]).
:- use_module(store,[normalise_store/2]).
% check if an operation is feasible
% UseInvariant = 1 : assume invariant
check_if_feasible_operation(OpName1,UseInvariant,TimeoutFactor,Res,NormalisedResultState) :-
    b_top_level_operation(OpName1),
	(debug_mode(on)
	 -> print(checking_feasibility(OpName1)),nl,tools:start_ms_timer(TIMER) ; true),
	if(check_if_feasible_operation2(OpName1,UseInvariant,TimeoutFactor,Res,State),true,
	  (add_internal_error('Failed:',check_if_feasible_operation2(OpName1,UseInvariant,TimeoutFactor,Res,State)),fail)),
	% TO DO: add missing variables and put into order
	if(normalise_store(State,NormalisedResultState),true,
	   NormalisedResultState=State),
	%b_interpreter:sort_variable_binding(NormalisedResultState,SortedStore),
	(debug_mode(on) -> tools:stop_ms_timer_with_msg(TIMER,feasible(OpName1)) ; true).


:- use_module(tools_timeout,[get_time_out_with_factor/2]).
check_if_feasible_operation2(OpName1,UseInvariant,TimeoutFactor,Res,ResultState) :-
	get_negated_guard(OpName1,PosGuard1,_NegGuard1,Precise),
	% format('Guard is ~w for ~w~n',[Precise,OpName1]),
	Precise = precise,
	!,
	% Warning: guard is under approximation for sequential composition, while loops, ...
	get_conj_inv_predicate([PosGuard1],UseInvariant,Pred),
	(debug_mode(on) -> translate:print_bexpr(Pred),nl ; true),
	% visualize_graph:print_predicate_dependency_as_graph_for_dot(Pred,'~/Desktop/out.dot'),
	(solve_predicate_with_chr(Pred,State,TimeoutFactor,[full_machine_state],RR)
	  -> %print(state(_State)),nl, print(result(RR)),nl,
	     ((RR=no_solution_found(unfixed_deferred_sets) ; RR = contradiction_found)
	      -> Res = impossible, ResultState = '$UNKNOWN_STATE'
	         %,format('Computing unsat core for infeasible ~w~n',[OpName1]),unsat_cores:unsat_core(Pred,Core),print('CORE: '),nl,translate:nested_print_bexpr_as_classicalb(Core),nl
	      ;  RR=solution(_) -> Res = possible,
	         (debug_mode(on) -> print('SOLUTION: '),translate:print_bstate(State),nl ; true),
	         ResultState = State
	      ;  simplify_for_output(RR,Res), ResultState = '$UNKNOWN_STATE' )
	   ; add_internal_error('Failed to compute feasibility for: ',OpName1),
	     translate:print_bexpr(Pred),nl,
	     Res = internal_error, ResultState = '$UNKNOWN_STATE').
check_if_feasible_operation2(OpName1,UseInvariant,TimeoutFactor,Res,ResultState) :-
  format('Guard computation is imprecise for ~w. Not using predicate solver (but B interpreter).~n',[OpName1]),
  % this version uses sap:testcase_path: this works for things like sequential composition and WHILE loops better
  get_time_out_with_factor(TimeoutFactor,TO),
  (UseInvariant=1 -> INV=invariant ; INV=typing),
  (testcase_path_timeout_catch(INV,TO,[OpName1],b(truth,pred,[]),_Constants,Ops,StateSequence,_TI,R1)
    -> (is_timeout(R1) -> Res=timeout ; Res = possible),
       (nonvar(StateSequence),StateSequence=[ResultState|_] -> true ; ResultState = '$UNKNOWN_STATE'),
       (debug_mode(off) -> true
         ; print(feasibility_result(OpName1,R1,Ops)),nl,
           (ResultState = '$UNKNOWN_STATE' -> true
             ; print('SOLUTION: '),translate:print_bstate(ResultState),nl)
        )
    ; Res=impossible,ResultState = '$UNKNOWN_STATE').


simplify_for_output(no_solution_found(enumeration_warning(_,_,_,_,_)),virtual_time_out).
simplify_for_output(no_solution_found(clpfd_overflow),overflow).
simplify_for_output(time_out,time_out).

