:- module(input_syntax_tree, [remove_raw_position_info/2
                             ,get_raw_position_info/2
                             ,extract_raw_identifiers/2
                             ,get_definition_name/2
                             ,raw_replace/3
                             ,raw_conjunct/2
                             ,create_fresh_identifier/3
                             ]).


:- use_module(error_manager).

get_raw_position_info(Compound,Info) :-
   compound(Compound),
   functor(Compound,_,Arity), Arity>0,
   arg(1,Compound,Info),
   Info = pos(_,_,_,_,_,_),!.
get_raw_position_info(Compound,Info) :-
   add_internal_error('Could not get position info: ',get_raw_position_info(Compound,Info)),
   Info = unknown.

:- use_module(tools,[safe_functor/4]).

remove_raw_position_info([],Output) :- !,Output=[].
remove_raw_position_info([I|Irest],[O|Orest]) :-
    !,remove_raw_position_info(I,O),
    remove_raw_position_info(Irest,Orest).
remove_raw_position_info(Input,Output) :-
    compound(Input),functor(Input,Functor,Arity),Arity>0,!,
    Arity2 is Arity-1,
    functor(Output,Functor,Arity2),
    remove_raw_position_info2(1,Arity,Input,Output).
remove_raw_position_info(Input,Input).


remove_raw_position_info2(Pos,Arity,Input,Output) :-
    Pos < Arity,!,Pos2 is Pos+1,
    arg(Pos2,Input,IA), arg(Pos,Output,OA),
    remove_raw_position_info(IA,OA),
    remove_raw_position_info2(Pos2,Arity,Input,Output).
remove_raw_position_info2(_Pos,_Arity,_Input,_Output).

:- use_module(library(ordsets),[ord_union/3, ord_member/2]).
extract_raw_identifiers(Ast,Ids) :-
    extract_raw_identifiers2(Ast,[],Unsorted,[]),
    sort(Unsorted,Ids).
extract_raw_identifiers2([],_) --> !,[].
extract_raw_identifiers2([I|Irest],Bound) -->
    !,extract_raw_identifiers2(I,Bound),
    extract_raw_identifiers2(Irest,Bound).
extract_raw_identifiers2(identifier(_,I),Bound) --> !,add_if_free(I,Bound).
extract_raw_identifiers2(definition(_,I,Args),Bound) --> !,
    add_if_free(I,Bound), extract_raw_identifiers2(Args,Bound).
extract_raw_identifiers2(Input,Bound) -->
    {raw_quantifier(Input,Ids,Subs)},!,
    %{append(Ids,Bound,NewBound)},
    {ord_union(Ids,Bound,NewBound)},
    extract_raw_identifiers2(Subs,NewBound).
extract_raw_identifiers2(Input,Bound) -->
    {compound(Input),functor(Input,_Functor,Arity),Arity>1,!},
    extract_raw_identifiers3(2,Arity,Input,Bound).
extract_raw_identifiers2(_,_) --> [].

add_if_free(Id,Bound,In,Out) :-
    ( ord_member(Id,Bound) -> In=Out
    ; otherwise           -> In=[Id|Out]).

extract_raw_identifiers3(Pos,Arity,Input,Bound) -->
    {Pos=<Arity,!,Pos2 is Pos+1, arg(Pos,Input,Arg)},
    extract_raw_identifiers2(Arg,Bound),
    extract_raw_identifiers3(Pos2,Arity,Input,Bound).
extract_raw_identifiers3(_Pos,_Arity,_Input,_Bound) --> [].

:- use_module(library(lists),[maplist/3]).
raw_quantifier(Expr,SortedIds,Subs) :-
    raw_quantifier_aux(Expr,RawIds,Subs),
    maplist(raw_id,RawIds,Ids),
    sort(Ids,SortedIds).
raw_id(identifier(_,ID),ID).

raw_quantifier_aux(forall(_,Ids,P),Ids,[P]).
raw_quantifier_aux(exists(_,Ids,P),Ids,[P]).
raw_quantifier_aux(comprehension_set(_,Ids,P),Ids,[P]).
raw_quantifier_aux(event_b_comprehension_set(_,Ids,E,P),Ids,[E,P]).
raw_quantifier_aux(lambda(Ids,P,E),Ids,[P,E]).
raw_quantifier_aux(general_sum(_,Ids,P,E),Ids,[P,E]).
raw_quantifier_aux(general_product(_,Ids,P,E),Ids,[P,E]).
raw_quantifier_aux(quantified_union(_,Ids,P,E),Ids,[P,E]).
raw_quantifier_aux(quantified_intersection(_,Ids,P,E),Ids,[P,E]).
raw_quantifier_aux(any(_,Ids,P,S),Ids,[P,S]).
raw_quantifier_aux(let(_,Ids,P,S),Ids,[P,S]).
raw_quantifier_aux(var(_,Ids,S),Ids,[S]).
raw_quantifier_aux(recursive_let(_,Id,E),[Id],[E]).

get_definition_name(expression_definition(_Pos,Name,_Args,_Body),Name).
get_definition_name(predicate_definition(_Pos,Name,_Args,_Body),Name).
get_definition_name(substitution_definition(_Pos,Name,_Args,_Body),Name).

% raw_replace(+InputAST,-Replaces,+OutputAST):
% InputAST:  A raw AST as it comes from the parser (without type information)
% Replaces:  A list of "replace(Id,Expr)" where Id the the identifier which
%            should be replaced by the expression Expr
% OutputAST: A raw AST where all matching identifiers are replaced by their
%            expression
% used for rewriting DEFINITION calls
raw_replace(Expr,[],Expr) :- !.
raw_replace(Old,Replaces,New) :-
    get_introduced_ids(Replaces,IntroducedIds), %print(intro(IntroducedIds)),nl,
    raw_replace2(Old,Replaces,IntroducedIds,New).
raw_replace2([],_,_,[]) :- !.
raw_replace2([E|Rest],Replaces,IntroducedIds,[N|NRest]) :- !,
    raw_replace2(E,Replaces,IntroducedIds,N),
    raw_replace2(Rest,Replaces,IntroducedIds,NRest).
raw_replace2(rewrite_protected(X),_Replaces,_,rewrite_protected(Y)) :- !,X=Y.
raw_replace2(identifier(_,Old),Replaces,_,New) :-
    memberchk(replace(Old,New),Replaces),!.
raw_replace2(Expr,_,IntroducedIds,_) :-
    quantifier_capture_warning(Expr,IntroducedIds),
    fail.
raw_replace2(Expr,Replaces,IntroducedIds,New) :-
    safe_functor(raw_replace2_expr,Expr,Functor,Arity),
    safe_functor(raw_replace2_new,New,Functor,Arity),
    raw_replace3(Arity,Expr,Replaces,IntroducedIds,New).
raw_replace3(0,_,_,_,_) :- !.
raw_replace3(N,Expr,Replaces,IntroducedIds,NExpr) :-
    arg(N,Expr,Old),
    arg(N,NExpr,New),
    raw_replace2(Old,Replaces,IntroducedIds,New),
    N2 is N-1,
    raw_replace3(N2,Expr,Replaces,IntroducedIds,NExpr).

:- use_module(library(ordsets),[ord_union/2, ord_intersection/3, ord_intersect/2]).
% this can happen e.g. for egt(x) == (#y.(y:NAT1 & y<100 & x<y)) when calling egt(y+1), i.e., replacing x by y+1 in body of DEFINITION
quantifier_capture_warning(Expr,IntroducedIds) :-
    raw_quantifier(Expr,SortedIds,_Body),
    ord_intersection(IntroducedIds,SortedIds,Clash),
    Clash \= [],
    % TO DO?: check if we really use a corresponding LHS identifier in Body: extract_raw_identifiers(Body,UsedIds), maplist(get_lhs_id,Replaces,LHSIds),
    %format('Quantifier ids: ~w~nIntroduced Ids: ~w~n',[SortedIds,IntroducedIds]),
    arg(1,Expr,Pos),
    add_warning(definition_variable_capture,'Quantifier may capture identifiers in DEFINITION arguments: ',Clash,Pos).

% compute which ids are introduced by performing the replacements
get_introduced_ids(Replaces,Ids) :- maplist(get_rhs_ids,Replaces,ListOfList),
   ord_union(ListOfList,Ids).
get_rhs_ids(replace(_,New),Ids) :- extract_raw_identifiers(New,Ids).
%get_lhs_id(replace(Id,_),Id).

raw_conjunct([],truth(none)).
raw_conjunct([H|T],R) :-
    ( T=[] -> H=R
    ; otherwise ->
        R=conjunct(none,H,RT),
        raw_conjunct(T,RT)
    ).

%raw_occurs(ID,Term) :- raw_id(Term,ID).
%raw_occurs(ID,Term) :- compound(Term), Term =.. [_Func,_Pos|Args],
%   member(A,Args), raw_occurs(ID,A).


% create_fresh_identifier(+PreferredName,+Ast,-Name):
%   Generates an identifier that does not occur in Ast
%   PreferredName: An atom with the preferred name of the identifier
%   Ast: An untyped AST
%   Name: An identifier (an atom) that does not occur free in AST,
%      If PreferredName does not occur free in AST, it's 'PreferredName',
%      otherwise it's 'PreferredName_N' where N is a natural number
create_fresh_identifier(PreferredName,Ast,Name) :-
    extract_raw_identifiers(Ast,UsedIds),
    ( memberchk(PreferredName,UsedIds) ->
        % Allready in use, find a fresh one by adding a number
        atom_codes(PreferredName,PreferredCodes),
        create_fresh_identifier_aux(PreferredCodes,UsedIds,1,Name)
    ; otherwise ->
        Name = PreferredName).
create_fresh_identifier_aux(PreferredCodes,UsedIds,I,Name) :-
    number_codes(I,ICodes),append(PreferredCodes,[95|ICodes],Codes),
    atom_codes(NewName,Codes),
    ( memberchk(NewName,UsedIds) ->
        % Allready in use, try again with a higher number
        I2 is I+1,
        create_fresh_identifier_aux(PreferredCodes,UsedIds,I2,Name)
    ; otherwise ->
        Name = NewName).
