% (c) 2012-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_strings, [empty_string/1,
                           string_length/2,
                           string_to_int_wf/4,
                           int_to_string/2,
                           int_to_dec_string/3,
                           string_is_int/2,
                           to_string/2,
                           string_append_wf/4,
                           string_split/3,
                           string_join/3,
                           string_chars/2,
                           string_codes/2,
                           format_to_string/3, convert_b_sequence_to_list_of_atoms/3,

                           % utilities:
                           split_atom_string/3,
                           generate_code_sequence/3
  ]).

% Strings in ProB are represented by terms of the form string(PrologAtom)

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(library(lists)).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides (external) functions to manipulate B strings.').

empty_string(string('')).

:- use_module(kernel_objects,[exhaustive_kernel_succeed_check/1,exhaustive_kernel_fail_check/1,
   exhaustive_kernel_check_wf/2,exhaustive_kernel_fail_check_wf/2, exhaustive_kernel_fail_check_wfdet/2]).

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_length(string(''),int(0)))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_length(string('a'),int(1)))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_length(string('aa'),int(2)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:string_length(string('a'),int(0)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:string_length(string('a'),int(2)))).

:- use_module(kernel_objects,[greater_than_equal/2]).

:- block string_length(-,-).
string_length(SA,int(L)) :-
   greater_than_equal(int(L),int(0)),
   string_len2(SA,L).
:- block string_len2(-,-).
string_len2(SA,L)  :-
   L==0,!,
   empty_string(SA).
string_len2(string(A),L) :-
   string_len3(A,L).
:- block string_len3(-,-).
string_len3(A,L)  :-
   L==0,
   !,
   empty_string_atom(A).
% in case A is not known and L=1 we could enumerate chars ??
string_len3(A,L) :-
   string_len4(A,L).
:- block string_len4(-,?).
string_len4(A,L) :-
   atom_length(A,LL), LL=L. % delay unification due to bug in SICStus atom_length
% bug in SICStus: dif(X,1), atom_length(a,X) succeeds in 4.2.0 and 4.2.1


% ----------------------------


:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_to_int_wf(string('11'),int(11),unknown,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(kernel_strings:string_to_int_wf(string('11'),int(1),unknown,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(kernel_strings:string_to_int_wf(string('1'),int(11),unknown,WF),WF)).
:- assert_must_fail((kernel_strings:string_to_int_wf(S,int(11),u,WF),kernel_strings:string_to_int_wf(S,int(12),u,WF))).

:- block string_to_int_wf(-,-,?,?).
string_to_int_wf(string(S),int(I),Span,WF) :-
   string_to_int2(S,I,Span,WF).

:- block string_to_int2(-,-,?,?).
string_to_int2(S,Res,Span,WF) :- var(S), % we know the number; we cannot construct the string as leading spaces/0s are ok
   % with an injective string_to_int conversion we could invert the function
   !,
   frozen(S,Goal),  %print(s2int(S,Res,Goal)),nl,
   (incompatible_goal(Goal,S,Res) ->  %print(prune(S,Res)),nl,
     fail ; true),
   strint_to_int3(S,Res,Span,WF).
string_to_int2(S,Res,Span,WF) :- strint_to_int3(S,Res,Span,WF).

:- use_module(kernel_waitflags,[add_wd_error_set_result/6]).
:- block strint_to_int3(-,?,?,?).
strint_to_int3(S,Res,Span,WF) :-
    atom_codes(S,C),
    on_exception(error(syntax_error(_),_),
         integer_number_codes(C,S,Res,Span,WF),
         add_wd_error_set_result('### Could not convert string to integer: ',S,Res,0,Span,WF)).
         %add_error_and_fail(external_functions,'### Could not convert string to integer: ',S)),

integer_number_codes(C,S,Res,Span,WF) :-
  number_codes(Num,C),
  (integer(Num) -> Res=Num
   ; %add_error_and_fail(external_functions,'### String represents a floating point number (expected integer): ',S)).
     add_wd_error_set_result('### String represents a floating point number (expected integer): ',S,Res,0,Span,WF)).

% check if another pending co-routine transforms the same string into another number
incompatible_goal((A,B),S,Res) :-
   (incompatible_goal(A,S,Res) -> true ; incompatible_goal(B,S,Res)).
incompatible_goal(kernel_strings:strint_to_int3(S2,Res2,_,_),S,Res) :-
   number(Res2),
    S==S2, Res2 \= Res.

% ----------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_string(int(0),string('0')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_string(int(10),string('10')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_string(int(-10),string('-10')))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:int_to_string(int(0),string('1')))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:int_to_string(int(1),string('01')))).

% difference to string_to_int: do not throw error when string cannot be converted to integer

:- block int_to_string(-,?).
int_to_string(int(I),S) :- int_to_string2(I,S).

:- block int_to_string2(-,?).
int_to_string2(Num,Res) :-
   number_codes(Num,C),
   atom_codes(S,C), Res=string(S).


:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string('0'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string('-10'),pred_true))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string('1267650600228229401496703205376'),pred_true))). %// 2^100
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string('1.0'),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string(''),pred_false))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_is_int(string('-'),pred_false))).

:- block string_is_int(-,?).
string_is_int(string(S),Res) :-
    string_is_int2(S,Res).

:- block string_is_int2(-,?).
string_is_int2(S,Res) :-
    atom_codes(S,C),
    on_exception(error(syntax_error(_),_),
           (number_codes(Num,C),
            (integer(Num)->Res=pred_true;Res=pred_false)),
         Res=pred_false).


:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(0),int(1),string('0.0')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(21),int(1),string('2.1')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(121),int(2),string('1.21')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(101),int(2),string('1.01')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(-101),int(2),string('-1.01')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(-101),int(3),string('-0.101')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(121),int(0),string('121')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:int_to_dec_string(int(121),int(-2),string('12100')))).

:- block int_to_dec_string(-,?,?), int_to_dec_string(?,-,?).
int_to_dec_string(int(I),int(Prec),S) :- int_to_dec_string2(I,Prec,S).

:- block int_to_dec_string2(-,?,?), int_to_dec_string2(?,-,?).
int_to_dec_string2(I,Prec,String) :-
    Prec=<0,
    !,
    IP is I * (10^abs(Prec)),
    int_to_string2(IP,String).
int_to_dec_string2(I,Prec,String) :- %Prec>0,
    PowTen is 10^Prec,
    IntVal is I // PowTen,
    number_codes(IntVal,IVC),
    ((IntVal=0, I<0) -> Prefix = [45|IVC] % need to add leading -
     ; Prefix = IVC),
    DecVal is abs(I) mod PowTen,
    number_codes(DecVal,DVC),
    length(DVC,Digits),
    NrZeros is Prec-Digits,
    length(Zeros,NrZeros),
    maplist(is_zero,Zeros),
    append(Zeros,DVC,Suffix),
    append(Prefix,[46|Suffix],Codes), % 46 is the dot .
    atom_codes(Atom,Codes),
    String = string(Atom).

is_zero(48). % ascii code of zero 0
% -------------------


:- use_module(kernel_tools,[ground_value_check/2]).

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:to_string(int(10),string('10')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:to_string(pred_true,string('TRUE')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:to_string([],string('{}')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:to_string(string('01'),string('01')))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:to_string(int(1),string('01')))).

:- block to_string(-,?).
to_string(int(I),S) :- !,
  int_to_string2(I,S).
to_string(string(S),Res) :- !,
  Res=string(S). % we already have a string; nothing needs to be done
to_string(Value,S) :- ground_value_check(Value,GrValue), to_string_aux(GrValue,Value,S).

:- block to_string_aux(-,?,?).
to_string_aux(_,Value,Str) :- to_string_aux(Value,Str).

:- use_module(preferences,[temporary_set_preference/3,reset_temporary_preference/2]).
:- use_module(translate,[translate_bvalue/2]).
% convert_to_avl
to_string_aux(Value,Str) :-
   normalise_value_for_to_string(Value,NValue),
   temporary_set_preference(expand_avl_upto,100000,CHNG),
   translate_bvalue(NValue,Atom),
   reset_temporary_preference(expand_avl_upto,CHNG),
   Str=string(Atom).


:- use_module(store,[normalise_value_for_var/3]).
% normalise_value_for_var normalises values for storing; for printing we need to do less work
% e.g., we do not need to normalise AVL values; we could add further cases for records ...
normalise_value_for_to_string(avl(A),R) :- !, R=avl(A).
normalise_value_for_to_string((A,B),R) :- !, R=(NA,NB),
    normalise_value_for_to_string(A,NA),
    normalise_value_for_to_string(B,NB).
normalise_value_for_to_string(A,R) :- normalise_value_for_var(to_string,A,R).



% -------------------

:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_append_wf(string('a'),string('b'),string('ab'),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_append_wf(string('0'),string('1'),string('01'),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_append_wf(string('aa'),string(''),string('aa'),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_strings:string_append_wf(string(''),string('aa'),string('aa'),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(kernel_strings:string_append_wf(string('0'),string('1'),string('1'),WF),WF)).

:- block string_append_wf(-,?,-,?), string_append_wf(?,-,-,?).
string_append_wf(string(A),string(B),string(C),WF) :- %print(app(A,B,C)),nl,
  app2(A,B,C,WF).
:- block app2(-,?,-,?), app2(?,-,-,?).
app2(A,B,C,_WF) :- %print(app2(A,B,C)),nl,
   nonvar(A),nonvar(B),!,atom_concat(A,B,CC), /* overcome bug in SICStus; delay unification */
   CC=C. %, print(app2(A,B,C)),nl.
app2(A,B,C,WF) :- atom_codes(C,CC), app3(A,B,CC,WF).

:- use_module(kernel_waitflags,[get_wait_flag/4]).
app3(A,B,CC,WF) :-
   (   nonvar(B) -> atom_codes(B,BB), append(AA,BB,CC), atom_codes(A,AA)
     ; nonvar(A) -> atom_codes(A,AA), append(AA,BB,CC), atom_codes(B,BB)
     ; length(CC,CLen), %print(enum(CC,CLen)),nl,
       get_wait_flag(CLen,'STRING_APPEND',WF,LWF),
       app4(A,B,CC,WF,LWF)
    ).
:- block app4(-,?,-,?,-), app4(?,-,-,?,-).
app4(A,B,CC,WF,_) :- (nonvar(A) ; nonvar(B)), !, % no need to enumerate
   app3(A,B,CC,WF).
app4(A,B,CC,_WF,_) :- %print(enumerating(CC)),nl,
   append(AA,BB,CC), % will be non-deterministic
   atom_codes(A,AA), atom_codes(B,BB).

% -------------------------------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_split(string('01001'),string('1'),[(int(1),string('0')),(int(2),string('00')),(int(3),string(''))]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_split(string('789'),string('1'),[(int(1),string('789'))]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_strings:string_split(string('aaa'),string('a'),[(int(1),string('a')),(int(2),string('a'))]))).

% function to split a string into a list of strings which were delimited by a separator
% WARNING: if the seperator is of length more than one, then first match-strategy will be used
:- block string_split(?,-,?), string_split(-,?,-).
string_split(string(A),string(B),R) :-
   string_split2(A,B,R).
:- block string_split2(?,-,?),string_split2(-,?,-).
string_split2(Atom,Seperator,SplitAtomList) :-
  (var(Atom) ; var(Seperator)),
  !, % currently : separator always known
  expand_custom_set_to_list(SplitAtomList,ExpandedSplitAtomList,Done,string_split2),
  string_split3(Atom,Seperator,SplitAtomList,Done,ExpandedSplitAtomList).
string_split2(Atom,Separator,SplitAtomList) :- % normal forward mode: atom and separator known:
  string_split_forward(Atom,Separator,SplitAtomList).

string_split_forward(Atom,Separator,SplitAtomList) :-
  split_atom_string(Atom,Separator,List), % safe_call ?
  convert_prolog_to_b_list(List,SplitAtomList).

:- block string_split3(?,-,?,?,?),string_split3(-,?,?,-,?). % we need to know the seperator: TO DO : improve this
string_split3(Atom,Seperator,SplitAtomList,Done,_ExpandedSplitAtomList) :-
   var(Done),
   !,
   string_split_forward(Atom,Seperator,SplitAtomList).
string_split3(Atom,Seperator,SplitAtomList,_Done,ExpandedSplitAtomList) :-
   ExpandedSplitAtomList \= [],  % split("",sep) --> [""] not the empty list
   sort(ExpandedSplitAtomList,SL),
   maplist(drop_index,SL,IL),
   convert_b_to_prolog_atoms(IL,PL,Done),
   atom_codes(Seperator,SepCodes),
   append(SepCodes,_,SepCodes2),
   split4(Done,SepCodes2,PL,Seperator,Atom,SplitAtomList).


:- use_module(probsrc(tools_strings),[ajoin/2]).

:- block split4(-,?,?,?,-,?), split4(-,?,?,-,?,?).
% unblock either when Done or when both Atom and Sperator known
split4(Done,_SepCodes2,_PL,Seperator,Atom,SplitAtomList) :-
   var(Done),
   !,
   % we can now compute forwards anyhow; ignore backwards direction
   string_split_forward(Atom,Seperator,SplitAtomList).
split4(_,SepCodes2,PL,Seperator,Atom,_) :-
      maplist(not_suffix_atom(SepCodes2),PL), % check that seperator occurs in no split atom: e.g. STRING_SPLIT(r,"_") = ["a","_","c"] should fail
       insert_sep(PL,Seperator,PL2),
       ajoin(PL2,Atom).

insert_sep([],_,[]).
insert_sep([H],_,R) :- !, R=[H].
insert_sep([H|T],Sep,[H,Sep|IT]) :- insert_sep(T,Sep,IT).

not_suffix_atom(SepCodes,Atom) :- \+ suffix_atom(SepCodes,Atom).
suffix_atom(SepCodes,Atom) :-
   atom_codes(Atom,AL),
   append(_,SepCodes,AL).

% ------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_join([(int(1),string('0')),(int(2),string('00')),(int(3),string(''))],string('1'),string('01001')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_join([(int(1),string('0')),(int(2),string('00'))],string('1'),string('0100')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_join([(int(1),string('0'))],string('1'),string('0')))).

:- block string_join(?,-,?), string_join(-,?,?).
string_join(SplitAtoms,string(Sep),Res) :- string_join2(SplitAtoms,Sep,Res).

% this is not reversible  ["a","b"],"_"  ->  "a_b"  but ["a_b"],"_" -> "a_b" has same result
:- block string_join2(?,-,?),string_join2(-,?,?).
string_join2(SplitAtomList,Seperator,Result) :-
  expand_custom_set_to_list(SplitAtomList,ExpandedSplitAtomList,Done,string_split2),
  string_join3(Result,Seperator,SplitAtomList,Done,ExpandedSplitAtomList).

:- use_module(kernel_objects,[equal_object/3]).
:- block string_join3(?,-,?,?,?),string_join3(?,?,?,-,?). % we need to know the seperator: TO DO : improve this
string_join3(Result,Seperator,_SplitAtomList,_Done,ExpandedSplitAtomList) :-
   ExpandedSplitAtomList \= [],  % split("",sep) --> [""] not the empty list
   sort(ExpandedSplitAtomList,SL),
   maplist(drop_index,SL,IL),
   convert_b_to_prolog_atoms(IL,PL,Done2),
   when(nonvar(Done2),
     (insert_sep(PL,Seperator,PL2),
      ajoin(PL2,Atom),
      equal_object(Result,string(Atom),string_join))).


% -----------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_chars(string(''),[]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_chars(string('010'),[(int(1),string('0')),(int(2),string('1')),(int(3),string('0'))]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_chars(string('a'),[(int(1),string('a'))]))).

:- block string_chars(-,-).
% to do: make it reversible
string_chars(Str,SeqRes) :- nonvar(Str), Str=string(A), ground(A),!,
   string_chars2(A,SeqRes).
string_chars(Str,SeqRes) :- when((ground(Str);ground(SeqRes)),string_chars1(Str,SeqRes)).

string_chars1(Str,SeqRes) :- nonvar(Str), Str=string(A), ground(A),!,
   string_chars2(A,SeqRes).
string_chars1(Str,SeqRes) :-
   expand_custom_set_to_list(SeqRes,ExpandedAtomList,Done,string_chars1),
   string_chars3(Str,SeqRes,ExpandedAtomList,Done).

:- use_module(kernel_objects,[equal_object_optimized/3]).
string_chars2(A,SeqRes) :- atom_codes(A,AA), generate_char_seq(AA,1,CharSeq),
   equal_object_optimized(CharSeq,SeqRes,string_chars2).

:- use_module(probsrc(tools_strings),[ajoin/2]).

:- block string_chars3(-,?,?,-).
string_chars3(Str,SeqRes,_ExpandedAtomList,_Done) :-
   nonvar(Str), Str=string(A), ground(A),
   !,
   string_chars2(A,SeqRes).
string_chars3(Str,_SeqRes,ExpandedAtomList,Done) :-
   nonvar(Done),
   !,
   sort(ExpandedAtomList,SL),
   maplist(drop_index,SL,IL),
   kernel_strings:convert_b_to_prolog_atoms(IL,PL,Done2),
   when(nonvar(Done2),
      (ajoin(PL,Atom),
       equal_object(Str,string(Atom),string_chars))).
string_chars3(Str,SeqRes,ExpandedAtomList,Done) :- % Str is only partially instantiated
   when((ground(Str);nonvar(Done)),string_chars3(Str,SeqRes,ExpandedAtomList,Done)).
generate_char_seq([],_,[]).
generate_char_seq([Code|T],Nr,[(int(Nr),string(CS))|TSeq]) :-
   atom_codes(CS,[Code]),
   N1 is Nr+1, generate_char_seq(T,N1,TSeq).

drop_index((int(_),R),R).

% ------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_codes(string(''),[]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:string_codes(string('010'),[(int(1),int(48)),(int(2),int(49)),(int(3),int(48))]))).

:- block string_codes(-,-).
string_codes(string(A),SeqRes) :- string_codes2(A,SeqRes).

:- use_module(custom_explicit_sets,[expand_custom_set_to_list/4]).

:- block string_codes2(-,-).
string_codes2(A,SeqRes) :-
   nonvar(A),
   !,
   string_codes4(A,SeqRes).
string_codes2(A,SeqRes) :-
   SeqRes==[],
   !,
   empty_string_atom(A).
string_codes2(A,SeqRes) :- expand_custom_set_to_list(SeqRes,SeqList,_,string_codes2),
   when((nonvar(A);ground(SeqList)), string_codes3(A,SeqList)).

string_codes3(A,SeqRes) :-
  nonvar(A),
  !,
  string_codes4(A,SeqRes).
string_codes3(A,SeqRes) :-
   sort(SeqRes,SSeqRes),
   extract_codes(SSeqRes,1,Codes),
   atom_codes(A,Codes).
string_codes4(A,SeqRes) :-
   atom_codes(A,AA), generate_code_sequence(AA,1,CodeSeq),
   equal_object_optimized(CodeSeq,SeqRes,string_codes4).

generate_code_sequence([],_,[]).
generate_code_sequence([Code|T],Nr,[(int(Nr),int(Code))|TSeq]) :-
   N1 is Nr+1, generate_code_sequence(T,N1,TSeq).

extract_codes([],_,[]).
extract_codes([(int(Nr),int(Code))|T],N,[Code|CT]) :-
  (Nr==N -> true ; add_error(extract_codes,'Unexpected index: ',(Nr,N))),
  N1 is N+1, extract_codes(T,N1,CT).


empty_string_atom('').

% ------------------------

:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:format_to_string(string('abc'),[],string('abc')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:format_to_string(string('abc~wfg'),[(int(1),string('de'))],string('abcdefg')))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(kernel_strings:format_to_string(string('abc~wfg~w'),[(int(1),string('de')),(int(2),string('h'))],string('abcdefgh')))).

:- block format_to_string(-,?,?).
format_to_string(string(FormatString),BSeqOfValues,Res) :-
    convert_b_sequence_to_list_of_atoms(BSeqOfValues,ListOfAtoms,Done),
    format_to_string_aux(Done,FormatString,ListOfAtoms,Res).

:- use_module(library(codesio),[format_to_codes/3]).
:- block format_to_string_aux(-,?,?,?), format_to_string_aux(?,-,?,?).
format_to_string_aux(_,FormatString,ListOfAtoms,Res) :-
   % print(format_to_codes(FormatString,Atoms,Codes)),nl,
   format_to_codes(FormatString,ListOfAtoms,Codes),
   atom_codes(Atom,Codes),
   Res = string(Atom).

:- use_module(custom_explicit_sets,[is_set_value/2, expand_custom_set_to_list_gg/4]).

% convert a B sequence into a list of atoms; pretty printing if necessary
:- block convert_b_sequence_to_list_of_atoms(-,?,?).
convert_b_sequence_to_list_of_atoms(BSeqOfValues,Res,Done) :-
    is_set_value(BSeqOfValues,convert_b_sequence_to_list_of_atoms),
    !,
    expand_custom_set_to_list_gg(BSeqOfValues,ESet,GG,kernel_strings), % could use _gg version
    (GG=guaranteed_ground -> GrESet=true ; ground_value_check(ESet,GrESet)),
    convert_aux(GrESet,ESet,Res,Done).
convert_b_sequence_to_list_of_atoms(SingleValue,[S],Done) :-
    translate_bvalue(SingleValue,XS),
    add_warning(kernel_strings,'B sequence expected, obtained single value: ',XS),
    ground_value_check(SingleValue,GrValue), to_string_aux(GrValue,SingleValue,string(S)), Done=GrValue.

:- block convert_aux(-,?,?,?), convert_aux(?,-,?,?).
convert_aux(_,ESet,ListOfAtoms,Done) :-
   sort(ESet,SortedESet),
   maplist(get_string,SortedESet,ListOfAtoms),
   Done=true.

get_string((_,string(S)),R) :- !,R=S.
get_string((_,X),R) :- !,to_string_aux(X,string(R)).
get_string(X,R) :-
   translate_bvalue(X,XS),
   add_warning(kernel_strings,'B sequence expected, obtained set containing: ',XS),
   to_string_aux(X,string(R)).


% ------------------------
% UTILITIES
% ------------------------


:- assert_must_succeed((kernel_strings:split_atom_string('ef,g',',',R), R==[ef,g])).
:- assert_must_succeed((kernel_strings:split_atom_string('ab,cd,ef,g',',',R), R==['ab','cd',ef,g])).
:- assert_must_succeed((kernel_strings:split_atom_string('ab','a',R), R==['','b'])).
:- assert_must_succeed((kernel_strings:split_atom_string('','a',R), R==[''])).
:- assert_must_succeed((kernel_strings:split_atom_string('STRING1','',R), R==['STRING1'])).
:- assert_must_succeed((kernel_strings:split_atom_string('mod274,mod276,mod277,mod282,mod283,mod284,mod285,mod286',',',R), R==[mod274,mod276,mod277,mod282,mod283,mod284,mod285,mod286])).

split_atom_string(Atom,Sep,SplitList) :- %print(split_atom_string(Atom,Sep)),nl,
  atom_chars(Sep,SepAscii),
  (SepAscii=[] -> SplitList = [Atom]
   ; SepAscii = [H|T], atom_chars(Atom,ListAscii),
     split3(ListAscii,H,T,Match,Match,SplitList)).

% MatchSoFar is passed in two variables: one to instantiate and one with the Result of the match
% this avoids calling reverse
split3([],_,_,MatchSoFarIn,MatchSoFarRes,R) :- !,
   MatchSoFarIn=[],  % match complete, ground tail of match
   atom_chars(Atom,MatchSoFarRes),R=[Atom].
split3([H|List],H,Sep,MatchSoFarIn,MatchSoFarRes,Res) :-
   append(Sep,Tail,List),
   !, % we have a match with a separator
   MatchSoFarIn=[], % match complete
   atom_chars(Atom,MatchSoFarRes),
   Res=[Atom|R2], split3(Tail,H,Sep,NewMatch,NewMatch,R2).
split3([H|T],HS,Sep,[H|MatchSoFarIn],MatchSoFarRes,Res) :- % no match
   split3(T,HS,Sep,MatchSoFarIn,MatchSoFarRes,Res).



% -----------------------

convert_prolog_to_b_list(PL,BL) :-
  convert_prolog_to_b_list_aux(PL,1,CPL),
  equal_object(CPL,BL,convert_prolog_to_b_list).

convert_prolog_to_b_list_aux([],_,[]).
convert_prolog_to_b_list_aux([H|T],Index,[(int(Index),CH)|CT]) :-
  convert_prolog_to_b_term(H,CH),
  I1 is Index+1, convert_prolog_to_b_list_aux(T,I1,CT).

convert_prolog_to_b_term(N,R) :-
   number(N),!,
   R=int(N).
convert_prolog_to_b_term(A,R) :-
   atomic(A),!,
   R=string(A).
convert_prolog_to_b_term(A,R) :-
   add_internal_error('Illegal Prolog term: ',convert_prolog_to_b_term(A,R)), R=A.


% a version that delays converting and sets Done to done when all B Atoms have been grounded
convert_b_to_prolog_atoms([],[],done).
convert_b_to_prolog_atoms([BAtom|T],[PrologAtom|PT],Done) :-
   convert_b_to_prolog_atoms_aux(BAtom,PrologAtom,DoneAtom),
   convert_b_to_prolog_atoms(T,PT,DoneT),
   both_done(DoneAtom,DoneT,Done).

:- block both_done(-,?,?), both_done(?,-,?).
both_done(_,_,done).

:- block convert_b_to_prolog_atoms_aux(-,?,?).
convert_b_to_prolog_atoms_aux(pred_true,'TRUE',done).
convert_b_to_prolog_atoms_aux(pred_false,'FALSE',done).
convert_b_to_prolog_atoms_aux(string(S),PrologAtom,Done) :-
   convert_b_to_prolog_atoms_aux2(S,PrologAtom,Done).
convert_b_to_prolog_atoms_aux(int(S),PrologAtom,Done) :-
   convert_b_to_prolog_atoms_aux2(S,PrologAtom,Done).

:- block convert_b_to_prolog_atoms_aux2(-,?,?).
convert_b_to_prolog_atoms_aux2(Atom,Atom,done).
