% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)
:- module(kernel_tools,[
             ground_value_check/2,
             bexpr_variables/2,
             ground_bexpr/1,
             ground_bexpr_check/2,
             value_variables/2,
             ground_value/1,
             ground_state_check/2, ground_state/1,
             ground_typedvals_check/2,
             %discard_ground_value_check/1,
             %map_over_bvalue/3,
             filter_cannot_match/4, get_template_for_filter_cannot_match/2,
             cannot_match/2, can_match/2,
             cannot_match_aggressive/2
             ]).


:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides utilities for term variables for B expressions and values.').

:- use_module(self_check).
:- use_module(error_manager).
:- use_module(bsyntaxtree,[safe_syntaxelement/5]).

% use when ground value check no longer need to instantiate result variable:
% seems to have performance issues
%discard_ground_value_check(X) :- nonvar(X),!,
%  (X==ground_value -> true ; print(illegal_ground_value_check(X)),nl).
%discard_ground_value_check(ground_value).

:- assert_must_succeed((kernel_tools:ground_value_check(int(X),V), var(V), X=1, nonvar(V))).
:- assert_must_succeed((kernel_tools:ground_value_check(string(X),V), var(V), X=a, nonvar(V))).
:- assert_must_succeed((kernel_tools:ground_value_check([int(1),int(X)],V), var(V), X=1, nonvar(V))).
% instantiate second argument to atomic-nonvar term if first argument ground B value
:- block ground_value_check(-,?).
ground_value_check([],R) :- !, R=ground_value.
ground_value_check(pred_true,R) :- !, R=ground_value.
ground_value_check(pred_false,R) :- !, R=ground_value.
ground_value_check(int(X),R) :- !, ground_atom(X,R).
ground_value_check(fd(X,_T),R) :- !,ground_atom(X,R). % avoid waking up co-routines with R=X.
ground_value_check(string(X),R) :- !, ground_atom(X,R).
ground_value_check((A,B),R) :- !, ground_value_check(A,GA), ground_value_check_aux(GA,B,R).
ground_value_check([H|T],R) :- !, ground_value_check(H,GA), ground_value_check_aux(GA,T,R).
ground_value_check(avl_set(_),R) :- !, % assume avl_set grounded
   R = ground_value.
ground_value_check(rec(F),R) :- !, ground_fields(F,R).
ground_value_check(closure(_P,_T,B),R) :- !,
   bexpr_variables(B,Vars),
   ground_value_check(Vars,R).
   %when(ground(Vars),R=ground_value).
ground_value_check(term(X),R) :- !, when(ground(X),R=ground_value).
ground_value_check(Val,R) :-  %print(uncov(Val,R)),nl,trace,
   (atomic(Val) -> R=ground_value ; when(ground(Val),R=ground_value)).

:- block ground_atom(-,?).
ground_atom(_,ground_value).  %we could do  a check ground_atom(X,R) :- (nonvar(R) -> nl,print(illegal_grval(X,R)),nl,nl ; R=X).
% we use this co-routine rather than direct unification to avoid unnecessarily waking up pending nonvar/ground checks
% indeed: if we unify X with another variable Y, this will wake up all when(nonvar(X) checks
% in test 1101 for prob_examples/examples/B/Systerel/C578.EML.014/612_001.mch this caused performance problems
%  we woke up another pending co-routine with the same waitflag priority, it had a check when((nonvar(X);ground(WF1),
%  WF1 was already ground: this meant we started evaluating another Prolog kernel call before finishing the current one

:- block ground_value_check_aux(-,?,?).
% :- block ground_value_check_aux(-,?,-). % this slows down test 884, 292, 293, 1456, 1737
%ground_value_check_aux(_,_,R) :- nonvar(R),!. % groundness check no longer needed
ground_value_check_aux(_,V,R) :- ground_value_check(V,R).

:- block ground_fields(-,?).
ground_fields([],R) :- !, R=ground_value.
ground_fields([field(F,V)|T],R) :- ground_fields_aux(F,V,T,R).

:- block ground_fields_aux(-,?,?,?).
ground_fields_aux(_,V,T,R) :- ground_fields(T,GrT), ground_value_check_aux(GrT,V,R).


:- block ground_state_check(-,?).
ground_state_check([],R) :- !, R=ground_value.
ground_state_check([bind(F,Val)|T],R) :- !, ground_state_check_aux(F,Val,T,R).
ground_state_check(BL,R) :- add_internal_error('Illegal bind list:',ground_state_check(BL,R)).

:- block ground_state_check_aux(-,?,?,?).
ground_state_check_aux(_,Val,T,R) :- ground_state_check(T,GrT), ground_value_check_aux(GrT,Val,R).


:- block ground_typedvals_check(-,?).
ground_typedvals_check([],R) :- !, R=ground_value.
ground_typedvals_check([typedval(Val,_Type,VarID,_EnumWarning)|T],R) :- !, ground_typedvals_check_aux(VarID,Val,T,R).
ground_typedvals_check(TL,R) :- add_internal_error('Illegal typedval list:',ground_typedvals_check(TL,R)).

:- block ground_typedvals_check_aux(-,?,?,?).
ground_typedvals_check_aux(_,Val,T,R) :- ground_typedvals_check(T,GrT), ground_value_check_aux(GrT,Val,R).


:- block ground_bexpr_check(-,?).
ground_bexpr_check(BExpr,R) :-
   bexpr_variables(BExpr,Vars),
   ground_value_check(Vars,R).

% -------------------------------


% attempt at a more efficient version of term_variables; avoiding traversing big avl_set's
%bexpr_variables(T,V) :- !, term_variables(T,V).
bexpr_variables(b(P,_,_),Vars) :-
   bexpr_variables_aux(P,[],Vars), ! . %, print(vars(P,Vars)),nl.
bexpr_variables(T,V) :- add_internal_error('Illegal call: ',bexpr_variables(T,V)),
   %trace, bexpr_variables(T,V),
   term_variables(T,V).

% check if a B expression or predicate is ground
ground_bexpr(b(P,_,_)) :- !,
   bexpr_variables_aux(P,[],[]).
ground_bexpr(T) :- add_internal_error('Illegal call: ',ground_bexpr(T)),
   ground(T).

% get free variables inside a B value
value_variables(V,Vars) :- var(V),!, Vars=[V]. % optimize a few frequent cases
value_variables(avl_set(_),Vars) :- !, Vars=[].
value_variables([],Vars) :- !, Vars=[].
value_variables(Val,Vars) :- %print(val(Val)),nl,
  value_variables(Val,[],Vars).

% check if a B value is ground
%version below seems a bit faster than ground_value(Val) :- value_variables(Val,[],[]).
ground_value(V) :- var(V),!,fail.
ground_value(pred_true) :- !.
ground_value(pred_false) :- !.
ground_value(int(N)) :- !, nonvar(N).
ground_value(global_set(N)) :- !, ground(N).
ground_value(freetype(N)) :- !, ground(N).
ground_value(string(S)) :- !, ground(S).
ground_value(fd(X,T)) :- !, nonvar(X),nonvar(T).
ground_value([]) :- !.
ground_value(avl_set(_)) :- !.
ground_value([H|T]) :- !, ground_value(H),ground_value(T).
ground_value((A,B)) :- !, ground_value(A),ground_value(B).
ground_value(freeval(ID,Case,Value)) :- !,
   ground(ID), ground(Case), % necessary ?
   ground_value(Value).
ground_value(rec(F)) :- !, record_variables(F,[],[]).
ground_value(closure(_P,_T,B)) :- !, ground_bexpr(B).  %% DO WE NEED TO CHECK P ? Probably not
ground_value(term(T)) :- !, ground(T).
ground_value(T) :- functor(T,F,N), format('~n*** uncovered value ~w/~w~n',[F,N]),
   ground(T).


% check if a B state is ground
ground_state(V) :- var(V),!,fail.
ground_state([]) :- !.
ground_state([bind(N,V)|T]) :- !, ground(N),
   ground_value(V), ground_state(T).
ground_state(S) :- add_internal_error('Illegal state: ',ground_state(S)),fail.

:- use_module(library(aggregate),[term_variables/3]).

bexpr_variables(b(P,_,_)) --> bexpr_variables_aux(P).

value_variables(V) --> {var(V)},!,term_variables_atomic(V). % slightly unnecessary to do two var checks
value_variables(avl_set(_)) --> !,[].
value_variables((A,B)) --> !, value_variables(A),value_variables(B).
value_variables([H|T]) --> !, value_variables(H),value_variables(T).
value_variables(closure(_P,_T,B)) --> !, bexpr_variables(B).  %% DO WE NEED TO CHECK P ? Probably not
value_variables(pred_false) --> !,[].
value_variables(pred_true) --> !,[].
value_variables(fd(X,T)) --> !, {check_nonvar(T)},term_variables_atomic(X).
value_variables([]) --> !,[].
value_variables(rec(F)) --> !, record_variables(F).
value_variables(freeval(ID,Case,Value)) --> !,
   value_variables(ID), value_variables(Case), % necessary ?
   value_variables(Value).
value_variables(string(S)) --> !,term_variables_atomic(S). % not sure this is useful
value_variables(int(I)) --> !,term_variables_atomic(I).
value_variables(global_set(S)) --> !, term_variables_atomic(S). % normally global_sets should always be ground
% TO DO: support freevals
value_variables(T) -->  %{ functor(T,F,N), print(uncov_value_variables(F,N)),nl},
                       term_variables(T).

check_nonvar(T) :- (var(T) -> add_internal_error('Unexpected variable: ',T) ; true).

record_variables(V) --> {var(V)},!,term_variables(V).
record_variables([]) --> [].
record_variables([field(F,V)|T]) -->
      term_variables_atomic(F), % is this necessary ?
      value_variables(V), record_variables(T).

:- use_module(tools, [exact_member/2]).
% an optimized version of term_variables/3 which assumes argument is either var or ground
term_variables_atomic(VarOrGround,Vars,NewVars) :-
      var(VarOrGround),
      \+ exact_member(VarOrGround,Vars),
      !,
      NewVars = [VarOrGround|Vars].
term_variables_atomic(_,Vars,Vars).

bexpr_variables_aux(value(V)) --> !,value_variables(V).
% 0-ary
bexpr_variables_aux(boolean_false) --> !,[].
bexpr_variables_aux(boolean_true) --> !,[].
bexpr_variables_aux(bool_set) --> !,[].
bexpr_variables_aux(empty_set) --> !,[].
bexpr_variables_aux(empty_sequence) --> !,[].
bexpr_variables_aux(falsity) --> !,[].
bexpr_variables_aux(integer_set) --> !,[].
bexpr_variables_aux(max_int) --> !,[].
bexpr_variables_aux(min_int) --> !,[].
bexpr_variables_aux(string_set) --> !,[].
bexpr_variables_aux(truth) --> !,[].
%unary:
bexpr_variables_aux(card(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(closure(A)) --> !, bexpr_variables(A).
%bexpr_variables_aux(closure1(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(reflexive_closure(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(convert_bool(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(domain(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(identifier(_)) --> !,[].
bexpr_variables_aux(integer(_)) --> !,[].
bexpr_variables_aux(first(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(last(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(max(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(min(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(perm(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(pow_subset(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(pow1_subset(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(range(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(reflexive_closure(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(rev(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(reverse(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(second_of_pair(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(seq(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(seq1(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(size(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(string(_)) --> !, [].
bexpr_variables_aux(unary_minus(A)) --> !, bexpr_variables(A).
% binary:
bexpr_variables_aux(add(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(conjunct(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(disjunct(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(domain_restriction(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(domain_subtraction(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(cartesian_product(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
%bexpr_variables_aux(comprehension_set(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(couple(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(equal(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(equivalence(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(function(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(greater_equal(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(greater(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(image(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(implication(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(insert_front(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(insert_tail(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(intersection(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(interval(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(lazy_lookup_expr(_)) --> !,[].
bexpr_variables_aux(lazy_lookup_pred(_)) --> !,[].
bexpr_variables_aux(less_equal(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(less(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(member(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(minus(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(modulo(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(multiplication(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(negation(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(not_equal(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(not_member(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(not_subset(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(not_subset_strict(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(overwrite(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(partial_function(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(range_restriction(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(range_subtraction(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(record_field(A,_FieldName)) --> !, bexpr_variables(A).
bexpr_variables_aux(restrict_front(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(restrict_tail(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(set_subtraction(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(subset(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(subset_strict(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(union(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
% ternary
bexpr_variables_aux(if_then_else(A,B,C)) --> !, bexpr_variables(A), bexpr_variables(B), bexpr_variables(C).
% with new local variables:
bexpr_variables_aux(comprehension_set(_Ids,A)) --> !, bexpr_variables(A).
bexpr_variables_aux(exists(_Ids,A)) --> !, bexpr_variables(A).
bexpr_variables_aux(event_b_comprehension_set(_Ids,A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(forall(_Ids,A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(lambda(_Ids,A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(lazy_let_expr(_ID,A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(lazy_let_pred(_ID,A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(lazy_let_subst(_ID,A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
%bexpr_variables_aux(let_predicate(_Ids,E,A)) --> !, bexpr_variables(E), bexpr_variables(A).
%bexpr_variables_aux(let_expression(_Ids,E,A)) --> !, bexpr_variables(E), bexpr_variables(A).
bexpr_variables_aux(general_sum(_Ids,A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(general_product(_Ids,A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(quantified_union(_Ids,A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(quantified_intersection(_Ids,A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
% with lists:
bexpr_variables_aux(sequence_extension(A)) --> !, l_bexpr_variables(A).
bexpr_variables_aux(set_extension(A)) --> !, l_bexpr_variables(A).
bexpr_variables_aux(struct(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(tail(A)) --> !, bexpr_variables(A).
bexpr_variables_aux(total_function(A,B)) --> !, bexpr_variables(A), bexpr_variables(B).
bexpr_variables_aux(external_function_call(_FunName,A)) --> !, l_bexpr_variables(A).
bexpr_variables_aux(external_pred_call(_FunName,A)) --> !, l_bexpr_variables(A).
% general case (could be used for all above; but is considerably slower)
bexpr_variables_aux(Expr) -->
     %{ functor(Expr,F,N), print(uncov_var_aux(F,N)),nl },
     {bsyntaxtree:safe_syntaxelement(Expr,Subs,_Names,_List,_Constant)},
     l_bexpr_variables(Subs).

l_bexpr_variables([]) --> [].
l_bexpr_variables([H|T]) --> bexpr_variables(H), l_bexpr_variables(T).




%:- use_module(custom_explicit_sets, [expand_custom_set_to_list_now/2]).
% dispatch. TODO: closure/3 and freetype
%:- meta_predicate map_over_bvalue(2,?,?).
%map_over_bvalue(_,[],[]) :- !.
%map_over_bvalue(Pred,[H|T],[Res|Tail]) :- !,
%    map_over_bvalue(Pred,H,Res),
%    map_over_bvalue(Pred,T,Tail).
%map_over_bvalue(Pred,(A,B),(ResA,ResB)) :- !,
%    map_over_bvalue(Pred,A,ResA),
%    map_over_bvalue(Pred,B,ResB).
%map_over_bvalue(Pred,rec(L),rec(Res)) :- !,
%    map_over_bvalue(Pred,L,Res).
%map_over_bvalue(Pred,avl_set(A),Res) :- !,
%    custom_explicit_sets:expand_custom_set_to_list_now(avl_set(A),L),
%    map_over_bvalue(Pred,L,Res).
%map_over_bvalue(Pred,global_set(A), Res) :- !,
%    custom_explicit_sets:expand_custom_set_to_list_now(global_set(A), L),
%    map_over_bvalue(Pred,L,Res).
%map_over_bvalue(Pred,field(Name,Value),field(Name,TValue)) :- !,
%    map_over_bvalue(Pred,Value,TValue).
%% apply
%map_over_bvalue(Pred,BValue,OutValue) :- !,
%    call(Pred,BValue,OutValue).

% ----------------------------------
:- assert_must_succeed((kernel_tools:filter_cannot_match([int(X),int(11),Y],int(12),R,F), R==[int(X),Y],F==true)).

% filter out from the set Set (list part) all elements that cannot match X
% Filtered=true means at least one element was filtered out
filter_cannot_match(Set,X,NewSet,Filtered) :- instantiated_enough(X),!,
   filter_cannot_match_aux(Set,X,NewSet,Filtered).
filter_cannot_match(S,_,S,false).

:- use_module(library(avl),[avl_domain/2, avl_height/2]).
:- use_module(custom_explicit_sets,[sorted_ground_normalised_list_to_avlset/3]).
:- use_module(performance_messages,[perfmessage/1]).
filter_cannot_match_aux(Var,_,Res,Filtered) :- var(Var),!,Res=Var,Filtered=false.
filter_cannot_match_aux([],_,Res,Filtered) :- !, Res=[],Filtered=false.
filter_cannot_match_aux([H|T],X,Res,Filtered) :- !,
    (cannot_match(X,H) -> Filtered=true,filter_cannot_match_aux(T,X,Res,_)
     ; Res=[H|RT], filter_cannot_match_aux(T,X,RT,Filtered)).
filter_cannot_match_aux(avl_set(Avl),X,R,Filtered) :- avl_height(Avl,H), %print(avl(H)),nl,
   (H<8 -> true ; perfmessage(avl_set_not_filtered(H,X)),fail),
   !,
   avl_domain(Avl,List),
   filter_cannot_match_aux(List,X,L2,Filtered),
   (Filtered=true
     -> sorted_ground_normalised_list_to_avlset(L2,R,filter_cannot_match_aux)
     ; R=avl_set(Avl)).
filter_cannot_match_aux(S,_,S,false).

% does it make sense to filter Set, or will cannot_match always fail
instantiated_enough(V) :- var(V),!,fail.
instantiated_enough(pred_true) :- !.
instantiated_enough(pred_false) :- !.
instantiated_enough(int(X)) :- !, nonvar(X).
instantiated_enough(fd(X,T)) :- !, nonvar(X),nonvar(T).
instantiated_enough(string(X)) :- !, nonvar(X).
instantiated_enough((A,B)) :- instantiated_enough(A) -> true ; instantiated_enough(B).
instantiated_enough(rec(Fs)) :- (member(field(N,V),Fs),nonvar(N),instantiated_enough(V) -> true).
% TO DO: free-types, ...


%delay_filter_cannot_match(Set,X,NewSet) :-
%   block_filter_cannot_match_aux(Set,X,NewSet).
%
%:- block block_filter_cannot_match_aux(-,?,?).
%block_filter_cannot_match_aux([],_,Res) :- !, Res=[].
%block_filter_cannot_match_aux([H|T],X,Res) :-
%    (cannot_match(X,H) -> block_filter_cannot_match_aux(T,X,Res)
%     ; Res=[H|RT], block_filter_cannot_match_aux(T,X,RT)).
%block_filter_cannot_match_aux(S,_,S).


% convert a typed expression into a value that can be used to filter a list of values with filter_cannot_match
get_template_for_filter_cannot_match(TExpr,ValueForFiltering) :- %print(getT(TExpr)),nl,
   get_template_for_filter3(TExpr,ValueForFiltering,NonVar), nonvar(NonVar).
get_template_for_filter3(b(E,_,_),Val,NonVar) :-
   get_template_for_filter_aux(E,Val,NonVar).
get_template_for_filter_aux(value(Val),Val,NonVar) :- !, (nonvar(Val) -> NonVar=nonvar ; true).
get_template_for_filter_aux(couple(A,B),(VA,VB),NonVar) :- !,
   get_template_for_filter3(A,VA,NonVar),
   get_template_for_filter3(B,VB,NonVar).
%get_template_for_filter_aux(identifier(_),_,_).
get_template_for_filter_aux(_,_,_).


:- assert_must_succeed((kernel_tools:can_match(int(1),_V))).
:- assert_must_fail((kernel_tools:cannot_match(int(1),_V))).
:- assert_must_succeed((kernel_tools:can_match(int(12),int(12)))).
:- assert_must_succeed((kernel_tools:cannot_match(int(1),int(2)))).
:- assert_must_succeed((kernel_tools:cannot_match(string(a),string(b)))).
:- assert_must_succeed((kernel_tools:can_match(string(a),string(a)))).
:- assert_must_succeed((kernel_tools:can_match(string(a),string(_)))).
:- assert_must_succeed((kernel_tools:can_match(string(_),string(b)))).
:- assert_must_succeed((kernel_tools:cannot_match([],[string(b)]))).
:- assert_must_succeed((kernel_tools:cannot_match([_],[]))).
:- assert_must_succeed((kernel_tools:cannot_match((int(11),string(a)),(int(11),string(b))))).
:- assert_must_succeed((kernel_tools:can_match((int(11),string(a)),(int(11),string(_))))).
:- assert_must_succeed((kernel_tools:cannot_match(pred_false,pred_true))).
:- assert_must_succeed((kernel_tools:can_match(pred_true,pred_true))).
:- assert_must_succeed((kernel_tools:can_match(record([field(a,pred_true)]),record([field(a,_)])))).
:- assert_must_succeed((kernel_tools:cannot_match(record([field(a,pred_false)]),record([field(a,pred_true)])))).

% can_match: quick check to see if the (AVL or List) element could match

can_match(A,B) :- \+ cannot_match(A,B).
cannot_match(A,B) :- \+ can_match_aux(A,B). % Note: probably better to wrap can_match into negation to avoid instantiations

:- use_module(clpfd_interface,[clpfd_can_match/2]).

can_match_aux(A,_) :- var(A),!. % TO DO: we could check for not_empty_set co-routine if A or B = []?
can_match_aux(_,B) :- var(B),!.
can_match_aux((A,B),(ElA,ElB)) :- !,can_match_aux(A,ElA), can_match_aux(B,ElB).
can_match_aux(int(A),R) :- !,
     % (\+ ground(R) -> true,print(ok(int(A),R)),nl ; print(unify(R,int(A))),nl,R=int(A)). % unification could trigger other constraint propagations and be expensive; we could just check that A is in domain
     (var(R) -> true
      ; R=int(B), clpfd_can_match(A,B)). %,(B=A -> true ; print(ko(A,B)),nl,fail)).%,fail,trace,B=A)).
can_match_aux(fd(NrA,GS),R) :- !,
      %(\+ ground(R) -> true ; R=fd(Nr,GS)). % ditto (see, e.g., test 985 - graph iso)
     (var(R) -> true ; R=fd(NrB,GS), clpfd_can_match(NrA,NrB)).
can_match_aux(pred_false,R) :- !, R=pred_false.
can_match_aux(pred_true,R) :- !, R=pred_true.
can_match_aux(string(A),R) :- !, %(\+ ground(R) -> true ; R=string(A)). % may instantiate LHS
     (var(R) -> true ; R=string(B), can_match_atomic(A,B)).
can_match_aux(term(A),R) :- !, %(\+ ground(R) -> true ; R=term(A)). % ditto
     (var(R) -> true ; R=term(B), can_match_atomic(A,B)).
can_match_aux(freeval(ID,Case,Value),R) :- !,
    (var(R) -> true
     ; R=freeval(IDR,CaseR,VR),
       (var(IDR) ; var(CaseR) -> true
        ; (ID,Case)=(IDR,CaseR),
          can_match_aux(Value,VR))).
can_match_aux([],R) :- !, R \= avl_set(node(_,_,_,_,_)), R\= [_|_].
can_match_aux(avl_set(A),R) :- !, can_match_avl_set(R,A).
can_match_aux(global_set(_),R) :- !,  R \= []. % global sets must be non-empty; TO DO: maybe quick check in case global_set is infinite; TO DO: check if two global_sets are identical
can_match_aux([_|_],R) :- !, R \= [].
can_match_aux(record(Fields1),record(Fields2)) :- !,can_match_fields(Fields1,Fields2).
can_match_aux(_,_). % catch-all other types

can_match_fields(A,B) :- (var(A);var(B)),!.
can_match_fields([],[]).
can_match_fields([F1|T1],[F2|T2]) :- can_match_field(F1,F2), can_match_fields(T1,T2).

can_match_field(A,B) :- (var(A);var(B)),!.
can_match_field(field(F1,V1),field(F2,V2)) :- F1==F2,!, can_match_aux(V1,V2).
can_match_field(F1,F2) :- add_internal_error('Non-matching fields: ',can_match_field(F1,F2)).

:- use_module(kernel_dif,[frozen_dif/2]).
can_match_atomic(A,B) :- var(A),!, \+ frozen_dif(A,B).
can_match_atomic(A,B) :- var(B),!, \+ frozen_dif(B,A).
can_match_atomic(A,B) :- A=B.

:- use_module(library(avl),[avl_min/2, avl_max/2, avl_size/2]).

can_match_avl_set([],A) :- !, A=empty.
can_match_avl_set([H|T],A) :- !, list_can_match_avl_set([H|T],0,A).
can_match_avl_set(avl_set(B),A) :- !, avl_min(A,Min), avl_min(B,Min), avl_max(A,Max),avl_max(B,Max).
can_match_avl_set(_,_).

list_can_match_avl_set(T,_,_) :- var(T),!.
list_can_match_avl_set([],Sze,A) :- !, avl_size(A,Sze).
list_can_match_avl_set([H|T],Sze,A) :- !,
   (custom_explicit_sets:quick_test_avl_membership(A,H,PredRes)
     -> PredRes=pred_true, S1 is Sze+1,
        list_can_match_avl_set(T,S1,A)
     ;  true). % TO DO: maybe also check T ?
list_can_match_avl_set(_,_,_).

% ------------------------

cannot_match_aggressive(A,B) :- \+ can_match_aggr(A,B).

% a more aggressive version; which may trigger co-routines for left-side A
can_match_aggr(A,_) :- var(A),!.
can_match_aggr(_,B) :- var(B),!.
can_match_aggr((A,B),(ElA,ElB)) :- !,can_match_aggr(A,ElA), can_match_aggr(B,ElB).
can_match_aggr(int(A),R) :- !, % R=int(A). % can trigger co-routines both sides; causes problem for 985
     (var(R) -> true
      ; R=int(B), (var(B) -> clpfd_can_match(A,B) ; A=B)). % only triggers co-routines for left side
can_match_aggr(fd(NrA,GS),R) :- !, % R=fd(Nr,GS). % ditto
     (var(R) -> true
      ; R=fd(NrB,GS), (var(NrB) -> clpfd_can_match(NrA,NrB) ; NrA=NrB)).
can_match_aggr(string(A),R) :- !, %R=string(A). % ditto
     (var(R) -> true
      ; R=string(B), (var(B) -> can_match_atomic(A,B) ; A=B)).
can_match_aggr(pred_false,R) :- !, R=pred_false.
can_match_aggr(pred_true,R) :- !, R=pred_true.
can_match_aggr(A,B) :- can_match_aux(A,B).



