% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kodkod_process, [register_problem/7, kodkod_request_intern/5,
                           get_original_problem/2,
                           clear_java_comp_time/0,
                           get_java_comp_time/1,
                           stop_kodkod/0
                           ]).

:- use_module(library(avl)).
:- use_module(library(lists)).
:- use_module(library(process)).
:- use_module(library(file_systems)).

:- use_module(probsrc(module_information),[module_info/2]).
:- use_module(probsrc(store), [lookup_value_for_existing_id/4]).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(kernel_objects), [equal_object/2]).
:- use_module(probsrc(custom_explicit_sets), [expand_custom_set_to_list/2]).
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(system_call)).
:- use_module(probsrc(preferences), [get_preference/2]).
:- use_module(probsrc(debug), [debug_format/3]).

:- use_module(kodkod_printer).

:- use_module(probsrc(translate),[print_bvalue/1,print_bstate/1]).

:- module_info(group,kodkod).
:- module_info(description,'This is the inter-process communication with the Java process that contains the Kodkod engine.').

:- dynamic known_problem/7, next_problem_id/1.

next_problem_id(0).

register_problem(Ids,KExpr,Idmap,Typemap,Valuemap,Predicate,ProblemId) :-
    startup_kodkod,
    get_new_problem_id(ProblemId),
    extract_all_types(Typemap,Types),
    extract_all_relations(Idmap,Relations),
    prepare_value_mapping(Valuemap,ProbToKodkodVal,KodkodToProbVal),
    prepare_id_mapping(Ids,Idmap,ProbToKodkodId,KodkodToProbId),
    assert( known_problem(ProblemId,Types,Relations,KExpr,
                          probkodkod(ProbToKodkodVal,ProbToKodkodId),
                          kodkodprob(KodkodToProbVal,KodkodToProbId),
                          Predicate) ),
    catch( send_problem(ProblemId),
           error(_,_),
           (reset_kodkod_process,send_problem(ProblemId))).

get_original_problem(Id,OrigPredicate) :-
    known_problem(Id,_,_,_,_,_,OrigPredicate).

extract_all_types(Typemap,Types) :-
    avl_range(Typemap,Relations),
    include(is_base_relation,Relations,BaseRelations),
    %print(extract_int_types(Relations)),nl,
    extract_int_types(Relations,IntTypes),
    %print('Result : '), print(IntTypes),nl,
    append(IntTypes,BaseTypes,Types),
    maplist(extract_type, BaseRelations, BaseTypes).

is_base_relation(relation(_,_,reldef(_,_,_,_,_))).
extract_type(relation(Id,Size,_Relspec),type(Id,Size)).

:- use_module(kodkod_tools,[pow2integer_relation_kodkod_name/1]).
extract_int_types(Relations,IntTypes) :-
    exclude(is_base_relation,Relations,IntRelations),
    extract_int_types2(IntRelations,IntTypes).
extract_int_types2([],[]) :- !.
extract_int_types2(IntRelations,[inttype(PowName,PowMin,PowMax,IntsetName,Min,Max)]) :-
    selectchk(relation(IntsetName,_,intsetrel(_Id,RMin,Max)),IntRelations,Rest),!,
    Min is min(RMin,0),
    extract_int_types2(Rest,RestResult),
    (RestResult = [inttype(PowName,PowMin,PowMax)] -> true
     ; RestResult = [] ->  kodkod_tools:pow2integer_relation_kodkod_name(PowName),
                           PowMin=Min, PowMax=Max % I have no idea if this is correct; TO DO: Daniel check this
     ). % What if RestResult is something else ???
extract_int_types2(IntRelations,[inttype(PowName,PowMin,PowMax)]) :-
    memberchk(relation(PowName,_,pow2rel(_Id,RPowMin,PowMax)),IntRelations),!,
    PowMin is min(RPowMin,0).

prepare_id_mapping(Ids,Idmap,ProbToKodkod,KodkodToProb) :-
    prepare_id_mapping2(Ids,Idmap,ProbToKodkod,KodkodToProbList),
    list_to_avl(KodkodToProbList, KodkodToProb).
prepare_id_mapping2([],_,[],[]).
prepare_id_mapping2([TId|Rest],Idmap,[PK|ProbToKodkod],[KP|KodkodToProb]) :-
    prepare_id_mapping3(TId,Idmap,PK,KP),
    prepare_id_mapping2(Rest,Idmap,ProbToKodkod,KodkodToProb).
prepare_id_mapping3(TId,Idmap,bid(BType,P)-K,K-bid(BType,P)) :-
    get_texpr_id(TId,P),
    get_texpr_type(TId,BType1),
    adapt_seq_type(BType1,BType),
    avl_fetch(P,Idmap,reldef(K,_Types,_SetOrSingle,_ExactOrSubset,_Set)).

adapt_seq_type(seq(T),S) :- !,S=set(couple(integer,T)).
adapt_seq_type(T,T).

prepare_value_mapping(Valuemap,ProbToKodkod,KodkodToProb) :-
    avl_to_list(Valuemap,ProbKodkodList),
    maplist(reverse_key_value, ProbKodkodList, KodkodProbList),
    list_to_avl(KodkodProbList,KodkodToProb),
    maplist(remove_type_info,ProbKodkodList, L1),
    list_to_avl(L1,ProbToKodkod).
reverse_key_value((T:P)-K, (T:K)-P).
remove_type_info((_T:P)-K, P-K).

get_new_problem_id(ProblemId) :-
    retract(next_problem_id(Id)),
    NextId is Id+1,
    assert(next_problem_id(NextId)),
    ProblemId = Id.

extract_all_relations(Idmap,Relations) :- avl_range(Idmap,Relations).

send_problem(Id) :-
    kodkod_process(_,KOut,_),
    known_problem(Id,Types,Relations,Formula,_ToKodkod,_ToProB,_OrigPredicate),
    get_preference(time_out,TO),
    get_preference(kodkod_symmetry_level,Symmetry),
    get_preference(kodkod_sat_solver,Solver),
    kodkod_printer:kkp_problem(KOut,TO,Symmetry,Solver,Id,problem(Types,Relations,Formula,xxx,none)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% communication with external java process
:- dynamic kodkod_process/3.

reset_kodkod_process :- retractall(kodkod_process(_,_,_)).

startup_kodkod :- kodkod_process(_,_,_), !.
startup_kodkod :-
    absolute_file_name(prob_lib('.'), KodkodDir),
    kodkod_jar(KodkodJar),

    atom_concat('-Djava.library.path=', KodkodDir, JLibpath),

    get_command_path('java',CmdPath),

    %format('Starting Kodkod JAR ~w using ~w~n',[KodkodJar,JLibpath]),
    get_preference(kodkod_sat_solver,Solver),

    process_create(CmdPath,
                   [JLibpath, '-Xmx1024M','-jar', KodkodJar,Solver],
                   [process(KProc),stdout(pipe(KIn)),stdin(pipe(KOut))]),
    check_correct_startup(KIn),

    assert(kodkod_process(KProc,KOut,KIn)).

check_correct_startup(S) :-
    catch( (read_line(S,Line),check_first_line(Line)),
           E,
           add_error_and_fail(kodkod_process, 'Exception while connecting to Kodkod process', E)).
check_first_line(Line) :-
    append("ProB-Kodkod started: ",Msg,Line),!,
    format('Kodkod module started up successfully (SAT solver ~s).~n',[Msg]).
check_first_line(_Line) :-
    add_error_and_fail(kodkod_process,'Unable to start-up Kodkod process.\n').

kodkod_jar(KodkodJar) :-
    absolute_file_name(prob_lib('probkodkod.jar'), KodkodJar),
    ( file_exists(KodkodJar,read) -> true
    ; otherwise ->
        add_error_and_fail(kodkod_process,'Unable to read Kodkod library. Please check your installation or disable the use of Kodkod')).

stop_kodkod :-
    kodkod_process(PId,KOut,KIn),
    call_cleanup( (write(KOut, 'stop.\n'), flush_output(KOut) ),
                  ( process_wait(PId,_), close(KOut), close(KIn),
                      retractall(kodkod_process(_,_,_)))).

%kodkod_problem(Id,Problem) :-
%    kodkod_process(_,KOut,_),
%    get_preference(time_out,TO),get_preference(kodkod_symmetry_level,Symmetry),get_preference(kodkod_sat_solver,Solver),
%    kkp_problem(KOut,TO,Symmetry,Solver,Id,Problem).

kodkod_request_intern(ProblemId, Signum, MaxNrOfSols, LocalState, State) :-
    known_problem(ProblemId,_Size,_Relations,_Formula,ToKodkod,ToProB,_OrigPredicate),
    extract_known_values(ToKodkod,LocalState,State,Known,UnknownList),
    copy_term(UnknownList,UnknownListCopy),
    list_to_avl(UnknownListCopy,Unknown),
    !,
    debug_println(19,send_kodkod_request(ProblemId,Signum,MaxNrOfSols,Known)),
    send_request(ProblemId, ToProB, Signum, MaxNrOfSols, Known, Unknown),
    %describe_sol(UnknownListCopy),
    UnknownList=UnknownListCopy.
    %print_bstate(State).

%describe_sol(List) :-
%    nl,write('Kodkod solution:'),nl,describe_sol2(List).
%describe_sol2([]).
%describe_sol2([Id-Value|Rest]) :-
%    describe_sol3(Id,Value),
%    describe_sol2(Rest).
%describe_sol3(Id,Value) :-
%    write(Id),write(': '),print_bvalue(Value),nl.


extract_known_values(probkodkod(ValueMap,Ids),LocalState,State,Out,Unknown) :-
    extract_known_values2(Ids,ValueMap,LocalState,State,Out,[],Unknown,[]).
extract_known_values2([],_ValueMap,_LocalState,_State,Known,Known,Unknown,Unknown).
extract_known_values2([bid(Type,Prob)-Kodkod|Rest],ValueMap,LocalState,State,
                      KnownIn,KnownOut,UnknownIn,UnknownOut) :-
    extract_known_value(Type,Prob,Kodkod,ValueMap,LocalState,State,KnownIn,KnownInter,UnknownIn,UnknownInter),
    extract_known_values2(Rest,ValueMap,LocalState,State,KnownInter,KnownOut,UnknownInter,UnknownOut).
extract_known_value(Type,Prob,Kodkod,ValueMap,LocalState,State,KnownIn,KnownOut,UnknownIn,UnknownOut) :-
    lookup_value_for_existing_id(Prob,LocalState,State,Value),
    ( ground(Value) ->
        map_prob_value_to_kodkod(Type,Value,ValueMap,KValue),
        KnownIn = [kout(Kodkod,KValue)|KnownOut],
        UnknownIn = UnknownOut
    ; otherwise ->
        KnownIn = KnownOut,
        UnknownIn = [Kodkod-Value|UnknownOut]),!.
extract_known_value(_Type,Prob,_Kodkod,_ValueMap,_LocalState,_State,_KnownIn,_KnownOut,_UnknownIn,_UnknownOut) :-
    add_error_and_fail(kodkod_process,'mapping ProB value to Kodkod failed', Prob).

map_prob_value_to_kodkod(set(_),Prob,ValueMap,Kodkod) :-
    !, expand_custom_set_to_list(Prob,List),
    map_prob_values_to_kodkod(List,ValueMap,Kodkod).
map_prob_value_to_kodkod(_Type,Prob,ValueMap,Kodkod) :-
    map_prob_values_to_kodkod([Prob],ValueMap,Kodkod).
map_prob_values_to_kodkod([],_,[]).
map_prob_values_to_kodkod([P|Prest],ValueMap,[tuple(K)|Krest]) :-
    map_prob_value_to_kodkod2(P,ValueMap,K,[]),
    map_prob_values_to_kodkod(Prest,ValueMap,Krest).
map_prob_value_to_kodkod2((A,B),ValueMap) -->
    !,map_prob_value_to_kodkod2(A,ValueMap),
    map_prob_value_to_kodkod2(B,ValueMap).
map_prob_value_to_kodkod2(int(I),_ValueMap) --> !,[I].
map_prob_value_to_kodkod2(Prob,ValueMap) --> [Kodkod],
    {avl_fetch(Prob,ValueMap,Kodkod)}.


:- use_module(probsrc(debug),[debug_println/2]).
send_request(ProblemId, ToProb, Signum, MaxNrOfSols,OutValues, In) :-
    kodkod_process(_,OStream,IStream),
    debug:debug_println(19,sending_kodkod_request(ProblemId,Signum,MaxNrOfSols)),
    kkp_request(ProblemId, Signum, MaxNrOfSols, OutValues, OStream),
    get_solutions_from_stream(OStream,IStream,ProblemId,MaxNrOfSols,ToProb,In).

read_rest(ProblemId,MaxNrOfSols,ToProb,In) :-
    kodkod_process(_,OStream,IStream),
    kkp_list(ProblemId,MaxNrOfSols,OStream),
    get_solutions_from_stream(OStream,IStream,ProblemId,MaxNrOfSols,ToProb,In).

get_solutions_from_stream(OStream,IStream,ProblemId,MaxNrOfSols,ToProB,In) :-
    flush_output(OStream),
    read_answer_from_kodkod(IStream,Answer),
    !,
    %print(answer_from_stream(Answer)),nl,
    process_answer(Answer,ProblemId,MaxNrOfSols,ToProB,In).

% read an answer but skip over statistics infos:
read_answer_from_kodkod(IStream,Answer) :-
    read(IStream,Answer1),
    (Answer1=stats(T,S,C,V,P) % special statistics info fact
     -> debug_format(20,'Kodkod Statistics: ~w ms translation, ~w ms solving, ~w clauses, ~w variables, ~w primary variables~n',[T,S,C,V,P]),
        read_answer_from_kodkod(IStream,Answer)
     ; Answer=Answer1).

process_answer(solutions(Sols,More,Times),ProblemId,MaxNrOfSols,ToProb,In) :-
    format('Times for computing solutions: ~w~n',[Times]),
    save_java_comp_time(Times),
    get_solutions(Sols,More,ProblemId,MaxNrOfSols,ToProb,In).
process_answer(sat_timeout,ProblemId,_,_ToProb,_In) :-
    add_warning(kodkod_process,'Kodkod SAT Solver Timeout for Problem Id: ',ProblemId),
    %throw(time_out). % can only be caught by time_out not by on_exception !
    throw(enumeration_warning(kodkod_timeout,ProblemId,'','',critical)).
process_answer(sat_non_incremental,ProblemId,_,_ToProb,_In) :-
    add_warning(kodkod_process,'Kodkod SAT Solver is non incremental and cannot compute all solutions for Problem Id: ',ProblemId),
    throw(enumeration_warning(kodkod_timeout,ProblemId,'','',critical)).
process_answer(unknown(_),ProblemId,_,_ToProb,_In) :-
    add_warning(kodkod_process,'Kodkod SAT Solver error occured for Problem Id: ',ProblemId),
    throw(enumeration_warning(kodkod_timeout,ProblemId,'','',critical)).


get_solutions(Sols,_More,_ProblemId,_,ToProb,In) :-
    member(Sol,Sols), get_solution(Sol,ToProb,In).
get_solutions(_Sols,more,ProblemId,MaxNrOfSols,ToProb,In) :-
    read_rest(ProblemId,MaxNrOfSols,ToProb,In).

get_solution([],_,_).
get_solution([Binding|BRest],kodkodprob(VMap,IdMap),BValues) :-
    ( Binding = b(Name,KValues) -> true
    ; Binding = s(Name,KValues) -> Type = set(_)),
    avl_fetch(Name,BValues,BValue),
    avl_fetch(Name,IdMap,bid(Type,_BName)),
    map_kodkod_prob(Type,KValues,VMap,Result),!,
    equal_object(BValue,Result),
    get_solution(BRest,kodkodprob(VMap,IdMap),BValues).

map_kodkod_prob(set(Type),Tuples,Map,Value) :-
    !,map_kodkod_prob2(Tuples,Type,Map,Value).
map_kodkod_prob(Type,KValue,Map,BValue) :-
    map_kodkod_prob2([KValue],Type,Map,[BValue]).
map_kodkod_prob2([],_,_,[]).
map_kodkod_prob2([Values|TRest],Type,Map,[BValue|BRest]) :-
    map_kodkod_prob3(Type,Map,BValue,Values,[]),
    map_kodkod_prob2(TRest,Type,Map,BRest).
map_kodkod_prob3(couple(TA,TB),Map,(BA,BB)) -->
    map_kodkod_prob3(TA,Map,BA),
    map_kodkod_prob3(TB,Map,BB).
map_kodkod_prob3(integer,_,int(Value)) --> !,[Value].
map_kodkod_prob3(Type,Map,BValue) -->
    [K],{avl_fetch(Type:K,Map,BValue),!}.

:- dynamic java_comp_time/1.
clear_java_comp_time :-
    retractall(java_comp_time(_)).
get_java_comp_time(Time) :-
    findall( T, (java_comp_time(Ts),member(T,Ts)), Times),
    sumlist(Times,Time).
save_java_comp_time(Times) :-
    assert(java_comp_time(Times)).
