% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(ltl_propositions,[check_ap/2, % testing atomic preposition in a state
                            check_atomic_property_formula/2,
                            check_transition/4, % testing whether a transition proposition can be executed from some state
                            check_enabled/2,
                            trans/3, % exported for use in ctl.pl
                            get_por_preferences/1
                            ]).

/* SICSTUS Libraries */
:- use_module(library(lists)).

/* ProB standard modules */
% B
:- use_module(probsrc(kernel_objects),[equal_object/2]).
:- use_module(probsrc(bsyntaxtree), [create_texpr/4, def_get_texpr_id/2]). % used for test cases
:- use_module(probsrc(bmachine),[b_get_machine_operation/4]).
:- use_module(probsrc(b_interpreter),[b_compute_expression_nowf/4,b_test_boolean_expression_wf/3]).
% state space modules
:- use_module(probsrc(state_space),[impl_trans_term/3,
                           visited_expression/2,
                           %transition/3,
                           state_space_add/2,retract_visited_expression/2  % used for test cases
                           ]).
:- use_module(probsrc(store),[empty_state/1,lookup_value/3,add_var_to_localstate/4]).
:- use_module(probsrc(specfile),[state_corresponds_to_initialised_b_machine/2]).
% utility modules
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(self_check)).
% Promela
:- use_module(probpromela(promela_tools),[eval_promela/4]).
% other
:- use_module(probsrc(eclipse_interface),[test_boolean_expression_in_node/2]).
:- use_module(probsrc(plugins),[plugin_evaluate_transition_predicate/3]).

/* ProB LTL modules */
:- use_module(probltlsrc(ltl_tools)).
:- use_module(probltlsrc(ltl_translate),[pp_ltl_formula/2]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides predicates for evaluating atomic and transition prepositions.').

% check an atomic proposition in a given node (i.e. state)
check_ap(bpred(Property),Node) :-
    !, test_boolean_expression_in_node(Node,Property).
check_ap(promela_pred(Predicate),Node) :-
    visited_expression(Node,State),!,
    eval_promela(Predicate,global,State,Result),!,
    ( Result == 0    -> fail
    ; number(Result) -> true
    ; otherwise      -> add_error(ltl,'Promela eval returned a non-boolean value: ',Result),fail).
check_ap(enabled(Op),Node) :- !,
    check_enabled(Op,Node).
check_ap(det(OpList),Node) :- !,
    check_if_maximum_one_enabled(OpList,Node).
check_ap(dlk(OpList),Node) :- !,
    check_if_all_disabled(OpList,Node).
check_ap(ctrl(OpList),Node) :- !,
    is_exactly_one_operation_enabled(OpList,Node).
check_ap(deadlock,Node) :- !, \+ trans(Node,_,_).
check_ap(sink,Node) :- !,
    \+ ( trans(Node,Next,_),Next \= Node ).
check_ap(stateid(SpecNode),Node) :- !, SpecNode = Node.

check_if_maximum_one_enabled(OpList,Node) :-
    check_if_maximum_one_enabled(OpList,0,Node).

check_if_maximum_one_enabled([],_NrEnabled,_Node).
check_if_maximum_one_enabled([Op|Ops],NrEnabled,Node) :-
    (check_enabled(Op,Node) ->
        NrEnabled == 0, NrEnabled1 = 1
    ;   NrEnabled1 = 0
    ), check_if_maximum_one_enabled(Ops,NrEnabled1,Node).

is_exactly_one_operation_enabled(OpList,Node) :-
    is_exactly_one_operation_enabled(OpList,0,Node).

is_exactly_one_operation_enabled([],NrEnabled,_Node) :-
    NrEnabled=1.
is_exactly_one_operation_enabled([Op|Ops],NrEnabled,Node) :-
    (check_enabled(Op,Node) ->
        NrEnabled == 0, NrEnabled1 = 1
    ;   NrEnabled1 = NrEnabled
    ), is_exactly_one_operation_enabled(Ops,NrEnabled1,Node).

check_if_all_disabled([],_Node).
check_if_all_disabled([Op|Ops],Node) :-
    \+check_enabled(Op,Node),
    check_if_all_disabled(Ops,Node).

check_enabled(Spec,SrcNode) :-
        check_transition(Spec,Transition,SrcNode,DestNode),
        trans(SrcNode,DestNode,Transition),!.


trans(A,B,Op) :- impl_trans_term(A,Op,B).
% THE FOLLOWING CODE IS BROKEN: TO DO FIX or REMOVE and probably remove use_por_for_ltl preference
%:- use_module(probltlsrc(ample_sets),[compute_ample_set2/2]).
%trans(A,B,Op) :-
%    ( (preference(use_por_for_ltl,true), animation_mode(b))
%        -> get_por_preferences(POROptions), % poor implementation (ltl predicates should not be called from here)
%           ample_sets: compute_ample_set2(A,POROptions), % TO DO : THIS HAS 3 ARGUMENTS !!!
%           transition(A,Op,B) % from state_space.pl
%        ;  impl_trans_term(A,Op,B) /* will automatically compute as required */
%    ).

:- use_module(probsrc(preferences),[get_preference/2]).
get_por_preferences(por(TYPE,UseEnableGraph,Depth,UsePGE)) :-
  get_preference(por,TYPE),
  get_preference(enable_graph,UseEnableGraph),
  get_preference(enable_graph_depth,Depth),
  get_preference(pge,UsePGE).

% a utility to check restricted atomic LTL properties on single nodes
check_atomic_property_formula(ap(ParsedFormula),NodeId) :- !,
   check_ap(ParsedFormula,NodeId).
%check_atomic_property_formula(ap(ParsedFormula),NodeId) :- !,
check_atomic_property_formula(and(P1,P2),NodeId) :- !,
   check_atomic_property_formula(P1,NodeId),
   check_atomic_property_formula(P2,NodeId).
check_atomic_property_formula(or(P1,P2),NodeId) :- !,
   (check_atomic_property_formula(P1,NodeId) ;
    check_atomic_property_formula(P2,NodeId)).
check_atomic_property_formula(implies(P1,P2),NodeId) :- !,
   (check_atomic_property_formula(P1,NodeId)
    -> check_atomic_property_formula(P2,NodeId) ; true).
check_atomic_property_formula(not(P1),NodeId) :- !,
   (check_atomic_property_formula(P1,NodeId) -> fail ; true).
check_atomic_property_formula(Formula,_) :- %print(Formula),nl,
   pp_ltl_formula(Formula,Text),!,
   add_error(check_atomic_property_formula,'LTL formula is not an atomic property: ',Text).
check_atomic_property_formula(Formula,N) :-
   add_internal_error('LTL formula invalid: ',check_atomic_property_formula(Formula,N)),fail.

% check_transition(Spec,Transition,SrcNode,DstNode)
%   Checks if a transition fulfils the specification, it fails if not.
%   Please note, that this predicate is called _before_ the actual transition
%   is known.
%  Spec: specification that encodes how the transition should look like
%        See below for the supported transitions
%  Transition: The transition itself as it is stored in the second arg of state_space:transition/3
%  SrcNode: The Id of the source state (first arg of transition/3)
%  DstNode: The Id of the destination state (third arg of transition/3)

:- assert_must_succeed( check_transition(bop(dummyop,2,0,[],[]),dummyop(int(1),int(2)),tsrc,tdst) ).
:- assert_must_fail( check_transition(bop(dummyop,2,0,[],[]),otherop(int(1),int(2)),tsrc,tdst) ).
:- assert_must_fail( check_transition(bop(dummyop,2,0,[],[]),dummyop(int(1)),tsrc,tdst) ).
:- assert_must_succeed(call_cleanup(( empty_state(E),
                                      state_space_add(tsrc,E),
                                      create_texpr(integer(2),integer,[],I2),
                                      check_transition(bop(dummyop,2,0,[2/I2],[]),
                                                       dummyop(int(1),int(2)),tsrc,tdst)),
                                    retract_visited_expression(tsrc,_))).
:- assert_must_fail(call_cleanup(( empty_state(E),
                                   state_space_add(tsrc,E),
                                   create_texpr(integer(2),integer,[],I2),
                                   check_transition(bop(dummyop,2,0,[1/I2],[]),
                                                    dummyop(int(1),int(2)),tsrc,tdst)),
                                 retract_visited_expression(tsrc,_))).
:- assert_must_succeed( check_transition(bop(dummyop,1,1,[],[]),'-->'(dummyop(int(1)),[int(2)]),tsrc,tdst) ).
:- assert_must_fail( check_transition(bop(dummyop,1,1,[],[]),dummyop(int(1)),tsrc,tdst) ).
:- assert_must_succeed(call_cleanup(( empty_state(E),
                                      state_space_add(tsrc,E),
                                      create_texpr(integer(2),integer,[],I2),
                                      check_transition(bop(dummyop,1,1,[],[1/I2]),
                                                       '-->'(dummyop(int(1)),[int(2)]),tsrc,tdst)),
                                    retract_visited_expression(tsrc,_))).
:- assert_must_fail(call_cleanup(( empty_state(E),
                                   state_space_add(tsrc,E),
                                   create_texpr(integer(2),integer,[],I2),
                                   check_transition(bop(dummyop,1,1,[],[1/I2]),
                                                    '-->'(dummyop(int(1)),[int(1)]),tsrc,tdst)),
                                 retract_visited_expression(tsrc,_))).

% bop(Name,NumberArgs,NumberResults,ArgPatterns,ResPatterns):
%   The transition is a B operation with name Name, NumberArgs parameters and NumberResults
%   result values.
%   ArgPatterns is a list of Terms of the form Index/Expr where Index is the
%     number of the parameter that should have the value of the expression Expr.
%   ResPatterns is a list analogous to ArgPatterns on the result
check_transition(bop(Name,NumberArgs,NumberResults,ArgPatterns,ResPatterns),
                 Transition,SrcNode,_DstNode) :-
    !, functor(Op,Name,NumberArgs),
    ( NumberResults == 0 -> Transition = Op
    ; otherwise          ->
        Transition = '-->'(Op,ReturnValues),
        length(ReturnValues,NumberResults),
        check_bop_res_patterns(ResPatterns,SrcNode,State,ReturnValues)),
    check_bop_arg_patterns(ArgPatterns,SrcNode,State,Op).
check_transition(btrans(TransPred),Transition,SrcNode,DstNode) :-
    !, check_eventb_transition(TransPred,Transition,SrcNode,DstNode).
check_transition(csp(Channel,Arity,Args),Transition,_SrcNode,_DstNode) :-
    !, check_csp_event(Channel,Arity,Args,Transition).
check_transition(xtl(Op,Arity,PatternArgs),Transition,_SrcNode,_DstNode) :-
    !, check_xtl_transition(Op,Arity,PatternArgs,Transition).
check_transition(plugin_transition_check(Id,Predicate),Transition,_SrcNode,_DstNode) :-
    !, plugin_evaluate_transition_predicate(Id,Predicate,Transition).
check_transition(Spec,_Trans,_SrcNode,_DstNode) :-
    functor(Spec,Name,Arity), add_error(ltl, 'Unknown transition specification: ', Name/Arity), fail.

:- block check_xtl_transition(?,?,?,-).
check_xtl_transition(Op,Arity,PatternArgs,Transition) :-
    ( Arity=0 ->
        functor(Transition,Op,_) /* in this case: pattern match any operation with right number of args */
    ;   %functor(Transition,Op,Arity),
        Transition =.. [Op|Args],
        check_xtl_arguments(Args,PatternArgs)
    ).

check_bop_arg_patterns([],_,_,_).
check_bop_arg_patterns([P|Prest],Node,State,Op) :-
    check_bop_arg_patterns2([P|Prest],Node,State,Op).
:- block check_bop_arg_patterns2(?,?,?,-).
check_bop_arg_patterns2([],_,_,_).
check_bop_arg_patterns2([Index/Expr|Prest],Node,State,Op) :-
    check_bop_arg_pattern(Index,Expr,Node,State,Op),
    check_bop_arg_patterns2(Prest,Node,State,Op).
check_bop_arg_pattern(Index,Expr,Node,State,Op) :-
    arg(Index,Op,OpValue),  % exception if Index is expression like 1+1 <----
    check_b_operation_value(OpValue,Expr,Node,State).

check_bop_res_patterns([],_,_,_).
check_bop_res_patterns([P|Prest],Node,State,OpResult) :-
    check_bop_res_patterns2([P|Prest],Node,State,OpResult).
:- block check_bop_res_patterns2(?,?,?,-).
check_bop_res_patterns2([],_,_,_).
check_bop_res_patterns2([Index/Expr|Prest],Node,State,OpResult) :-
    check_bop_res_pattern(Index,Expr,Node,State,OpResult),
    check_bop_res_patterns2(Prest,Node,State,OpResult).
check_bop_res_pattern(Index,Expr,Node,State,OpResult) :-
    nth1(Index,OpResult,OpValue),
    check_b_operation_value(OpValue,Expr,Node,State).

check_b_operation_value(OpValue,Expr,Node,State) :-
    ( var(State) ->
        visited_expression(Node,RawState),
        state_corresponds_to_initialised_b_machine(RawState,State)
    ; otherwise -> true),
    empty_state(LS),
    b_compute_expression_nowf(Expr,LS,State,PatternValue),
    equal_object(PatternValue,OpValue),!.


:- block check_eventb_transition(?,-,?,?).
check_eventb_transition(event(Name),Transition,_SrcNode,_DstNode) :- !,
    functor(Transition,Name,_).
check_eventb_transition(event(Name,Predicate,PoststateAccess),Transition,SrcNode,DstNode) :- !,
    functor(Transition,Name,_),
    visited_expression(SrcNode,CurState),
    state_corresponds_to_initialised_b_machine(CurState,CurBState),
    parameter_values(Transition,LocalState1),
    add_post_state_values(PoststateAccess,DstNode,LocalState1,LocalState),
    b_test_boolean_expression_wf(Predicate,LocalState,CurBState).

add_post_state_values([],_DstNode,State,State) :- !.
add_post_state_values(Access,Node,InState,OutState) :-
    visited_expression(Node,DstState1),
    state_corresponds_to_initialised_b_machine(DstState1,DstState),
    add_post_state_values2(Access,DstState,InState,OutState).
add_post_state_values2([],_DstState,State,State).
add_post_state_values2([poststate(Orig,Primed)|Prest],DstState,InState,OutState) :-
    lookup_value(Orig,DstState,Value),
    add_var_to_localstate(Primed,Value,InState,InterState),
    add_post_state_values2(Prest,DstState,InterState,OutState).

parameter_values(Transition,LocalState) :-
    Transition =.. [Name|Arguments],
    b_get_machine_operation(Name,_,Parameters,_),
    empty_state(Empty),
    add_parameter_values2(Parameters,Arguments,Empty,LocalState).
add_parameter_values2([],[],State,State).
add_parameter_values2([TParameter|Prest],[Argument|Arest],InState,OutState) :-
    def_get_texpr_id(TParameter,Parameter),
    add_var_to_localstate(Parameter,Argument,InState,InterState),
    add_parameter_values2(Prest,Arest,InterState,OutState).

/*
:- assert_must_succeed(( ltl:split_operation('ch.2.3','.',ch,A), A=[2,3] )).
:- assert_must_succeed(( ltl:split_operation('ch.2.3','.',ch,[2,3]) )).
:- assert_must_succeed(( ltl:split_operation('ch','.',ch,[]) )).
:- assert_must_succeed(( ltl:split_operation('ch.','.',ch,[A]), A='' )).
split_operation(OpAtom, Sep,  OpName, OpArgs) :-
   atom_codes(OpAtom,ListAscii), atom_codes(Sep,[SepACode]),
   split_op(ListAscii,SepACode,HeadA, ArgsA),
   atom_codes(Head,HeadA), OpName=Head,
   l_name(Args,ArgsA), Args=OpArgs.
l_name([],[]).
l_name([H|T],[AH|AT]) :- name(H,AH),  % using name to generate numbers if possible
                               l_name(T,AT).

split_op([],_,[],[]). % no separator found
split_op([H|T],Sep,OpName,Args) :-
   (H=Sep -> Args=[FirstArg|OtherArgs],OpName=[],split_args(T,Sep,FirstArg,OtherArgs)
    ; OpName = [H|CT], split_op(T,Sep,CT,Args)).

split_args([],_Sep,[],[]).
split_args([H|T],Sep,FirstArg,OtherArgs) :-
   (H=Sep -> FirstArg=[], OtherArgs=[NextArg|TA],split_args(T,Sep,NextArg,TA)
    ; FirstArg = [H|CT], split_args(T,Sep,CT,OtherArgs)).
*/


check_xtl_arguments([],[]).
check_xtl_arguments([],[H|T]) :-
   print('### Extra arguments in LTL/CTL pattern: '),print([H|T]),nl,fail.
check_xtl_arguments([H|T],[]) :- print(ltl_check_xtl_arguments_ignoring([H|T])),nl.
check_xtl_arguments([H|T],[PatH|PatT]) :-
    check_xtl_arg(PatH,H,T,RemArgs),
    check_xtl_arguments(RemArgs,PatT).

check_xtl_arg('_',_,TArgs,RemArgs) :- !,RemArgs=TArgs.
check_xtl_arg('?',_,TArgs,RemArgs) :- !,RemArgs=TArgs.
check_xtl_arg(X,Y,TArgs,RemArgs) :- atomic(Y),!,X=Y,RemArgs=TArgs.
check_xtl_arg(X,Y,TArgs,RemArgs) :- nonvar(Y),!,Y=..[Functor|YArgs], Functor=X,
  append(YArgs,TArgs,RemArgs).
check_xtl_arg(X,Y,TArgs,RemArgs) :- var(Y),
   print(variable_for_ltl_pattern(X)),nl,
   X=Y,RemArgs=TArgs.
