% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check an LTL formula on a concrete path

:- module(ltl_verification, [evaluate_ltl_formula/6,evaluate_ltl_fairness/6]).

/* SICSTUS libraries */
:- use_module(library(lists)).
:- use_module(library(between),[numlist/3]).
:- use_module(library(ordsets),[ord_intersection/3]).

/* ProB standard modules */
:- use_module(probsrc(error_manager),[add_error/3]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides predicates for checking an LTL formula on a concrete path.').

/*
 * evaluate_ltl_formula(Formula,StateTransitions,Type,CallbackAP,CallbackTP,Result):
 *  Formula: AST of the LTL formula
 *  StateTransitions: List of the form [strans(StateId,TransId)] which describes the
 *    sequence of states and transitions between them. If there is no next state in
 *    the sequence, TransId can be 'none'
 *  Type: form of the counter-example. If it's a lasso-like form, Type should be
 *    loop(ReEntry) where ReEntry is the offset of the first state of the loop
 *    in the sequence, starting with 0.
 *    If Type is not loop(_), the sequence is regarded as finite
 *  CallbackAP: Callback predicate to check an atomic proposition. Callbacks are of
 *    the form CallbackAP(AP,StateId) where AP is the specification of an atomic
 *    proposition ap(AP) in the AST of the formula and StateId the the ID of the
 *    state in which it should be evaluated.
 *    If the call succeeds, the proposition is regarded as true, if not as false.
 *  CallbackTP: Like CallbackAP, CallbackTP is a predicate to check a proposition
 *    TP on a transition as it appears as action(TP) in the LTL formula.
 *    The call is of the form CallbackTP(TP,StateId,TransID).
 *  Result: true, false or unknown
 *    A formula can only be unknown if the sequence is finite (without a loop).
 */
:- meta_predicate evaluate_ltl_formula(+,+,+,2,3,-).
evaluate_ltl_formula(Formula,StateTransitions,Type,CallbackAP,CallbackTP,Result) :-
    check_sequence(StateTransitions,Type),
    length(StateTransitions,N),
    Path = path(StateTransitions,N,Type,CallbackAP,CallbackTP),
    eval_ltl_formula(Formula,0,Path,Result),
    % In case of an unknown result, check if the sequence if finite
    check_result(Type,Result,Formula).

check_sequence(StateTransitions,Type) :-
    ( Type=loop(_),last(StateTransitions,strans(_,none)) ->
        add_error(ltl_verification,'last element of lasso sequence has no transition id',
                  StateTransitions)
    ; otherwise -> true).

check_result(loop(_),unknown,Formula) :-
    !,add_error(ltl_verification,
                'evaluate_ltl_formula computed unknown result for infinite sequence',
                Formula).
check_result(loop(_),deadlock,Formula) :-
    !,add_error(ltl_verification,
                'evaluate_ltl_formula computed unknown result for dead locking sequence',
                Formula).
check_result(_Type,_Result,_Formula).

eval_ltl_formula(false,_Index,_Path,false).
eval_ltl_formula(true,_Index,_Path,true).
eval_ltl_formula(ap(AP),Index,Path,Value) :-
    access_index_in_path(Index,Path,State,_Trans),
    path_ap(Path,Call),
    ( call(Call,AP,State) -> Value=true
    ; otherwise           -> Value=false).
eval_ltl_formula(action(TP),Index,Path,Value) :-
    access_index_in_path(Index,Path,State,Trans),
    ( Trans=none ->
        ( eval_ltl_is_deadlock(Index,Path) -> Value=false
        ; otherwise                        -> Value=unknown)
    ; otherwise ->
        path_tp(Path,Call),
        ( call(Call,TP,State,Trans) -> Value=true
        ; otherwise                 -> Value=false)).
eval_ltl_formula(not(A),Index,Path,Value) :-
    eval_ltl_formula(A,Index,Path,VA),
    ltl_eval_not(VA,Value),!.
eval_ltl_formula(and(A,B),Index,Path,Value) :-
    eval_ltl_formula(A,Index,Path,VA),
    eval_ltl_formula(B,Index,Path,VB),
    ltl_eval_and(VA,VB,Value),!.
eval_ltl_formula(or(A,B),Index,Path,Value) :-
    eval_ltl_formula(A,Index,Path,VA),
    eval_ltl_formula(B,Index,Path,VB),
    ltl_eval_or(VA,VB,Value),!.
eval_ltl_formula(implies(A,B),Index,Path,Value) :-
    eval_ltl_formula(A,Index,Path,VA),
    eval_ltl_formula(B,Index,Path,VB),
    ltl_eval_implies(VA,VB,Value),!.
eval_ltl_formula(finally(A),Index,Path,Value) :-
    eval_ltl_formula(until(true,A),Index,Path,Value).         % TODO: Do not use normalisation
eval_ltl_formula(globally(A),Index,Path,Value) :-
    eval_ltl_formula(not(finally(not(A))),Index,Path,Value).  % TODO: Do not use normalisation
eval_ltl_formula(weakuntil(A,B),Index,Path,Value) :-
    eval_ltl_formula(or(globally(A),until(A,B)),Index,Path,Value).  % TODO: Do not use normalisation
eval_ltl_formula(release(A,B),Index,Path,Value) :-
    eval_ltl_formula(not(until(not(A),not(B))),Index,Path,Value).   % TODO: Do not use normalisation
eval_ltl_formula(once(A),Index,Path,Value) :-
    eval_ltl_formula(since(true,A),Index,Path,Value).      % TODO: Do not use normalisation
eval_ltl_formula(historically(A),Index,Path,Value) :-
    eval_ltl_formula(not(once(not(A))),Index,Path,Value).  % TODO: Do not use normalisation
eval_ltl_formula(trigger(A,B),Index,Path,Value) :-
    eval_ltl_formula(not(since(not(A),not(B))),Index,Path,Value). % TODO: Do not use normalisation
eval_ltl_formula(next(A),Index,Path,Value) :-
    last_index_in_path(Path,Last),
    ( Index<Last ->                                           % we're are somewhere in the path,
        Next is Index+1,eval_ltl_formula(A,Next,Path,Value)   %  just use the next position
    ; path_type(Path,loop(ReEntry)) ->                        % End of the path: If it's a loop,
        eval_ltl_formula(A,ReEntry,Path,Value)                %  use the start of the lasso as next
    ; eval_ltl_is_deadlock(Index,Path) ->                     % End of the path (no loop): In a
        Value=false                                           %  deadlock X(...) is always false
    ; otherwise ->                                            % End of the path: We don't know the
        Value=unknown).                                       %  next state
eval_ltl_formula(yesterday(A),Index,Path,Value) :-
    ( Index > 0 ->
        Y is Index-1,eval_ltl_formula(A,Y,Path,Value)
    ; otherwise ->                                            % Y(...) is always false in the
        Value=false).                                         %   initial state.
eval_ltl_formula(until(Inv,Goal),Index,Path,Value) :-
    future_indices(Index,Path,Indices),
    eval_until(Indices,Inv,Goal,Path,Value).
eval_ltl_formula(since(Inv,Goal),Index,Path,Value) :-
    eval_since(Index,Inv,Goal,Path,Value).
eval_ltl_is_deadlock(Index,Path) :-
    eval_ltl_formula(ap(deadlock),Index,Path,true).

access_index_in_path(Index,Path,State,Transition) :-
    path_statetransitions(Path,StateTransitions),
    nth0(Index,StateTransitions,strans(State,Transition)).
last_index_in_path(Path,Last) :-
    path_length(Path,N),Last is N-1.

path_statetransitions(path(StateTransitions,_N,_Type,_AP,_TP),StateTransitions).
path_type(path(_ST,_N,Type,_AP,_TP),Type).
path_length(path(_ST,N,_Type,_AP,_TP),N).
path_ap(path(_ST,_N,_Type,AP,_TP),AP).
path_tp(path(_ST,_N,_Type,_AP,TP),TP).

eval_until([],_,_,Path,Result) :-
    last_index_in_path(Path,Last),
    ( path_type(Path,loop(_)) -> Result=false
    ; eval_ltl_is_deadlock(Last,Path) ->
        Result=false
    ; otherwise ->
        Result=unknown).
eval_until([H|T],Inv,Goal,Path,Result) :-
    eval_ltl_formula(Goal,H,Path,GValue),
    ( GValue = true  -> Result=true
    ; otherwise ->
        eval_ltl_formula(Inv,H,Path,IValue),
        eval_until(T,Inv,Goal,Path,NValue),
        ltl_eval_until(IValue,GValue,NValue,Result),!).

future_indices(Index,Path,Result) :-
    last_index_in_path(Path,Last),
    numlist(Index,Last,Rest),
    ( path_type(Path,loop(ReEntry)), Index>ReEntry ->
        I2 is Index-1,
        numlist(ReEntry,I2,Before),
        append(Rest,Before,Result)
    ; otherwise ->
        Result = Rest).

eval_since(Index,Inv,Goal,Path,Result) :-
    eval_ltl_formula(Goal,Index,Path,GValue),
    ( Index = 0     -> Result=GValue
    ; GValue = true -> Result=true
    ; otherwise ->
        eval_ltl_formula(Inv,Index,Path,IValue),
        ( GValue=false,IValue=false -> Result=false
        ; otherwise ->
            Y is Index-1,
            eval_since(Y,Inv,Goal,Path,NValue),
            ltl_eval_until(IValue,GValue,NValue,Result),!  % same logic as until
        )
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% evalute a fairness condition on a given path

/*
 * evaluate_ltl_fairness(WeakStrong,StateTransitions,Type,FairCall,Result):
 *  WeakStrong: weak or strong: Which fairness type should be checked
 *  StateTransitions: [strans(StateId,TransId)] like in evaluate_ltl_formula/6
 *  Type: like in evaluate_ltl_formula/6, for all types but loop(_) the is result always fair
 *  FairCall: Callback predicate to check which transitions occur that match the Spec.
 *    The call has the form FairCall(+StateId,TransId,FairId).
 *    StateId is an ID of a state and TransId the ID of a transition starting from that
 *    state. If the predicate fails, it is assumed that the transition is not subject to
 *    fairness. It it succeeds, the predicate should return a term that is used to
 *    destinguish between actions.
 *    E.g. it could return 'op' for both op(1) and op(2) such that all op(...) transitions
 *    are treated as equal.
 *    The call can be used to enumerate all transition IDs with their fairness ID that
 *    start in the state and match the fairness criteria.
 *  Result: 'fair' if the sequence is fair, or unfair(Ids) if the sequence is unfair where
 *    Ids is the list of FairId that are enabled (permanently in the case of weak fairness)
 *    in the loop but never executed.
 */
:- meta_predicate evaluate_ltl_fairness(+,+,+,3,2,-).
:- meta_predicate find_enabled_fairids_for_state(3,-,-).
:- meta_predicate find_enabled_fairids(-,3,-,-).
:- meta_predicate eval_fairness(-,-,3,2,-).
:- meta_predicate is_executed(2,-,-).
:- meta_predicate is_executed2(2,-,-).

evaluate_ltl_fairness(WeakStrong,StateTransitions,loop(ReEntry),FairCallEnabled,FairCallExecuted,Result) :-
    !,append_length(Loop,StateTransitions,ReEntry),
    eval_fairness(WeakStrong,Loop,FairCallEnabled,FairCallExecuted,Result).
evaluate_ltl_fairness(_WeakStrong,_ST,_Type,_FairCallEn,_FairCallEx,fair). % Fairness is only relevant for loops
eval_fairness(Type,Loop,FairCallEnabled,FairCallExecuted,Result) :-
    maplist(strans_state,Loop,States),
    find_enabled_fairids(Type,FairCallEnabled,States,RelevantIds),
    exclude(is_executed(FairCallExecuted,Loop),RelevantIds,NotExecuted),
    ( NotExecuted = [] -> Result=fair
    ; otherwise        -> Result=unfair(NotExecuted)).
strans_state(strans(StateId,_TransId),StateId).

find_enabled_fairids(weak,FairCall,[StateId|RLoop],Ids) :-
    find_enabled_fairids_for_state(FairCall,StateId,Unsorted),
    sort(Unsorted,InitialIds),
    find_enabled_fairids_weak(RLoop,FairCall,InitialIds,Ids).
find_enabled_fairids(strong,FairCall,Loop,Ids) :-
    maplist(find_enabled_fairids_for_state(FairCall),Loop,LEnabled),
    append(LEnabled,Unsorted),
    sort(Unsorted,Ids).

find_enabled_fairids_weak([],_FairCall,Ids,Ids).
find_enabled_fairids_weak(_Loop,_FairCall,[],[]) :- !.
find_enabled_fairids_weak([StateId|RLoop],FairCall,InitialIds,Ids) :-
    find_enabled_fairids_for_state(FairCall,StateId,LocalIds),
    % we use intersect here, only those actions should be tested
    % for "executed in loop" that are enabled in each state of the loop
    ord_intersection(InitialIds,LocalIds,Ids1),
    find_enabled_fairids_weak(RLoop,FairCall,Ids1,Ids).

find_enabled_fairids_for_state(FairCall,StateId,Enabled) :-
    findall(FairId,call(FairCall,StateId,_TransId,FairId),Unsorted),
    sort(Unsorted,Enabled).


is_executed(FairCall,Loop,FairId) :-
    somechk(is_executed2(FairCall,FairId),Loop).
is_executed2(FairCall,FairId,strans(StateId,_TransId)) :-
    call(FairCall,StateId,FairId),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% determine unknown values by giving rules for known values

compute_unknown(CorePred,Arity,TargetPred) :-
    enumerate_true_false_unknown(Arity,InputArgs),
    compute_unknown_values(CorePred,InputArgs,Result),
    ToStore =.. [TargetPred | Result],
    % print( assert(ToStore) ),nl,
    assert( ToStore ),
    fail.
compute_unknown(_CorePred,_Arity,_TargetPred).

enumerate_true_false_unknown(0,[]).
enumerate_true_false_unknown(Arity,[Value|Rest]) :-
    Arity > 0, member(Value,[true,false,unknown]),
    A2 is Arity-1, enumerate_true_false_unknown(A2,Rest).

compute_unknown_values(CorePred,InputArgs,Result) :-
    length(InputArgs,N), N1 is N+1,
    length(Args,N1),
    append(IArgs,[R],Args),
    Call =.. [CorePred | Args],
    findall( R, (maplist(enumerate_unknown,InputArgs,IArgs),
                    call(Call)),
             Outputs),
    sort(Outputs,SOutputs),
    ( SOutputs = [O]   -> true
    ; SOutputs = [_|_] -> O=unknown),
    append(InputArgs,[O],Result).

enumerate_unknown(true,true).
enumerate_unknown(false,false).
enumerate_unknown(unknown,V) :- member(V,[true,false]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% basic rules

:- public ltl_eval_not_base/2, ltl_eval_and_base/3, ltl_eval_or_base/3.
:- public ltl_eval_implies_base/3, ltl_eval_until_base/4.

ltl_eval_not_base(true, false).
ltl_eval_not_base(false,true).

ltl_eval_and_base(false,_,    false).
ltl_eval_and_base(true, false,false).
ltl_eval_and_base(true, true, true).

ltl_eval_or_base(true, _,    true).
ltl_eval_or_base(false,true, true).
ltl_eval_or_base(false,false,false).

ltl_eval_implies_base(false,_,    true).
ltl_eval_implies_base(true, true, true).
ltl_eval_implies_base(true, false,false).

ltl_eval_until_base(_,    true, _,true).
ltl_eval_until_base(false,false,_,false).
ltl_eval_until_base(true, false,N,N).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% lifted to unknown values

:- dynamic ltl_eval_not/2, ltl_eval_and/3, ltl_eval_or/3, ltl_eval_implies/3.
:- dynamic ltl_eval_until/4.
:- compute_unknown(ltl_eval_not_base,1,ltl_eval_not).
:- compute_unknown(ltl_eval_and_base,2,ltl_eval_and).
:- compute_unknown(ltl_eval_or_base,2,ltl_eval_or).
:- compute_unknown(ltl_eval_implies_base,2,ltl_eval_implies).
:- compute_unknown(ltl_eval_until_base,3,ltl_eval_until).
