% (c) 2014-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

% A model checker for safety properties.

% At this point the property intended to be checked has been recognized as a safety property.
% As a consequence the negation of the property is represented by a finite automaton which
% accepts all bad prefixes violating the safety property. The algorithm implemented in this module
% traverses on-the-fly the state space and simulates at the same time movement through the automaton
% in regard to the evaluation of the atomic propositions. The algorithm halts when either a finite
% state of the automaton is visited (bad prefix for the property has been found and will be reported as a counter-example)
% or when all nodes of the state space has been visited.

:- module(safety_mc,[start_mc_safety_property/3]).

%:- use_module(probsrc(bmachine), [b_parse_machine_predicate/2]).
%:- use_module(probsrc(eclipse_interface),[test_boolean_expression_in_node/2]).
:- use_module(probsrc(error_manager), [add_error_fail/3, add_warning/2, add_error_and_fail/2]).
%:- use_module(probsrc(preferences),[set_preference/2, get_preference/2]).
:- use_module(probsrc(specfile),[animation_mode/1]).
:- use_module(probsrc(debug),[debug_println/2]).
:- use_module(probsrc(tools),[cputime/1]).
:- use_module(probsrc(bsyntaxtree),[is_truth/1]).

%%% state space exploration modules
:- use_module(probsrc(state_space),[visited_expression_id/1,
                                    transition/4, visited_expression/2, not_all_transitions_added/1, retract_open_node/1,
                                    set_context_state/1, clear_context_state/0]).
%% :- use_module(probsrc(state_space_open_nodes), [retract_open_node_direct/1]).

%%% ltl modules
%:- use_module(probltlsrc(ltl),[init_states/1]).
:- use_module(probltlsrc(ltl_safety),[aut_transition/3, aut_finite_state/1, aut_initial_state/1]).
:- use_module(probltlsrc(state_space_explorer),[compute_transitions_opt/3,get_b_optimisation_options/2]).

%%% library modules
:- use_module(library(file_systems)).
:- use_module(library(ordsets)).
:- use_module(library(random)).
:- use_module(library(lists)).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides predicates for automata-based model checking of safety properties.').

:- dynamic visited_pair/4, visited_pair_df/2, accepting_state/3.

%% profiler modules
%% :- use_module('../../extensions/profiler/profiler.pl').
%% :- use_module('../../extensions/profiler/profiler_te.pl').

%% :- enable_profiling(add_state_at_end/3).
%% :- enable_profiling(pop_state_from_end/1).
%% :- enable_profiling(check_for_aut_transition/4).
%% :- enable_profiling(compute_all_product_transitions/4).

% the automata-based safety model checker
start_mc_safety_property(SearchMode,StartNodes,Result) :-
  prepare_mc_safety_property,
  animation_mode(MODE),
  get_b_optimisation_options(MODE,Optimizations),
  %% perform_static_analyses(MODE,Optimizations),
  initial_states_mc_safety_property(StartNodes,Optimizations,ProductInitStates),
  (ProductInitStates == [], non_empty_automaton
    -> % add_warning(run_mc_safety_property,'No initial states of the product transition system have been generated.'),
       Result = none,
       print('All open nodes visited! Nodes analysed: '), print(0), nl
  ; otherwise ->
      (\+ aut_finite_state('accept_all')
         -> add_warning(run_mc_safety_property,'No accept_all state has been found. The property is possibly not a safety property.')
          ; true),
      run_mc_safety_property(SearchMode,ProductInitStates,Optimizations,Result)
  ).

run_mc_safety_property(breadth_first,ProductInitStates,Optimizations,Result) :-
  cputime(T1),
  run_mc_safety_property_bf(ProductInitStates,Optimizations,IntermediateResult),
  cputime(T2),
  print('% Overall Checking Time: '), D is T2-T1, print(D), print(' ms'),nl,
  cputime(T1_CE),
  get_bf_search_result(IntermediateResult,Result),
  cputime(T2_CE),
  print('% Time to get the result from model checking: '), D_CE is T2_CE-T1_CE, print(D_CE), print(' ms'),nl.
run_mc_safety_property(random,ProductInitStates,Optimizations,Result) :-
  cputime(T1),
  (run_mc_safety_property_df(ProductInitStates,Optimizations,Result) ->
    cputime(T2),
    print('Counter-example has been found'),nl,
    print(counter_example(Result)),nl
  ; otherwise ->
    cputime(T2),
    get_state_space_statistics(TS,Product),
    print('All nodes visited. No counter-example has been found.'), nl,
    print('Number of states checked: '), print(Product), print(', where '),
    print('number of states of the transition system is: '), print(TS), nl,
    Result = none
  ),
  print('% Overall Checking Time: '), D is T2-T1, print(D), print(' ms'),nl.
run_mc_safety_property(Otherwise,_ProductInitStates,_,_Result) :-
  add_error_fail(run_mc_safety_property,'Unknown search mode: ', Otherwise).

/* The algorithm for performing breadth-first search LTL safety model checking */
run_mc_safety_property_bf(ProductInitStates,Optimizations,Result) :-
  maplist(create_and_add_init_state,ProductInitStates),
  Nr = 0,
  open_search(Nr,Optimizations,Result).

open_search(Nr,Optimizations,Result) :-
  get_next_node(CurId,AutState,ActionId),!,
  (aut_finite_state_reached(AutState) -> % bad prefix has been found
    Result = counter_example_found,
    print(found_error(counter_example_found,CurId,AutState)),nl
  ; otherwise -> % continue the search
    Nr1 is Nr +1,
    compute_state_space_transitions_if_necessary(CurId,Optimizations,ActionsAndIDs),
    compute_all_product_transitions(CurId,AutState,ActionId,ActionsAndIDs),
    open_search(Nr1,Optimizations,Result)
  ).
open_search(Nr,_Optimizations,Res) :-
  Res = none,
  print('All open nodes visited! Nodes analysed: '), print(Nr), nl.

compute_state_space_transitions_if_necessary(CurId,Optimizations,ActionsAndIDs) :-
  compute_state_space_transitions_if_necessary_aux(CurId,Optimizations),
  findall((ActId,ActionAsTerm,DestId),
        transition(CurId,ActionAsTerm,ActId,DestId),
        ActionsAndIDs).

compute_state_space_transitions_if_necessary_aux(CurId,Optimizations) :-
  (not_all_transitions_added(CurId),retract_open_node(CurId)
   -> % if node is still not retracted from queue
    set_context_state(CurId),
    visited_expression(CurId,CurState),
    compute_transitions_opt(Optimizations,CurId,CurState),
    clear_context_state
  ; otherwise ->
    true % nothing to do
  ).


compute_all_product_transitions(CurId,AutState,ActionId,ActionsAndIDs) :-
  member((ActionId,_Term,DestId),ActionsAndIDs),
  check_for_aut_transition(AutState,DestId,NextAutState,NextActionId),
  (aut_finite_state_reached(NextAutState) ->
    create_and_add_state_to_queue(DestId,NextAutState,NextActionId,CurId,ActionId),
    assert(accepting_state(DestId,NextAutState,NextActionId)) % save accepting state (for building later the counter-example)
  ; otherwise ->
    \+ visited_pair(DestId,NextAutState,_,_),
    create_and_add_state_to_queue(DestId,NextAutState,NextActionId,CurId,ActionId),
    fail
  ).
compute_all_product_transitions(_CurId,_AutState,_ActionId,_ActionsAndIDs).

%%% getting the initial states with respect to the the safety property
initial_states_mc_safety_property(StartNodes,Optimizations,ProductInitStatesSet) :-
  findall(Init,aut_initial_state(Init),InitStates),
  get_product_init_states(StartNodes,InitStates,Optimizations,ProductInitStates),
  list_to_ord_set(ProductInitStates,ProductInitStatesSet),
  debug_println(9,init_states(ProductInitStatesSet)).

get_product_init_states([],_InitStates,_Optimizations,[]).
get_product_init_states([StartNode|Nodes],InitStates,Optimizations,Res) :-
  (ltl_safety:tps_occurring -> compute_state_space_transitions_if_necessary(StartNode,Optimizations,_ActionsAndIDs);true),
  get_init_product(StartNode,InitStates,IStates),
  append(IStates,Res1,Res),
  get_product_init_states(Nodes,InitStates,Optimizations,Res1).

get_init_product(_StartNode,[],[]).
get_init_product(StartNode,[Init|Inits],Res) :-
  findall(Succ,get_successor_state(StartNode,Init,Succ),L),
  append(L,Res1,Res),
  get_init_product(StartNode,Inits,Res1).

get_successor_state(StartNode,Init,(StartNode,Next,ActionId)) :-
  aut_transition(Init,Label,Next),
  check_transition(Label,StartNode,ActionId).
%---------------------------------------------------------------------

%%% utility predicates
prepare_mc_safety_property :-
  retractall(product(_,_,_)),
  retractall(visited_pair(_,_,_,_)),
  retractall(visited_pair_df(_,_)),
  retractall(accepting_state(_,_,_)).

% automata relevant functions
aut_finite_state_reached(AutState) :-
  aut_finite_state(AutState),
  AutState = 'accept_all',!.

% Checking whether the automaton expresses the trivial property 'true'.
% If so, a counter-example [inital_state] is returned
trivial_automaton :-
  aut_transition(P,L,Q),
  (P=='accept_init',(L = ap(bpred(Pred))->is_truth(Pred);fail),Q =='accept_init').

non_empty_automaton :-
  aut_transition(_P,_L,_Q).

% state space statistics
get_state_space_statistics(TS,Product) :-
  findall((X,Y),visited_pair_df(X,Y),L),
  debug_println(9,product_states(L)),
  length(L,Product),
  findall(ID,visited_expression_id(ID),IDL),
  length(IDL,TS).
%-------------------------------------------

%%% Queue operations
:- dynamic product/3.
create_and_add_state_to_queue(CurId,AutState,NextActionId,PrevState,ActionId) :-
  assert(visited_pair(CurId,AutState,PrevState,ActionId)),
  add_state_at_end(CurId,AutState,NextActionId),
  debug_println(9,add_node_to_queue(product(CurId,AutState,NextActionId))).

create_and_add_init_state((CurId,AutState,ActionId)) :-
  assert(visited_pair(CurId,AutState,root,none)),
  add_state_at_end(CurId,AutState,ActionId),
  (aut_finite_state_reached(AutState) -> assert(accepting_state(CurId,AutState,ActionId)); true),
  debug_println(9,add_node_to_queue(product(CurId,AutState,ActionId))).

get_next_node(CurId,AutState,ActionId) :-
  pop_state_from_end(Node),
  Node = product(CurId,AutState,ActionId),
  debug_println(9,node_to_be_explored(Node)).

add_state_at_end(State,AutState,ActionId) :-
  assertz(product(State,AutState,ActionId)).
%add_state_at_front(State,AutState,ActionId) :-
%  asserta(product(State,AutState,ActionId)).
pop_state_from_end(Node) :-
  retract(product(State,AutState,ActionId)),!,
  Node = product(State,AutState,ActionId).
%-------------------

/* An algorithm for checking safety LTL properties in depth-first manner.
   First tryouts showed that the depth-first search is much more effective than the breadth-first mode. */

run_mc_safety_property_df(InitStates,Optimizations,Result) :-
  member((CurId,AutState,ActionId),InitStates),
  (trivial_automaton -> % the automaton is trivial , i.e. we have checked the formula 'false' and need to return one of the initial states as a counter-example
    %% CEPath = [atom(CurId,CurId,none)],
    Result=model([atom(CurId,CurId,none)],no_loop)
  ; otherwise ->
    mc_safety_property_init((CurId,AutState,ActionId),Optimizations,CEPath,CEPath,Result)
  ).

% at the init state we already checked for next automaton transitions
mc_safety_property_init((CurId,AutState,ActionId),Optimizations,History,HTail,Result) :-
  (transition(CurId,_Op,ActionId,CurId) -> true ; assert(visited_pair_df(CurId,AutState))),
  mc_safety_property_next_state((CurId,AutState,ActionId),Optimizations,History,HTail,Result).

% model_checker: get_open_node_to_check(DFMODE,NodeID)
mc_safety_property((CurId,AutState,ActionId),Optimizations,History,HTail,Result) :-
  assert(visited_pair_df(CurId,AutState)),
  debug_println(9,current_history(CurId,History)),
  check_for_aut_transition(AutState,CurId,NextAutState,ActionId),
  mc_safety_property_next_state((CurId,NextAutState,ActionId),Optimizations,History,HTail,Result).

mc_safety_property_next_state((CurId,AutState,ActionId),_Optimizations,History,[atom(CurId,CurId,ActionId1)],Result) :-
  aut_finite_state(AutState),
  AutState = 'accept_all',!,
  ( nonvar(ActionId) -> ActionId1=ActionId; ActionId1 = none),
  print(found_error(counter_example_found,CurId,History)),nl,
  Result = model(History,no_loop).
mc_safety_property_next_state((CurId,AutState,ActionId),Optimizations,History,HTail,Result) :-
  compute_state_space_transitions_if_necessary(CurId,Optimizations,ActionsAndIDs),
  random_permutation(ActionsAndIDs,ActionsAndIDs1),
  HTail = [atom(CurId,CurId,ActionId)|HTail2],
  get_next_state(ActionsAndIDs1,ActionId,DestId),
  \+visited_pair_df(DestId,AutState),
  (ltl_safety:tps_occurring -> compute_state_space_transitions_if_necessary(DestId,Optimizations,_DstActionsAndIDs);true),
  mc_safety_property((DestId,AutState,_NewActionId),Optimizations,History,HTail2,Result).

check_for_aut_transition(AutState,CurId,NextAutState,ActionId) :-
  get_randomised_next_aut_states(AutState,Trans),
  member((Label,NextAutState),Trans),
  check_transition(Label,CurId,ActionId).

get_randomised_next_aut_states(AutState,Res) :-
  % here we want to test first the non-truth automata transitions in order to terminate earlier in case of a counter-example
  % this is a type of selective randomising (the idea is to execute as soon as possible transitions that do not have the label 'true')
  findall((Labl,Next),(aut_transition(AutState,Labl,Next),\+is_truth_label(Labl)),L),
  random_permutation(L,L1),
  findall((Labl,Next),(aut_transition(AutState,Labl,Next),is_truth_label(Labl)),LTruth),
  append(L1,LTruth,Res).

get_next_state(ActionsAndIDs,ActionId,DstId) :-
  member((ActionId,_Term,DstId),ActionsAndIDs).

%%% for getting the result from LTL Safety Model Check
get_bf_search_result(none,Result) :- !, Result = none.
get_bf_search_result(counter_example_found,Result) :- !,
  get_counter_example_from_product(Result),
  debug_println(9,counter_example(Result)).
get_bf_search_result(Arg,_Result) :-
  add_error_fail(run_mc_safety_property_bf, 'Unknow result: ', Arg).

get_counter_example_from_product(Result) :-
  accepting_state(State,AutState,ActionId),!,
  (nonvar(ActionId) -> Trans = ActionId; Trans = none),
  get_counter_example_list(State,AutState,[atom(State,State,Trans)],ResList),
  Result = model(ResList,no_loop).
get_counter_example_from_product(_Result) :-
  add_error_and_fail(safety_mc,'No accepting state was found (this is an internal error in the model checker).').

get_counter_example_list(State,AutState,CurRes,Result) :-
  visited_pair(State,AutState,root,none),!,
  Result = CurRes.
get_counter_example_list(State,AutState,CurRes,Result) :-
  visited_pair(State,AutState,PrevState,ActionId),!,
  transition(PrevState,_ActionAsTerm,ActionId,State),
  aut_transition(PrevAutState,_,AutState),
  get_counter_example_list(PrevState,PrevAutState,[atom(PrevState,PrevState,ActionId)|CurRes], Result).
%------------------------------------------------------------------------------------------------------------------

is_truth_label(true).

:- use_module(ltl_propositions,[check_enabled/2]).
:- use_module(probltlsrc(ltl_propositions), [check_transition/4, check_ap/2]).
%%% synchronising the automaton transitions with the successor states of the model
check_transition(true,_StateId,_ActionId) :- !.
check_transition(false,_StateId,_ActionId) :- !, fail.
check_transition(ap(AP),StateId,_ActionId) :- !,
  check_ap_at_state(AP,StateId).
check_transition(tp(TP),StateId,ActionId) :- !,
  ltl_propositions: check_enabled(TP,StateId),
  transition(StateId,Op,ActionId,DestId),
  ltl_propositions:check_transition(TP,Op,StateId,DestId).
check_transition(not(AF),StateId,ActionId) :- !,
  check_transition_fail(AF,StateId,ActionId).
check_transition(and(AF,AG),StateId,ActionId) :- !,
  check_transition(AF,StateId,ActionId),
  check_transition(AG,StateId,ActionId).
check_transition(or(AF,AG),StateId,ActionId) :- !,
  (check_transition(AF,StateId,ActionId) ;
   check_transition(AG,StateId,ActionId)).
check_transition(Label,_StateId,_ActionId) :-
  add_error_fail(check_for_aut_transition,'Unknown Buchi automaton transition label: ', Label).

check_transition_fail(true,_StateId,_ActionId) :- !, fail.
check_transition_fail(false,_StateId,_ActionId) :- !.
check_transition_fail(ap(AP),StateId,_ActionId) :- !,
  (check_ap_at_state(AP,StateId)-> fail; true).
check_transition_fail(tp(TP),StateId,ActionId) :- !,
  transition(StateId,Op,ActionId,DestId),
  \+ltl_propositions:check_transition(TP,Op,StateId,DestId).
check_transition_fail(not(AF),StateId,ActionId) :- !,
  check_transition(AF,StateId,ActionId).
check_transition_fail(and(AF,AG),StateId,ActionId) :- !,
  (check_transition_fail(AF,StateId,ActionId) ;
   check_transition_fail(AG,StateId,ActionId) ).
check_transition_fail(or(AF,AG),StateId,ActionId) :- !,
  check_transition_fail(AF,StateId,ActionId),
  check_transition_fail(AG,StateId,ActionId).
check_transition_fail(Label,_StateId,_ActionId) :-
  add_error_fail(check_for_aut_transition,'Unknown Buchi automaton transition label: ', Label).

check_ap_at_state(deadlock,StateId) :- !,
  animation_mode(MODE),
  get_b_optimisation_options(MODE,Optimizations),
  compute_state_space_transitions_if_necessary_aux(StateId,Optimizations),
  \+transition(StateId,_ActionId,_ActionAsTerm,_DestId).
check_ap_at_state(AP,StateId) :-
  ltl_propositions: check_ap(AP,StateId).

