% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(model_checker,[is_deadlocked/1, open_search/11,
  expired_static_analysis_time/1,
  proof_info_model_checking_incomplete/5, model_checking_is_incomplete/6,
  perform_static_analyses/3, compute_transitions_opt/3,
  find_invariant_error/4]).

:- use_module(module_information).
:- module_info(group,model_checker).
:- module_info(description,'The core model checker (does not treat LTL).').

:- use_module(b_interpreter,[state_violates_assertions/2]).
:- use_module(error_manager,[real_error_occurred/0]).

:- use_module(specfile,[animation_mode/1,csp_mode/1]).
:- use_module(tools,[print_message/1]).
:- use_module(state_space,[visited_expression/2,transition/3, operation_not_yet_covered/1]).

:- use_module(probporsrc(static_analysis),[compute_dependendency_relation_of_all_events_in_the_model/3]).
:- use_module(probpgesrc(pge_algo), [compute_transitions_pge/2,compute_operation_relations_for_pge_if_necessary/2]).

:- use_module(state_space_exploration_modes,[depth_breadth_first_mode/1]).
:- use_module(bmachine,[b_machine_has_constants/0]).
:- use_module(state_space,[multiple_concrete_constants_exist/0]).

/* open_search: search for errors/goals among open nodes */
/* LimitNr: max number of nodes to check, NodesAnalysed: number of nodes checked */
open_search(LimitNr,NodesAnalysed,LimitTime,Res,FindDeadlocks,FindInvViolations,
                           FindGoal,FindAssViolations,StopAtFullCoverage,PartialOrderReduction,PartialGuardEvaluation) :-
    animation_mode(MODE),
    initialise_optimisations(MODE,PartialOrderReduction,PartialGuardEvaluation,FindInvViolations,Optimisations),
    perform_static_analyses(MODE,FindInvViolations,Optimisations),
    depth_breadth_first_mode(DFMODE),
    preferences:temporary_set_preference(do_invariant_checking,true,CHNG),
    statistics(walltime,[ReferenceTime,_]),
    get_model_check_stats(States,_Transitions,_ProcessedTotal,Percentage),
    % MEMO: used to cache expanded concrete_constants values
    % particularly useful, e.g., for ../prob_examples/public_examples/B/PerformanceTests/LiftLargeWithLargeCst.mch
    ((\+ b_machine_has_constants ; multiple_concrete_constants_exist)
       -> MEMO = '$disabled' % it could still be useful to cache one constant setup ??
        ; true),
    open_search10(0,LimitNr,refstatus(ReferenceTime,Percentage,States,ReferenceTime),
                           NodesAnalysed,LimitTime,Res,FindDeadlocks,FindInvViolations,
                           FindGoal,FindAssViolations,
                           StopAtFullCoverage,Optimisations,MODE,DFMODE,MEMO),
    %tools_printing:print_term_summary(memo(MEMO)),nl,
    preferences:reset_temporary_preference(do_invariant_checking,CHNG),
    clear_context_state.


:- use_module(library(random)).
open_search10(Nr,LimitNr,RefStatus,NodesAnalysed,LimitTime,Res,
              FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,
              SFC,Optimisations,MODE,DF,MEMO) :-
  %print(open_search(Nr,LimitNr,RefStatus)),nl,
  Nr < LimitNr,!,
  (\+ number(LimitTime)
   -> open_search_do_one_node(Nr,LimitNr,RefStatus,NodesAnalysed,LimitTime,Res,
					  FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,
					  SFC,Optimisations,MODE,DF,MEMO)
   ;  statistics(walltime,[NewTime,_]),
	  ( (Nr mod 1000 =:= 0,Nr\=0)
		 ->  RefStatus = refstatus(RefTime,RefPerc,PrevStates,StartTime),
		     Delta is NewTime-RefTime,
		     (Delta>20000  % more than 20 seconds since last message
		      % ; random(0,20000,X), X=<Delta)
		     -> get_model_check_stats(States,Transitions,ProcessedTotal,Perc),
		        format('~n* ~w states checked (~1f% of total ~w), ~w transitions,',[ProcessedTotal,Perc,States,Transitions]),
			    statistics(memory_used,M), tools:print_mb(M),
			    DeltaStates is (States-PrevStates),
			    NodesPerSec is DeltaStates*1000 / (NewTime-RefTime),
			    format(', current ~2F states/sec~n',[NodesPerSec]),
			    RefStatus1 = refstatus(NewTime,Perc,States,StartTime),
			    (Perc>RefPerc -> P5 is (Perc/100) **5,
			                     %DeltaPerc is P5 - ((RefPerc/100) ** 5),
			                     FullDeltaTime is NewTime - StartTime,
			                     FullDeltaTimeMinutes is FullDeltaTime / 60000.0,
			                     TimeRemaining is (FullDeltaTime/P5 - NewTime) / 60000,
			                     EstimatedStates is integer(States / P5),
			                     format('* walltime ~1f minutes, estimating ~1f minutes remaining with ~w states in total~n',[FullDeltaTimeMinutes,TimeRemaining,EstimatedStates])
			                  ; true)
			  ; RefStatus1 = RefStatus)
		 ;   RefStatus1 = RefStatus),
	  (NewTime>LimitTime, Nr>0
		     -> Res = [timeout,Nr], NodesAnalysed=Nr
		     ;  open_search_do_one_node(Nr,LimitNr,RefStatus1,NodesAnalysed,LimitTime,Res,
					  FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,
					  SFC,Optimisations,MODE,DF,MEMO))
	).
open_search10(N,_LimitNr,_RefStatus,N,_,no ,_,_,_,_,_,_,_,_,_MEMO).

get_model_check_stats(States,Transitions,ProcessedTotal,Percentage) :-
     get_state_space_stats(States,Transitions,ProcessedTotal),
	 Percentage is (100*ProcessedTotal / States).

:- use_module(preferences,[preference/2]).
:- use_module(specfile,[prepare_state_for_specfile_trans/3]).
open_search_do_one_node(Nr,LimitNr,RefStatus,NodesAnalysed,LimitTime,Res,
                           FindDeadlocks,FindInvViolations,
                           FindGoal,FindAssViolations,
                           StopAtFullCoverage,Optimisations,MODE,DFMODE,MEMO) :-
  get_next_node_to_check(FindInvViolations,ID,MODE,DFMODE),!,
  set_context_state(ID), % print(treating_node(ID)),nl,trace,
  (preferences:preference(ignore_hash_collisions,true),
   preferences:preference(forget_state_space,true)
    -> retract_visited_expression(ID,CurState) %, retractall(id_to_marker(ID,_)) not necessary
    ;  visited_expression(ID,CurState)),
  (FindInvViolations=1 % compute invariant status
    -> specfile: prepare_state_for_specfile_trans(CurState,MEMO,PCurState),
       user:compute_invariant_if_necessary(ID,PCurState,MODE)
     ; CurState=PCurState),
    %print(processed_invariant(ID)), statistics(memory_used,M), tools:print_mb(M),nl,
  !,
  compute_transitions_opt(Optimisations,ID,PCurState),
  %print_message(mc(ID)),
    %print(computed_all_transitions(ID)), statistics(memory_used,M2), tools:print_mb(M2),nl,
   % !,
  %user:do_one_gui_event,
  (open_search_found_error(ID,PCurState,Res,FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopAtFullCoverage,MODE)
   ->  print_found_error(Res,ID), clear_context_state,
       NodesAnalysed=Nr,
       user:tcltk_execute_trace_to_node(ID)  % TO DO: allow to disable this when -nocounter option given
    ;  N1 is Nr + 1,
	   open_search10(N1,LimitNr,RefStatus,NodesAnalysed,LimitTime,Res,
		FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopAtFullCoverage,
		Optimisations,MODE,DFMODE,MEMO)
  ).
open_search_do_one_node(N,_LimitNr,_RefStatus,N,_,all,_,_,_,_,_,_,_,_,_MEMO) :-
    print_message('All open nodes visited').

print_found_error(state_error(X),ID) :- nonvar(X),functor(X,F,_),!, % only print functor; term can be big
   format('% Found "~w" in state id ~w~n',[F,ID]).
print_found_error(X,ID) :-  format('% Found "~w" error in state id ~w~n',[X,ID]).


%% Optimisation options (see first argument 'Optimisations'):
compute_transitions_opt(Optimisations,ID,CurState) :-
  Optimisations = opt(PORPrefs,PGEPref),
  compute_transitions(PORPrefs,PGEPref,ID,CurState).

compute_transitions(PORPrefs,_PGEPref,ID,CurState) :-
  PORPrefs = por(POROption,_UseEnableGraph,_Depth,_UsePGE),
  (POROption==ample_sets;POROption==ample_sets2),!,
  user: compute_ample_set(ID,CurState,PORPrefs).
compute_transitions(_,PGEPref,ID,CurState) :-
  PGEPref \= off,!,
  pge_algo: compute_transitions_pge(ID,CurState).
compute_transitions(_,_,ID,CurState) :-
    user: compute_all_transitions(ID,CurState).

:- use_module(preferences,[get_preference/2]).
:- use_module(specfile,[b_or_z_mode/1]).
:- use_module(state_space_exploration_modes,[get_open_node_to_check/2]).
%%get_next_node_to_check(1,ID,_) :- state_space:not_invariant_checked(ID). % if there is an existing node whose invariant we have not yet checked: check it now
get_next_node_to_check(_,ID,AnimMODE,DFMODE) :-
  %(nonvar(ID) -> print(nonvar(ID)),nl ; true),
  get_open_node_to_check(DFMODE,ID),
  (b_or_z_mode(AnimMODE) -> assert(not_invariant_checked(ID)) ; true).


:- use_module(state_space).
open_search_found_error(ID,CurState,Error,
                       _FindDeadlocks,1,_FindGoal,_FindAssViolations,_StopAtFullCoverage,MODE) :-
    find_invariant_error(MODE,ID,CurState,Error).
open_search_found_error(ID,_CurState,deadlock,1,_,_,_,_,_) :-
   is_deadlocked(ID). %, print(deadlock(ID)),nl.
open_search_found_error(ID,CurState,assertion_violation,_,_,_,1,_,_MODE) :-
   state_violates_assertions(ID,CurState).
open_search_found_error(ID,_CurState,ERR,_,_,_,_,_,_MODE) :-
   state_space:state_error(ID,_,Error), Error \== invariant_violated,
   ERR = state_error(Error).
open_search_found_error(_ID,_CurState,general_error_occurred,_,_,_,_,_,_MODE) :-
   error_manager:real_error_occurred.  /* do we need to reset_errors_occurred before model check ?? */
open_search_found_error(ID,_CurState,goal_found,_,_,1,_,_,MODE) :-
   (user:test_goal_in_node(ID) ; is_csp_goal_state(ID,MODE)).
open_search_found_error(_ID,_CurState,full_coverage,_,_,_,_,1,_MODE) :-
   \+ operation_not_yet_covered(_).

find_invariant_error(b,ID,_CurState,invariant_violation) :- !, % here we know that invariant information has been computed
    (invariant_violated(ID) -> true).
%find_invariant_error(b,_ID,CurState,invariant_false) :- !,
%    preference(double_evaluation_when_analysing,true), % DOUBLE_EVALUATION
%    CurState \= root, CurState \= concrete_constants(_),
%    specfile:expand_const_and_vars_to_full_store(CurState,FullState), %print(eval_neg_inv(_ID)),nl,
%    (b_interpreter:state_satisfies_negation_of_invariant(FullState)->true).
find_invariant_error(cspm,ID,_CurState,csp_error) :- !, % we know that only csp_can be possible
    is_csp_error_state(ID).
find_invariant_error(MODE,ID,CurState,invariant_violation) :-
    user:check_invariantKO(MODE,ID,CurState).
find_invariant_error(csp_and_b,ID,_CurState,csp_error) :- !, % xtl mode cannot apply
    is_csp_error_state(ID).
find_invariant_error(MODE,ID,CurState,xtl_error) :-
    is_xtl_error_state(ID,CurState,MODE).

is_deadlocked(ID) :-
  preferences:preference(forget_state_space,true),!, /*  transitions not added */
  (retract(state_space:transition(ID,_,_,_)) -> fail ; true).
is_deadlocked(ID) :- % ID must be ground upon call
 %%  visited_expression_id(ID),
  deadlocked_state(ID),
  \+ max_reached_or_timeout_for_node(ID).
/* \+((transition(ID,_,NID),NID \= ID)).   <-- to detect more liberal form of DEADLOCK */

is_csp_error_state(ID) :-
  transition(ID,io(_,error,_),_).  /* <--- THE CSP has raised the error channel */

:- use_module(xtl_interface,[xtl_property/2, promela_property/2]).
is_xtl_error_state(_ID,CurState,AM) :-
  ( AM=xtl     -> xtl_property(CurState,unsafe)   /* <--- XTL has found unsafe configuration */
  ; AM=promela -> promela_property(CurState,unsafe)
%  ; AM=smv     -> smv_property(CurState,unsafe)
  ).

is_csp_goal_state(ID,MODE) :- csp_mode(MODE),
  visited_expression_id(ID),
  transition(ID,io(_,goal,_),_).


% ---------------------------------------

:- use_module(tools, [ajoin/2]).
:- use_module(symmetry_marker,[hash_model_checking_imprecise/2]).
%model_checking_is_unsound(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term)
model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,'') :-
    proof_info_model_checking_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg).
model_checking_is_incomplete(_FindInv,_FindDeadlocks,_FindGoal,_FindAss,Msg,Term) :-
   (hash_model_checking_imprecise(ID,BasicType)
    ->  Msg = '*** Hashing may be imprecise for, e.g., for variable ',
        translate:pretty_type(BasicType,S),
        ajoin([ID,' : ',S],Term)).
model_checking_is_incomplete(1,_FindDeadlocks,_FindGoal,_FindAss,Msg,Term) :-
    (time_out_for_invariant(ID)
      -> Msg = '*** Timeouts occured during invariant computation, e.g., state with id = ', Term=ID).
model_checking_is_incomplete(_FindInv,_FindDeadlocks,_FindGoal,1,Msg,Term) :-
    (time_out_for_assertions(ID)
      -> Msg = '*** Timeouts occured during evaluation of ASSERTIONS, e.g., state with id = ', Term=ID).
model_checking_is_incomplete(_FindInv,_FindDeadlocks,_FindGoal,_FindAss,Msg,Term) :-
    (max_reached_or_timeout_for_node(ID)
      -> Msg = '*** Not all transitions were computed, e.g., state with id = ', Term=ID).

:- use_module(bmachine, [b_machine_has_proven_invariants/0, b_machine_has_unproven_assertions/0]).
proof_info_model_checking_incomplete(1,_FindDeadlocks,_FindGoal,0,Msg) :-
    b_machine_has_proven_invariants,
    preferences:get_preference(use_po,true),
    b_machine_has_unproven_assertions,
    Msg = '*** Using proof information (-p PROOF_INFO TRUE) is potentially incomplete when not checking theorems (-noass) !'.

:- dynamic expired_static_analysis_time/1.

% disable POR if not applicable + merge POR/PGE info into one atom for easier lookup in compute_transitions
initialise_optimisations(MODE,PartialOrderReduction,PartialGuardEvaluation,FindInvViolations,Optimisations) :-
  MODE == b,
  !,
  ((PartialOrderReduction == ample_sets; PartialOrderReduction == ample_sets2) ->
    get_preference(enable_graph,UseEnableGraph),
    get_preference(enable_graph_depth,Depth),
    PORPrefs = por(PartialOrderReduction,UseEnableGraph,Depth,PartialGuardEvaluation),
    Optimisations = opt(PORPrefs,PartialGuardEvaluation),
    (FindInvViolations == 1 -> assert(ample_sets:check_cycle_proviso);true)
  ; PartialGuardEvaluation \= off ->
    Optimisations = opt(off,PartialGuardEvaluation)
  ; otherwise ->
    Optimisations = opt(off,off)).
initialise_optimisations(_,_,_,_,opt(off,off)).

perform_static_analyses(MODE,FindInvViolations,opt(PORPrefs,PGEPref)) :-
  MODE == b,
  !,
  retractall(expired_static_analysis_time(_)),
  statistics(runtime,[T1,_]),
  perform_static_analyses_aux(PORPrefs,PGEPref,FindInvViolations),
  statistics(runtime,[T2,_]),
  D is T2-T1,
  assert(expired_static_analysis_time(D)).
perform_static_analyses(_,_,_).

perform_static_analyses_aux(off,off,_FindInvViolations) :- !.
perform_static_analyses_aux(PORPrefs,PGEPref,FindInvViolations) :-
  PORPrefs = por(POROption,_UseEnableGraph,_Depth,_),
  (POROption==ample_sets;POROption==ample_sets2),!,
  compute_dependendency_relation_of_all_events_in_the_model(FindInvViolations,PORPrefs,_EnableGraph),
  compute_operation_relations_for_pge_if_necessary(PGEPref,FindInvViolations).
perform_static_analyses_aux(_,PGEPref,FindInvViolations) :-
  compute_operation_relations_for_pge_if_necessary(PGEPref,FindInvViolations).
perform_static_analyses_aux(_,_,_).
