% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(partition_detection, [detect_partitions/2, detect_all_partitions_in_predicate/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module detects set partition predicates.').

:- use_module(library(lists)).
:- use_module(bsyntaxtree).
:- use_module(bmachine_structure,[select_section/5]).

detect_partitions(In,Out) :-
    select_section(properties,Old,New,In,Out),
    detect_all_partitions_in_predicate(Old,New).

% detect all set partition predicates at "top-level", not inside inner predicates (bool(.), comprehension_set,...)
% this is assumed to be checked by b_ast_cleanup during bottom-up construction
detect_all_partitions_in_predicate(Predicate,NewPredicate) :-
    ( detect_partitions_in_predicate(Predicate,P1) ->
        detect_all_partitions_in_predicate(P1,NewPredicate)
    ; otherwise ->
        Predicate=NewPredicate %, print('detected: '),translate:print_bexpr(NewPredicate),nl
        %detect_partitions_inside_predicate(Predicate,NewPredicate) % now done in b_ast_cleanup
    ).

% detect partitions of set of the type:
%  x /\ y = {} & x \/ y = z
detect_partitions_in_predicate(Predicate,NewPredicate) :-
    conjunction_to_list(Predicate,PList),
    detect_partitions_in_predicates(PList,PNew),
    conjunct_predicates(PNew,NewPredicate).
detect_partitions_in_predicates(Predicates,NewPredicates) :-
    find_union(Superset,Partitions,Predicates,P1),
    create_pairs(Partitions,Pairs),
    get_texpr_type(Superset,Type),
    find_all_intersections(Pairs,Type,RestPairs,P1,P2),
    RestPairs=[],
    !,
    safe_create_texpr(partition(Superset,Partitions),pred,Part),
    NewPredicates=[Part|P2].

:- use_module(bsyntaxtree,[get_texpr_set_type/2]).
% should we also find A = {c1,c2,...cn} with all_different(c1,...,cn) ??
find_union(Superset,Partitions,In,Out) :-
    create_texpr(equal(A,B),pred,_,Equal),
    get_texpr_set_type(A,Type),
    get_texpr_set_type(B,Type),
    select(Equal,In,Out),
    ( is_union(A,Partitions) -> Superset=B
    ; is_union(B,Partitions) -> Superset=A).
is_union(Expr,Partitions) :-
    get_texpr_expr(Expr,union(_,_)),
    collect_union(Expr,Partitions,[]).
collect_union(Expr) -->
    {get_texpr_expr(Expr,union(A,B)),!},
    collect_union(A),
    collect_union(B).
collect_union(Expr) --> [Expr].

create_pairs(Parts,Pairs) :-
    create_pairs2(Parts,Pairs,[]).
create_pairs2([]) --> [].
create_pairs2([A|Rest]) -->
    create_pairs3(Rest,A),
    create_pairs2(Rest).
create_pairs3([],_) --> [].
create_pairs3([B|Rest],A) -->
    [distinct(A,B)],
    create_pairs3(Rest,A).

find_all_intersections(Pairs,Type,RestPairs,In,Out) :-
    find_all_intersections2(In,Pairs,Type,RestPairs,Out).
find_all_intersections2([],Pairs,_,Pairs,[]).
find_all_intersections2([Expr|Erest],Pairs,Type,RestPairs,Out) :-
    is_disjoint_pred(Expr,Type,A,B),
    remove_pair(A,B,Pairs,InterPairs),!,
    find_all_intersections2(Erest,InterPairs,Type,RestPairs,Out).
find_all_intersections2([Expr|Erest],Pairs,Type,RestPairs,[Expr|Out]) :-
    find_all_intersections2(Erest,Pairs,Type,RestPairs,Out).

is_disjoint_pred(b(Expr,_,_),Type,A,B) :-
  is_disjoint_pred2(Expr,Type,A,B).
is_disjoint_pred2(equal(X,Y),_,A,B) :- % detect A /\ B = {}  or  {} = A /\ B
    ( get_texpr_expr(X,empty_set) -> Inter=Y
    ; get_texpr_expr(Y,empty_set) -> Inter=X),
    get_texpr_expr(Inter,intersection(A,B)).
is_disjoint_pred2(not_equal(X,Y),set(Type),A,B) :- get_texpr_type(X,Type),
   A = b(set_extension([X]),set(Type),_),
   B = b(set_extension([Y]),set(Type),_).
is_disjoint_pred2(negation(NE),Type,A,B) :-
    is_disjoint_neg_pred(NE,Type,A,B).

is_disjoint_neg_pred(b(Expr,_,_),Type,A,B) :-
  is_disjoint_neg_pred2(Expr,Type,A,B).
is_disjoint_neg_pred2(not_equal(X,Y),_,A,B) :- % detect A /\ B /= {}  or  {} /= A /\ B
    ( get_texpr_expr(X,empty_set) -> Inter=Y
    ; get_texpr_expr(Y,empty_set) -> Inter=X),
    get_texpr_expr(Inter,intersection(A,B)).
is_disjoint_neg_pred2(equal(X,Y),set(Type),A,B) :- get_texpr_type(X,Type),
   A = b(set_extension([X]),set(Type),_),
   B = b(set_extension([Y]),set(Type),_).
is_disjoint_neg_pred2(negation(NE),Type,A,B) :-
    is_disjoint_pred(NE,Type,A,B).

remove_pair(A,B,In,Out) :-
    remove_all_infos(A,FA),
    remove_all_infos(B,FB),
    ( select(distinct(FA,FB),In,Out)
    ; select(distinct(FB,FA),In,Out)),!.
