% (c) 2015-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(b_ast_cleanup_rewrite_rules,[rewrite_rule_with_rename/7,normalization_rule_with_rename/7]).


:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'A rewriting engine for transformations/simplifications on the AST.').


:- use_module(probsrc(bsyntaxtree),[get_texpr_type/2, get_texpr_expr/2,
        replace_id_by_expr/4, extract_pos_infos/2,syntaxtransformation/5,
        safe_syntaxelement/5]).
:- use_module(probsrc(tools), [foldl/4]).
:- use_module(rewrite_rules_db, [rewrite_rule_mandatory/8, rewrite_rule_normalize/8]).
:- use_module(probsrc(btypechecker),[unify_types_strict/2]).
:- use_module(library(lists)).

% rename any quantified variables inside RHS of rules to avoid variable capture
% replace_id_by_expr/4 will perform replace

% TO DO:  can we find a more elegant/efficient way to perform alpha-renaming/avoid name clashes?
% if we replace all quantified variables by variables like '_zzzz_unary' which cannot be used by the user: are we safe provided these identifiers cannot escape !?

:- meta_predicate apply_rewrite_rule_with_rename(8,-,-,-,-,-,-,-).
:- meta_predicate mycall(8,-,-,-,-,-,-,-,-).

rewrite_rule_with_rename(Expr,Type,Infos, FullNewExpr,NewType, FullNewInfos, Name) :-
   apply_rewrite_rule_with_rename(rewrite_rule_mandatory,Expr,Type,Infos, FullNewExpr,NewType, FullNewInfos, Name).
normalization_rule_with_rename(Expr,Type,Infos, FullNewExpr,NewType, FullNewInfos, Name) :-
   apply_rewrite_rule_with_rename(rewrite_rule_normalize,Expr,Type,Infos, FullNewExpr,NewType, FullNewInfos, Name).

apply_rewrite_rule_with_rename(RULEPRED,Expr,Type,Infos, FullNewExpr,NewType, FullNewInfos, Name) :-
   syntaxtransformation(Expr,Subs,_Names,NewSubs,ExprWithVars),
   create_fresh_ids(Subs,1,NewSubs,Renamings), %debug:debug_nl_time(renamings(RULEPRED,Renamings,Expr)),
   mycall(RULEPRED,ExprWithVars,RuleType,Infos, ArgTypes,NewCopy,NewCType, NewCInfos, Name),
   my_unify_types(Type,RuleType),
   get_argument_types(Expr,ExprArgTypes),
   %print(e(Name,Expr,ExprArgTypes)),nl,
   maplist(my_unify_types,ArgTypes,ExprArgTypes),
    (ground(RuleType) -> true ; print(unified(Type,RuleType)),nl),
   foldl(perform_rename,Renamings,b(NewCopy,NewCType,NewCInfos),NewTExpr),
   (extract_pos_infos(Infos,Pos)
    -> update_pos(NewTExpr,Pos,b(FullNewExpr,NewType,FullNewInfos))
     ; NewTExpr=b(FullNewExpr,NewType,FullNewInfos)
    ).

mycall(rewrite_rule_mandatory,ExprWithVars,RuleType,Infos, ArgTypes,NewCopy,NewCType, NewCInfos, Name) :- !,
   rewrite_rule_mandatory(ExprWithVars,RuleType,Infos, ArgTypes,NewCopy,NewCType, NewCInfos, Name).
mycall(rewrite_rule_normalize,ExprWithVars,RuleType,Infos, ArgTypes,NewCopy,NewCType, NewCInfos, Name) :- !,
   rewrite_rule_normalize(ExprWithVars,RuleType,Infos, ArgTypes,NewCopy,NewCType, NewCInfos, Name).
mycall(RULEPRED,ExprWithVars,RuleType,Infos, ArgTypes,NewCopy,NewCType, NewCInfos, Name) :-
   call(RULEPRED,ExprWithVars,RuleType,Infos, ArgTypes,NewCopy,NewCType, NewCInfos, Name).

my_unify_types(T1,T2) :- (unify_types_strict(T1,T2) -> true ; format('*** Unify failed: ~w = ~w~n',[T1,T2]),fail).

get_argument_types(EXPR,ArgTypes) :- safe_syntaxelement(EXPR,Subs,_,_,_),
   maplist(get_texpr_type,Subs,ArgTypes).

perform_rename(rename(ID,Arg),Expr,NewExpr) :-
   bsyntaxtree:replace_id_by_expr(Expr,ID,Arg,NewExpr).

create_fresh_ids([],_,[],[]).
create_fresh_ids([Arg|T],Nr,Args,RT) :-
   is_ground_b_pattern(Arg),!, Args = [Arg|IT],
   N1 is Nr+1,
   create_fresh_ids(T,N1,IT,RT).
create_fresh_ids([Arg|T],Nr,[b(identifier(FRESHID),Type,[])|IT],[rename(FRESHID,Arg)|RT]) :-
   get_texpr_type(Arg,Type), get_fresh_id(Nr,FRESHID),
   N1 is Nr+1,
   create_fresh_ids(T,N1,IT,RT).
:- use_module(probsrc(tools_strings), [string_concatenate/3]).
get_fresh_id(Nr,FRESHID) :- string_concatenate('___$FRESH$_',Nr,FRESHID).

% for those patterns we do not need to add fresh variables, i.e., not containing identifiers
is_ground_b_pattern(V) :- nonvar(V), get_texpr_expr(V,VE), nonvar(VE),is_ground_b_pattern_aux(VE).
is_ground_b_pattern_aux(boolean_false).
is_ground_b_pattern_aux(boolean_true).
is_ground_b_pattern_aux(bool_set).
is_ground_b_pattern_aux(empty_sequence).
is_ground_b_pattern_aux(empty_set).
is_ground_b_pattern_aux(integer(_)).
is_ground_b_pattern_aux(integer_set(_)).
is_ground_b_pattern_aux(max_int).
is_ground_b_pattern_aux(min_int).
is_ground_b_pattern_aux(string_set).
is_ground_b_pattern_aux(value(_)).
 % TO DO: extend this

:- use_module(probsrc(bsyntaxtree),[transform_bexpr/3]).

:- public add_pos/3.
add_pos(Pos,b(E,T,I),b(E,T,NewInfo)) :-
  (memberchk(nodeid(_),I) -> NewInfo=I ; append(Pos,I,NewInfo)).

% also the updating of the position info is a bit ad-hoc; can we do this more elegantly ?
update_pos(E,Pos,Res) :- !,
   transform_bexpr(b_ast_cleanup_rewrite_rules:add_pos(Pos),E,Res).


% ----------------------------------
% stuff below should be put into seperate file

