% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(static_ordering,[sort_ids_by_usage/3, sort_ids_by_usage/4, reorder_state/3]).

:- use_module(module_information).
:- module_info(group,tools).
:- module_info(description,'Sort a list of identifiers according to the number of occurences inside conjuncts of a predicate. Mimics the static ordering heursitic of SAT solvers').

:- use_module(bsyntaxtree,[find_identifier_uses/3,def_get_texpr_id/2]).

:- use_module(error_manager).

:- dynamic used/2.
% put most used IDs towards the front; these will be enumerated first
sort_ids_by_usage(IDs,Predicate,SortedIDs) :-
   sort_ids_by_usage(IDs,Predicate,SortedIDs,warnings).
sort_ids_by_usage(IDs,Predicate,SortedIDs,WARN) :-
    retractall(used(_,_)),
    count_uses(Predicate),
    sort_according_to_usage(IDs,SIDs,WARN),
    %% print(SIDs),nl, %%
    project_used_id(SIDs,SortedIDs).

count_uses(Var) :- var(Var),!, add_internal_error('Variable AST:',count_uses(Var)),fail.
count_uses(b(conjunct(LHS,RHS),pred,_)) :- !,
    count_uses(LHS), count_uses(RHS).
count_uses(X) :-  % translate:print_bexpr(X),nl,
   find_identifier_uses(X,[],IDs),
   %% print(usage(IDs)),nl,%%
   add_usage(IDs).
add_usage([]).
add_usage([ID|T]) :- (retract(used(ID,Nr)) -> true ; Nr=0),
   N1 is Nr+1, assert(used(ID,N1)),
   add_usage(T).

sort_according_to_usage([],[],_).
sort_according_to_usage([TID|T],Res,WARN) :-
    def_get_texpr_id(TID,ID),
    (used(ID,Nr) -> true %%print(used(ID,Nr)),nl
     ; (WARN==warnings
        -> add_message(sort_according_to_usage,'### Unused identifier: ',ID)
         ; true),
        Nr=0),
    sort_according_to_usage(T,TSorted,WARN), % recurse first for stability of sorting
    insert_used_id(TSorted,TID,Nr,Res).
insert_used_id([],TID,Nr,[(TID,Nr)]).
insert_used_id([(H,HNr)|T],TID,Nr,Res) :-
    (Nr>=HNr -> Res = [(TID,Nr),(H,HNr)|T]
       ; Res=[(H,HNr)|RT], insert_used_id(T,TID,Nr,RT)
    ).

project_used_id([],[]).
project_used_id([(A,_)|T],[A|PT]) :- project_used_id(T,PT).


% --------------


:- use_module(library(lists),[select/3]).
:- use_module(bsyntaxtree,[def_get_texpr_id/2]).

% reorder state into original order, to avoid issues with state_packing or ltsmin
reorder_state([],Rest,Rest).
reorder_state([TID|T],InState,[bind(ID,V)|OutState]) :- def_get_texpr_id(TID,ID),
   select(bind(ID,V),InState,Rest), reorder_state(T,Rest,OutState).

