% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(static_symmetry_reduction, [static_symmetry_reduction_possible/5, perform_ssr/8]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides support for symmetry reduction on deferred sets.').

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(debug,[debug_println/2]).



% peform MACE style static symmetry reduction for those global constants
% that have not already been fixed
% e.g., for constants aa,bb,cc,dd of type ID and bb/=cc  --> nrs of bb,cc will be fixed as 1 and 2; we will ensure that numbers of aa:1..3 and dd:1..4 and that dd=4 only if aa=3
% TO DO: extend this kind of symmetry reduction to other constants (e.g., total fun over fixed domain f: ... --> ID)
% TO DO: ensure this is also used in enabling analysis, symbolic_model checker, ...
% TO DO: support detection of injections


:- use_module(b_global_sets,[b_fd_type/3, is_b_global_constant/3]).
:- use_module(bmachine,[b_get_properties_from_machine/1, b_get_disjoint_constants_of_type/3]).


% check if we can apply static symmetry reduction for a global deferred set GS
% means we have a deferred set GS with fixed element (DisjointConstants)
% and some remaining deferred set elements (Other) whose index starts at FirstAvailableNewIdx
static_symmetry_reduction_possible(GS,FirstAvailableNewIdx,Low,Up,Other) :-
  b_fd_type(GS,Low,Up),
  (b_get_disjoint_constants_of_type(GS, DisjointConstants, AllConstantsIDs) -> true
    ; DisjointConstants=[], % find_constants_of_type(Constants,GS,AllConstantsIDs,_GSConstants)
      AllConstantsIDs=[] % otherwise at least one disjoint constant would have been returned
  ),
  findall(Nr,is_b_global_constant(GS,Nr,_Cst),Ns),
  (Ns=[] -> FirstAvailableNewIdx=1
   ; max_member(Max,Ns),
     length(Ns,Max), % check contiguous, DisjointConstants are numbered 1..Max
     FirstAvailableNewIdx is Max+1
  ),
  sort(AllConstantsIDs,SAll),
  sort(DisjointConstants,SDis),
  ord_subtract(SAll,SDis,Other),
  % Other \= [],  % If Other = [] one could still use partition symmetry reduction
  debug_println(4,static_symmetry_possible(GS,FirstAvailableNewIdx,Up,Other,SDis)).


% -----------------------------

:- use_module(store,[lookup_value_for_existing_id/3]).
:- use_module(bsyntaxtree,[member_in_conjunction/2, get_texpr_expr/2,get_texpr_id/2]).
:- use_module(fd_utils_clpfd,[in_fd/3]).
:- use_module(clpfd_interface,[force_post_constraint/1, force_clpfd_inlist/2]).
:- use_module(tools,[exact_member/2]).

perform_ssr([],_PrevNrs,_InitialF,FirstNew,GS,_Low,Up,ConstantsState) :-  FirstNew < Up,
   %  if partition exists -> we could enforce order on partitions for Indexes in FirstAvailableNewIdx..Up
   b_get_properties_from_machine(Properties),
   member_in_conjunction(C,Properties),
   get_texpr_expr(C,partition(Set,DisjSets)),
   global_set_identifier(Set,GS),
   maplist(try_get_value(ConstantsState),DisjSets,DisjSetVals), % TO DO: try and evaluate at runtime in case DisjSets are not all identifiers but expressions such as {aa,bb}
   !,
   print(partition_symmetry_reduction(DisjSetVals,FirstNew,GS,Up)),nl,
   % when(ground(DisjSetVals), format('SETS=~w~n',[DisjSetVals])),
   partition_sym_reduction(DisjSetVals,FirstNew).
perform_ssr([],PrevNrs,InitialF,FirstNew,GS,_,_,_) :-
   debug_println(4,done_ssr(GS,PrevNrs,InitialF,FirstNew)).
perform_ssr([OtherID|T],PrevNrs,InitialF,FirstNew,GS,Low,Up,ConstantsState) :-
   % adding static symmetry breaking constraints for OtherID of type GS
   lookup_value_for_existing_id(OtherID,ConstantsState,fd(Nr,GS)),
   debug_println(4,ssr(OtherID,Nr,FirstNew,Up)),
   fd_utils_clpfd:in_fd(Nr,Low,FirstNew),
   clpfd_interface:force_post_constraint( (Nr#>InitialF) #<=> Trigger),
   check_contiguous(Trigger,Nr,PrevNrs),
   (FirstNew<Up -> F1 is FirstNew+1 ; F1=Up),
   perform_ssr(T,[Nr|PrevNrs],InitialF,F1,GS,Low,Up,ConstantsState).



try_get_value(ConstantsState,TID,DisjSetVals) :-
   get_texpr_id(TID,ID),lookup_value_for_existing_id(ID,ConstantsState,DisjSetVals).

:- block check_contiguous(-,?,?),check_contiguous(?,-,?). % to do: do not delay on Nr
check_contiguous(0,_Nr,_PrevNrs). % we have not used a new deferred set number
check_contiguous(1,Nr,PrevNrs) :- % we have used a new number, check that previous number also used
   %print(check_used(Nr,PrevNrs)),nl,
   N1 is Nr-1,
   (tools:exact_member(N1,PrevNrs) -> true
     ; when(ground(PrevNrs),clpfd_interface:force_clpfd_inlist(N1,PrevNrs))).


global_set_identifier(C,GS) :- get_texpr_expr(C,BE), global_set_identifier2(BE,GS).
global_set_identifier2(identifier(GlobalSet),GlobalSet).
global_set_identifier2(value(global_set(GlobalSet)),GlobalSet). % generated by Z

% --------------------


% This predicate implements a symmetry reduction on a partition of the deferred sets
% DS = P1 \/ P2 ... Pn  &  P1 /\ P2 = {} ...
% where we may have some constants already allocated fixed numbers (1..n)
% and other constants ranging from 1..(n+m) and where n+m+1 is the FirstFreelyChoosableIndex
% for any index >= FirstFreelyChoosableIndex we impose that these must be allocated in order to
% the partitions

% For example: if FirstFreelyChoosableIndex 1 and we have two sets, then we will only allow:
%  {}, {1,2,3}   ;   {1}, {2,3}  ;  {1,2}, {3}  ;  {1,2,3} , {}  as partitions and not e.g.
%  {2}, {1,3} ...
% Idea: if fd(Nr,_) appears in partition k, then all numbers from FirstFreelyChoosableIndex..Nr are prohibited in later partitions

% partition_sym_reduction(List of Sets, FirstFreelyChoosableIndex)

partition_sym_reduction([],_).
partition_sym_reduction([PSet1|OtherPSets],FirstNew) :-
   check_partition_order(PSet1,OtherPSets,FirstNew),
   partition_sym_reduction(OtherPSets,FirstNew).

:- block check_partition_order(-,?,?).
check_partition_order([],_,_) :- !.
check_partition_order([H|T],OtherSets,FirstNew) :-  (T,OtherSets) \== ([],[]),
   !,
   check_partition_order_for_el(H,[T|OtherSets],FirstNew),
   check_partition_order(T,OtherSets,FirstNew).
check_partition_order(global_set(_),_,_) :- !. % all other sets will be forced to be empty
%check_partition_order(A,_,_) :- print(uncov_pset(A)),nl,fail. % TO DO: treat avl_set ..
check_partition_order(_,_,_).

:- block check_partition_order_for_el(-,?,?).
check_partition_order_for_el(fd(Nr,_),OtherSets,FirstNew) :-
   check_partition_order_for_el2(Nr,OtherSets,FirstNew).
:- block check_partition_order_for_el2(-,?,?).
check_partition_order_for_el2(Nr,OtherSets,FirstNew) :-  % print(sym_red(Nr,FirstNew,OtherSets)),nl,
   Nr > FirstNew,!, % sym. reduction applicable
   maplist(prohibit_indices(FirstNew,Nr),OtherSets).
check_partition_order_for_el2(_,_,_).

:- use_module(library(clpfd),[fdset_interval/3, fdset_complement/2, in_set/2]).
prohibit_indices(FirstNew,Nr,Set) :-
     fdset_interval(Int,FirstNew,Nr),
     fdset_complement(Int,NotInt),
     %print(prohibit2(Set,NotInt)),nl,
     prohibit_indices2(Set,NotInt).
:- block prohibit_indices2(-,?).
prohibit_indices2([],_) :- !.
prohibit_indices2([H|T],NotInt) :- !,   %print(prohibit(H,T,NotInt)),nl,
   H = fd(Nr,_),
   Nr in_set NotInt,
   prohibit_indices2(T,NotInt).
prohibit_indices2(A,_) :- % TO DO: treat avl_set, ..
   print(uncov_prohibit_indices2(A)),nl.
