% (c) 2014-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(ctigar,[ctigar_symbolic_model_check/1]).

:- use_module(probsrc(module_information)).
:- module_info(group,symbolic_model_checker).
:- module_info(description,'An infinite-state symbolic model checker based on CTIGAR.').

:- use_module(library(heaps)).
:- use_module(library(avl)).
:- use_module(library(ordsets)).
:- use_module(library(lists)).

:- use_module(extension('counter/counter'),[counter_init/0,
                                            new_counter/1,
                                            get_counter/2]).

:- use_module(probsrc(preferences), [get_preference/2]).
:- use_module(probsrc(bmachine),[b_get_invariant_from_machine/1,
                        %b_get_initialisation_from_machine/2,
                        b_get_properties_from_machine/1,
                        b_get_machine_operation/4,
                        b_specialized_invariant_for_op/2,
                        get_proven_invariant/2]).
:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2,
                            disjunct_predicates/2,
                            conjunction_to_list/2,
                            create_implication/3,
                            create_negation/2
                            %create_exists/3, replace_id_by_expr/4
                            ]).
:- use_module(probsrc(translate), [translate_bexpression/2, print_bexpr/1]).
:- use_module(probsrc(state_space),[state_space_reset/0]).
:- use_module(probsrc(unsat_cores), [unsat_core_with_fixed_conjuncts/3]).
:- use_module(probsrc(weakest_preconditions), [weakest_precondition/3]).

:- use_module(symbolic_model_checker(predicate_handling)).
:- use_module(symbolic_model_checker(solver_handling)).
:- use_module(symbolic_model_checker(mic_generation)).
:- use_module(symbolic_model_checker(predicate_abstraction)).

% algorithm for computing the minimal inductive clause
mic_algorithm(down).
% mic_algorithm(ctgDown).

ctigar_symbolic_model_check(Res) :-
    statistics(walltime,[Start,_]),
    catch(symbolic_model_check_aux(Res),solver_and_provers_too_weak,Res=solver_and_provers_too_weak),
    statistics(walltime,[End,_]),
    Took is End - Start,
    get_counter(ic3solvercalls,Calls),
    format('Result of CTIGAR based model checking: ~w,~nCTIGAR took ~w ms and ~w solver calls~n',[Res,Took,Calls]).

symbolic_model_check_aux(Result) :-
    % count number of solver calls
    counter_init, new_counter(ic3solvercalls),
    % fetch the invariant we want to check
    b_get_invariant_from_machine(Invariant),
    % and the predicate describing the initial state
    get_initial_state_predicate(Init),
    % debug output
    translate_bexpression(Init,PPInit), translate_bexpression(Invariant,PPInvariant),
    format('Symbolic Model Checking:~nProperty: ~w~nInitial Predicate: ~w~n',[PPInvariant,PPInit]),
    symbolic_model_check(Invariant,Init,Result).

% check base cases (length 0 and 1) first - as in Bradleys implementation
symbolic_model_check(Property,Init,ResultOut) :-
    % ce of length 0? check if Init => Invariant
    create_implication(Init,Property,Predicate),
    % debug output
    translate_bexpression(Predicate,PPPredicate),
    format('Checking: ~w~n',[PPPredicate]),
    solve_negated(Predicate,Result),
    functor(Result,ForPrint,_),
    format('Counter-Example of Length 0 exists: ~w~n', [ForPrint]),
    Result = solution(Sol), !,
    create_state_predicate(Sol,StatePred),
    replay_counter_example([StatePred]),
    ResultOut = counterexample_found.
symbolic_model_check(Property,Init,ResultOut) :-
    % ce of length 1? check if Init & Transition => Invariant'
    prime_predicate(Property,PrimedProperty),
    get_single_transition_predicate(OpName,Transition),
    conjunct_predicates([Init,Transition],LHS),
    create_implication(LHS,PrimedProperty,Predicate),
    % debug output
    translate_bexpression(Predicate,PPPredicate),
    format('Checking: ~w~n',[PPPredicate]),
    solve_negated(Predicate,Result),
    functor(Result,ForPrint,_),
    format('Counter-Example of Length 1 (transition ~w) exists: ~w~n', [OpName,ForPrint]),
    Result = solution(Sol), !,
    handle_counter_example(Sol,[]),
    ResultOut = counterexample_found.
% recursively check other cases. this corresponds to method "check" of Bradleys reference implementation
% internal data structures taken from Bradleys original paper:
% frame(F_i as in the paper (+transition), set of clauses to check for convergence)
%   i/K (= level) is the key to the AVL tree
% obligation(state / predicate, primary inputs / predicate, list of successor states)
%   level is the key to the priority queue
symbolic_model_check(Property,Init,Result) :-
    K = 1,
    % two initial frames:
    % F_0 holds the initial condition
    % F_1 holds the property
    % further frames will be added by extend later on
    % we store the transition in the frames as well in order to spare passing it around
    b_get_properties_from_machine(PropertiesOnConstants),
    F_0 = frame([Init,PropertiesOnConstants],[]), F_1 = frame([Property,PropertiesOnConstants],[]),
    empty_avl(E),
    avl_store(0,E,F_0,FramesT),
    avl_store(1,FramesT,F_1,Frames),
    get_abstract_domain(Domain),
    catch(symbolic_model_check_fx(K,Frames,Domain,Property,Init,Result),
          ic3_counter_example(Solution,Succs),
          (handle_counter_example(Solution,Succs),Result=counterexample_found)).

symbolic_model_check_fx(K,Frames,Domain,Property,Init,Result) :-
    % initially, K is 1
    % the algorithm tries to maintain the following invariants:
    % (1) ! i . i > 0 . I => F_i
    % (2) ! i . i > 0 . F_i => P (P = Property to check / Invariant)
    % (3) ! i . i > 0 . clauses(F_{i+1}) <: clauses(F_i)
    % (4) ! i . 0 <= i < k . F_i & T => F_{i+1}'
    extend(Frames,K,[Property],Frames2),
    % strengthen establishes (1) - (4)
    % if impossible => counter example
    strengthen(K,Frames2,Domain,Property,FramesOut,RefinedDomain), !,
    % propagate_clauses assumes (1) - (4)
    % pushes clauses from one frame to a later one (as far as possible)
    propagate_clauses(K,FramesOut,FramesOut2),
    (check_convergence(FramesOut2)
     -> Result = property_holds
     ;  K2 is K + 1,
        symbolic_model_check_fx(K2,FramesOut2,RefinedDomain,Property,Init,Result)).

:- public print_frames/1. % for debugging
print_frames(Frames) :-
    avl_to_list(Frames,List),
    print_frames_list(List).
print_frames_list([]).
print_frames_list([K-frame(_,Clauses)|T]) :-
    length(Clauses,L),
    format('frame ~w holds ~w clauses:~n',[K,L]),
    %print_clauses(Clauses),
    print_frames_list(T).

:- public print_clauses/1. % for debugging
print_clauses([]).
print_clauses([C|Cs]) :-
    disjunct_predicates(C,DC),
    translate:print_bexpr(DC),nl,
    print_clauses(Cs).

extend(FramesIn,K,Content,FramesOut) :-
    % any new frame holds the properties in addition to the content (usually the transition)
    NextFrameId is K + 1,
    b_get_properties_from_machine(PropertiesOnConstants),
    avl_store(NextFrameId,FramesIn,frame([PropertiesOnConstants|Content],[]),FramesOut).

propagate_clauses(K,Frames,FramesOut) :-
    % for i = 1 .. K call propagate_clauses_on_level
    for(I,1,K), fromto(Frames,In,Out,FramesOut)
    do propagate_clauses_on_level(I,In,Out).
propagate_clauses_on_level(I,Frames,FramesOut) :-
    % pushes from F_I to F_{I+1}
    IP1 is I + 1,
    clauses_on_level(I,Frames,_F_From,ClausesFrom),
    in_solver_on_level(I,Frames,IS),

    include(inductive_clause(IS),ClausesFrom,InductiveClauses),

    add_clauses_to_level(IP1,Frames,InductiveClauses,FramesOut).

inductive_clause(F_I,ClauseDisjuncts) :-
    get_transition_predicate(T),
    disjunct_predicates(ClauseDisjuncts,Clause),
    % checks if a single clause is inductive relative to F_i
    % by verifying if F_i => C' (F_i includes T)
    % i.e. there is a contradiction in not C' & F_i
    create_negation(Clause,NegClause),
    prime_predicate(NegClause,NegCPrime),
    conjunct_predicates([NegCPrime,T|F_I],Pred),
    solve(Pred,Result), !,
    Result = contradiction_found(_).

strengthen(K,Frames,Domain,Property,FramesOut,RefinedDomain) :-
    findall(op(OpName,Trans),get_single_transition_predicate(OpName,Trans),Transitions),
    (foreach(Transition,Transitions),
    fromto(Frames,TFIn,TFOut,FramesOut), fromto(Domain,TDIn,TDOut,RefinedDomain),
    param(K), param(Property)
    do strengthen_loop(Transition,K,Property,TFIn,TFOut,TDIn,TDOut)).

strengthen_loop(op(OpName,Transition),K,Property,FramesIn,FramesOut,DomainIn,DomainOut) :-
    get_preference(use_po,UsePO),
    strengthen_get_property(UsePO,OpName,Property,PrimedNegProperty),
    in_solver_on_level(K,FramesIn,IS),
    conjunct_predicates([Transition,PrimedNegProperty|IS],RPred),
    solve(RPred,Result),
    strengthen_aux(Result,Transition,K,Property,FramesIn,FramesOut,DomainIn,DomainOut).

strengthen_get_property(false,_OpName,Property,PrimedNegProperty) :-
    create_negation(Property,NegProperty),
    prime_predicate(NegProperty,PrimedNegProperty).
strengthen_get_property(true,OpName,FullProperty,Primed) :-
    (b_specialized_invariant_for_op(OpName,Property)
     -> true ; Property = FullProperty),
    (get_proven_invariant(OpName,Proven)
     -> true ; Proven = b(truth,pred,[])),
    create_negation(Property,NegProperty),
    conjunct_predicates([Proven,NegProperty],ToPrime),
    prime_predicate(ToPrime,Primed).

% strengthen searches for a counter example to the property
% it searches for a state that is permitted by F_i that is the predecessor
% of a state that violates the property
% if there is a contradiction, we are done, because there is no counter example
strengthen_aux(contradiction_found(_),_TransitionPredicate,_K,_Property,Frames,Frames,Domain,Domain).
% if there is a counter example, we need to strenghten the frames
% in order to make it spurious, i.e. (1) - (4) have to be reestablished
strengthen_aux(solution(Wit),TransitionPredicate,K,Property,Frames,FramesOut,Domain,DomainOut) :-
    % extract predecessor
    create_state_predicate(Wit,Pre),
    create_succ_state_predicate(Wit,SStatePred),
    create_primary_inputs_predicate(Wit,Inputs),
    K2 is K-2,
    % finds the highest N to which not Pre can be added
    % this can fail if the counter example is not spurious
    inductively_generalize(Pre,[Pre],K2,K,Frames,FramesT,N), !,
    N2 is N + 1,
    empty_heap(H),
    generalize_and_add(concrete,TransitionPredicate,K,H,N2,Frames,Domain,RefinedDomain,Pre,Inputs,[SStatePred],Obls),
    % Obls is a heap of proof obligations that need to hold
    % push generalization handles those that are on level =< K
    push_generalization(Obls,K,FramesT,FramesT2,RefinedDomain,ReRefinedDomain),
    strengthen(K,FramesT2,ReRefinedDomain,Property,FramesOut,DomainOut).

% generalize Pre and / or split it into multiple obligations
generalize_and_add(concrete,TP,I,HIn,N,Frames,Domain,Domain,State,Inputs,[SuccState|SuccStates],HOut) :-
    concrete_state_to_abstract_state(Domain,State,AbsCTI),
    write(absstate(AbsCTI)),nl,
    % check if lifting is possible / if query is actually unsat
    in_solver_on_level(I,Frames,IS),
    prime_predicate(SuccState,TPrime),
    create_negation(TPrime,NotTPrime),
    conjunct_predicates([NotTPrime,TP|IS],Conjunction),
    conjunct_predicates([AbsCTI,Conjunction],VerifyQuery),
    solve(VerifyQuery,Result),
    write(result(Result)),nl,
    Result = contradiction_found(_), !,
    % query is unsat and lifting succeeds
    unsat_core_with_fixed_conjuncts(AbsCTI,Conjunction,Core),
    I2 is I-1, M is N+1,
    NewObligation = obligation(Core,TP,Inputs,[SuccState|SuccStates],I2,M),
    add_to_heap(HIn,N,NewObligation,HOut).
generalize_and_add(concrete,TP,I,HIn,N,Frames,Domain,RefinedDomain,State,Inputs,[SuccState|SuccStates],HOut) :-
    findall(WP,
            (b_get_machine_operation(_,_,_,D),weakest_precondition(D,SuccState,WP)),
            WeakestPreconditions),
    conjunct_predicates(WeakestPreconditions,BigConjunction),
    conjunction_to_list(BigConjunction,AllConjuncts),
    list_to_ord_set(AllConjuncts,OrdAllConjuncts),
    ord_union(OrdAllConjuncts,Domain,NewDomain),
    length(Domain,LD), length(NewDomain,NLD),
    (NLD = LD -> write('diverges'),nl,trace ; true),
    generalize_and_add(concrete,TP,I,HIn,N,Frames,NewDomain,RefinedDomain,State,Inputs,[SuccState|SuccStates],HOut).

push_generalization(Obligations,_K,Frames,Frames,Domain,Domain) :-
    empty_heap(Obligations). % no proof obligations left
push_generalization(Obligations,K,Frames,Frames,Domain,Domain) :-
    min_of_heap(Obligations,N,_S),
    N > K, !. % only proof obligations on a higher level left - fine for level K
push_generalization(Obligations,K,Frames,FramesOut,Domain,DomainOut) :-
    % grep the proof obligations with the lowest level - somewhat a heuristic
    min_of_heap(Obligations,N,obligation(S,T,Inputs,Succs,I,M)),
    in_solver_on_level(N,Frames,IS),
    prime_predicate(S,SPrimed),
    conjunct_predicates([SPrimed,T|IS],Pred),
    solve(Pred,Result), !,
    push_generalization_aux(Result,T,N,obligation(S,T,Inputs,Succs,I,M),Obligations,K,Frames,FramesT,Domain,RefinedDomain,ObligationsOut),
    push_generalization(ObligationsOut,K,FramesT,FramesOut,RefinedDomain,DomainOut).

push_generalization_aux(solution(Pre),_T,N,obligation(S,T,Inputs,Succs,_I,_M),Obligations,K,Frames,FramesOut,Domain,RefinedDomain,ObligationsOut) :-
    N2 is N - 2,
    create_state_predicate(Pre,PrePred), !,
    inductively_generalize(PrePred,[S|Succs],N2,K,Frames,FramesOut,M), !,
    M2 is M + 1,
    generalize_and_add(concrete,T,K,Obligations,M2,Frames,Domain,RefinedDomain,PrePred,Inputs,[S|Succs],ObligationsOut).
push_generalization_aux(contradiction_found(_),_T,N,obligation(S,T,Inputs,Succs,I,M),Obligations,K,Frames,FramesOut,Domain,Domain,ObligationsOut) :-
    inductively_generalize(S,[S|Succs],N,K,Frames,FramesOut,Lvl), !,
    delete_from_heap(Obligations,N,obligation(S,T,Inputs,Succs,I,M),ObligationsT),
    M2 is M + 1, NLvl is Lvl + 1,
    add_to_heap(ObligationsT,NLvl,obligation(S,T,Inputs,Succs,I,M2),ObligationsOut).

:- public print_heap/1. % for debugging
print_heap(Obls) :-
    heap_to_list(Obls,List),
    write('Content of Obligations Heap:'), nl,
    print_heap2(List).
print_heap2([]).
print_heap2([Key-Datum|T]) :-
    Datum = obligation(Predicate,_,Succs),
    translate_bexpression(Predicate,PPPredicate),
    format('On level ~w state predicate ~w~n',[Key,PPPredicate]),
    format('Successors:~n',[]),
    print_succs(Succs),
    print_heap2(T).
print_succs([]).
print_succs([P|T]) :-
    write('   '), print_bexpr(P),nl,
    print_succs(T).

handle_counter_example(Solution,Succs) :-
    create_state_predicate(Solution,StatePred),
    create_succ_state_predicate(Solution,SuccStatePred), !,
    format('Counterexample found. Trace:~n',[]),
    print_succs([StatePred,SuccStatePred|Succs]),
    (create_constants_state_predicate(Solution,ConstantsStatePred)
     -> replay_counter_example([ConstantsStatePred,StatePred,SuccStatePred|Succs])
     ;  replay_counter_example([StatePred,SuccStatePred|Succs])).
replay_counter_example(A) :-
    state_space_reset,
    replay_aux(A,_).
:- use_module(probsrc(b_trace_checking),[find_successor_state/4]).
replay_aux([],_).
replay_aux([P|Ps],CurId) :-
    find_successor_state(CurId,P,SuccID,_),
    replay_aux(Ps,SuccID).

inductively_generalize(S,Succs,Min,_K,Frames,_FramesOut,_N) :-
    % if min < 0 and sat(F_0 & T & neg S & S')
    % fail to report counter example
    Min < 0,
    in_solver_on_level(0,Frames,IS),
    prime_predicate(S,SPrimed),
    create_negation(S,SNegated),
    get_transition_predicate(T),
    conjunct_predicates([SNegated,SPrimed,T|IS],Pred),
    solve(Pred,Result),
    Result = solution(Solution), !,
    throw(ic3_counter_example(Solution,Succs)).
inductively_generalize(S,_,Min,K,Frames,FramesOut,N) :-
    I is max(1, Min + 1),
    inductively_generalize_aux(I,S,K,Frames,FramesOut,N).
inductively_generalize_aux(I,S,K,Frames,FramesOut,N) :-
    I =< K,
    in_solver_on_level(I,Frames,IS),
    prime_predicate(S,SPrimed),
    create_negation(S,SNegated),
    get_transition_predicate(T),
    conjunct_predicates([SNegated,SPrimed,T|IS],Pred),
    solve(Pred,Result), !,
    (Result = solution(_) -> N is I - 1, generate_clause(S,N,K,Frames,FramesOut) ;
     Result = contradiction_found(_) -> I2 is I + 1, inductively_generalize_aux(I2,S,K,Frames,FramesOut,N)).
inductively_generalize_aux(I,S,K,Frames,FramesOut,K) :-
    I > K,
    generate_clause(S,K,K,Frames,FramesOut).

generate_clause(S,I,K,Frames,FramesOut) :-
    mic(S,I,K,Frames,FramesT,Q),
    translate_bexpression(S,PS),
    translate_bexpression(Q,PQ),
    format('minimal inductive clause computation:~n~w to~n~w~n',[PS,PQ]),
    conjunction_to_list(Q,Conjuncts),
    maplist(create_negation,Conjuncts,NegConjuncts),
    generate_clause_aux(1,NegConjuncts,I,FramesT,FramesOut).
generate_clause_aux(J,NegS,I,Frames,FramesOut) :-
    J =< I + 1, !,
    add_clauses_to_level(J,Frames,[NegS],FramesT),
    J2 is J + 1,
    generate_clause_aux(J2,NegS,I,FramesT,FramesOut).
generate_clause_aux(_,_,_,Frames,Frames).

mic(S,I,_K,Frames,Frames,Q) :-
    mic_algorithm(down), !,
    conjunction_to_list(S,Conjuncts),
    down(Conjuncts,I,Frames,ConjunctsOut),
    conjunct_predicates(ConjunctsOut,Q).
mic(S,I,K,FramesIn,FramesOut,Q) :-
    mic_algorithm(ctgDown), !,
    conjunction_to_list(S,Conjuncts),
    ctgDown(Conjuncts,I,K,FramesIn,FramesOut,ConjunctsOut),
    conjunct_predicates(ConjunctsOut,Q).

check_convergence(Frames) :-
    avl_to_list(Frames,L),
    check_convergence_aux(L).
check_convergence_aux([_-F1,K2-F2|Fs]) :-
    % the algorithm converges if two frames hold the same set of clauses
    F1 = frame(_,Clauses1),
    F2 = frame(_,Clauses2),
    % assumes ordered sets
    (ord_symdiff(Clauses1, Clauses2, []) ; check_convergence_aux([K2-F2|Fs])).
