:- module(data_generator,[generate_data/1]).

% Run from the prob_prolog root directory like sics -l src/synthesis/deep_learning/data_generator.pl
% The data is of the following form:
% Example,Example,..,GroundTruth,component1,component2,...
% whereat a single Example is a set of tuples of input and output for each value like ((v1_input,v1_output)~(v2_input,v2_output)~...~(vn_input,vn_output))

% TODO: generate data for preconditions and invariants to synthesize predicates instead of expressions

:- ensure_loaded('../../prob_cli') ,
   set_search_pathes ,
   init_eclipse_preferences ,
   external_functions:reset_argv ,
   set_prefs.

:- use_module(library(lists),[append/2,maplist/3]).
:- use_module(library(random),[random/3]).
:- use_module(library(codesio),[write_term_to_codes/3]).
:- use_module(library(file_systems),[file_members_of_directory/3,
                                     directory_members_of_directory/2,
                                     directory_exists/1]).

:- use_module(probsrc(bmachine),[b_get_machine_operation/4,
                                 b_load_machine_from_file/1,
                                 b_machine_precompile/0,
                                 b_machine_reset/0]).
:- use_module(probsrc(solver_interface),[solve_predicate/3]).
:- use_module(probsrc(translate),[translate_bexpression/2]).
:- use_module(probsrc(preferences),[set_preference/2]).
:- use_module(probsrc(bsyntaxtree),[conjunct_predicates/2]).
:- use_module(probsrc(before_after_predicates),[before_after_predicate_for_operation/2]).

:- use_module(synthesis(synthesis_util),[b_get_typed_invariant_from_machine/1,
                                         get_valid_inputs_for_operation/3,
                                         create_equality_nodes_from_example/2,
                                         get_output_nodes_from_bindings/2]).

:- module_info(group,synthesis).
:- module_info(description,'This module generates data to train a neural network that predicts the library components that are necessary to synthesize a program for given I/O examples.').

prob_examples_root_dir('~/STUPS/prob_examples/').

min_amount_of_examples(2).
max_amount_of_examples(10).

file_to_save('src/synthesis/deep_learning/data.csv').

generate_data(Data) :-
    get_classical_b_machines(FileList) ,
    file_to_save(FilePath) ,
    open(FilePath,write,Stream) ,
    generate_data_from_machine_paths(Stream,FileList,Data) ,
    close(Stream).

get_classical_b_machines(FileList) :-
    prob_examples_root_dir(ProBExamplesDir) ,
    get_classical_b_machines(ProBExamplesDir,FileList).

% Get a list of file paths from the sub directories of the given directory.
get_classical_b_machines(Directory,FileList) :-
    directory_exists(Directory) ,
    directory_members_of_directory(Directory,SubDirectories) ,
    get_classical_b_machines_aux(SubDirectories,FileList).
get_classical_b_machines(Directory,[]) :-
    \+directory_exists(Directory).

get_classical_b_machines_aux(Directory,FileList) :-
    get_classical_b_machines_aux(Directory,[],FileList).

% Get a list of file paths from the list of directories and recursively crawl each sub directory.
get_classical_b_machines_aux([],Acc,Acc).
get_classical_b_machines_aux([_-AbsoluteSubDirectoryPath|T],Acc,FileList) :-
    file_members_of_directory(AbsoluteSubDirectoryPath,'*.mch',CurrentFileList) ,
    get_classical_b_machines(AbsoluteSubDirectoryPath,RecursiveFileList) ,
    append([CurrentFileList,RecursiveFileList,Acc],NewAcc) ,
    get_classical_b_machines_aux(T,NewAcc,FileList).

generate_data_from_machine_paths(Stream,FileList,Data) :-
    generate_data_from_machine_paths(Stream,FileList,[],Data).

generate_data_from_machine_paths(_,[],Acc,Acc).
generate_data_from_machine_paths(Stream,[_-AbsoluteFilePath|T],Acc,Data) :-
    format("~w~n~n",[AbsoluteFilePath]) ,
    generate_data_from_machine_path(Stream,AbsoluteFilePath,MachineData) ,
    append(MachineData,Acc,NewAcc) ,
    generate_data_from_machine_paths(Stream,T,NewAcc,Data).

% Load the machine for a given path and generate data for this machine.
generate_data_from_machine_path(Stream,AbsoluteFilePath,Data) :-
    b_machine_reset ,
    set_preference(try_kodkod_on_load,false) ,
    b_load_machine_from_file(AbsoluteFilePath) ,
    b_machine_precompile ,
    get_all_operation_bodies_from_current_machine(OperationBodies) ,
    generate_data_from_operation_substitutions(Stream,OperationBodies,Data).
% machine could not be loaded, e.g. because of a well-definedness error
generate_data_from_machine_path(_,_,[]).

get_all_operation_bodies_from_current_machine(OperationBodies) :-
    findall((OperationName,OperationBody),b_get_machine_operation(OperationName,_,_,OperationBody),OperationBodies).

% Generate data from the substitutions of the given machine operations, i.e., generate
% I/O examples that satisfy the precondition and infer the ground truth from the substitution.
generate_data_from_operation_substitutions(Stream,OperationBodies,Data) :-
    generate_data_from_operation_substitutions(Stream,OperationBodies,[],Data).

generate_data_from_operation_substitutions(_,[],Acc,Acc).
generate_data_from_operation_substitutions(Stream,[(OperationName,OperationBody)|T],Acc,Data) :-
    format('.',[]) ,
    get_random_amount_of_examples(AmountOfExamples) ,
    get_valid_inputs_for_operation(OperationName,AmountOfExamples,InputExamples) ,
    get_outputs_for_input_examples(InputExamples,OperationName,InputOutputTuples) ,
    get_ground_truth_from_operation_body(OperationBody,GroundTruth) , ! ,
    ((InputExamples \= [[]] , InputExamples \= [] , InputOutputTuples \= [] , GroundTruth \= [])
    ->  write_single_data_to_file(Stream,InputOutputTuples,GroundTruth) ,
        generate_data_from_operation_substitutions(Stream,T,[(InputOutputTuples,GroundTruth)|Acc],Data)
    ;   % skip this operation if no I/O examples could be generated or no ground truth can be inferred
        generate_data_from_operation_substitutions(Stream,T,Acc,Data)).

% create a csv line for a single data of I/O examples and ground truth (currently using ~ as the delimiter)
write_single_data_to_file(Stream,[],GroundTruth) :-
    write_list_to_csv(Stream,_,['GroundTruth~']) ,
    write_list_to_csv(Stream,'~',GroundTruth) ,
    write_list_to_csv(Stream,_,['\n']).
write_single_data_to_file(Stream,[(Input,Output)|T],GroundTruth) :-
    interleave_to_tuple_atom(Input,Output,IOValueTuples) , % single example with tuples of input and output for each value
    write(Stream,'(') ,
    write_list_to_csv(Stream,',',IOValueTuples) ,
    write(Stream,')~') ,
    write_single_data_to_file(Stream,T,GroundTruth).

write_list_to_csv(_,_,[]).
write_list_to_csv(Stream,_,[H]) :-
    write(Stream,H). % last value without commata
write_list_to_csv(Stream,Delimiter,[H|T]) :-
    T \= [] ,
    write(Stream,H) ,
    write(Stream,Delimiter) ,
    write_list_to_csv(Stream,Delimiter,T).

get_random_amount_of_examples(AmountOfExamples) :-
    min_amount_of_examples(MinAmountOfExamples) ,
    max_amount_of_examples(TempMaxAmountOfExamples) ,
    MaxAmountOfExamples is TempMaxAmountOfExamples + 1 ,
    random(MinAmountOfExamples,MaxAmountOfExamples,AmountOfExamples).

get_outputs_for_input_examples(InputExamples,OperationName,TranslatedInputOutputTuples) :-
    get_outputs_for_input_examples(InputExamples,OperationName,[],InputOutputTuples) ,
    translate_input_output_tuples(InputOutputTuples,TranslatedInputOutputTuples).

% Compute the corresponding output for each input example and return a list of I/O tuples.
get_outputs_for_input_examples([],_,Acc,Acc).
get_outputs_for_input_examples([InputExample|T],OperationName,Acc,InputOutputTuples) :-
    create_equality_nodes_from_example(InputExample,EqualityNodes) ,
    conjunct_predicates(EqualityNodes,EqualityPredicate) ,
    before_after_predicate_for_operation(OperationName,BeforeAfterPredicate) ,
    conjunct_predicates([EqualityPredicate,BeforeAfterPredicate],BeforeAfterPredicateWithEqualities) ,
    solve_predicate(BeforeAfterPredicateWithEqualities,_,Solution) ,
    Solution = solution(SolutionBindings) , ! ,
    get_output_nodes_from_bindings(SolutionBindings,OutputExample) ,
    get_outputs_for_input_examples(T,OperationName,[(InputExample,OutputExample)|Acc],InputOutputTuples).
get_outputs_for_input_examples([_|T],OperationName,Acc,InputOutputTuples) :-
    get_outputs_for_input_examples(T,OperationName,Acc,InputOutputTuples).

translate_input_output_tuples(InputOutputTuples,TranslatedInputOutputTuples) :-
    translate_input_output_tuples(InputOutputTuples,[],TranslatedInputOutputTuples).

translate_input_output_tuples([],Acc,Acc).
translate_input_output_tuples([(Input,Output)|T],Acc,TranslatedInputOutputTuples) :-
    maplist(translate_bexpression,Input,TranslatedInput) ,
    maplist(translate_bexpression,Output,TranslatedOutput) ,
    translate_input_output_tuples(T,[(TranslatedInput,TranslatedOutput)|Acc],TranslatedInputOutputTuples).

get_ground_truth_from_operation_body(OperationBody,UsedComponents) :-
    get_used_library_components_from_substitution(OperationBody,UsedComponents).

get_used_library_components_from_substitution(b(Substitution,_,_),UsedComponents) :-
    Substitution =.. [Functor,Assignments] ,
    (Functor = parallel ; Functor = sequence) ,
    maplist(get_used_library_components_from_substitution,Assignments,NestedListOfComponents) ,
    append(NestedListOfComponents,UsedComponents).
get_used_library_components_from_substitution(b(Node,_,_),UsedComponents) :-
    Node =.. [Functor,_,_,Assignments] ,
    (Functor = let ; Functor = any) ,
    is_list(Assignments) ,
    maplist(get_used_library_components_from_substitution,Assignments,NestedListOfComponents) ,
    append(NestedListOfComponents,UsedComponents).
get_used_library_components_from_substitution(b(Node,_,_),UsedComponents) :-
    Node =.. [Functor,_,_,Assignment] ,
    (Functor = let ; Functor = any) ,
    \+ is_list(Assignment) ,
    get_used_library_components_from_substitution(Assignment,UsedComponents).
get_used_library_components_from_substitution(b(precondition(_,Substitution),_,_),UsedComponents) :-
    get_used_library_components_from_substitution(Substitution,UsedComponents).
get_used_library_components_from_substitution(b(if(SubstitutionList),_,_),[if|UsedComponents]) :-
    maplist(get_used_library_components_from_substitution,SubstitutionList,NestedListOfComponents) ,
    append(NestedListOfComponents,UsedComponents).
get_used_library_components_from_substitution(b(if_elsif(Condition,Substitution),_,_),UsedComponents) :-
    get_used_library_components_from_substitution(Condition,UsedComponents1) ,
    get_used_library_components_from_substitution(Substitution,UsedComponents2) ,
    append(UsedComponents1,UsedComponents2,UsedComponents).
get_used_library_components_from_substitution(b(select([b(select_when(_,Assignments),_,_)]),_,_),UsedComponents) :-
    is_list(Assignments) ,
    maplist(get_used_library_components_from_substitution,Assignments,NestedListOfComponents) ,
    append(NestedListOfComponents,UsedComponents).
get_used_library_components_from_substitution(b(select([b(select_when(_,Assignment),_,_)]),_,_),UsedComponents) :-
    \+ is_list(Assignment) ,
    get_used_library_components_from_substitution(Assignment,UsedComponents).
get_used_library_components_from_substitution(b(assign(_,Assignments),_,_),UsedComponents) :-
    maplist(get_used_library_components_from_substitution,Assignments,NestedListOfComponents) ,
    append(NestedListOfComponents,UsedComponents).
get_used_library_components_from_substitution(b(identifier(_),_,_),[]).
get_used_library_components_from_substitution(b(assign_single_id(_,Arg),_,_),UsedComponents) :-
    get_used_library_components_from_substitution(Arg,UsedComponents).
get_used_library_components_from_substitution(b(Bool,boolean,_),[constant-boolean]) :-
    Bool = boolean_true ; Bool = boolean_false.
get_used_library_components_from_substitution(b(integer(_),integer,_),[constant-integer]).
get_used_library_components_from_substitution(b(string(_),string,_),[constant-string]).
get_used_library_components_from_substitution(b(Node,Type,_),[constant-Type]) :-
    member(Node,[set_extension(_),rec(_),sequence_extension(_)]).
get_used_library_components_from_substitution(b(value(_),Type,_),[constant-Type]).
get_used_library_components_from_substitution(b(Node,_,_),[Component|UsedComponents]) :-
    Node =.. [Component,Arg] ,
    \+ member(Component,[identifier,parallel,sequence]) ,
    get_used_library_components_from_substitution(Arg,UsedComponents).
get_used_library_components_from_substitution(b(Node,_,_),[Component|UsedComponents]) :-
    Node =.. [Component,Arg1,Arg2] ,
    \+ member(Component,[assign_single_id,assign]) ,
    get_used_library_components_from_substitution(Arg1,UsedComponents1) ,
    get_used_library_components_from_substitution(Arg2,UsedComponents2) ,
    append(UsedComponents1,UsedComponents2,UsedComponents).
get_used_library_components_from_substitution(_,[]).

interleave_to_tuple_atom(A,B,Interleaved) :-
    interleave_to_tuple_atom(A,B,[],Interleaved).

interleave_to_tuple_atom([],_,Acc,Acc).
interleave_to_tuple_atom(_,[],Acc,Acc).
interleave_to_tuple_atom([Ai|T1],[Bi|T2],Acc,Interleaved) :-
    write_term_to_codes(Ai,CAi,[]) ,
    write_term_to_codes(Bi,CBi,[]) ,
    append([CAi,[44],CBi,[41]],TempCodes) ,
    atom_codes(TupleAtom,[40|TempCodes]) ,
    interleave_to_tuple_atom(T1,T2,[TupleAtom|Acc],Interleaved).
