:- module(location_vars_to_program,[location_vars_to_program/9]).

:- use_module(synthesis(synthesis_util)).
:- use_module(synthesis(logging)).

:- use_module(probsrc(bmachine_eventb),[prime_variable/2]).
:- use_module(probsrc(bsyntaxtree),[conjunct_predicates/2,get_texpr_info/2,add_texpr_infos/3]).
:- use_module(probsrc(module_information),[module_info/2]).

:- use_module(library(sets),[subset/2]).
:- use_module(library(lists),[append/2,remove_dups/2,maplist/3,delete/3]).

:- module_info(group,synthesis).
:- module_info(description,'This module defines the translation from a distribution of location variables to a B/Event-B AST.').

% Convert location values to a program represented as an ast.
% Collect all location variables that are part of the solution (no deadcode).
location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,Parameter,solution(Solution),UsedLVars,AST) :-  
    % split by temporary input and output variables
    findall(Out,(member(Out,LocationVars) , Out = b(identifier(Name),_,_) ,
         atom_concat(lo,_,Name)),LocationVarsOut) ,
    delete(LocationVars,LocationVarsOut,LocationVarsIn) ,
    % get output location variable defined in the last program line
    (location_vars_to_program_aux(Type,M1,Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,UsedLVars,AST)
    -> true ; synthesis_log(location_vars_to_program_and_fail(Type,Library,LocationVarsIn,LocationVarsOut,Parameter,Solution))) , !.
location_vars_to_program_aux(action,M1,Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,UsedLVars,AST) :-
    length(CurrentVars,OutputAmount) ,
    M0 is M1 - OutputAmount ,
    % Translate the program of each output variable and conjunct them. For constraint solving we need
    % conjunctions which are replaced with parallel executions after finding the final solution.
    get_program_for_output_var(Library,M0,M1,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,TempUsedLVars,ASTList) ,
    remove_dups(TempUsedLVars,UsedLVars) ,
    conjunct_predicates(ASTList,AST).
location_vars_to_program_aux(Type,M1,Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,UsedLVars,AST) :-
    Type \= action ,
    % Output values of the program are also located to the last line (root), but we want to find the output
    % location variable. Here, we just have one output whose location variable is named 'lo0'.
    member(binding(Root,int(M1),_),Solution) , atom_concat(lo,Suffix,Root) , \+atom_concat(value,_,Suffix) ,
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,Root,UsedLVars,AST).

% For synthesis of actions:
% If we have more than one output we need to collect the partial programs and insert
% assignments with primed variables.
get_program_for_output_var(_,M0,M0,_,_,_,_,_,[],[]).
get_program_for_output_var(Library,M0,ProgramLine,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,UsedLVars,[Assignment|T]) :-
    member(binding(Root,int(ProgramLine),_),Solution) , atom_concat(lo,_,Root) ,
    % get the output location mapped to this line of the program
    get_location_var_node_by_info_term(LocationVarsOut,synthesis(program_line(ProgramLine)),RootNode) ,
    % and get the corresponding machine var using the info term synthesis/2
    get_texpr_info(RootNode,RootInfo) ,
    member(synthesis(machinevar,MachineVar),RootInfo) ,
    member(b(identifier(MachineVar),MachineVarType,MachineVarInfo),CurrentVars) ,
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,Root,TempUsedLVars,AST) ,
    prime_variable(b(identifier(MachineVar),MachineVarType,MachineVarInfo),PrimedVar) ,
    Assignment = b(equal(PrimedVar,AST),pred,[]) ,
    NProgramLine is ProgramLine - 1 ,
    get_program_for_output_var(Library,M0,NProgramLine,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,R,T) ,
    append(TempUsedLVars,R,UsedLVars).

set_type(b(member(b(Elm,EType,EInfo),Set),NodeType,MInfo),b(member(b(Elm,EType,EInfo),NSet),NodeType,MInfo)) :-
    set_type_aux(EType,Set,NSet).
set_type(b(Node,Type,Info),b(NewNode,Type,Info)) :-
    Node =.. [Head,Arg] , Head \= identifier ,
    set_type(Arg,NArg) ,
    NewNode =.. [Head,NArg].
set_type(b(Node,Type,Info),b(NewNode,Type,Info)) :-
    Node =.. [Head,Arg1,Arg2] ,
    set_type(Arg1,NArg1) ,
    set_type(Arg2,NArg2) ,
    NewNode =.. [Head,NArg1,NArg2].
set_type(Node,Node).

set_type_aux(Type,b(set_extension(Set),_,SInfo),b(set_extension(NSet),set(Type),SInfo)) :-
    maplist(set_type_aux(Type),Set,NSet).
set_type_aux(Type,b(Node,_,Info),b(Node,Type,Info)).

% if_then_else
location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,LocationVarName,[LocationVarNode|RUsedLVars],IfThenElseAst) :-
    LocationVarNode = b(identifier(LocationVarName),_,Info) ,
    member(LocationVarNode,LocationVarsOut) ,
    sets:subset([synthesis(type,Type),synthesis(Component,output)],Info) ,
    atom_concat(if_then_else,_,Component) , ! ,
    % find the location variable of the statement's condition
    get_location_var_node_by_info_term(LocationVarsIn,synthesis(Component,condition),b(identifier(ConditionName),_,_)) ,
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,ConditionName,ConditionUsedLVars,ConditionAst) ,
    % find location variables for truth and falsity outputs of the if_then_else statement
    get_location_var_node_by_info_term(LocationVarsIn,synthesis(Component,true_out),b(identifier(TrueOutName),_,_)) ,
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,TrueOutName,TrueOutUsedLVars,TrueOutAst) ,
    get_location_var_node_by_info_term(LocationVarsIn,synthesis(Component,false_out),b(identifier(FalseOutName),_,_)) ,
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,FalseOutName,FalseOutUsedLVars,FalseOutAst) ,
    append([ConditionUsedLVars,TrueOutUsedLVars,FalseOutUsedLVars],RUsedLVars) ,
    IfThenElseAst = b(if_then_else(ConditionAst,TrueOutAst,FalseOutAst),Type,[]).
% machine constants or operation parameters
location_to_ast(_,_,_,LocationVarsOut,_,_,LocationVarName,[LocationVarNode],IdNode) :-
    LocationVarNode = b(identifier(LocationVarName),_,Info) ,
    member(LocationVarNode,LocationVarsOut) ,
    (sets:subset([synthesis(_,output),synthesis(machineconstant,IdNode)],Info) ;
     sets:subset([synthesis(_,output),synthesis(operationparameter,IdNode)],Info)) , !.
% constant sets like NATURAL or INTEGER
location_to_ast(_,_,_,LocationVarsOut,_,_,LocationVarName,[LocationVarNode],AST) :-
    LocationVarNode = b(identifier(LocationVarName),_,Info) ,
    member(LocationVarNode,LocationVarsOut) ,
    sets:subset([synthesis(ConstantName,output),synthesis(type,Type)],Info) ,
    atom_concat(constant_set,_,ConstantName) , ! ,
    % export as infinite set
    get_single_constant_or_set_operator_id(Type,_,AST).
location_to_ast(_,_,_,LocationVarsOut,_,_,LocationVarName,[LocationVarNode],Node) :-
    LocationVarNode = b(identifier(LocationVarName),_,Info) ,
    member(LocationVarNode,LocationVarsOut) ,
    member(synthesis(ConstantName,output),Info) ,
    member(synthesis(ParamType,ParamName),Info) ,
    (ParamType = param_single ; ParamType = param_set) , 
    atom_concat(constant,_,ConstantName) , ! , 
    member(synthesis(type,Type),Info) ,
    (Type = set(global(Independent)) ; Type = global(Independent)) ,
    (ParamType = param_set
    ->  Node = b(set_extension([b(identifier(ParamName),global(Independent),Info)]),set(global(Independent)),Info)
    ;   ParamType = param_single , 
        Node = b(identifier(ParamName),global(Independent),Info)).
% synthesized constants that have been enumerated by the solver
location_to_ast(_,_,_,LocationVarsOut,_,Solution,LocationVarName,[LocationVarNode],Node) :-
    LocationVarNode = b(identifier(LocationVarName),_,Info) ,
    member(LocationVarNode,LocationVarsOut) ,
    member(synthesis(ConstantName,output),Info) ,
    atom_concat(constant,_,ConstantName) , ! ,
    % get constant solution, either integer or boolean
    member(binding(ConstantName,Value,_),Solution) ,
    member(synthesis(type,Type),Info) ,
    Node = b(value(Value),Type,Info).
% output
% one input parameter
location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,LocationVarName,[LocationVarNode|RUsedLVars],AstNode) :-
    LocationVarNode = b(identifier(LocationVarName),_,InfoOut) ,
    member(LocationVarNode,LocationVarsOut) ,
    sets:subset([synthesis(Component,output),synthesis(type,Type)],InfoOut) ,
    synthesis_util:delete_numbers_from_atom(Component,ComponentName) ,
    member(ComponentName,[negation,card,size,max,min,general_union,general_intersection,reverse,range,domain,skip,convert_bool,first,last]) , ! ,
    % find input location variables
    get_location_var_node_by_info_term(LocationVarsIn,synthesis(Component,input),b(identifier(InName),_,_)) ,
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,InName,RUsedLVars,InputAST) ,
    Node =.. [ComponentName,InputAST] ,
    single_input_location_to_ast_aux(ComponentName,b(Node,Type,InfoOut),InputAST,AstNode).
% two input parameters
location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,LocationVarName,UsedLVars,AST) :-
    LocationVarNode = b(identifier(LocationVarName),_Type,InfoOut) ,
    member(LocationVarNode,LocationVarsOut) , ! ,
    member(synthesis(Component,output),InfoOut) ,
    % find input location variables
    get_location_var_node_by_info_term(LocationVarsIn,synthesis(Component,left_input),b(identifier(InName1),_,_)) ,
    get_location_var_node_by_info_term(LocationVarsIn,synthesis(Component,right_input),b(identifier(InName2),_,_)) ,
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,InName1,
                    LeftUsedLVars,LeftInputAST) ,
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,InName2,
                    RightUsedLVars,RightInputAST) ,
    synthesis_util:delete_numbers_from_atom(Component,ComponentName) ,
    % get the type from the output nodes information
    member(synthesis(type,Type),InfoOut) ,
    Node =.. [ComponentName,LeftInputAST,RightInputAST] ,
    append([[LocationVarNode],LeftUsedLVars,RightUsedLVars],UsedLVars) ,
    set_type(b(Node,Type,InfoOut),AST).
% Input values that are located in the first lines of a synthesized program, thus, the location is
% less than the parameter amount.
location_to_ast(_,CurrentVars,LocationVarsIn,_,Parameter,Solution,LocationVarName,[],b(identifier(VarName),Type,Info)) :-
    member(binding(LocationVarName,int(Value),_),Solution) ,
    Value < Parameter , ! ,
    get_varname_for_input_enumeration(LocationVarsIn,Value,VarName) ,
    member(b(identifier(VarName),Type,Info),CurrentVars).
% An input location that is connected to another operator's output location.
location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,LocationVarName,UsedLVars,AST) :-
    member(binding(LocationVarName,int(Value),_),Solution) ,
    Value >= Parameter ,
    % temporary output variable
    LocationOut = b(identifier(LocationOutName),_,_) ,
    % get output defined at corresponding location
    member(binding(LocationOutName,int(Value),_),Solution) ,
    LocationVarName \= LocationOutName ,
    member(LocationOut,LocationVarsOut) ,
    location_to_ast(Library,CurrentVars,LocationVarsIn,LocationVarsOut,Parameter,Solution,
                    LocationOutName,UsedLVars,AST).

% The argument 'Value' defines the line for which we want to find the corresponding machine variable.
% When creating the behavioral constraint we mark those input location variables with the term synthesis(input(Value))
% so that we are now able to find a specific input variable and determine its name (also in the info: synthesis(machinevar,VarName))
% without assuming any order of the variables (!).
get_varname_for_input_enumeration([],_,_) :- fail.
get_varname_for_input_enumeration([LocationVar|_],Value,VarName) :-
    LocationVar = b(_,_,Info) ,
    member(synthesis(input(Value)),Info) ,
    member(synthesis(machinevar,VarName),Info) , !.
get_varname_for_input_enumeration([_|T],Value,VarName) :-
    get_varname_for_input_enumeration(T,Value,VarName).

% "skip" component is not translated explicitly, i.e. we just return the input parameter its input is mapped to but
% we need to add the skip component node's info to the parameter node
single_input_location_to_ast_aux(skip,b(_,_,Info),ComponentAstNode,ComponentAstNodeWithInfo) :-
    add_texpr_infos(ComponentAstNode,Info,ComponentAstNodeWithInfo).
single_input_location_to_ast_aux(Component,AstNode,_,AstNode) :-
    Component \= skip.
