% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(unsat_cores,[unsat_core/2,
                       unsat_core_with_time_limit/3,
                       unsat_core_with_fixed_conjuncts/3,
                       minimum_unsat_core/2,
                       minimum_unsat_core_with_fixed_conjuncts/3,
                       tcltk_unsat_core_properties/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,misc).
:- module_info(description,'Computation of unsatisfiable cores of predicates.').

:- use_module(library(lists),[append/2,min_member/3]).

:- use_module(bsyntaxtree,[conjunction_to_list/2,
                           conjunct_predicates/2,
                           create_negation/2]).
:- use_module(solver_interface, [solve_predicate/4]).
:- use_module(preferences, [get_preference/2]).
:- use_module(debug, [debug_println/2]).
:- use_module(self_check, [assert_must_succeed/1,assert_must_fail/1]).

:- assert_must_succeed((unsat_core(b(falsity,pred,[]),Core), Core == b(falsity,pred,[]))).
:- assert_must_succeed((unsat_core(b(conjunct(b(falsity,pred,[]),b(truth,pred,[])),pred,[]),Core), Core == b(falsity,pred,[]))).

tcltk_unsat_core_properties(list(CorePredStrings)) :-
   bmachine:b_get_properties_from_machine(Properties), % TO DO: also provide for unsat component
   print('Computing (efficient) unsat core for PROPERTIES'),nl,
   unsat_core_chr(Properties,CorePredicate),
   print(finished),nl,
   findall(CS, (b_interpreter:member_conjunct(C,CorePredicate,_), % TO DO: move member_conjunct to another module
                translate:translate_bexpression(C,CS)), CorePredStrings),
   print(CorePredStrings),nl.

:- use_module(preferences, [temporary_set_preference/3, reset_temporary_preference/2]).
% calls unsat_core but with CHR and other preferences set
unsat_core_chr(Pred,Core) :-
    temporary_set_preference(use_chr_solver,true,CHNG1),
    call_cleanup(unsat_core(Pred,Core),
                 reset_temporary_preference(use_chr_solver,CHNG1)).

% TODO: cleanup / rewrite predicate in order to minimize cores
% i.e. implications could be rewritten into conjuncts

:- assert_must_fail((minimum_unsat_core(b(conjunct(b(falsity,pred,[]),
  b(conjunct(b(less(b(integer(1),integer,[]),b(identifier(x),integer,[])),pred,[]),
  b(less(b(identifier(x),integer,[]),b(integer(1),integer,[])),pred,[])),pred,[])),pred,[]),Core),
   Core == b(less(b(integer(1),integer,[]),b(identifier(x),integer,[])),pred,[]))).
:- assert_must_fail((minimum_unsat_core(b(conjunct(b(falsity,pred,[]),
                                           b(conjunct(b(less(b(integer(1),integer,[]),b(identifier(x),integer,[])),pred,[]),b(less(b(identifier(x),integer,[]),b(integer(1),integer,[])),pred,[])),pred,[])),pred,[]),Core),
                                           Core == b(less(b(identifier(x),integer,[]),b(integer(1),integer,[])),pred,[]))).
% there should be two cores
:- assert_must_succeed((findall(Core, unsat_core(b(conjunct(b(falsity,pred,[]),
                                      b(conjunct(b(less(b(integer(1),integer,[]),b(identifier(x),integer,[])),pred,[]),b(less(b(identifier(x),integer,[]),b(integer(1),integer,[])),pred,[])),pred,[])),pred,[]),Core),
                                      ListOfCores), length(ListOfCores,2))).

% call if you know how much time (in ms) it took to find inconsistency in Predicate
unsat_core_with_time_limit(Predicate,CorePredicate,TimeToFindContradiction) :-
    number(TimeToFindContradiction), TimeToFindContradiction>=0,
    NewTimeOut is TimeToFindContradiction+30, % in case we have 0; just add a few ms
    temporary_set_preference(time_out,NewTimeOut,CHNG1),
    % this time will be multiplied by 2 below
    call_cleanup(unsat_core(Predicate,CorePredicate),
                 reset_temporary_preference(time_out,CHNG1)).

unsat_core(Predicate,CorePredicate) :-
    uc_conjunction_to_list(Predicate,PredicateList),
    unsat_core_with_fixed_conjuncts_lists(PredicateList,[],CoreList),
    conjunct_predicates(CoreList,CorePredicate).

% strip outermost existential quantifiers so that we can examine the conjunction inside
uc_conjunction_to_list(b(exists(_V,Pred),pred,_),List) :- !,
   uc_conjunction_to_list(Pred,List).
% TO DO: add let_predicate (decomposed further below) ? forall ?
uc_conjunction_to_list(Pred,List) :- conjunction_to_list(Pred,List).

unsat_core_with_fixed_conjuncts(Predicate,FixedConjuncts,Core) :-
    uc_conjunction_to_list(Predicate,PredicateList),
    conjunction_to_list(FixedConjuncts,FixedList),
    unsat_core_with_fixed_conjuncts_lists(PredicateList,FixedList,CoreList),
    conjunct_predicates(CoreList,Core).

unsat_core_with_fixed_conjuncts_lists(PredicateList,FixedList,CoreList) :-
    get_preference(unsat_core_algorithm,divide_and_conquer), !,
    temporary_set_preference(allow_enumeration_of_infinite_types,false,OldValueOfPref),
    call_cleanup(unsat_core_with_fixed_conjuncts_lists_divide_and_conquer(PredicateList,FixedList,CoreList),
                 reset_temporary_preference(allow_enumeration_of_infinite_types,OldValueOfPref)).
unsat_core_with_fixed_conjuncts_lists(PredicateList,FixedList,CoreList) :-
    temporary_set_preference(allow_enumeration_of_infinite_types,true,OldValueOfPref),
    call_cleanup(unsat_core_with_fixed_conjuncts_lists_linear(PredicateList,[],FixedList,CoreList),
                 reset_temporary_preference(allow_enumeration_of_infinite_types,OldValueOfPref)).


% simple unsat core computation: remove conjuncts till satisfiability is reached
unsat_core_with_fixed_conjuncts_lists_linear([],DefinitelyInCoreClauses,_FixedConjuncts,DefinitelyInCoreClauses).
unsat_core_with_fixed_conjuncts_lists_linear([C|Cs],DefinitelyInCoreClauses,FixedConjuncts,Core) :-
    % remove C and check for satisfiability
    append([Cs,DefinitelyInCoreClauses,FixedConjuncts],ToCheck),
    conjunct_predicates(ToCheck,ConjunctionToCheck),
    % print('REMOVING: '), translate:print_bexpr(C),nl,
    solve_predicate(ConjunctionToCheck,_State,2,Res),
    debug_println(4,solve_predicate_result(Res)),
    progress_info(Res),
    solution_found(Res), !, % a solution has been found or a time-out occured - C can not be removed; corresponds to solve_result: contradiction_found or no_solution_found(unfixed_deferred_sets)
    % TO DO: if outermost predicate generates virtual_time_out(Reason) then also accepts this
    % TO DO: maybe make treatment uniform with divide_and_conquer_recur_condition which works on solver_result return value
    (can_be_further_decomposed(C,CA,CB)
      -> % print('% Split: '), translate:print_bexpr(CA),nl, print('%        '),translate:print_bexpr(CB),nl,
         unsat_core_with_fixed_conjuncts_lists_linear([CA,CB|Cs],DefinitelyInCoreClauses,FixedConjuncts,Core)
      ;  unsat_core_with_fixed_conjuncts_lists_linear(Cs,[C|DefinitelyInCoreClauses],FixedConjuncts,Core)
    ).
unsat_core_with_fixed_conjuncts_lists_linear([C|Cs],DefinitelyInCoreClauses,FixedConjuncts,Core) :-
    % C can be removed - here we introduce a choice point to not remove C.
    % that way, further cores could be computed by backtracking (i.e. for finding the minimal one)
    unsat_core_with_fixed_conjuncts_lists_linear(Cs,DefinitelyInCoreClauses,FixedConjuncts,Core)
    ;  % Possible TO DO: we could do a can_be_further_decomposed check and put either CA or CB in or both in core
    unsat_core_with_fixed_conjuncts_lists_linear(Cs,[C|DefinitelyInCoreClauses],FixedConjuncts,Core).


solution_found(Res) :-
    Res \= contradiction_found,
    Res \= no_solution_found(_).  %% change to e.g. error to find unsat error core
%solution_found(Res) :- Res \= error. % TO DO: provide parameter to say what kind of core we are looking for

can_be_further_decomposed(TE,A,B) :- can_be_further_decomposed(TE,positive,A,B).
can_be_further_decomposed(b(E,pred,Info),NegPos,A,B) :- can_be_further_decomposed_aux(E,Info,NegPos,A,B).
can_be_further_decomposed_aux(conjunct(A,B),_,positive,A,B).
can_be_further_decomposed_aux(disjunct(A,B),_,negative,A,B).
can_be_further_decomposed_aux(implication(A,B),_,negative,NA,B) :- % A=>B  <-->  not(A) or B
    create_negation(A,NA).
can_be_further_decomposed_aux(negation(D),_,NegPos,NA,NB) :- neg(NegPos,PosNeg),
    can_be_further_decomposed(D,PosNeg,A,B),
    create_negation(A,NA),
    create_negation(B,NB).
can_be_further_decomposed_aux(let_predicate(V,E,P),Info,NegPos,LEA,LEB) :-
   can_be_further_decomposed(P,NegPos,A,B),
   LEA = b(let_predicate(V,E,A),pred,Info),
   LEB = b(let_predicate(V,E,B),pred,Info).
can_be_further_decomposed_aux(lazy_let_pred(V,E,P),Info,NegPos,LEA,LEB) :-
   can_be_further_decomposed(P,NegPos,A,B),
   LEA = b(lazy_let_pred(V,E,A),pred,Info),
   LEB = b(lazy_let_pred(V,E,B),pred,Info).

neg(positive,negative). neg(negative,positive).

% a binary search algorithm in order to only need log(n) instead of n solver calls
% Idea: split list of conjuncts in half, C = [A,B]
unsat_core_with_fixed_conjuncts_lists_divide_and_conquer([],_,_) :- !, fail. % we split to far!
unsat_core_with_fixed_conjuncts_lists_divide_and_conquer([C],_FixedConjuncts,[C]) :- !. % if there is only one conjunct left, it has to belong to the core
unsat_core_with_fixed_conjuncts_lists_divide_and_conquer(Clauses,FixedConjuncts,Core) :-
    split_in_half(Clauses,A,B),
    % print('SPLIT: '), print(A),nl,print(B),nl,
    unsat_core_with_fixed_conjuncts_lists_divide_and_conquer_split(A,B,FixedConjuncts,Core).

unsat_core_with_fixed_conjuncts_lists_divide_and_conquer_split(A,B,FixedConjuncts,Core) :-
    append(A,FixedConjuncts,AllConjuncts),
    conjunct_predicates(AllConjuncts,APred),
    % print('SOLVING LHS: '), translate:print_bexpr(APred),nl,
    solve_predicate(APred,_AState,2,ARealRes),
    progress_info(ARealRes),
    % print(result(ARealRes)),nl,
    !, % do not backtrack into solving or multiple results can be found leading to duplicated cores
    unsat_core_with_fixed_conjuncts_lists_divide_and_conquer_recur1(ARealRes,A,B,FixedConjuncts,Core).

progress_info(ARes) :- progress_info_aux(ARes),!,flush_output(user_output).
progress_info_aux(time_out) :- write('T').
progress_info_aux(no_solution_found(unfixed_deferred_sets)) :- write('-').
progress_info_aux(contradiction_found) :- write('-').
progress_info_aux(solution(_)) :- write('+').
progress_info_aux(no_solution_found(_)) :- write('V').
progress_info_aux(clpfd_overflow) :- write('O').

unsat_core_with_fixed_conjuncts_lists_divide_and_conquer_recur1(ARealRes,A,_B,FixedConjuncts,Core) :-
        divide_and_conquer_recur_condition(ARealRes), % A is (virtually / effectively) unsat. Recur on A.
        unsat_core_with_fixed_conjuncts_lists_divide_and_conquer(A,FixedConjuncts,Core).
unsat_core_with_fixed_conjuncts_lists_divide_and_conquer_recur1(ARealRes,A,B,FixedConjuncts,Core) :-
        append(B,FixedConjuncts,AllConjuncts),
        conjunct_predicates(AllConjuncts,BPred),
        % print('SOLVING RHS: '), translate:print_bexpr(BPred),nl,
        solve_predicate(BPred,_BState,2,BRealRes),
        progress_info(BRealRes),
        % print(result(BRealRes)),nl,
        !,
        unsat_core_with_fixed_conjuncts_lists_divide_and_conquer_recur2(ARealRes,BRealRes,A,B,FixedConjuncts,Core).

unsat_core_with_fixed_conjuncts_lists_divide_and_conquer_recur2(_ARealRes,BRealRes,_A,B,FixedConjuncts,Core) :-
    divide_and_conquer_recur_condition(BRealRes), % B is (virtually / effectively) unsat. Recur on B.
    !,
    unsat_core_with_fixed_conjuncts_lists_divide_and_conquer(B,FixedConjuncts,Core).
unsat_core_with_fixed_conjuncts_lists_divide_and_conquer_recur2(ARealRes,BRealRes,A,B,FixedConjuncts,Core) :-
    % the interesting case: if both are sat
    \+ divide_and_conquer_recur_condition(ARealRes),
    \+ divide_and_conquer_recur_condition(BRealRes),
    % minimize A assuming usat core includes B
    append(B,FixedConjuncts,FixedAndB),
    unsat_core_with_fixed_conjuncts_lists_divide_and_conquer(A,FixedAndB,ACore),
    % now ACore & B is unsat and minimal but not minimum core
    % minimize B assuming ACore
    append(ACore,FixedConjuncts,FixedAndA),
    unsat_core_with_fixed_conjuncts_lists_divide_and_conquer(B,FixedAndA,BCore),
    append(ACore,BCore,Core).

divide_and_conquer_recur_condition(contradiction_found).
divide_and_conquer_recur_condition(no_solution_found(X)) :-
  % to do: if the outermost conjunct already generates a time_out or critical enumeration warning then this should be enabled as a recur condition
    X \= enumeration_warning(_,_,_,_,critical).
%divide_and_conquer_recur_condition(time_out).

split_in_half(C, A, B) :-
    length(C,L),
    Half is L // 2,
    length(A, Half),
    append(A, B, C).

% there is currently no known good algorithm to compute the minimum unsat core
% we can just compute every core and take the minimum one out of the list
:- assert_must_succeed((minimum_unsat_core(b(conjunct(b(falsity,pred,[]),
                                           b(conjunct(b(less(b(integer(1),integer,[]),b(identifier(x),integer,[])),pred,[]),b(less(b(identifier(x),integer,[]),b(integer(1),integer,[])),pred,[])),pred,[])),pred,[]),Core),
                                           Core == b(falsity,pred,[]))).
minimum_unsat_core(Predicate,CorePredicate) :-
    uc_conjunction_to_list(Predicate,PredicateList),
    findall(CoreList,unsat_core_with_fixed_conjuncts_lists(PredicateList,[],CoreList),Cores),
    %print(all_cores(Cores)),nl,
    min_member(compare_length,MinList,Cores),
    conjunct_predicates(MinList,CorePredicate).

minimum_unsat_core_with_fixed_conjuncts(Predicate,FixedConjuncts,CorePredicate) :-
    uc_conjunction_to_list(Predicate,PredicateList),
    conjunction_to_list(FixedConjuncts,FixedList),
    findall(CoreList,unsat_core_with_fixed_conjuncts_lists(PredicateList,FixedList,CoreList),Cores),
    min_member(compare_length,MinList,Cores),
    conjunct_predicates(MinList,CorePredicate).

:- use_module(library(terms),[term_size/2]).
compare_length(List1,List2) :-
    length(List1,L1),
    length(List2,L2),
    (   L1 < L2 -> true
      ; L1 = L2 -> term_size(List1,T1), term_size(List2,T2), T1 < T2).
