% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(fuzzfile, [load_fuzzfile/1, reset_fuzzfile/0, get_z_definitions/1]).

:- dynamic current_fuzzfile/2.

:- use_module(library(lists)).

:- use_module(probsrc(tools)).
:- use_module(probsrc(error_manager)).

:- use_module(probsrc(module_information)).
:- module_info(group,proz).
:- module_info(description,'This module includes a parser for files generated by the fuzz type checker for Z specifications.').

reset_fuzzfile :- retractall(current_fuzzfile(_,_)).

get_z_definitions(Defs) :-
    current_fuzzfile(_,Defs).

load_fuzzfile(File) :-
    reset_fuzzfile,

    (readfile(File,Content) -> true;
    add_error(fuzzfile,'Error while reading file', File),fail),

    (zparser(Defs,Content,[]) -> true;
    add_error(fuzzfile,'Error while parsing fuzz file', File), fail),

    assert(current_fuzzfile(File,Defs)).


%*******************************************************************************
% read file content
readfile(Filename,Content) :-
    open(Filename,read,Stream),
    readbytes(Stream,Content),
    close(Stream).
readbytes(Stream,Res) :-
    get_code(Stream,C),
    (C= -1 -> Res=[] ; Res=[C|Rest],readbytes(Stream,Rest)).


%*******************************************************************************
% Parser
zparser(Items) --> wss, zparser2(Items).
zparser2([I|Rest]) --> zitem(I), !, zparser(Rest).
zparser2([]) --> [].

wss --> whitespace, !, wss.
wss --> "".

zitem(Item) --> "(", !, zitem2(Item).
zitem(String) --> identifier(String).

zitem2([]) --> ")",!.
zitem2(Item) --> itemname(IName), wss, zitem3(IName,Item), wss, ")",!.

zitem3('LIST',List) --> !,zparser(List).
zitem3('NAME',name(Name,Deco)) --> !, identifier(Name), wss, string(Deco).
zitem3('REF',ref(Name,Params)) --> !,
    parsenumber(_), wss, zitem(Name), wss, zitem(Params).
zitem3('SREF',sref(Name,Deco,Params,Renamings)) --> !,
    parsenumber(_), wss, identifier(Name), wss, string(Deco),
    wss, zitem(Params), wss, zitem(Renamings).
zitem3('NUMBER',number(N)) --> !,
    parsenumber(_), wss, [34], parsenumber(N), [34].
zitem3(IName,Item) --> parsenumber(_), zparser(Items),
    {to_lower(IName,LIName), Item =.. [LIName | Items]}.

parsenumber(Number) --> digit(D), parsenumber2(Rest), {number_codes(Number,[D|Rest])}.
parsenumber2([D|Rest]) --> digit(D),!,parsenumber2(Rest).
parsenumber2([]) --> [].

itemname(N) --> letter(L), itemname2(Rest), {atom_codes(N,[L|Rest])}.
itemname2([L|Rest]) --> alphanum(L), !, itemname2(Rest).
itemname2([]) --> [].

string(Atom) --> [34], stringcontent(String), [34], {atom_codes(Atom,String)}.

stringcontent([C|Rest]) --> [C], {C \= 34}, !, stringcontent(Rest).
stringcontent([]) --> [].

identifier(Id) --> [34], idcontent(String), {atom_codes(Id,String)}.

idcontent(String) --> [C], idcontent2(C, String).
idcontent2(34, []) --> !. % closing quotation mark
idcontent2(92, [D|String]) --> % backslash
    [C],{deescape(C,D)},!,idcontent(String).
idcontent2(C, [C|String]) --> idcontent(String).

deescape(95,95). % underscore

:- dynamic letter/3, digit/3, alphanum/3, whitespace/2, to_lower_char/2.
precompiled_rule(Rule) :-
    member(L,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"),
    member(Rule,[letter(L,[L|R],R),alphanum(L,[L|T],T)]).
precompiled_rule(Rule) :-
    member(D,"0123456789"),
    member(Rule,[digit(D,[D|R],R),alphanum(D,[D|T],T)]).
precompiled_rule(whitespace([W|T],T)) :-
    member(W," \t\n\r").
precompiled_rule(to_lower_char(Upper,Lower)) :-
    nth0(N,"ABCDEFGHIJKLMNOPQRSTUVWXYZ",Upper),
    nth0(N,"abcdefghijklmnopqrstuvwxyz",Lower).
store_rules :-
    precompiled_rule(R),
    assert(R),
    fail.
store_rules.
:- store_rules.

to_lower(UItem,LItem) :- name(UItem,Up),to_lower2(Up,Low),name(LItem,Low).
to_lower2([],[]).
to_lower2([Up|URest],[Low|LRest]) :-
    ( to_lower_char(Up,Low) -> true
    ; otherwise             -> Up = Low),
    to_lower2(URest,LRest).
